#ifndef _ALMSRV_H_
#define _ALMSRV_H_

#include "project.def"
#include "rpcdefs.h"
#include <time.h>

#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif

#define CASHOST_FILE "cashosts.csv"
#define MAXALARMS 100
#define DEFAULT_ALARM_WINDOW 3
#define DEFAULT_ALARM_HEARTBEAT 1200
#define DEFAULT_INSTANT_WINDOW 4
#define DEFAULT_TERMINATION_WINDOW 4

#define almNEWALARM     0x01
#define almHEARTBEAT    0x02
#define almOSCILLATION  0x04
#define almDATACHANGE   0x08
#define almINSTANT      0x10
#define almTERMINATE    0x40
#define almTEST         0x80

typedef struct AlarmMsgStruct
{
  char tag[16];
  char device[16];
  char alarmTag[16];
  UINT32 alarmCode;
  UINT32 timestamp;
  UINT32 alarmMask;
  BYTE alarmData[6];
  BYTE alarmDataFormat;
  BYTE alarmDataArraySize;
  BYTE severity;
  BYTE descriptor;
  UINT16 alarmSystem;
} AMS;                    /* AMS sent out upon request */
#define AMS_SIZE  72

typedef struct ADStag    /* Alarm Definition Structure */
{
   char alarmTag[16];
   UINT32 alarmCode;
   UINT32 alarmMask;
   short severity;
   BYTE alarmDataFormat;
   BYTE alarmDataArraySize;
   UINT32 timeRes;
   short hitsChannel;
   short hitsModule;
   char alarmText[40];
   char deviceText[38];
   UINT16 alarmSystem;  /* take 2 bytes from deviceText to hold the system ID */
   char dataText[40];
   char url[40];
   UINT32 ndefs;
} ADS;
#define ADS_SIZE  200
#define MAX_ADS_TABLE_SIZE 100 /* for use with appendAlarmInfoTable() */

typedef struct AMRBtag
{
  UINT32 almStartTime;
  UINT32 almStopTime;
  UINT32 almMinSeverity;
  UINT32 reserved;
} AMRB;                   /* Alarm Message Request Block */
#define AMRB_SIZE  16

typedef struct ALARMstruct/* local alarm list          */
{
  UINT32 timestamp;       /* our old friend                                */
  UINT32 alarmCode;       /* standardized return code (implies severity)   */
  int    removalBlocked;  /* flag to keep alarm in table for CAS readout   */
  BYTE   alarmData[6];    /* alarm specific data (database interpreted)    */
  BYTE   alarmStatus;     /* corresponds to the descriptor                 */
  BYTE   clrCount;        /* record of clear requests                      */
  struct ALARMstruct *next;
} ALARM;

#define ALMARM_WATCH_FILE "almwatch.csv"
typedef struct AlarmWatchStruct
{
  char eqm[8];
  char dev[32];
  char prp[32];
  int  siz;
  int  fmt;
  int  atyp;              /* array type */
  int  asys;              /* alarm system id */
  int  sev;               /* of hi and low (warn is sev - 2) */
  int  chkNormal;         /* TRUE => check vs normal; FALSE = check threshold */
  UINT32 mask;            /* if fmt is an integer type */
  UINT32 normal;          /* normal non-alarm state value (after mask) */ 
  int  cnt;
  int  cntThreshold;      /* set alarm when counter > threshold */
  float hi;               /* high threshold */
  float lo;               /* low threshold */
  float hiwarn;           /* high warn threshold */
  float lowarn;           /* low warn threshold */
  struct AlarmWatchStruct *nxt;
} ALARM_WATCH;

extern ALARM_WATCH *almWatchList;
extern int nAlmWatchList;
/**
 * \struct ALARM_WATCH_THRESHOLDS
 * \brief Defines a TINE Alarm Watch Threshold object
 *
 * - hi is the high value threshold used for setting value_too_high alarms
 * - hiwarn is the high warning value threshold used for setting warn_too_high alarms
 * - lo is the low value threshold used for setting value_too_low alarms
 * - lowarn is the low warning value threshold used for setting warn_too_low alarms
 */
typedef struct 
{
  int cnt;
  float hi;
  float hiwarn;
  float lo;
  float lowarn;
  UINT32 mask;
  UINT32 normal;
} ALARM_WATCH_THRESHOLDS;

extern ALARM *almList;

/* appendAlarmWatchTable() */
/**
 * \brief Inserts a property to be monitored into the local alarm server's Watch Table
 * 
 * Certain alarms are to be set whenever the value of a property exceeds a defineable 
 * threshold.  Such alarms can be managed automatically by the local alarm server if the
 * alarm criteria are entered into the alarm watch table.  This can be achieved by calling
 * this routine (or supplying a startup configuration file almwatch.csv).
 *
 * \param eqm is the 6-character local equipment identifier name, which is internal to server.
 * \param prp is the property which is to be called by the local alarm server
 * \param dev is the device name associated with the property to be called by the local alarm server
 * \param siz is the data array size to be called by the local alarm server
 * \param fmt is the TINE data format to be called by the local alarm server
 * \param atyp is the TINE data array type to be applied to the property called by the local alarm server
 * \param sev is the severity of the alarm issued when the data returned by the call exceed the
 *  given thresholds.
 * \param sys is the alarm system identifier to be associated with the alarm.
 * \param thr is an ALARM_WATCH_THRESHOLDS object specifying the threshold criteria for setting
 *  the alarm
 *  
 * \note Information can also be entered into the alarm watch table by supplying the startup
 * configuration file almwatch.csv, where the input parameters are given in the relevant columns.
 * This is frequently the preferred way of inputting such alarm information, since if does not
 * involve hard-coding such alarm criteria.
 *
 * \return 0 upon success, otherwise a TINE error code.
 *
 */
EXPORT int appendAlarmWatchTable(char *eqm,char *prp,char *dev,int siz,int fmt,int atyp, int sev,int sys,ALARM_WATCH_THRESHOLDS *thr);

/* MaxNumAlarmDefinitions */
/**
 * \brief Determines the maximum alarm definition table size when using the appendAlarmInfoTable()
 * API call to register alarm table information.
 * 
 * If an alarm definition configuration file \<local name\>-alarms.csv is used, this variable
 * plays no role.  When the API call appendAlarmInfoTable() is used to register alarm table
 * information, this variable is used to pre-allocate the appropriate alarm definition table
 * size.
 *
 * \b Default: 100
 *  
 * \b Example:

  MaxNumAlarmDefinitions = 500;  

 */
extern EXPORT int MaxNumAlarmDefinitions;
/* appendAlarmInfoTable() */
/**
 * \brief Inserts an alarm definition into the alarm definition table.
 * 
 * As an alternative to the \<local name\>-alarms.csv configuration file,
 * the front end server can make use of this API call in order to fill in
 * the alarm definition table describing locally generated alarms.
 * This is particularly useful for embedded platforms where there is no 
 * file system, or where a TINE server is used as a translation layer and 
 * needs to map a given alarm system onto the TINE alarm system.
 *
 * \param eqm is the 6-character local equipment identifier name, which is internal to server.
 * \param ads is a pointer to an Alarm Definition Structure (ADS) containing the
 * alarm table information which is to be appended to the alarm definition table.
 *  
 * \return 0 upon success, otherwise a TINE error code.
 *
 * \include eg_appendAlarmInfoTable.c
 * 
 */
EXPORT int appendAlarmInfoTable(char *eqm,ADS *ads);

/* SetDeviceAlarm() */
/**
 * \brief Inserts an alarm into the local alarm server table.
 * 
 * Analogous to SetAlarmEx(), this call inserts alarms into the local alarm server table.
 * A server can set an alarm for a specific device by giving its device name.  This is sometime
 * preferred over specifying its device number as in SetAlarm() or SetAlarmEx().
 * For further information, please see the discussion of the local alarm server.
 *
 * \param eqm is the 6-character local equipment identifier name, which is internal to server.
 * \param devname is the registered device name of the device for which the alarm is to be set.
 * \param almCode is the registered alarm code which defines the alarm
 * \param almData is (6-byte) the optional alarm data set, which supplies additional information for the alarm.
 * \param almFlags is the optional alarm flag which is used in the alarm descriptor information.
 *  
 * \return 0 upon success, otherwise a TINE error code.
 *
 * \include eg_SetDeviceAlarm.c
 */
EXPORT int SetDeviceAlarm(char *eqm,char *devname,UINT32 almCode,BYTE *almData,BYTE almFlags);

/* SetAlarmsEx() */
/**
 * \brief Inserts an alarm into the local alarm server table.
 * 
 * This is the extended call corresponding to SetAlarm(), the difference being 
 * the additional argument flags which can be used to set the alarm descriptor.
 * A server can set an alarm for a specific device by giving its device number.  
 * For further information, please see the discussion of the local alarm server.
 *
 * \param eqm is the 6-character local equipment identifier name, which is internal to server.
 * \param devNr is the device number of the device for which the alarm is to be set.
 * \param almCode is the registered alarm code which defines the alarm
 * \param almData is (6-byte) the optional alarm data set, which supplies additional information for the alarm.
 * \param almFlags is the optional alarm flag which is used in the alarm descriptor information.
 *  
 * \return 0 upon success, otherwise a TINE error code.
 *
 * \include eg_setAlarmEx.c
 */
EXPORT int SetAlarmEx(char *eqm,short devNr,UINT32 almCode,BYTE *almData,BYTE almFlags);

/* SetAlarm() */
/**
 * \brief Inserts an alarm into the local alarm server table.
 * 
 * A server can set an alarm for a specific device by giving its device number.  
 * For further information, please see the discussion of the local alarm server.
 *
 * \param eqm is the 6-character local equipment identifier name, which is internal to server.
 * \param devNr is the device number of the device for which the alarm is to be set.
 * \param almCode is the registered alarm code which defines the alarm
 * \param almData is (6-byte) the optional alarm data set, which supplies additional information for the alarm.
 *  
 * \return 0 upon success, otherwise a TINE error code.
 *
 * \include eg_SetAlarm.c
 */
EXPORT int SetAlarm(char *eqm,short devNr,UINT32 almCode,BYTE *almData);

EXPORT int getAlarmsEx(char *eqpname,AMS *ams,short num,short minSeverity,time_t starttime,time_t stoptime,short startnum);
EXPORT int getAlarms(char *tag,AMS *ams,short num,short minSeverity);
int numAlarmsEx(char *eqpname,UINT32 *almTimeStamp,short *almSeverity,short *NrAtAlmTimeStamp,short *NrAtAlmSeverity);
int numAlarms(char *tag,UINT32 *almTimeStamp,short *almSeverity);

/* ClearAlarm() */
/**
 * \brief Instructs the local alarm server table that the given alarm is to be cleared.
 * 
 * A server can clear alarms for specific device by giving its device number.  
 * Note that all alarms for the device specified are cleared.  Also note, 
 * that "clearing" an alarm means that its "clear" counter is incremented.  
 * It is not removed immediately from the local alarm list. For further information, 
 * please see the discussion of the local alarm server.
 *
 * \param eqm is the 6-character local equipment identifier name, which is internal to server.
 * \param devNr is the device number of the device for which all alarms is to be cleared.
 *  
 * \include eg_ClearAlarm.c
 */
EXPORT void ClearAlarm(char *eqm,short devNr);
/* ClearDeviceAlarm() */
/**
 * \brief Instructs the local alarm server table that the given alarm is to be cleared.
 * 
 * A server can clear alarms for specific device by giving its registered device name.  
 * Note that all alarms for the device specified are cleared.  Also note, 
 * that "clearing" an alarm means that its "clear" counter is incremented.  
 * It is not removed immediately from the local alarm list. For further information, 
 * please see the discussion of the local alarm server.
 *
 * \param eqm is the 6-character local equipment identifier name, which is internal to server.
 * \param devname is the registered device name of the device for which all alarms is to be cleared.
 *  
 * \include eg_ClearDeviceAlarm.c
 */
EXPORT void ClearDeviceAlarm(char *eqm,char *devname);

/* RemoveDeviceAlarm() */
/**
 * \brief Instructs the local alarm server table that the given alarm is to be marked for removal.
 * 
 * A server can remove alarms for specific device by giving its registered device name.  
 * Note that only the alarm specified for the device specified is removed.  Also 
 * note, that "removing" an alarm means that termination bit will be set.  It is 
 * subsequently marked for removal from the local alarm list.  For further information, 
 * please see the discussion of the local alarm server.
 *
 * \param eqm is the 6-character local equipment identifier name, which is internal to server.
 * \param devname is the registered device name of the device for which all alarms is to be removed.
 * \param code is the alarm code, which identifies the specific alarm to be removed
 *
 * \include eg_RemoveDeviceAlarm.c
 */ 
EXPORT void RemoveDeviceAlarm(char *eqm,char *devname,UINT32 code);

/* RemoveAlarm() */
/**
 * \brief Instructs the local alarm server table that the given alarm is to be marked for removal.
 * 
 * A server can remove alarms for specific device by giving its device number.  
 * Note that only the alarm specified for the device specified is removed.  Also 
 * note, that "removing" an alarm means that termination bit will be set.  It is 
 * subsequently marked for removal from the local alarm list.  For further information, 
 * please see the discussion of the local alarm server.
 *
 * \param eqm is the 6-character local equipment identifier name, which is internal to server.
 * \param devNr is the device number of the device for which all alarms is to be removed.
 * \param code is the alarm code, which identifies the specific alarm to be removed
 *
 * \include eg_RemoveAlarm.c
 */ 
EXPORT void RemoveAlarm(char *eqm,short devNr,UINT32 code);
short findSeverity(char *name,UINT32 cc);
int RegisterSeverity(int almCode,int almSeverity);
int fillHRTstamp(BYTE *HRTstamp);
void getAlarmInfo(void);
int getAlarmDefs(char *eqm,ADS **ads);
int numAlarmDefs(char *eqm);
ADS *getAlarmTable(char *eqm,UINT32 code);
int getAlarmInfoTable(char *eqname);
int initAlarmServer(void);
EXPORT int GetAlarm(char *eqpname,char *devname,AMS *ams);
int SetAlarmByDeviceNumber(char *name,short devNr,UINT32 almCode,BYTE *almData,BYTE flags,time_t t);
int SetHrtAlarm(char *name,short devNr,UINT32 almCode,BYTE *almData,BYTE flags,time_t t);
int isCASInUserList(void);

extern char *tagList[];
extern int numStoredAlarms;

/* AlarmWindow */
/**
 * \brief Determines the minimum time an alarm is to remain in the local alarm table 
 * before it can be removed.
 * 
 * IO routines can sometimes set and clear alarms at such a high frequency that 
 * an alarm could go unnoticed if it appeared and disappeared before a client had 
 * a chance to read it.  The AlarmWindow value determines the amount of time (in seconds) 
 * an alarm must remain in the local alarm list before it can be removed.  The default 
 * is 3 seconds, which is generally fine for a central alarm reader polling at 1 second 
 * intervals.
 *
 * \b Default: 3 sec
 *  
 * \b Example:

  AlarmWindow = 10;  // keep alarms in the table a bit longer

 */
extern EXPORT int AlarmWindow;

/* AlarmHeartbeat */
/**
 * \brief Determines the time interval at the 'HEARTBEAT' flag will be re-issued 
 * for persistent alarms.
 * 
 * The purpose of an alarm is to alert operators or maintenance staff to a problem.  
 * If an alarm persists and is not dealt with, its timestamp is adjusted to the current 
 * time at intervals given by the AlarmHeartbeat value.  This will 
 * call attention to the alarm.  It also places the alarm at regular intervals in an 
 * archive of alarms.  The developer can set this to any desired value (usually at 
 * initialization time).
 *
 * \b Default: 1200 sec
 *  
 * \b Example:

  AlarmHeartbeat = 3600;  // heartbeat once per hour

 */
extern EXPORT int AlarmHeartbeat;

/* AlarmInstantWindow */
/**
 * \brief Determines the minimum time an alarm is to remain in the local alarm table 
 * before it can be removed.
 * 
 * Transient alarms (with no duration) are immediately marked for termination as soon 
 * as they are set.  The AlarmInstantWindow determines the amount of time in seconds 
 * an "Instant" Alarm is to remain in the alarm list (default 4 seconds).
 *
 * \b Default: 4 sec
 *  
 * \b Example:

  AlarmInstantWindow = 10;  // keep transient alarms in the table a bit longer

 */
extern EXPORT int AlarmInstantWindow;

/* TerminationWindow */
/**
 * \brief Determines the minimum time an expired alarm is to remain in the local alarm table. 
 * 
 * Terminated alarms remain in the local alarm table for an amount of time specified by
 * this variable.  Thus a client monitoring the local alarm list will be able tosee that 
 * the alarm was set and has finished.
 *
 * \b Default: 4 sec
 *  
 * \b Example:

  TerminationWindow = 10;  // keep terminated alarms in the table a bit longer

 */
extern EXPORT int TerminationWindow;
extern int NTAGS;
extern int NSEVERITYCODE;

/* MaxNumberAlarms */
/**
 * \brief Determines the maximum number of alarms which can be maintained in the
 * local alarm server's alarm list.
 * 
 * Each registered device can have an associated alarm link list.  The length of this 
 * list will not exceed the value entered for MaxNumAlarms (default 100).  The default 
 * is generally more than enough.  Primarily this variable serves as a safety valve 
 * against sudden alarm avalanches.
 *
 * \b Default: 100
 *  
 * \b Example:

  MaxNumberAlarms = 500;  

 */
extern EXPORT int MaxNumberAlarms;

/* AlarmTestState */
/**
 * \brief Determines the minimum time an alarm is to remain in the local alarm table 
 * before it can be removed.
 * 
 * When developing the alarm logic for a particular server is frequently useful to 
 * mark the alarms being set with the almTEST bit.  This can either be done on an 
 * individual basis using the SetAlarmEx() call, or one can set the global variable 
 * AlarmTestState to a non-zero value.
 *
 * \b Default: FALSE
 *  
 * \b Example:

  AlarmTestState = TRUE;  // just testing ...

 */
extern int AlarmTestState;
extern int ALMdebug;
extern int gRequireReadFromCAS;

extern int gIsCentralAlarmServer;
EXPORT void isCentralAlarmServer(int value);

#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif

#endif /* _ALMSRV_H_ */
