#ifndef _HISTORY_H_
#define _HISTORY_H_

#include "project.def"
#include "prolog.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include "servdef.h"

#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif

#ifndef MAX_HISTORY_RECORDS
# define MAX_HISTORY_RECORDS 20
#endif

typedef struct
{
  CONTRACT c;
  int  pollingRate;               /* polling rate in msec */
  int  archiveRate;               /* archive rate in msec */
  int  depthShort;             /* for short term storage */
  int  depthLong;               /* for long term storage */
  int  dataPtr;
  int  heartbeat;            /* archive heartbeat in sec */
  float pTolerance;                 /* percent tolerance */
  float aTolerance;                /* absolute tolerance */
  int  recordIndex;            /* local index identifier */
  int  recordLength;
  int  isScheduled;                     /* schedule flag */
  double lastArchiveTime;
  char ArchiveFileName[32]; /* current working file name */
  double *timestamp;                  /* timestamp array */
  PrpRedirBlk *prd;    /* re-director to CHS if not NULL */
  BYTE *data;
} HistoryRecord;

/**
 * \struct HistorySpecification
 * \brief Defines a TINE History Specification object
 *
 * - pollingRate is the polling rate in msec
 * - archiveRate is the archive rate in sec
 * - depthShort is the ring buffer depth for short term storage
 * - depthLong is the depth in months for long term disk storage
 * - heartbeat is the archive heartbeat in sec
 * - pTolerance is the percent tolerance filtering criterion
 * - aTolerance is the absolute tolerance filtering criterion
 * - rhsServerName is the Remote History Server Name
 * - rhsPropertyName is the associated archive property at the
 *   Remote History Server
 */
typedef struct
{
  int  pollingRate;              /* polling rate in msec */
  int  archiveRate;               /* archive rate in sec */
  int  depthShort;             /* for short term storage */
  int  depthLong;               /* for long term storage */
  int  heartbeat;            /* archive heartbeat in sec */
  float pTolerance;                 /* percent tolerance */
  float aTolerance;                /* absolute tolerance */
  char *rhsServerName;             /* Remote Server Name */
  char *rhsPropertyName;         /* Remote Property Name */
} HistorySpecification;

#ifdef UNIX
# define SH_DENYNONE (S_IWUSR|S_IRUSR|S_IRGRP|S_IROTH)
#endif
#ifdef VXWORKS
# undef SH_DENYNONE
# define SH_DENYNONE (0644)
#endif
#define HISTORY_FILENAME "history.csv"

#define HISTORY_BUFFER_SIZE 1024
#define STS_DEADBAND 1

extern char ArchiveLocation[128];
extern HistoryRecord *historyRecords;
extern int nHistoryRecords;
extern EXPORT int minHistoryPollingRate;
extern EXPORT int maxHistoryRecords;

/* useHighResolutionTimeStamp */
/**
 * \brief Instructs the local history server to use the TINE timestamp 
 * (UTC + milliseconds stored as double) when saving data to disk.
 *
 * The local history subsystem can timestamp the data records stored to
 * disk with a higher resolution timestamp than the traditional UTC variable.
 * This results in more disk space being used by an archive record, as the
 * timestamp is now 8 bytes per archived record instead of 4.  It also 
 * means that to retrieve data records along with the high resolution timestamp,
 * the data should be pulled with the format type CF_DBLDBL.
 *
 * \b Default: FALSE
 */
extern EXPORT int useHighResolutionTimeStamp;
/* useMonthlyHistoryFiles */
/**
 * \brief Instructs the local history server to keep monthly archive files instead of daily
 * files.
 *
 * The local history subsystem maintains one file per record per day.  In some cases it 
 * might be advantageous to keep one file per record per month, for instance when many
 * small records are being maintained.  You can influence this behavior be setting this
 * flag to TRUE.
 */
extern EXPORT int useMonthlyHistoryFiles;

/* AppendHistoryInformation() */
/**
 * \brief Inserts a local history element into the local history server
 * 
 * A server can instruct the local history server to keep a history of the given
 * property by utilizing this call.  The local history server will periodically call
 * the property as specified according to the following input parameters.
 *
 * \param eqm is the 6-character local equipment identifier name, which is internal to server.
 * \param prp is the requested property for which a history is to be kept
 * \param dev is the device name to be associated with the property name supplied as the second
 *  parameter.
 * \param len is the length of the local history call.
 * \param fmt is the TINE format of the local history call.
 * \param idx is the local history index to be identified with this local history element 
 *  (Note: this must be unique with this server process and within the history data repository).
 * \param hspec is a pointer to a HistorySpecification structure, where the relevant filtering
 *  criteria are given.
 *  
 * \note In lieu of calling AppendHistoryInformation(), it is recommended to make use of the
 * startup configuration file 'history.csv', in which all local history critieria can be 
 * introduced.  This has the advantage that no information is hard-coded.  Indeed by introducing
 * a 'history.csv' file one can immediately instruct a server to maintain local history 
 * information (even if the server is developed and maintained by someone else!).
 *
 * \return 0 upon success, otherwise a TINE error code.
 *
 * \b Example:
 *
 * \include eg_AppendHistoryInformation.c
 */
EXPORT int AppendHistoryInformation(char *eqm,char *prp,char *dev,int len,int fmt,int idx,HistorySpecification *hspec);
int LoadHistoryInformation(char *fname,char *expName,char *eqpName);
int getRemoteHistoryServer(char *eqp,char *prp);
void historyInit(char *expName,char *eqpName);
void historyCycle(void);
int loadLTSData(int i,int index,double dstarttime,double dendtime,
                 float lwrLimit, float uprLimit,
                 void *datbuf,int datfmt,int bufsize,int sample);
EXPORT int setSTSData(char *eqp,char *prp, char *dev,int index,BYTE *data,int fmt,int arraysize);
int getLTSData(char *eqp,char *prp, char *dev,int index,
               double dstarttime,double endtime,
               float lwrLimit, float uprLimit,
               BYTE *data,int fmt,int arraysize,int sample,int recsreadFlag);
int getSTSData(char *eqp,char *prp, char *dev,int index,
               double dstarttime,double dendtime,
               float lwrLimit, float uprLimit,
               BYTE *data,int fmt,int arraysize,int sample,int recsreadFlag);
void assignNumberOfPoints(BYTE *data,int fmt,int siz,int *npoints,int *nfound,int recsreadFlag);
int isInRange(BYTE *data,int fmt,int number,float lwrLimit,float uprLimit);
time_t getStartTimeFromDepth(char *eqp,char *prp, int depth);

/* getLastStoredData() */
/**
 * \brief Retreives the last stored data set from disk for the given input parameters
 * 
 * A server can retrieve the last stored data set from disk for any given local history
 * property.  This is frequently useful to insure that a server restart will reload data 
 * various sets, for instance when data are collected according to intermittent triggers, 
 * etc.
 *
 * \param eqm is the 6-character local equipment identifier name, which is internal to server.
 * \param prp is the requested property for which a history is to be kept
 * \param dev is the device name to be associated with the property name supplied as the second
 *  parameter.
 * \param idx is the local history index to be identified with this local history element 
 *  (Note: this must be unique with this server process and within the history data repository).
 * \param data is the buffer to receive the data.  It must be pre-allocated to at least the
 *  array length specified in the 'len' parameter.
 * \param fmt is the TINE format of the local history call.
 * \param len is the length of the local history call.
 *  
 * \return 0 upon success, otherwise a TINE error code.
 *
 * \b Example:
 *
 * \include eg_getLastStoredData.c
 */
EXPORT int getLastStoredData(char *eqm,char *prp, char *dev,int idx,BYTE *data,int fmt,int len);
EXPORT int GetLocalHistoryData(char *eqp,char *prp, char *dev,DTYPE *dout,DTYPE *din);
int HISTQueryFunction(CONTRACT *con,char *dataIn,char *data);

#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif

#endif /* _HISTORY_H_ */
