/** \file almsrv.c
 * \brief TINE Local Alarm Server plus API routines
 *
 * The API routines offered here pertain directly to the TINE Local Alarm Server maintained
 * on the running server.  A server can of course run without setting alarms.  However a 
 * control system will suffer if important information relating to the health of the system
 * is surpressed.  One can set a certain category of alarms (threshold alarms) without issuing
 * any API call, but instead supplying a startup configuration file 'almwatch.csv' with 
 * instructions as to which properties to call and into what range the data returned must fall
 * in order to function properly.  Otherwise calls to ClearAlarm() and SetAlarm() for instance
 * should be made inside a server's IO loop where necessary.
 *
 */
#ifndef NO_DOXYGEN
#include "prolog.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include "servdef.h"
#include "errors.h"
#include "database.h"
#include "almsrv.h"
#include "rpctypes.h"

#define findDeviceNr(eqm,dev) GetDeviceNumber(eqm,dev)
extern char FecDBpath[80];
short ftoi(char *fmt);

int numStoredAlarms = 0;
int NSeverityCodes = 0;
int AlarmServerInitialized = 0;
int AlarmTestState = 0;
int ALMdebug = 0;
ALARM_WATCH *almWatchList = NULL;
int nAlmWatchList = 0;
int gRequireReadFromCAS = FALSE;
int gIsCentralAlarmServer = FALSE;
EXPORT void isCentralAlarmServer(int value) { gIsCentralAlarmServer = value; }
static int alarmInfoTableFile = FALSE;

#define NEW_ALARM_WINDOW 3

#define idx_too_high      0
#define idx_warn_too_high 1
#define idx_too_low       2
#define idx_warn_too_low  3
#define idx_invalid_data  4
ADS almWatchTableADS[] = 
{
  { "Value Too HIGH",value_too_high,0,14,BFMT(CF_FLOAT),1,0,0,0,"measured value is above threshold","",0,"last measured value","",4 },
  { "Value Too High",warn_too_high,0,12,BFMT(CF_FLOAT),1,0,0,0,"measured value is above warn threshold","",0,"last measured value","",4 },
  { "Value Too LOW",value_too_low,0,14,BFMT(CF_FLOAT),1,0,0,0,"measured value is below threshold","",0,"last measured value","",4 },
  { "Value Too Low",warn_too_low,0,12,BFMT(CF_FLOAT),1,0,0,0,"measured value is below warn threshold","",0,"last measured value","",4 },
  { "Invalid Data",invalid_data,0,12,BFMT(CF_LONG),1,0,0,0,"measured value not at set point","",0,"last measured value","",4 }
};
int csvAlmWatchFileGetFormat(char *fmtstr);
CSVCOLUMN csvAlmWatchFileColumns[] = 
{
  { "LOCAL_NAME","",0,CF_TEXT,6,(BYTE *)OFFSETIN(ALARM_WATCH,eqm),NULL, 0 },
  { "DEVICE_NAME","",0,CF_TEXT,32,(BYTE *)OFFSETIN(ALARM_WATCH,dev),NULL, 0 },
  { "PROPERTY","",0,CF_TEXT,32,(BYTE *)OFFSETIN(ALARM_WATCH,prp),NULL, 0 },
  { "SIZE","",0,CF_LONG,1,(BYTE *)OFFSETIN(ALARM_WATCH,siz),NULL, 0 },
  { "FORMAT","",0,CF_NULL,1,NULL,csvAlmWatchFileGetFormat, 0 },
  { "SEVERITY","",0,CF_LONG,1,(BYTE *)OFFSETIN(ALARM_WATCH,sev),NULL, 0 },
  { "ALARM_SYSTEM","0",0,CF_LONG,1,(BYTE *)OFFSETIN(ALARM_WATCH,asys),NULL, 0 },
  { "MASK","0",0,CF_LONG,1,(BYTE *)OFFSETIN(ALARM_WATCH,mask),NULL, 0 },
  { "NORMAL","0",0,CF_LONG,1,(BYTE *)OFFSETIN(ALARM_WATCH,normal),NULL, 0 },
  { "COUNT_THRESHOLD","0",0,CF_LONG,1,(BYTE *)OFFSETIN(ALARM_WATCH,cntThreshold),NULL, 0 },
  { "HIGH","0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(ALARM_WATCH,hi),NULL, 0 },
  { "LOW","0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(ALARM_WATCH,lo),NULL, 0 },
  { "HIGHWARN","0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(ALARM_WATCH,hiwarn),NULL, 0 },
  { "LOWWARN","0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(ALARM_WATCH,lowarn),NULL, 0 }
};
CSVDB csvAlmWatchFileDb =
{
  csvAlmWatchFileColumns,
  sizeof(csvAlmWatchFileColumns)/sizeof(csvAlmWatchFileColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(ALARM_WATCH)
};
int csvAlmWatchFileGetFormat(char *fmtstr)
{
  almWatchList[csvAlmWatchFileDb.cnt].fmt = BFMT(ftoi(fmtstr));
  almWatchList[csvAlmWatchFileDb.cnt].atyp = getArrayType(fmtstr);
  /* while we're at it, set the next pointer ... */
  if (csvAlmWatchFileDb.cnt < csvAlmWatchFileDb.siz - 1)
    almWatchList[csvAlmWatchFileDb.cnt].nxt = &almWatchList[csvAlmWatchFileDb.cnt+1];
  return 0;
}
int GetAlarmWatchTable(void)
{
  int i, cc;
  char almfn[64];

  strncpy(almfn,ALMARM_WATCH_FILE,64);
  if (strlen(gFecDBSubDir) > 0) sprintf(almfn,"%s%s",gFecDBSubDir,ALMARM_WATCH_FILE);
  if ((cc=csvReadFile(FecDBpath,almfn,&csvAlmWatchFileDb,(void **)&almWatchList)) != 0) ccerr(cc);
  nAlmWatchList = csvAlmWatchFileDb.siz;

  for (i=0; i<nAlmWatchList; i++)
  {
    if (almWatchList[i].mask != 0) almWatchList[i].chkNormal = TRUE;
  }
err:
  feclog("ALM: watch file : %s", cc ? erlst[cc] : "found");
  return cc;
}
void checkAlarmsInWatchTable(void)
{
  ALARM_WATCH *aw;
  int j,cc,cnt;
  short devNr;
  ExportListStruct *el;
  char sDevName[36];
  float fval;
  UINT32 ival;
  static BYTE almdata[6];
  static DTYPE din,dout;
  static int done = 0;

  if (gEqpFcnBusySemaphore) return;
  if (!done) memset(&din,0,sizeof(DTYPE));
  for (aw=almWatchList; aw != NULL; aw = aw->nxt)
  {
    if ((el=getExportListItem(aw->eqm)) == NULL) continue;
    if (el->EqpFcn == NULL) continue;
    if ((devNr=GetDeviceNumber(aw->eqm,aw->dev)) < 0)
    {
      if (NGdebug) dbglog("LAS: %s is not a registered device",aw->dev);
      continue;
    }
    dout.dArrayLength = aw->siz;
    dout.dFormat = LFMT(aw->fmt);
    dout.data.vptr = tmpWorkArea;
    dout.dTag[0] = 0;
    strncpy(sDevName,aw->dev,32);
    gEqpFcnBusySemaphore = 1;
    gEqpFcnAccessor = CA_ALARM;
    cc = (*el->EqpFcn)(sDevName,aw->prp,&dout,&din,CA_READ|CA_ALARM);
    gEqpFcnAccessor = 0;
    gEqpFcnBusySemaphore = 0;
    if (cc != 0)
    {
      if (NGdebug)
        dbglog("LAS: %s %s : <%d> %s",
               sDevName,aw->prp,cc, cc > 0 && cc < numErr ? erlst[cc] : "undefined");
      continue;
    }
    cnt = 0;
    for (j=0; j<aw->siz; j++)
    {
      if (aw->atyp == AT_CHANNEL) devNr = j;
      if (aw->atyp != AT_WAVEFORM && aw->siz == el->EqpNumModules) devNr = j; /* assume array of devices */
      switch (LFMT(aw->fmt))
      {
        case CF_BYTE: 
          ival = (UINT32)dout.data.bptr[j]; 
          fval = (float)dout.data.bptr[j]; 
          break;
        case CF_SHORT:
          ival = (UINT32)dout.data.sptr[j]; 
          fval = (float)dout.data.sptr[j]; 
          break;
        case CF_LONG: 
          ival = (UINT32)dout.data.lptr[j]; 
          fval = (float)dout.data.lptr[j]; 
          break;
        case CF_FLOAT:
          ival = (UINT32)dout.data.fptr[j]; 
          fval = (float)dout.data.fptr[j]; 
          break;
        case CF_DOUBLE:
          ival = (UINT32)dout.data.dptr[j]; 
          fval = (float)dout.data.dptr[j]; 
          break;
        default: fval = 0;
      }
      if (aw->chkNormal)
      {
        if (aw->normal != (ival & aw->mask))
        {
          cnt++;
          memcpy(almdata,&ival,4);
          aw->cnt++;
          almWatchTableADS[idx_invalid_data].alarmSystem = aw->asys;
          almWatchTableADS[idx_invalid_data].severity = aw->sev;
          if (aw->cnt > aw->cntThreshold) SetAlarm(aw->eqm,devNr,invalid_data,almdata);
        }
      }
      else if (fval > aw->hiwarn) 
      {
        cnt++;
        memcpy(almdata,&fval,4);
        aw->cnt++;
        almWatchTableADS[idx_too_high].alarmSystem = aw->asys;
        almWatchTableADS[idx_too_high].severity = aw->sev;
        almWatchTableADS[idx_warn_too_high].alarmSystem = aw->asys;
        almWatchTableADS[idx_warn_too_high].severity = aw->sev - 2;
        if (aw->cnt > aw->cntThreshold) 
          SetAlarm(aw->eqm,devNr,fval > aw->hi ? value_too_high : warn_too_high,almdata);
      }
      else if (fval < aw->lowarn) 
      {
        cnt++;
        memcpy(almdata,&fval,4);
        aw->cnt++;
        almWatchTableADS[idx_too_low].alarmSystem = aw->asys;
        almWatchTableADS[idx_too_low].severity = aw->sev;
        almWatchTableADS[idx_warn_too_low].alarmSystem = aw->asys;
        almWatchTableADS[idx_warn_too_low].severity = aw->sev - 2;
        if (aw->cnt > aw->cntThreshold) 
          SetAlarm(aw->eqm,devNr,fval < aw->lo ? value_too_low : warn_too_low,almdata);
      }
      else 
      {
        if (aw->cnt) /* if there is or was an alarm somewhere, run through RemoveAlarm() for the others */
        {
          RemoveAlarm(aw->eqm,devNr,invalid_data);
          RemoveAlarm(aw->eqm,devNr,value_too_high);
          RemoveAlarm(aw->eqm,devNr,value_too_low);
          RemoveAlarm(aw->eqm,devNr,warn_too_high);
          RemoveAlarm(aw->eqm,devNr,warn_too_low);
        }
      }
    }
    /* only clear the watch table counter when a complete pass has no alarms */
    if (cnt == 0) aw->cnt = 0;
  }
  return;
}
void getAlarmInfo(void)
{
  ExportListStruct *el;
  char str[64];

  for (el=ExportList; el != NULL; el=el->next)
  {
    if (el->adsTable != NULL) continue;
    if (getAlarmInfoTable(el->EqpName) < 0)
    {
      sprintf(str,"No alarm information for %s\n>",el->EqpName);
      feclog(str);
    }
  }
}
int csvAlarmInfoGetFormat(char *fmtstr);
CSVCOLUMN csvAlarmInfoColumns[] = 
{
  { "ALARM_TAG","",0,CF_TEXT,16,(BYTE *)OFFSETIN(ADS,alarmTag),NULL,0 },
  { "ALARM_CODE","0",0,CF_LONG,1,(BYTE *)OFFSETIN(ADS,alarmCode),NULL,0 },
  { "ALARM_MASK","0",0,CF_LONG,1,(BYTE *)OFFSETIN(ADS,alarmMask),NULL,0 },
  { "SEVERITY","1",0,CF_SHORT,1,(BYTE *)OFFSETIN(ADS,severity),NULL,0 },
  { "DATA_FORMAT","",0,CF_NULL,1,NULL,csvAlarmInfoGetFormat,0 },
  { "DATA_ARRAYSIZE","0",0,CF_BYTE,1,(BYTE *)OFFSETIN(ADS,alarmDataArraySize),NULL,0 },
  { "TIME_RESOLUTION","0",0,CF_LONG,1,(BYTE *)OFFSETIN(ADS,timeRes),NULL,0 },
  { "HITS_CHANNEL","0",0,CF_SHORT,1,(BYTE *)OFFSETIN(ADS,hitsChannel),NULL,0 },
  { "HITS_MODULE","0",0,CF_SHORT,1,(BYTE *)OFFSETIN(ADS,hitsModule),NULL,0 },
  { "ALARM_TEXT","not supplied",0,CF_TEXT,40,(BYTE *)OFFSETIN(ADS,alarmText),NULL,0 },
  { "DEVICE_TEXT","not supplied",0,CF_TEXT,38,(BYTE *)OFFSETIN(ADS,deviceText),NULL,0 },
  { "DATA_TEXT","not supplied",0,CF_TEXT,40,(BYTE *)OFFSETIN(ADS,dataText),NULL,0 },
  { "URL","not supplied",0,CF_TEXT,40,(BYTE *)OFFSETIN(ADS,url),NULL,0 },
  { "ALARM_SYSTEM","0",0,CF_SHORT,1,(BYTE *)OFFSETIN(ADS,alarmSystem),NULL,0 }
};
CSVDB csvAlarmInfoDb =
{
  csvAlarmInfoColumns,
  sizeof(csvAlarmInfoColumns)/sizeof(csvAlarmInfoColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(ADS)
};
int csvAlarmInfoGetFormat(char *fmtstr)
{
  if (fmtstr == NULL) return 0;
  ((ADS *)csvAlarmInfoDb.tgt)[csvAlarmInfoDb.cnt].alarmDataFormat = BFMT(ftoi(fmtstr));
  return 0;
}
int getAlarmInfoTable(char *eqm)
{
  int i,n,cc,hasAlarmSystem=0;
  ADS *ads = NULL;
  char fn[32],almfn[64];
  ExportListStruct *el;

  for (i=0; i<2; i++)
  {
    memset(fn,0,32);
    strncpy(fn,eqm,6);
    strlwr(fn);
#   if !defined(MSDOS)
    strcat(fn,i == 0 ? "-alarms.csv" : "al.csv");
#   else
    strcat(fn,i == 0 ? "-a.csv" : "al.csv");
#   endif
    strncpy(almfn,fn,64);
    if (strlen(gFecDBSubDir) > 0) sprintf(almfn,"%s%s",gFecDBSubDir,fn);
    if ((cc=csvReadFile(FecDBpath,almfn,&csvAlarmInfoDb,(void **)&ads)) == 0) break;
  }
  if (cc != 0) return cc;
  n = csvAlarmInfoDb.siz;
  for (i=0; i<n; i++) 
  {
    ads[i].ndefs = n;
    if (ads[i].alarmSystem) hasAlarmSystem = 1;
  }
  if ((el=getExportListItem(eqm)) != NULL)
  {
    el->adsTable = ads;
    el->alarmSystem = hasAlarmSystem;
  }
  feclog("read ALM file for %s",eqm);
  alarmInfoTableFile = TRUE;
  return 0;
}
static char gCASName[16] = "";
CSVCOLUMN csvCASNameColumns[] =
{
  { "CAS","",0,CF_TEXT,16,(BYTE *)gCASName,NULL,0 }
};
CSVDB csvCASNameDb =
{
  csvCASNameColumns,
  sizeof(csvCASNameColumns)/sizeof(csvCASNameColumns[0]),
  0,
  NULL,
  NULL,
  NULL,
  0,
  NULL,
  0
};
char *getCASUserName(void)
{
  int cc = 0;
  if ((cc=csvReadFile(FecDBpath,CASHOST_FILE,&csvCASNameDb,NULL)) != 0) ccerr(cc);
  feclog("ALM: recognizing %s as Central Alarm Server",gCASName);
err:
  return gCASName;
}
int isCASInUserList(void)
{
  NAME16 usrlst[10];
  int i, n = 10;
  if (strlen(gCASName) == 0) return 0;
  GetCallerInfo(usrlst,NULL,NULL,NULL,&n);
  for (i=0; i<10 && i<n; i++) if (!strncmp(gCASName,usrlst[i].name,16)) break;
  return i < 10 ? -1 : 0;
}
#ifdef MSDOS
#pragma argsused
#endif
int RegisterSeverity(int almCode,int almSeverity)
{
  return 0;
}
#ifdef MSDOS
#pragma argsused
#endif
void enforceAlarm(short cc)
{
 /* method to make sure this alarm goes out */
  return;
}

void clearCASAlarmList(void)
{
  ExportListStruct *el;
  int cc = 0;
  char devName[256];

  for (el=ExportList; el != NULL; el=el->next)
  {
    if (!strcmp(el->EqpName,SRVEQM_NAME)) continue;
    sprintf(devName,"/%s/CAS/%s",gDeviceContext,el->tagName);
    cc = ExecLink(devName,"REMOVEALARMS",NULL,NULL,CA_WRITE|CA_RETRY);
  }
  feclog("ALM: clear CAS Alarm List : %s",erlst[cc]);
}
int initAlarmServer(void)
{
  getCASUserName();
  GetAlarmWatchTable();
  if (!gIsCentralAlarmServer) clearCASAlarmList();
  AlarmServerInitialized = 1;
  return 0;
}
int verifyDeviceNr(ExportListStruct *el,short devNr)
{
 /* verify that the device number passed in the API is valid */
  if (el == NULL || devNr >= el->EqpNumModules || devNr < 0) return -1;
  return 0;
}
void RefreshMRT(UINT32 t,ExportListStruct *e)
{
  if (e == NULL) return;
  if (t == e->almMRT) 
  {
    e->nAlmMRT++;
  }
  else 
  {
    e->almMRCT = e->almMRT;
    e->nAlmMRCT = e->nAlmMRT;
    e->almMRT = t;
    e->nAlmMRT = 1;
  }
}
void RefreshHSV(short s,ExportListStruct *e) 
{
  if (e == NULL) return;
  if (s == e->almHSV)
  {
    e->nAlmHSV++;
  }
  else if (s > e->almHSV)
  { 
    e->almHSV = (BYTE)s; 
    e->nAlmHSV = 1; 
  }
}
/* Set High Resolution Time Alarm */
int SetHrtAlarm(char *name,short devNr,UINT32 almCode,BYTE *almData,BYTE flags,time_t t)
{
  return SetAlarmByDeviceNumber(name,devNr,almCode,almData,flags,t);
}
int SetAlarmByDeviceNumber(char *name,short devNr,UINT32 almCode,BYTE *almData,BYTE flags,time_t t)
{
  short swpfmt,swpdim,cc = 0,adatasize = 0;
  ALARM *alm = NULL;
  ALARM *lsa = NULL;  /* least significant alarm */
  short lsc,sc = findSeverity(name,almCode),tmpsc = sc;
  ADS *ads;
  ExportListStruct *el;

  if (!AlarmServerInitialized) return -1;
  if (AlarmTestState) flags |= almTEST; /* declared test state */
  /* Instant alarms pickup the TERMINATE signal immediately */
  if (flags & almINSTANT) flags |= almTERMINATE;
  /* fill in ALARMStruct for the tag via link list: */
  if ((el=getExportListItem(name)) == NULL) ccerr(0);
  if (verifyDeviceNr(el,devNr)) ccerr(out_of_range);
  if ((ads=getAlarmTable(name,almCode)) != NULL && almData != NULL)
  {
    if ((swpfmt=LFMT(ads->alarmDataFormat)) == CF_BIT)
    {
      swpfmt = CF_BYTE;
      swpdim = MIN((ads->alarmDataArraySize+7)/8,6);
    }
    else if (swpfmt == CF_HITTS)
    {
      swpfmt = CF_BYTE;
      swpdim = 6;
    }
    else swpdim = MIN(ads->alarmDataArraySize,6);
    adatasize = swpdim*getFormatSize(swpfmt);
    PrepOutgoingAlarmData(almData,swpdim,swpfmt);
  }
  for (alm=el->almLst[devNr]; alm != NULL; alm=alm->next)
  {
    if ((lsc=findSeverity(name,alm->alarmCode)) < tmpsc) { lsa = alm; tmpsc = lsc; }
    if (almCode == alm->alarmCode &&          /* it's in the list */
        !(alm->alarmStatus & almTERMINATE))   /* and not marked for termination */
    {
      if (alm->alarmStatus & almNEWALARM && 
          t <= (time_t)alm->timestamp + NEW_ALARM_WINDOW) /* a double SetAlarm() ? */
      {
        alm->clrCount = 0;                     /* reset clear count      */
        return 0;
      }
      if (almData != NULL && memcmp(almData,alm->alarmData,adatasize))
      {                                        /* alarm data have changed */
        alm->alarmStatus &= ~((BYTE)almNEWALARM); /* remove NEWALARM */
        alm->alarmStatus |= almDATACHANGE;
        alm->timestamp = (UINT32)t;
        RefreshMRT(alm->timestamp,el);
        memcpy(alm->alarmData,almData,adatasize); /* copy the new data ! */
        if (ALMdebug)
        {
          dbglog("ALM: datachange set for %s <%d> (0x%x)",
            el->EqpName,(int)alm->alarmCode,alm->alarmStatus);
        }
      }
      else
      {                                        /* alarm has been reset */
        if (alm->clrCount > 1)
        {
          alm->alarmStatus |= almOSCILLATION;  /* appears to be oscillating */
          if (ALMdebug)
          {
            dbglog("ALM: oscillation set for %s <%d> (0x%x)",
              el->EqpName,(int)alm->alarmCode,alm->alarmStatus);
          }
          alm->alarmStatus &= ~((BYTE)almNEWALARM); /* remove NEWALARM */
          alm->timestamp = (UINT32)t;
          RefreshMRT(alm->timestamp,el);
        }
      }
      alm->clrCount = 0;                     /* reset clear count      */
      return 0;                              /* already in list, so return */
    }
  }
  flags |= almNEWALARM;                        /* alarm not in list */
  if (el->nalarms < MaxNumberAlarms)          /* still room to put the alarm in list */
  {
    if ((alm=(ALARM *)SystemCalloc(1,sizeof(ALARM))) == NULL) ccerr(out_of_server_memory);
    if (gRequireReadFromCAS) alm->removalBlocked = TRUE; /* must be cleared by CAS readout */
    alm->timestamp = (UINT32)t;
    alm->alarmCode = almCode;
    alm->alarmStatus |= flags;
    if (adatasize > 0) memcpy(alm->alarmData,almData,adatasize);
    alm->next = (ALARM *)el->almLst[devNr];
    el->almLst[devNr] = alm;
    el->nalarms++;
    RefreshMRT(alm->timestamp,el);
    RefreshHSV(sc,el);
    if (ALMdebug)
      dbglog("ALM: alarm set for %s <%d> (0x%x)",el->EqpName,(int)alm->alarmCode,alm->alarmStatus);
    return 0;
  }
  else if (lsa != NULL)  /* replace the least significant alarm */
  {
    alm = lsa;
    if (ALMdebug)
    {
      dbglog("ALM: replace alarm %s (%d) with %s (%d)",
	      el->EqpName,(int)alm->alarmCode,
	      el->EqpName,(int)almCode);
    }
    alm->timestamp = (UINT32)t;
    alm->alarmCode = almCode;
    alm->alarmStatus = flags;
    if (adatasize > 0)
    {
      memset(alm->alarmData,0,6);
      memcpy(alm->alarmData,almData,adatasize);
    }
    RefreshMRT(alm->timestamp,el);
    RefreshHSV(sc,el);
  }
  ccerr(max_alarms_exceeded);
err:
  switch (cc)
  {
    case 0: return 0;
    case out_of_server_memory:
    case max_alarms_exceeded:
      enforceAlarm(cc);
      break;
    default:
      if (alm) SystemFree((char *)alm);
      break;
  }
  return cc;
}
/* and in the background... */
ALARM *RemoveAlarmFromList(ExportListStruct *el, int n, ALARM *alm)
{
  ALARM *a0, *a1, *tmp = alm;

  if (el == NULL) return alm->next;
  if (n >= el->EqpNumModules) return alm->next;
  for (a1=a0=el->almLst[n]; a1 != NULL; a0=a1,a1=a1->next)
  {
    if (a1 == alm) /* found it */
    {
      if (a1 == a0) el->almLst[n] = a1->next;
      else a0->next = a1->next;
      tmp = a1->next;
      if (a1->timestamp == el->almMRT) el->nAlmMRT--;
      if (a1->timestamp == el->almMRCT) el->nAlmMRCT--;
      if (el->almHSV == findSeverity(el->EqpName,a1->alarmCode))
      {
        if (--(el->nAlmHSV) <= 0)
        {
          el->nAlmHSV = 0; /* reset it */
          el->almHSV = 0;
        }
      }
      SystemFree((char *)a1); a1 = NULL;
      el->nalarms--;

      break;
    }
  }
  return tmp;    /* return the next link in the list */
}

void SurveyAlarms(struct timeval *tv)
{
  short sc;
  int n, findHSV;
  ALARM *alm;
  ExportListStruct *el;
  static time_t interim = 0;

  if (!AlarmServerInitialized) return;
  if (tv->tv_sec == interim) return;   /* give it some breathing room */
  interim = tv->tv_sec;

  if (nAlmWatchList > 0) checkAlarmsInWatchTable();
  for (el=ExportList; el != NULL; el=el->next)
  {
    if (tv->tv_sec > (time_t)el->almMRT) /* adjust most recent complete timestamp */
    {
      el->almMRCT = el->almMRT;
      el->nAlmMRCT = el->nAlmMRT;
    }
    findHSV = (el->almHSV == 0) ? TRUE : FALSE;
    for (n=0;el->nalarms && n<el->EqpNumModules; n++)
    {
      for (alm=el->almLst[n]; alm != NULL;)
      {
        if ((alm->alarmStatus & almTERMINATE)) /*  time to clear the alarm ? */
        {
          if (alm->removalBlocked) 
          {
            alm = alm->next;
            continue;
          }
          if (alm->timestamp + TerminationWindow < (UINT32)tv->tv_sec) /* okay to clear the alarm */
          {
            if (ALMdebug)
              dbglog("ALM: clear %s <%d> (0x%x)",el->EqpName,(int)alm->alarmCode,alm->alarmStatus);
            alm = RemoveAlarmFromList(el,n,alm);
            continue;
          }
          else if ((alm->alarmStatus & almINSTANT) &&
                    alm->timestamp + AlarmInstantWindow < (UINT32)tv->tv_sec)
          {
            if (ALMdebug)
              dbglog("ALM: clear Instant alarm %s <%d> (0x%x)",el->EqpName,(int)alm->alarmCode,alm->alarmStatus);
            alm = RemoveAlarmFromList(el,n,alm);
            continue;
          }
        }
        else                                   /* check alarm status */
        {
          if ((alm->clrCount > 3) &&
               alm->timestamp + AlarmWindow < (UINT32)tv->tv_sec)
          {
            if (ALMdebug)
              dbglog("ALM: termination for %s <%d> (0x%x)",el->EqpName,(int)alm->alarmCode,alm->alarmStatus);
            alm->alarmStatus &= ~((BYTE)almNEWALARM); /* remove NEWALARM */
            alm->alarmStatus |= almTERMINATE;
            alm->timestamp = tv->tv_sec;
            RefreshMRT(alm->timestamp,el);
          }
          else if (alm->timestamp + AlarmHeartbeat < (UINT32)tv->tv_sec) /*  set the heartbeat */
          {
            if (ALMdebug)
              dbglog("ALM: heartbeat set for %s <%d> (0x%x)",el->EqpName,(int)alm->alarmCode,alm->alarmStatus);
            alm->alarmStatus &= ~((BYTE)almNEWALARM); /* remove NEWALARM */
            alm->alarmStatus |= almHEARTBEAT;
            alm->timestamp = tv->tv_sec;
            RefreshMRT(alm->timestamp,el);
          }
        }
        if (findHSV)
        {
          if ((sc=findSeverity(el->EqpName,alm->alarmCode)) > el->almHSV)
          {
            el->almHSV = sc; el->nAlmHSV = 1;
          }
          else if (sc > 0 && sc == el->almHSV) el->nAlmHSV++;
        }
        alm = alm->next;
      }
    }
  }
  return;
}
/* caller subscribes to NALARMS which calls (REFRESH mode) */
int numAlarmsEx(char *eqpname,UINT32 *almTimeStamp,short *almSeverity,short *NrAtAlmTimeStamp,short *NrAtAlmSeverity)
{
  ExportListStruct *el;

  *almTimeStamp = 0; *almSeverity = 0;

  if ((el=getExportListItem(eqpname)) == NULL) return 0;
  *almTimeStamp = el->almMRCT;
  *almSeverity  = el->almHSV;
  *NrAtAlmTimeStamp = el->nAlmMRCT;
  *NrAtAlmSeverity = el->nAlmHSV;
  return el->nalarms;
}
int numAlarms(char *eqpname,UINT32 *almTimeStamp,short *almSeverity)
{
  return numAlarmsEx(eqpname,almTimeStamp,almSeverity,NULL,NULL);
}
int numAlarmDefs(char *eqm)
{
  ADS *ads;
  ExportListStruct *el;

  if ((el=getExportListItem(eqm)) == NULL) return -non_existent_elem;
  if ((ads=el->adsTable) != NULL) return (int)ads[0].ndefs;
  return 0;
}
int getAlarmDefs(char *eqm,ADS **ads)
{
  ExportListStruct *el;

  if ((el=getExportListItem(eqm)) == NULL) return 0;
  *ads = el->adsTable;
  return numAlarmDefs(eqm);
}
short findSeverity(char *eqm, UINT32 cc)
{
  ADS *ads;
  ExportListStruct *start,*stop,*el;
  int i;

  if (eqm == NULL)
  {
    start = ExportList; stop = NULL;
  }
  else
  {
    if ((start=getExportListItem(eqm)) == NULL) return 0;
    stop = start->next;
  }
  for (el=start; el != stop; el=el->next)
  {
    if ((ads=el->adsTable) != NULL) 
    {
      for (i=0; i<(int)(ads[0].ndefs) && ads[i].alarmCode != (UINT32)cc; i++);
      if (i<(int)ads[0].ndefs) return ads[i].severity;
    }
  }
  switch (cc)
  {
    case invalid_data: return almWatchTableADS[idx_invalid_data].severity;
    case value_too_high: return almWatchTableADS[idx_too_high].severity;
    case value_too_low: return almWatchTableADS[idx_too_low].severity;
    case warn_too_high: return almWatchTableADS[idx_warn_too_high].severity;
    case warn_too_low: return almWatchTableADS[idx_warn_too_low].severity;
  }
  return 0;
}
ADS *getAlarmTable(char *eqm,UINT32 code)
{
  ADS *ads;
  ExportListStruct *el;
  int i;

  if ((el=getExportListItem(eqm)) != NULL)
  {
    if ((ads=el->adsTable) != NULL) 
    {
      for (i=0; i<(int)ads[0].ndefs; i++) if ((UINT32)code == ads[i].alarmCode) return &ads[i];
    }
    switch (code)
    {
      case invalid_data: return &almWatchTableADS[idx_invalid_data];
      case value_too_high: return &almWatchTableADS[idx_too_high];
      case value_too_low: return &almWatchTableADS[idx_too_low];
      case warn_too_high: return &almWatchTableADS[idx_warn_too_high];
      case warn_too_low: return &almWatchTableADS[idx_warn_too_low];
    }
  }
  return NULL;
}
int getAlarmsEx(char *eqpname,AMS *ams,short num,short minSeverity,time_t starttime,time_t stoptime,short startnum)
{
  int sc,dv,i=0,n=0;
  ALARM *alm;
  ADS *ads;
  ExportListStruct *el;
  char defaultDeviceName[36];

  if ((el=getExportListItem(eqpname)) == NULL) return 0;
  for (dv=0; el->nalarms && dv<el->EqpNumModules; dv++)
  {
    for (alm=el->almLst[dv]; alm != NULL; alm=alm->next)
    {
      if (alm->removalBlocked && isCASInUserList()) alm->removalBlocked = FALSE;
      if ((time_t)alm->timestamp >= starttime && (time_t)alm->timestamp <= stoptime &&
          (sc=findSeverity(eqpname,alm->alarmCode)) >= minSeverity )
      {
        if (i++ < startnum) continue;
        /* fill in ams[n] here ... */
        strncpy(ams[n].tag,el->tagName,16);
        sprintf(defaultDeviceName,"DEVICE %d",dv);
        if (el->EqpModuleName) strncpy(ams[n].device,el->EqpModuleName[dv].name,16);
        if (strlen(ams[n].device) == 0) strncpy(ams[n].device,defaultDeviceName,16);
        if ((ads=getAlarmTable(eqpname,alm->alarmCode)) != NULL)
        {
          strncpy(ams[n].alarmTag,ads->alarmTag,16);
          ams[n].alarmMask = ads->alarmMask;
          ams[n].alarmDataFormat = ads->alarmDataFormat;
          ams[n].alarmDataArraySize = ads->alarmDataArraySize;
          ams[n].alarmSystem = ads->alarmSystem;
        }
        ams[n].timestamp = alm->timestamp;
        ams[n].alarmCode = alm->alarmCode;
        memcpy(ams[n].alarmData,alm->alarmData,6);
        ams[n].severity = sc;
        ams[n].descriptor = alm->alarmStatus;
        if (ams[n].alarmCode < (UINT32)numErr && strlen(ams[n].alarmTag) == 0)
          strncpy(ams[n].alarmTag,erlst[ams[n].alarmCode],16);
        if (++n >= num) return n;
      }
    }
  }
  return n;
}
EXPORT int getAlarms(char *eqpname,AMS *ams,short num,short minSeverity)
{
  return getAlarmsEx(eqpname,ams,num,minSeverity,0,0x7fffffffL,0);
}
EXPORT int GetAlarm(char *eqpname,char *devname,AMS *ams) /* get highest alarm for given device */
{
  int sc,scmax=-1,dv;
  ALARM *alm;
  ADS *ads;
  ExportListStruct *el;

  if ((el=getExportListItem(eqpname)) == NULL) return non_existent;
  if ((dv=GetDeviceNumber(eqpname,devname)) < 0) return illegal_equipment_number;
  for (alm=el->almLst[dv]; alm != NULL; alm=alm->next)
  {
    if ((sc=findSeverity(eqpname,alm->alarmCode)) > scmax)
    {
      scmax = sc;
      /* fill in ams[n] here ... */
      strncpy(ams->tag,el->tagName,16);
      strncpy(ams->device,devname,16);
      if ((ads=getAlarmTable(eqpname,alm->alarmCode)) != NULL)
      {
        strncpy(ams->alarmTag,ads->alarmTag,16);
        ams->alarmMask = ads->alarmMask;
        ams->alarmDataFormat = ads->alarmDataFormat;
        ams->alarmDataArraySize = ads->alarmDataArraySize;
        ams->alarmSystem = ads->alarmSystem;
      }
      ams->timestamp = alm->timestamp;
      ams->alarmCode = alm->alarmCode;
      memcpy(ams->alarmData,alm->alarmData,6);
      ams->severity = sc;
      ams->descriptor = alm->alarmStatus;
    }
  }
  return 0;
}
#endif /* NO_DOXYGEN */

/* Local Alarm Server API Calls follow: */
EXPORT int AlarmWindow = DEFAULT_ALARM_WINDOW;
EXPORT int AlarmHeartbeat = DEFAULT_ALARM_HEARTBEAT;
EXPORT int AlarmInstantWindow = DEFAULT_INSTANT_WINDOW;
EXPORT int MaxNumberAlarms = MAXALARMS;
EXPORT int TerminationWindow =  DEFAULT_TERMINATION_WINDOW;
EXPORT int MaxNumAlarmDefinitions = MAX_ADS_TABLE_SIZE;
/* provide API registration as alternative to -alarms.csv file : */
#ifndef SMALL_TINE_LIB /* for MSDOS, etc. ... */
EXPORT int appendAlarmInfoTable(char *eqm,ADS *ads)
{
  ExportListStruct *el;
  int i,n,cc=0,hasAlarmSystem=0;
  
  if (alarmInfoTableFile) ccerr(not_allowed);
  if (eqm == NULL || ads == NULL) ccerr(argument_list_error);
  if ((el=getExportListItem(eqm)) == NULL) ccerr(non_existent_elem);
  if ((el->adsTable == NULL) && 
      (el->adsTable=(ADS *)SystemCalloc(MaxNumAlarmDefinitions,sizeof(ADS))) == NULL)
    ccerr(out_of_server_memory);
  if ((i=el->adsTable[0].ndefs) >= MaxNumAlarmDefinitions) ccerr(resources_exhausted);
  ads->ndefs = 0; /* in case this is not already initialized to zero ! */
  memcpy(&el->adsTable[i],ads,sizeof(ADS));
  n = el->adsTable[0].ndefs + 1;
  for (i=0; i<n; i++) 
  {
    el->adsTable[i].ndefs = n;
    if (el->adsTable[i].alarmSystem) hasAlarmSystem = 1;
  }
  el->alarmSystem = hasAlarmSystem;
err:
  return cc;
}
EXPORT int appendAlarmWatchTable(char *eqm,char *prp,char *dev,int siz,int fmt,int atyp,int sev,int sys,ALARM_WATCH_THRESHOLDS *thr)
{
  int cc = 0;
  ALARM_WATCH *aw;

  if (eqm == NULL || prp == NULL || dev == NULL || thr == NULL) ccerr(argument_list_error);
  if ((aw=(ALARM_WATCH *)SystemCalloc(1,sizeof(ALARM_WATCH))) == NULL) ccerr(out_of_server_memory);
  strncpy(aw->eqm,eqm,EQP_NAME_SIZE);
  strncpy(aw->prp,prp,EQP_PROP_SIZE);
  strncpy(aw->dev,dev,32);
  aw->siz = siz;
  aw->fmt = fmt;
  aw->atyp = atyp;
  aw->sev = sev;
  aw->asys = sys;
  aw->cntThreshold = thr->cnt;
  if (thr->mask != 0)
  {
    aw->mask = thr->mask;
    aw->normal = thr->normal;
    aw->chkNormal = TRUE;
  }
  aw->hi = thr->hi;
  aw->hiwarn = thr->hiwarn;
  aw->lo = thr->lo;
  aw->lowarn = thr->lowarn;
  aw->nxt = almWatchList;
  almWatchList = aw;
  nAlmWatchList++;
  feclog("LAS: append %s %s to alarm watch table",dev,prp);
err:
  return cc;
}
#endif /* SMALL_TINE_LIB */
EXPORT void ClearAlarm(char *name,short devNr)
{
  ALARM *alm;
  ExportListStruct *el;
  int start,stop,i;

  if (!AlarmServerInitialized) return;

  if ((el=getExportListItem(name)) == NULL) return;
  if (devNr == -1)                        /* clear them all */
  {
    start = 0; stop = el->EqpNumModules;
  }
  else
  {
    if (verifyDeviceNr(el,devNr)) return;
    start = devNr; stop = devNr + 1;
  }
  for (i=start; i<stop; i++)
  {
    for (alm=el->almLst[i]; alm != NULL; alm=alm->next)
    {
      if (alm->alarmStatus & almINSTANT) alm->clrCount = 8; /* it'll be cleared immediately */
      else alm->clrCount = MIN(alm->clrCount+1,8); /* increment the clear count */
    }
  }
  return;                                 /* wasn't set */
}
EXPORT void ClearDeviceAlarm(char *name,char *devname)
{
  short devNr = -1;
  if (strcmp(devname,"ALL")) devNr = findDeviceNr(name,devname);
  ClearAlarm(name,devNr);
}
EXPORT void RemoveAlarm(char *name,short devNr,UINT32 code)  /* clear it immediately */
{
  ALARM *alm;
  ExportListStruct *el;

  if (!AlarmServerInitialized) return;
  if ((el=getExportListItem(name)) == NULL) return;
  if (verifyDeviceNr(el,devNr)) return;

  for (alm=el->almLst[devNr]; alm != NULL; alm=alm->next)
  {
    if (code == alm->alarmCode)
    {
      alm->clrCount = 8;
    }
  }
  return;                                 /* wasn't set */
}
EXPORT void RemoveDeviceAlarm(char *name,char *devname,UINT32 code)  /* clear it immediately */
{
  short devNr = findDeviceNr(name,devname);
  RemoveAlarm(name,devNr,code);
}
EXPORT int SetAlarm(char *name,short devNr,UINT32 almCode,BYTE *almData)
{
  return SetAlarmEx(name,devNr,almCode,almData,almNEWALARM);
}
EXPORT int SetAlarmEx(char *name,short devNr,UINT32 almCode,BYTE *almData,BYTE flags)
{
  time_t t = time(NULL);
  return SetAlarmByDeviceNumber(name,devNr,almCode,almData,flags,t);
}
EXPORT int SetDeviceAlarm(char *name,char *devname,UINT32 almCode,BYTE *almData,BYTE flags)
{
  short devNr = findDeviceNr(name,devname);
  return SetAlarmEx(name,devNr,almCode,almData,flags);
}
