/** \file database.c
 * \brief Database routines used for name resolution, etc.
 *      
 */
#include "cprolog.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "database.h"
#include "servdef.h"
#include "srvip.h"
#include "rpctypes.h"
#include "errors.h"
#include "brdcast.h"


int gIsEquipmentNameServer = FALSE;

EXPORT void isEquipmentNameServer(int value) { gIsEquipmentNameServer = value; }
static BYTE zero[6] = {0,0,0,0,0,0};
/** 
 * \internal
 * \brief General client-side database path
 * 
 *  Given by the environment variable setting "TINE_HOME" or (deprecated) 
 *  CONTROLDB.  This specifies the location of the 'cshost.csv' file (among others).
 *
 *  Declared in database.h, initialized in initFecTable().
 *
 * \todo refacter the name to TineHomePath
 *
 */
char ControlDBpath[80];

/** 
 * \internal
 * \brief Path of the RPC database.
 * 
 *  Declared in database.h, initialized in initFecTable().
 */
char gRPCDataBaseFullName[128];

/**
 * \internal
 * \brief Path of the FEC database.
 * 
 * Declared in database.h, initialized in initFecTable().
 */ 
char gFECDataBaseFullName[128];

/**
 * \internal
 * \brief Number of valid entries in RPCData
 * \see RPCData
 */
int RPCNumRecsRead=0;

/**
 * \internal
 * \brief Number of valid entries in RPCFec
 * \see RPCFec
 */
int RPCNumFecsRead=0;

/**
 * \internal
 * \brief Maximum size of the RPCData table
 *
 * This is more or less hard-wired to the MaxNumConnections.  The
 * original intent was to let this grow dynamically as needed, but
 * there are some unfortunate side-effects on some OSes without good
 * MMUs.  So the value is currently being set up front.
 *
 * \see RPCData
 */
int RPCallocNum = 0;

/**
 * \internal
 * \brief Maximum size of the RPCFec table
 *
 * This is more or less hard-wired to the MaxNumConnections.  The
 * original intent was to let this grow dynamically as needed, but
 * there are some unfortunate side-effects on some OSes without good
 * MMUs.  So the value is currently being set up front.
 *
 * \see RPCFec
 */
int FECallocNum=0;

/**
 * \internal
 * \brief Flag whether at least one ENS is known or not.
 */
int NameServerLoaded = 0;

/**
 * \internal
 * \brief Index of the currently active ENS 
 *
 * Array index of NameServer[] array which is currently
 * active.
 */
int curNameServer = 0;

/** 
 * \internal
 * \brief Array of RPC Data.
 * 
 * This array is used as a local cache which stores
 * information of known Context/ExportName pairs. 
 * The array is used to map Context/ExportName to
 * Equipment name, FEC name and other data.  
 * 
 * Declared in servdef.h. 
 * Allocated in initFecTable().
 * 
 * \note The element with index 0 always refers to the current name server.
 */
RPCDataStruct *RPCData = NULL;

/** 
 * \internal
 * \brief RPCDataStruct which holds the ENS data
 *
 * This is the static information which is used as the 0th
 * entry of the RPCData table when at least one ENS is configured.
 *
 * \note If there is an ENS, then it is the 0th element in the cached
 * address tables AND the 0th element in the connection table.
 *
 */
static RPCDataStruct NameServerData; /**< Name Server data.*/

/** 
 * \internal
 * \brief Connection Table Entry which holds the ENS data
 *
 * This is the static information which is used as the 0th
 * entry of the connection table when at least one ENS is configured.
 *
 * \note If there is an ENS, then it is the 0th element in the cached
 * address tables AND the 0th element in the connection table.
 *
 */
ConnTableEntry connNameServer;

/** 
 * \internal
 * \brief Flag which determines whether to use the loopback address when client and
 * server are on the same computer.
 *
 * Some OSes (UNIX, Windows) will use a UNIX socket (pipe) or MMF as a transport
 * between client and server if it is determined that the client is on the same 
 * machine as the server.  This can be over-ridden if gUseLoopback is set to TRUE.
 *
 */
int gUseLoopback = FALSE;

/** 
 * \internal
 * \brief Determine how best to communicate with this FEC
 *
 * Sets the .Prot field to the best available communication protocol.
 *
 * \note If .Prot = 0, the address will again be queried from the ENS.
 * Thus a series of timeouts as seen in the TINE engine will set the .Prot
 * field explicitly to 0 and thereby force an address (re)-resolution.
 *
 */
int fixFecProtocol(RPCFecStruct *fec)
{
  int defaultProtocol = 0;
# if !defined(VXWORKS)
  int myip = getmyipaddr();
# else
  int myip = 0;
# endif
# if defined(MAKE_TINE32_DLL)
  int hWnd;
# endif
# if defined(IPXNW) || defined(IPXBSD)
  int TransportTime = 0;
  if (memcmp(fec->Network,zero,4))
  {
    defaultProtocol = IPX;
#   ifdef IPXNW
    if (IPXGetLocalTarget(fec->Network,
        &fec->ImmediateAddress,
        &TransportTime)) return ipx_socket_error;
#   endif
  }
# endif
# if !defined(VXWORKS)
  if (myip == 0) feclog("cannot determine local ip address!");
# else
  myip = myip;
# endif
  if (memcmp(fec->IPh_addr,zero,4) || !strcmp(fec->Name,SRVFEC_NAME)) defaultProtocol = UDP;
# if defined(UNIX) && !defined(NO_TINE_PIPE)
  if (!gUseLoopback && (!memcmp(&myip,fec->IPh_addr,4) || !strncmp(fec->IP,LOOPBACKADDR,16)))
  {
    memset(fec->Network,0,4);
    defaultProtocol = PIPE;
  }
# endif
# if defined(MAKE_TINE32_DLL)
  if (!gUseLoopback && (!memcmp(&myip,fec->IPh_addr,4) || !strncmp(fec->IP,LOOPBACKADDR,16)))
  {
		defaultProtocol = MMF;
    if ((hWnd=(int)FindWindow(fec->Name,fec->Name)) != (int)NULL)
    {
      memcpy(fec->ImmediateAddress,&hWnd,4); 
		  return 0;
    }
  }
# endif
  fec->Prot = defaultProtocol;
  if (defaultProtocol)
  {
    if (NGdebug)
      dbglog("Communicating with %s via %s\n>",fec->Name,getProtocolAsString(defaultProtocol));
    return 0;
  }
  return illegal_protocol;
}

/**
 * \internal
 * \brief Locate an export name and context in the FEC table.
 * 
 * This function tries to locate an entry in the address cache: 
 * RPCFec[] and RPCData[] according to the specified export name and context.
 * Fetches FEC data by calling GetAddressFromNameServer() and/or GetAddressFromNameFile(), which 
 * creates or overwrites an entry in RPCFec and/or RPCData and return its index.
 * 
 * \param expname Exported device server name
 * \param context Context
 * \param rpcidx [output] index of corresponding entry in RPCData[] or -1. May be set to NULL if no result desired.
 * \return index in RPCFec[] or negative TINE return code.
 * \todo make expname and context const 
 * \see GetAddressFromNameServer 
 * \see GetAddressFromNameFile
 */
int locateTagName(char *expname,char *context,int *rpcidx)
{
  char fecname[FEC_NAME_SIZE];
  int i,cc,idx;
  if ((i=GetAddressFromNameServer(context,expname,fecname,&idx)) < 0 &&
      (i=GetAddressFromNameFile(context,expname,fecname,&idx)) < 0) return i;
  if ((cc=fixFecProtocol(&RPCFec[i])) != 0) return -cc;
  if (rpcidx != NULL) *rpcidx = idx;
  return i;
}

/**
 * \internal
 * \brief Locates the specified FEC name in the FEC table
 *
 * No longer used.  Address (re)-resolution generally requires a lookup of the
 * device server name (export name) and not just the fec name.
 * \deprecated 
 * \todo remove
 */
int locateFecName(char *fecname)
{
  int i,cc;
  if ((i=GetAddressFromNameServer(NULL,NULL,fecname,NULL)) < 0 &&
      (i=GetAddressFromNameFile(NULL,NULL,fecname,NULL)) < 0) return i;
  if ((cc=fixFecProtocol(&RPCFec[i])) != 0) return -cc;
  return i;
}

/**
 * \internal
 * \brief Switches to the next name server.
 * 
 * Increments the global variable curNameServer and
 * copies data from NameServer[curNameServer] to RPCFec[0] and RPCData[0].
 * \see curNameServer
 * \see NameServer
 * \see RPCFec
 * \see RPCData
 */
void ToggleNameServer(void)
{
  curNameServer = (curNameServer+1) % numNameServers;
  memcpy(&RPCFec[0],&NameServer[curNameServer],sizeof(RPCFecStruct));
  strncpy(RPCData[0].FecName,NameServer[curNameServer].Name,FEC_NAME_SIZE);
  memcpy(&RPCFec[1],&NameServer[curNameServer],sizeof(RPCFecStruct));
  strcpy(RPCFec[1].Name,"GENS");
  RPCFec[1].PortOffset += 101;
  feclog("ENS: toggle name server");
}

int fecNameId = -1;
char fecNameTarget[FEC_NAME_SIZE+4];
char tagNameTarget[ELEM_NAME_SIZE+4];
BYTE fecAddressBuf[sizeof(RPCDataStruct)+sizeof(RPCFecStruct)];
void fillinFecName(int id,int cc)
{
  int i;
  static int ens_to = 0;
  fecNameId = -1;             /* release semaphore */
  if (NGdebug > 1) dbglog("asyncLocateFec callback %d <%d>",id,cc);
  for (i=0; i<RPCNumFecsRead && strnicmp(fecNameTarget,RPCFec[i].Name,FEC_NAME_SIZE); i++);
  if (i == RPCNumFecsRead)
  {
    if (RPCNumFecsRead < FECallocNum) RPCNumFecsRead++;
    else return;
  }
  if (cc == 0)
  {
    ens_to = 0;
    prepFecaddr(&RPCFec[i],fecAddressBuf);
    fixFecProtocol(&RPCFec[i]);
    feclog("assigned %s",fecNameTarget);
  }
  else
  {
    if (ens_to++ > 5) ToggleNameServer();
  }
  return;
}
void fillinTagName(int id,int cc)
{
  int i,j,k;
  static int ens_to = 0;
  char preCallExp[20],preCallEqm[8];
  fecNameId = -1;             /* release semaphore */
  if (NGdebug > 1) dbglog("asyncLocateTag callback %d <%d>",id,cc);
  for (j=0; j<RPCNumRecsRead && strnicmp(tagNameTarget,RPCData[j].ExportName,ELEM_NAME_SIZE); j++);
  if (j == RPCNumRecsRead)
  {
    if (RPCNumRecsRead < RPCallocNum) RPCNumRecsRead++;
    else return;
  }
  if (cc == 0)
  {
    ens_to = 0;
    strncpy(preCallExp,RPCData[j].ExportName,16); preCallExp[16] = 0;
    strncpy(preCallEqm,RPCData[j].EqName,6); preCallEqm[6] = 0;
    prepEqpaddr(&RPCData[j],fecAddressBuf);
    if (strnicmp(preCallExp,RPCData[j].ExportName,16) || strncmp(preCallEqm,RPCData[j].EqName,8))
    {
      feclog("ENS: name change in assignment %s %s vs %s %s\n",
        preCallExp,preCallEqm,RPCData[j].ExportName,RPCData[j].EqName);
    }
    for (i=0; i<RPCNumFecsRead && strnicmp(RPCData[j].FecName,RPCFec[i].Name,FEC_NAME_SIZE); i++);
    if (i == RPCNumFecsRead)
    {
      if (RPCNumFecsRead < FECallocNum) RPCNumFecsRead++;
      else return;
    }
    /* avoid allignment problems on some machines: */
    memcpy(fecAddressBuf,&fecAddressBuf[RPCDATASTRUCT_SIZE],RPCFECSTRUCT_SIZE);
    prepFecaddr(&RPCFec[i],fecAddressBuf);
    fixFecProtocol(&RPCFec[i]);
    for (k=0; k<nConnectionTableEntries; k++)
    {
      if (connTable[k]->RPCindex == j && !strnicmp(connTable[k]->name,RPCData[j].ExportName,16))
      {
        connTable[k]->FECindex = i;  /* The device server got assigned to a new FEC ! */
      }
    }
    feclog("ENS: %s [FEC %s] added",tagNameTarget,RPCFec[i].Name);
  }
  else
  {
    if (ens_to++ > 5) ToggleNameServer();
  }
  return;
}
int asyncLocateFecName(char *fecname)
{
  int cc;
  DTYPE dout;
  char scratch[64];

  if (!NameServerLoaded || fecNameId != -1) return -not_ready;

  if (NGdebug > 1) dbglog("Locate FEC %s",fecname);
  sprintf(scratch,"%s/#1",CSHOSTS);
  strncpy(fecNameTarget,fecname,FEC_NAME_SIZE);
  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = RPCFECSTRUCT_SIZE;
  memset(dout.dTag,0,8);
  dout.data.vptr = fecAddressBuf;
  if ((fecNameId=AttachLink(scratch,fecNameTarget,&dout,NULL,
       CA_READ,500,fillinFecName,CM_SINGLE)) < 0)
  {
    cc = -fecNameId;
    fecNameId = -1;
    return cc;
  }
  return fecNameId;
}
int asyncLocateTagName(char *expname,char *context)
{
  int cc;
  DTYPE dout;
  char scratch[64],cntxt[32] = "";

  if (!NameServerLoaded || fecNameId != -1) return -not_ready;

  if (NGdebug > 1) dbglog("Locate export %s",expname);
  strncpy(tagNameTarget,expname,ELEM_NAME_SIZE);

  if (context != NULL) strncpy(cntxt,context,EQP_CONTEXT_SIZE);
  if (strlen(cntxt) == 0) sprintf(cntxt,"#2");
  sprintf(scratch,"%s/%s",CSHOSTS,cntxt);
  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = RPCFECSTRUCT_SIZE+RPCDATASTRUCT_SIZE;
  memset(dout.dTag,0,8);
  dout.data.vptr = fecAddressBuf;

  if ((fecNameId=AttachLink(scratch,expname,&dout,NULL,
       CA_READ,500,fillinTagName,CM_SINGLE)) < 0)
  {
    cc = -fecNameId;
    fecNameId = -1;
    return cc;
  }
  return fecNameId;
}
#ifndef MSDOS
int appendAddressToCache(char *fecname,char *ip, short portOffset)
{
  int i,cc = 0;
  DWORD haddr;

  initClient();
  for (i=0; i<RPCNumFecsRead && strnicmp(fecname,RPCFec[i].Name,FEC_NAME_SIZE); i++);
  if (i < RPCNumFecsRead && RPCFec[i].Prot) ccerr(0);
  strncpy(RPCFec[i].Name,fecname,FEC_NAME_SIZE);
  strncpy(RPCFec[i].IP,ip,16);
  haddr = inet_addr(RPCFec[i].IP);
  memcpy(RPCFec[i].IPh_addr,&haddr,4);
  RPCFec[i].PortOffset = (BYTE)portOffset;
  RPCFec[i].Port[0] = portOffset&0xff;
  RPCFec[i].Port[1] = (portOffset>>8)&0xff;
  fixFecProtocol(&RPCFec[i]);
  RPCNumFecsRead++;
err:
  return cc;
}
int appendServerToCache(char *fecname,char *context,char *expname, char *eqpname)
{
  int i,cc = 0;

  if (fecname == NULL || strlen(fecname) == 0) ccerr(argument_list_error);
  if (expname == NULL || strlen(expname) == 0) ccerr(argument_list_error);
  if (eqpname == NULL || strlen(eqpname) == 0) ccerr(argument_list_error);
  initClient();
  for (i=0; i<RPCNumRecsRead; i++)
  {
    if (context != NULL && strlen(context) &&
        strnicmp(context,RPCData[i].Context,EQP_CONTEXT_SIZE)) continue;
    if (!strnicmp(expname,RPCData[i].ExportName,ELEM_NAME_SIZE)) break;
  }
  if (i < RPCNumRecsRead) ccerr(0);
  strncpy(RPCData[i].FecName,fecname,FEC_NAME_SIZE);
  if (context != NULL) strncpy(RPCData[i].Context,context,EQP_CONTEXT_SIZE);
  strncpy(RPCData[i].ExportName,expname,ELEM_NAME_SIZE);
  strncpy(RPCData[i].EqName,eqpname,EQP_NAME_SIZE);
  RPCNumRecsRead++;
err:
  return cc;
}
#endif

/**
 * \internal
 * \brief Retrieve FEC data from name server or the local cache.
 *
 * This function first attempts to find the requested server name in its local
 * address cache.  If an complete entry is found (including transport protocol)
 * the entry's index is returned.
 * If not, the function requests the FEC data from the name server. 
 * If multiple name servers are known, it takes the first one which returns a result.
 * If the FEC is already buffered locally in RPCFec, the buffer is overweritten. 
 * Otherwise a new entry is created. If the FEC is already buffered locally in RPCData, 
 * the buffer is overwritten. Otherwise a ne entry is created.  
 * 
 * Special case: If the supplied export name equals "NETWORK", this indicates a (multicast)
 * request for the targetted server to respond directly.  This is for cases where the
 * ENS address is not known (diskless server, with no apriori knowledge of the ENS).
 * Usually, this technique is used to find the ENS itself.  Once known, address resolution
 * can proceed normally.  Note that server must be configured to respond to NETWORK
 * address requests, and that the ENS is usually the only server prepared to do so.
 * Also note that there are in such cases potentially two new entries in RPCFec as well as in RPCData. 
 * If the initially request is indeed to find the ENS, then the special slot 0 is filled with the
 * returned address.  Irregardless of this, the next available slot is filled with the full address
 * information of the responding server.
 * 
 * \param context Context 
 * \param expname Export name
 * \param fecname Returns FEC name or empty string. 
 * \param rpcidx  Not used if set to NULL. Otherwise, either returns the index of the corresponding 
 * element in RPCData[] or -1. 
 * \return I. <0 means error code. II. 0 has duplicate meaning:
 * a) fecname was found in buffer, but FEC has Protocol 0, b) FEC Index 0.
 * III. >=0 means FEC index in RPCFec[];
 * 
 * \see RPCData
 * \see RPCFec
 * 
 * \todo Remove gotos and jumps. (unless they're needed)
 * \todo context and expname should be const (not changed by this function).
 * \todo RPCFec has limited memory & is never cleaned => may reach limit if many FECs are used.
 * \todo Same for RPCData.
 * \todo remove dead code
 * \todo Like GetAddressFromNameFile, this function is two functions.
 */
int GetAddressFromNameServer(char *context,char *expname,char *fecname,int *rpcidx)
{
  int retry = 1;  /* Jump-to-err-then-jump-back-counter; max.number of backjumps for the whole loop */
  char scratch[64]; 
  int n,i=0,j,cc=0,isNew=FALSE,isServiceRequest=FALSE;
  char cntxt[32] = "";

  if (rpcidx != NULL) *rpcidx = -1;
  if (!strcmp(expname,SRVEXP_NAME)) isServiceRequest = TRUE;
  if (!NameServerLoaded && !isServiceRequest) return -not_ready;
  
  /* Loop as often as there are name servers.
     variable n is not used.
  */
  for (n=0; n<numNameServers || isServiceRequest; n++)
  {
retryRequest:
    if (expname == NULL)                  /* resolve fecname only */
    {
      if (isServiceRequest) ccerr(not_allowed);
      /* Jump to err */
      if (fecname == NULL || strlen(fecname) == 0) ccerr(non_existent_fec);
      /* Find fecName in RPCFec[] */
      for (i=0; i<RPCNumFecsRead && strnicmp(fecname,RPCFec[i].Name,FEC_NAME_SIZE); i++);
      /* If RPCFec found and FEC's protocol != 0 => Jump to err, return 0*/
      if (i < RPCNumFecsRead && RPCFec[i].Prot) ccerr(0); /* return what's been cached */

      /* Used in doConsumer in case ExecLink is used */
      isNameServerRequest = TRUE;
      cc = tExecLink(0,(char *)fecAddressBuf,CSHOSTS,"#1",fecname,(short)(RPCFECSTRUCT_SIZE),CF_STRUCT,500);      
      isNameServerRequest = FALSE;
      
      /* Jump to err if name call returned error */
      if (cc) ccerr(cc);
      
      /* else if this is a new entry (no fecName found)  */
      if (i == RPCNumFecsRead)
      {
      	/* If buffer available */
        if (RPCNumFecsRead < FECallocNum) 
        {
          RPCNumFecsRead++;
          isNew = TRUE;
        }
        else ccerr(buffer_too_small);
      }
      /* Copy data to new FEC entry */
      prepFecaddr(&RPCFec[i],fecAddressBuf);
      BytesToHex(RPCFec[i].Node,scratch,12);
      if (isNew) feclog("ENS: FEC %s [%s %s] added",RPCFec[i].Name,scratch,RPCFec[i].IP);
    }
    else                /* resolve both tagname and fecname */
    {
      /* overwrite fecName with null string */
      /* if{} tests fecname == NULL, else{} doesn't.*/	
      fecname[0] = 0;
      
      if (strlen(expname) == 0) ccerr(non_existent_elem);
      
      /* if expname == SRVEXP_NAME this is a signal to find the address via multicast */
      if (NGdebug && isServiceRequest) dbglog("using network services as FEC address");
      
      /* Search RPCData for matching context + export name  */
      for (j=0; j<RPCNumRecsRead; j++)
      {
      	/* if context is given, then it has to match; otherwise ignore it */
        if (context != NULL && strlen(context) &&
            strnicmp(context,RPCData[j].Context,EQP_CONTEXT_SIZE)) continue;
        if (!strnicmp(expname,RPCData[j].ExportName,ELEM_NAME_SIZE)) break;
      }
      
      if (j < RPCNumRecsRead)
      { /* EQM Entry found: Search RPCFec for matching FEC name */
        for (i=0; i<RPCNumFecsRead && strnicmp(RPCData[j].FecName,RPCFec[i].Name,FEC_NAME_SIZE); i++);
        /* i< RPCNumFecsRead means FEC entry found. */
        if (rpcidx != NULL) *rpcidx = j;	
        /* Jump out if FEC entry is complete (has protocol information) */
        if (i < RPCNumFecsRead && RPCFec[i].Prot) ccerr(0); /* returned cached entry */
      }
      
      if (isServiceRequest) /* Asking the "NETWORK" for the address */
      {
        if (RPCNumRecsRead == 0 && j == 0) 
        { 
          /* I don't have a configured ENS (GENS)! 
             Leave room for them, because I'm probably trying to find it
          */
          RPCNumFecsRead++; RPCNumRecsRead++; 
        }
        if (RPCNumFecsRead >= FECallocNum) ccerr(out_of_client_memory);
        if (RPCNumRecsRead >= RPCallocNum) ccerr(out_of_client_memory);
        /* "NETWORK" is not yet in the cached tables so add it */
        i = RPCNumFecsRead;
        strncpy(RPCFec[i].Name,SRVFEC_NAME,FEC_NAME_SIZE);
        strncpy(RPCFec[i].IP,gtNetSrvAddr,16);
        memcpy(RPCFec[i].IPh_addr,&tNetSrvAddr.s_addr,sizeof(struct in_addr));
        i = RPCNumRecsRead;	
        strncpy(RPCData[i].FecName,SRVFEC_NAME,FEC_NAME_SIZE);
        strncpy(RPCData[i].ExportName,SRVEXP_NAME,ELEM_NAME_SIZE);
        strncpy(RPCData[i].EqName,SRVEQM_NAME,EQP_NAME_SIZE);
        if (rpcidx != NULL) *rpcidx = i;
        RPCNumFecsRead++; /* add "NETWORK" */
        RPCNumRecsRead++; 
        feclog("NETWORK: added network services to the FEC table");
        if (NGdebug) dbglog("using network services as FEC address");
        ccerr(0); /* Jump out & return 0 */
      } 
      /* Buffer context */
      if (context != NULL) strncpy(cntxt,context,EQP_CONTEXT_SIZE);      
      /* Replace context with "#2" if empty. Legacy ENS request (#2 => return both FEC and EQP addresses) */
      if (strlen(cntxt) == 0) sprintf(cntxt,"#2");
            
      /* Name server request */
      isNameServerRequest = TRUE;
      cc = tExecLink(0,(char *)fecAddressBuf,CSHOSTS,cntxt,expname,(short)(RPCFECSTRUCT_SIZE+RPCDATASTRUCT_SIZE),CF_STRUCT,500);
      isNameServerRequest = FALSE;      
      /* Jump to err if not successful */
      if (cc) ccerr(cc);
      /* If no RPCFec in cache, create a new one */
      if (j == RPCNumRecsRead)
      {
        if (RPCNumRecsRead < RPCallocNum) 
        {
          RPCNumRecsRead++;
          isNew = TRUE;
        }
        else ccerr(buffer_too_small);
      }
      /* Fill RPCFec entry with data */
      prepEqpaddr(&RPCData[j],fecAddressBuf);
      if (context != NULL && strnicmp(context,RPCData[j].Context,EQP_CONTEXT_SIZE) != 0)
      {
        /* context was apparently decorated */
        strncpy(RPCData[j].Context,context,EQP_CONTEXT_SIZE);
      }
      if (rpcidx != NULL) *rpcidx = j;
      /* Look for FEC name in RPCFec[] */
      for (i=0; i<RPCNumFecsRead && strnicmp(RPCData[j].FecName,RPCFec[i].Name,FEC_NAME_SIZE); i++);
      /* No entry found => create a new one. */
      if (i == RPCNumFecsRead)
      {
        if (RPCNumFecsRead < FECallocNum) 
        {
          RPCNumFecsRead++;
          isNew = TRUE;
        }
        else ccerr(buffer_too_small);
      }      
      /* avoid alignment problems on some machines: */
      memcpy(fecAddressBuf,&fecAddressBuf[RPCDATASTRUCT_SIZE],RPCFECSTRUCT_SIZE);
      /* i must be > 0! Do not under any circumstances overwrite the ENS or GENS addresses! */
      if (i > 1) prepFecaddr(&RPCFec[i],fecAddressBuf);
      strncpy(fecname,RPCFec[i].Name,FEC_NAME_SIZE);
      if (isNew) feclog("ENS: %s [FEC %s, EQP %s] added",RPCData[j].ExportName,RPCData[j].FecName,RPCData[j].EqName);
    }
err:
    /* Jump back */
    if ((cc == link_not_open || cc == connection_timeout) && retry-- > 0) 
      goto retryRequest;  
    /* Jump out */
    if (cc == 0) break;                      /* finished */
    feclog("ENS: /%s/%s [%s] : %s",
      context != NULL ? context : "",
	    expname != NULL ? expname : "",
      fecname != NULL ? fecname : "",erlst[cc]);
    ToggleNameServer();
  }
  /* i is the found/new index in RPCFec[] */
  return cc ? -cc : i;                      /* return FEC index */
}

/**
 * \brief Get Data and FEC data from FEC and RPC-databases, or only FEC data.
 * 
 *  <ul>
 *  <li>Function 1. Find an entry in the local RPC Database file which corresponds to
 *  a given export name. Overwrite context and fecname. Store this entry in
 *  RPCData and write its index to rpcidx. Then Function 2.</li>
 *  <li>Function 2. Find an entry in the local FEC Database file which corresponds to
 *  a given fec name.</li>
 *  </ul>
 *  
 * If expname is set to NULL, the function is used only as Function 2. 
 * In this case fecname is an input parameter. otherwise it is used as 
 * Function 1 and Function 2, with fecname as an output parameter.
 * 
 * \param context Function 1: context (output). Function 2: Not used.  
 * \param expname Decides if this function 1 and 2 or only function 2 is used. Function 1: export name. Function 2: NULL.
 * \param fecname Function 1: Fec name (output). Function 2: FEC name (input).
 * \param rpcidx Function 1: Index of found entry in RPCData. Function 2: Not used.
 * 
 * \see gRPCDataBaseFullName
 * \see gFECDataBaseFullName
 * \see RPCData
 * \see RPCFec
 * .
 * \todo Remove gotos and jumps.
 * \todo Make two functions instead of two-in-one.
 */
int GetAddressFromNameFile(char *context,char *expname,char *fecname,int *rpcidx)
{
  FILE *fp = NULL;
  DWORD haddr;
  int i = -1,cc = 0,done = 0,nam_col=-1,fec_col=-1,eqm_col=-1,cxt_col=-1,csv_col=-1;
  int chkAmbiguous = FALSE, hasIncremented = FALSE;
  char str[256],hdr[256],scratch[64], notedContext[32], *contxt = "";
  UINT16 poff;

  if (rpcidx != NULL) *rpcidx = -1;
  if (fecname == NULL) ccerr(non_existent_fec);
  if (expname == NULL) goto acquireFecName;     /* only interested in FEC address*/
  if (strlen(expname) == 0) ccerr(non_existent_elem);
  if (context != NULL && stricmp(context,"DEFAULT")) contxt = context;
  if ((fp=fopen(gRPCDataBaseFullName,"rt")) == NULL) ccerr(no_such_file);
  for (done=0; fgets(str,255,fp) != NULL;)   /* scan file       */
  {
    strtrm(str);
    if (strlen(str) == 0) continue;          /* carriage return */
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
       
    /* Read header. */
    /* done means header has been read */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((nam_col=findcol(hdr,"NAME")) < 0) ccerr(no_such_column);
      if ((fec_col=findcol(hdr,"FEC_NAME")) < 0) ccerr(no_such_column);
      if ((eqm_col=findcol(hdr,"EQP_MODULE")) < 0) ccerr(no_such_column);
      if ((cxt_col=findcol(hdr,"CONTEXT")) < 0) ccerr(no_such_column);
      done = 1;
      continue;
    }
    if (!namcmp(expname,str,nam_col,ELEM_NAME_SIZE))            /* found entry in file */
    {
      if (chkAmbiguous)
      {
        getColumnStr(cxt_col,str,scratch,EQP_CONTEXT_SIZE);
        if (strnicmp(notedContext,scratch,EQP_CONTEXT_SIZE))
        { /* same server different context */
          getColumnStr(fec_col,str,scratch,FEC_NAME_SIZE);
          if (strnicmp(RPCData[i].FecName,scratch,EQP_CONTEXT_SIZE))
          { /* pointing to different FECs */
            if (hasIncremented) RPCNumRecsRead--;
            ccerr(ambiguous);
          }
        }
        continue;
      }
      for (i=0; i<RPCNumRecsRead; i++)
      {
        if (strnicmp(expname,RPCData[i].ExportName,ELEM_NAME_SIZE)) continue;
        if (!strncmp(expname,GLOBALSRV,ELEM_NAME_SIZE) && strlen(contxt) == 0) break;
        if (strnicmp(contxt,RPCData[i].Context,EQP_CONTEXT_SIZE)) continue;
        break;
      }
      if (i == RPCNumRecsRead)
      {
        hasIncremented = TRUE;
        if (RPCNumRecsRead < RPCallocNum) RPCNumRecsRead++;
        else ccerr(buffer_too_small);
      }
      getColumnStr(nam_col,str,RPCData[i].ExportName,ELEM_NAME_SIZE);
      getColumnStr(fec_col,str,RPCData[i].FecName,FEC_NAME_SIZE);
      getColumnStr(eqm_col,str,RPCData[i].EqName,EQP_NAME_SIZE);
      getColumnStr(cxt_col,str,scratch,EQP_CONTEXT_SIZE);
      if (strlen(contxt) > 0)
      {
        if (strnicmp(contxt,scratch,EQP_CONTEXT_SIZE)) 
        {
          if (hasIncremented) RPCNumRecsRead--;
          i = -1; /* not a match */
          continue;
        }
        strncpy(RPCData[i].Context,scratch,EQP_CONTEXT_SIZE);
      }
      else 
      {
        strncpy(notedContext,scratch,EQP_CONTEXT_SIZE);
        RPCData[i].Context[0] = 0;
        chkAmbiguous = TRUE;
      }
      feclog("EFL: %s [FEC %s, EQP %s] added",
	      RPCData[i].ExportName,RPCData[i].FecName,RPCData[i].EqName);
      if (rpcidx != NULL) *rpcidx = i;
      if (!chkAmbiguous) break;  /* finished if context given */
    }
  }
  if (i < 0) ccerr(non_existent);
  strncpy(fecname,RPCData[i].FecName,FEC_NAME_SIZE);
  fclose(fp); fp = NULL;

acquireFecName:
  if (strlen(fecname) == 0) ccerr(non_existent_fec);
  if ((fp=fopen(gFECDataBaseFullName,"r")) == NULL) ccerr(no_such_file);
  for (done=0; fgets(str,255,fp) != NULL;)   /* scan file       */
  {
    strtrm(str);
    if (strlen(str) == 0) continue;          /* carriage return */
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((csv_col=findcol(hdr,"FEC_NAME")) < 0) ccerr(no_such_column);
      done = 1;
      continue;
    }
    if (!namcmp(fecname,str,csv_col,FEC_NAME_SIZE))           /* found entry in file */
    {
      for (i=0; i<RPCNumFecsRead && strnicmp(fecname,RPCFec[i].Name,FEC_NAME_SIZE); i++);
      if (i == RPCNumFecsRead)
      {
        if (RPCNumFecsRead < FECallocNum) RPCNumFecsRead++;
        else ccerr(buffer_too_small);
      }
      getColumnStr(csv_col,str,RPCFec[i].Name,FEC_NAME_SIZE);
      if ((csv_col=findcol(hdr,"FEC_NETWORK")) < 0) ccerr(no_such_column);
      getColumnStr(csv_col,str,scratch,16);
      HexToBytes(scratch,RPCFec[i].Network,8);
      if ((csv_col=findcol(hdr,"FEC_NODE")) < 0) ccerr(no_such_column);
      getColumnStr(csv_col,str,scratch,16);
      HexToBytes(scratch,RPCFec[i].Node,12);
      if ((csv_col=findcol(hdr,"IP_ADDR")) < 0) ccerr(no_such_column);
      getColumnStr(csv_col,str,RPCFec[i].IP,16);
      haddr = inet_addr(RPCFec[i].IP);
      memcpy(RPCFec[i].IPh_addr,&haddr,4);
      feclog("EFL: FEC %s [%s %s] added",RPCFec[i].Name,scratch,RPCFec[i].IP);
      if ((csv_col=findcol(hdr,"PORT_OFFSET")) >= 0)
      {
        getColumnStr(csv_col,str,scratch,16);
        poff = atoi(scratch);
        RPCFec[i].Port[0] = poff&0xff;
        RPCFec[i].Port[1] = (poff>>8)&0xff;
        RPCFec[i].PortOffset = (BYTE)poff;
      }
      else RPCFec[i].PortOffset = 0;
      break;
    }
  }
  if (i < 0) cc = non_existent;
err:
  if (fp) fclose(fp);
  if (cc)
  {
    if (expname == NULL) expname = "(null)";
    if (fecname == NULL) fecname = "(null)";
    feclog("EFL: /%s/%s [%s] : %s",contxt,expname,fecname,erlst[cc]);
  }
  return cc ? -cc : i;
}

int dbaseloaded = 0;
int MaxRPCAllocation = -1;
int reserveFecAddressSpace(void)
{
  int cc = 0;
  if (RPCData != NULL && RPCFec != NULL) 
  {
    /* already reserved */
    return 0;
  }
  if (RPCallocNum == 0) RPCallocNum = MaxNumConnections;
  if (FECallocNum == 0) FECallocNum = MaxNumConnections;
  if (MaxRPCAllocation > 0) RPCallocNum = FECallocNum = MaxRPCAllocation;

  if ((RPCData == NULL) && /* previously allocated ? */
     !(RPCData=(RPCDataStruct *)SystemCalloc(RPCallocNum,sizeof(RPCDataStruct))))
    ccerr(out_of_client_memory);

  if ((RPCFec == NULL) && /* previously allocated ? */
      !(RPCFec=(RPCFecStruct *)SystemCalloc(FECallocNum,sizeof(RPCFecStruct))))
    ccerr(out_of_client_memory);

  feclog("Reserved %d client connection link entries",RPCallocNum);
  feclog("Reserved %d client connection addr entries",FECallocNum);

err:
  return cc;
}
/**
 * \internal 
 * \brief Initializes different things relating to name resolution.
 * 
 * <ul>
 * <li>Initializes ControlDBpath,gRPCDataBaseFullName,gFECDataBaseFullName</li>
 * <li>Allocates RPCData and RPCFec</li>
 * <li>Initialize connTable with name server entry.</li>
 * </ul>
 * 
 * \see ControlDBpath,gRPCDataBaseFullName,gFECDataBaseFullName
 * \see dbaseloaded
 */
int initFecTable(void)
{
  char *ptr;
  int i,cc = 0;

  if (dbaseloaded) return 0;

  /* Determine gRPCDataBaseFullName and gFECDataBaseFullName from environment variables. */ 
  ControlDBpath[0] = 0;
  if (gIsEquipmentNameServer)
  {
    feclog("Server is the ENS: ignore environment settings");
  }
  else if ((ptr=getenv("TINE_HOME")) != NULL || (ptr=getenv("CONTROLDB")) != NULL)
  {
    strncpy(ControlDBpath,ptr,80);
#   ifdef FS_DELIMITER
    if (((i=(int)strlen(ControlDBpath)) > 0) &&
        (ControlDBpath[i-1] != FS_DELIMITER))  /* no delimiter ! */
    {
      ControlDBpath[i] = FS_DELIMITER; ControlDBpath[i+1] = 0;
    }
#   endif
  }
  sprintf(gRPCDataBaseFullName,"%s%s",ControlDBpath,EQPDB_FILE);
  sprintf(gFECDataBaseFullName,"%s%s",ControlDBpath,FECDB_FILE);
  
  if ((cc=reserveFecAddressSpace()) != 0) ccerr(cc);
  RPCNumRecsRead = RPCNumFecsRead = 0;
  addNameServerToTables();
  dbaseloaded = 1;
err:
  if (cc) nConnectionTableEntries = 0;
  return cc;
}

/**
 * \internal 
 * \brief Calls GetNameServerAddress() and Initializes NameServerData and 
 * connNameServer. Sets both output parameters to the MaxNumConnections.
 * 
 * Uses the global variable NameServerLoaded as a lock which ensures
 * that this function is called only once.
 * 
 * \see MaxNumConnections
 * \see NameServerData
 * \see connNameServer
 * \see GetNameServerAddress()
 * 
 * \todo Why are global pointers containing allocated data overwritten with pointers to fixed arrays???
 * \todo name is misleading.
 * \todo nConnectionTableEntries is set although connTable is not touchded
 * \todo what is a function for which returns constants?
 */
int addNameServerToTables(void)
{
  int cc=0;

  if (!NameServerLoaded)
  {
    if ((cc=GetNameServerAddress()) != 0) ccerr(cc);
    strncpy(NameServerData.FecName,NameServer[0].Name,FEC_NAME_SIZE);
    strncpy(NameServerData.ExportName,CSHOSTS,ELEM_NAME_SIZE);
    strncpy(NameServerData.EqName,CSHOSTS,6);
    strncpy(connNameServer.name,CSHOSTS,ELEM_NAME_SIZE);
    strncpy(connNameServer.deviceName,"#0",EQP_DEVICE_SIZE);
    strncpy(connNameServer.prop,"SIZES",EQP_PROP_SIZE);
    connNameServer.EqNo = 0;
    connNameServer.formatOut = CF_SHORT;
    connNameServer.formatIn = CF_NULL;
    connNameServer.RPCindex = 0;
    connNameServer.sizeOut = 2;
    connNameServer.sizeIn = 0;
    connNameServer.RPCprotocol = 5;
    connNameServer.mode = CM_REGISTER;
    connNameServer.blknumIn = 1;
    connNameServer.numblksIn = 1;
    connNameServer.nextDataPtr = 0;
    connTable[0] = &connNameServer; /* Hardwire first call */
    nConnectionTableEntries = 1;
    if (RPCData != NULL || RPCFec != NULL)
    {
      memcpy(&RPCData[0],&NameServerData,sizeof(RPCDataStruct));
      memcpy(&RPCFec[0],&NameServer[0],sizeof(RPCFecStruct));
      connTable[0] = &connNameServer; /* Hardwire first call */
      /* add the GENS information as well ... */
      memcpy(&RPCFec[1],&NameServer[0],sizeof(RPCFecStruct));
      strcpy(RPCFec[1].Name,"GENS");
      RPCFec[1].PortOffset += 101;
      memcpy(&RPCData[1],&NameServerData,sizeof(RPCDataStruct));
      strcpy(RPCData[1].FecName,"GENS");
      strcpy(RPCData[1].EqName,"GRPEQM");
      strcpy(RPCData[1].ExportName,"GENS");
      RPCNumRecsRead = RPCNumFecsRead = 2;
    }
    NameServerLoaded = 1;
  }
err:
  return cc;
}
