/** \file dblib.c
 *
 * \brief general database routines:
 *
 * dblib contains functions which can be used to access CVS files and 
 * extract fields.
 * 
 * It also contains some string and conversion utilities.
 * 
 * \todo clean redevelopment desirable
 */
#ifndef NO_DOXYGEN
#include "prolog.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#if defined(UNIX) && !defined(FREEBSD)
# include <alloca.h>
#endif
#include "servdef.h"
#include "database.h"
#include "errors.h"

#define LOINT(l)       ((int)(l))
#define HIINT(l)       ((int)(((SINT32)(l) >> 16) & 0xFFFF))

unsigned int HexToDec(char *str);
void BytesToHex(BYTE *b,char *s,int n)
{
  int v;

  while (n--)
  {
    *s++ = (v = (*b&0xf0)>>4) < 10 ? '0' + v : 'A' + v - 10;
    *s++ = (v = *b++&0xf) < 10 ? '0' + v : 'A' + v - 10;
  }
  *s = 0;
  strrev(s);
  return;
}

/**
 * \internal
 * \brief Converts a string into a byte array.
 * 
 * Each digit is interpreted as a hexadecimal number.
 * 
 * \note This function also prepends up to 13 '0's to the input string. 
 * 
 * \param s String containing a hexadecimal number (input and output).
 * \param b Pointer to resulting byte array.
 * \param nibs Double length of b and ALSO buffer length (not string length) of s.
 * \todo This is two functions in one: a) Prepend zeros to a number, b) convert to hex
 */
void HexToBytes(char *s,BYTE *b,int nibs)
{
  char buf[20];
  int l,i;

  if ((l=(int)strlen(s)) < nibs)
  {
    strrev(s); strncat(s,"000000000000",nibs-l);
    strrev(s);
  }
  for (i=0; i<nibs/2; i++)
  {
    strncpy(buf,&s[i*2],2); buf[2] = 0;
    b[i] = (unsigned char)strtol(buf,NULL,16);
  }
}
unsigned int HexToDec(char *str)
{
  int i,d;
  unsigned int off,value;

  off=1;value=0;
  for (i=(int)strlen(str)-1;i>=0;i--)
  {
    if (str[i]>='A') d=str[i]-'A'+10;else d=str[i]-'0';
    value=value+off*d;
    off=off*16;
  }
  return value;
}
#if !defined(UNIX) && !defined(VXWORKS) && !defined(NIOS)
/* provide this posix routine for those OSes which don't have it */
EXPORT char *strtok_r(char *s, const char *delim, char **ptrptr)
{
  char *str, *c;
  if (delim == NULL || ptrptr == NULL) return NULL;
  if (s != NULL) /* first call */
  {
    /* start at the beginning of the string given */
    *ptrptr = s + strspn(s,delim); /* skip initial delimiters */
  }
  if (*ptrptr == NULL) return NULL; /* I must be at the end */
  str = *ptrptr; /* this is the current starting point */
  if ((c=strpbrk(str,delim)) != NULL) /* move along */
  {
    *ptrptr = c + strspn(c,delim);
    *c = 0;
  }
  else /* no more tokens */
  {
    *ptrptr = NULL;
  }
  return str;
}
#endif

/**
 * \internal
 * \brief Removes whitespace from the beginning and the end of a string.
 * 
 * Also removes newline characters from the end of the string.
 *
 * \note Modifies the input string in situ.
 * 
 * \return Pointer to modified string.
 */
EXPORT char *strtrm(char *s)
{
  char *start, *end;
  unsigned int len;

  if (s == NULL) return s;  /* NULL given, give it back */
  if ((len=(int)strlen(s)) == 0) return s; /* empty string given, give it back */
  /* find first non-whitespace */
  for (start=s; isspace(*start) && len > 0; start++, len--); 
  /* find last non-whitespace */
  for (end=start+len-1; isspace(*end) && len > 0; end--, len--) *end = 0;
  /* Move to the left */
  memmove(s, start, len + 1);
  /* return a pointer to the modified string as a curtesy */
  return s;
}

/**
 * \internal
 * \brief Counts the lines in a text file.
 * 
 * This function assumes that the first line is a header line,
 * which is not counted. It does not count comment lines.
 * Comment lines are lines which start with '%", '#', or ';'.
 * 
 * \return 
 * <ul> 
 * <li>number of lines</li>
 * <li>-1 if file is empty</li>
 * <li>0 if the file contains only a header line</li>
 * <li>-no_such_file if the file could not be openend</li>
 * </ul> 
 */
int GetDataBaseSize(char *filename)
{
  char str[256];
  int line = -1;
  FILE *fp;

  if (filename == NULL || strlen(filename) == 0) return -argument_list_error;
  if ((fp=fopen(filename,"r")) == NULL) return -no_such_file;
  while (fgets(str,255,fp) != NULL)   /* scan file       */
  {
    strtrm(str);
    if (strlen(str) == 0) continue;          /* carriage return */
    if (strchr("%;#",str[0])) continue;      /* comment line(s) */
    line++;
  }
  fclose(fp);
  return line;
}
int csvNameServerFillPortOffset(char *po);
int csvNameServerFillHAddr(char *ip);

/**
 * \internal
 * \brief Array of FEC data.
 * This array serves as a local cache for DEC addresses.
 * Each entry in RPCFec is identified by a FEC name and contains
 * the information which is needed to access that FEC.
 * 
 * Declared in servdef.h, allocated in initFecTable(), deallocated in freeServerMemory().
 * \note Index 0 in this array always refers to the current name server. 
*/
RPCFecStruct *RPCFec = NULL;

/**
 * \internal
 * \brief Array of Name server data.
 * 
 * This array contains information about available name servers.
 * 
 * Declared in servdef.h, allocated in SetNameServerAddress(), deallocated in freeServerMemory().
 * 
 * \see numNameServers
 */
RPCFecStruct *NameServer = NULL;

/**
 * \internal
 * \brief Number of valid entries in NameServer.
 * 
 * Declared in database.c, servdef.h.
 * 
 * \see NameServer
 */
int numNameServers = 0;

CSVCOLUMN csvNameServerColumns[] =
{
  { "FEC_NAME","",0,CF_TEXT,FEC_NAME_SIZE,(BYTE *)OFFSETIN(RPCFecStruct,Name),NULL,0 },
  { "FEC_NETWORK","0",0,CF_BYTE,8,(BYTE *)OFFSETIN(RPCFecStruct,Network),NULL,0 },
  { "FEC_NODE","0",0,CF_BYTE,12,(BYTE *)OFFSETIN(RPCFecStruct,Node),NULL,0 },
  { "PORT_OFFSET","0",0,CF_NULL,1,NULL,csvNameServerFillPortOffset,0 },
  { "IP_ADDR","",0,CF_TEXT,16,(BYTE *)OFFSETIN(RPCFecStruct,IP),csvNameServerFillHAddr,0 },
};

CSVDB csvNameServerDb =
{
  csvNameServerColumns,
  sizeof(csvNameServerColumns)/sizeof(csvNameServerColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(RPCFecStruct)
};

int csvNameServerFillHAddr(char *ip)
{
  char str1[32],str2[32];
  DWORD haddr = inet_addr(ip);
  memcpy(NameServer[csvNameServerDb.cnt].IPh_addr,&haddr,4);
  if (NGdebug)
  {
    BytesToHex(NameServer[csvNameServerDb.cnt].Network,str1,4);
    BytesToHex(NameServer[csvNameServerDb.cnt].Node,str2,6);
    dbglog("Using Name Server %s [%s %s %s]",
            NameServer[csvNameServerDb.cnt].Name,
            str1,str2,
            NameServer[csvNameServerDb.cnt].IP);
  }
  return 0;
}

int csvNameServerFillPortOffset(char *po)
{
  int poff = atoi(po);
  NameServer[csvNameServerDb.cnt].Port[0] = poff&0xff;
  NameServer[csvNameServerDb.cnt].Port[1] = (poff>>8)&0xff;
  NameServer[csvNameServerDb.cnt].PortOffset = (BYTE)poff;
  return 0;
}

/**
 * \internal
 * \brief Indicates if NameServer has already been initialized.
 * 
 * Roughly equivalent to NameServerLoaded.
 * \see SetNameServerAddress, GetNameServerAddress
 */
int gotNameServerAddress = 0;

/**
 * \internal
 * \brief Initialize Name server information
 * 
 * <ol>
 * <li>If the envronment variables TINE_ENS or CONTROLENS are set, use name
 * server information from there.</li>
 * <li>If not, read name server information from cshosts file which is expected to
 * be in a directory indicated by TINE_HOME or CONTROLDB. Therby changes/sets ControlDBpath.</li>
 * </ol> 
 * 
 * Only called by GetAllocSizesFromNameServer()
 * 
 * \see GetAddressFromNameServer()
 * \see SetNameServerAddress()
 * \see ControlDBpath
 * 
 * \todo ControlDBpath is already initialized in initFecTable.
 */
int GetNameServerAddress(void)
{
# ifdef FS_DELIMITER
  int i;
# endif
  int cc;
  short tmpPortOffset;
  char *ptr;

  if (gIsEquipmentNameServer)
  {
    feclog("Server is the ENS: not using a name server!");
    feclog("TINE HOME : working directory");
    numNameServers = 0;
    return not_implemented;
  }
  if ((ptr=getenv("TINE_ENS")) != NULL || (ptr=getenv("CONTROLENS")) != NULL) 
    SetNameServerAddress(ptr);
  if (gotNameServerAddress) return 0;
  ControlDBpath[0] = 0;
  if ((ptr=getenv("TINE_HOME")) != NULL || (ptr=getenv("CONTROLDB")) != NULL) 
    strcpy(ControlDBpath,ptr);
# ifdef FS_DELIMITER
  if (((i=(int)strlen(ControlDBpath)) > 0) &&
      (ControlDBpath[i-1] != FS_DELIMITER))  /* no delimiter ! */
  {
    ControlDBpath[i] = FS_DELIMITER; ControlDBpath[i+1] = 0;
  }
# endif
  feclog("TINE HOME : [%s]",ControlDBpath);
  numNameServers = 0;
  csvNameServerColumns[3].ptr = (BYTE *)&tmpPortOffset;
  if ((cc=csvReadFile(ControlDBpath,CSHOSTS_FILE,&csvNameServerDb,(void **)&NameServer)) == 0)
  {
    gotNameServerAddress = 1;
    numNameServers = csvNameServerDb.siz;
  }
  else
  {
    feclog("name server address unresolved : %s",erlst[cc]);
  }
  return cc;
}

/**
 * \internal
 * \brief Determines the index of a specified column name in a header line.
 * 
 * Some alternate spellings are tolerated.
 * 
 * \return Column index, starting with 0.
 * 
 */
int findcol(char *str,char *tag)
{
  char tagAlt0[96],tagAlt1[96],tagAlt2[96],altc=0,brkc=0;
  char *c;
  int len,i,k;

  if (str == NULL || tag == NULL) return -1;
  /* effect a string-insensitive compare */
  strncpy(tagAlt0,tag,96);
  strupr(tagAlt0);  
  if ((c=strchr(tagAlt0,' ')) != NULL) 
  {
    brkc = ' ';
    altc = '_';
  }
  else if ((c=strchr(tagAlt0,'_')) != NULL) 
  {
    brkc = '_';
    altc = ' ';
  }
  if (brkc != 0) /* prepare alternates */
  {
    for (i=0,k=0; i<(int)strlen(tagAlt0) && i<95; i++)
    {
      if (tagAlt0[i] != brkc)
      {
        tagAlt1[i] = tagAlt0[i];
        tagAlt2[k++] = tagAlt0[i];
      }
      else /* tagAlt0[i] == brkc */
      {
        tagAlt1[i] = altc;
      }
    }
    tagAlt1[i] = 0; tagAlt2[k] = 0;
  }
  for (i=0,c=str; strlen(c); i++,c++)
  {
    while (*c == '\"') { c = strchr(++c,'\"'); if (c++ == NULL) return -1; }
    while (isspace(*c) != 0) c++;
    strupr(c);
    len = (int)strcspn(c,",\n"); while (isspace(c[len-1]) != 0) len--;
    if (!strncmp(c,tagAlt0,MAX(len,(int)strlen(tagAlt0)))) return i;
    if (brkc != 0) 
    {
      if (!strncmp(c,tagAlt1,MAX(len,(int)strlen(tagAlt1)))) return i;
      if (!strncmp(c,tagAlt2,MAX(len,(int)strlen(tagAlt2)))) return i;
    }
    if ((c=strpbrk(c,",\n")) == NULL) return -1;
  }
  return -1;
}

/**
 * \internal
 * \brief Extracts a partial string from a comma-separated line.
 * 
 * This function scans a comma-separated line and extracts the column
 * which is indexed by col.  The extracted string is copied into
 * the supplied buffer (and is thread safe!)
 * 
 * \param col Column number. Indexing starts at 0.
 * \param str Comma-separated line.
 * 
 * \return 0 upon success or a TINE error code.
 * 
 */
int getColumnStr(int col,char *str,char *buf,int bufsiz)
{
  int colcnt=0, cc = 0;
  char *c;
  int sslen, FindEndOfString = FALSE;

  if (str == NULL || buf == NULL) ccerr(argument_list_error);

  /* find column position 'col' (count the non-embedded commas) : */
  /* c points to current position in string; DO NOT scan past the string length !*/
  for (c=str; strlen(c) > 0 && colcnt < col; c++)
  {
    /* scan through the string ... */
    if (FindEndOfString)
    {
      /* jump ahead to it */
      if ((c=strchr(c,'\"')) == NULL) ccerr(string_expected);
      FindEndOfString = FALSE; /* found it */
    }
    else
    {
      if (*c == ',') colcnt++;
      else if (*c == '\"') FindEndOfString = TRUE; /* quoted string */
    }
  }

  /* did we jump out because strlen(c) == 0 ? */
  if (colcnt < col) ccerr(no_such_column);

  while (isspace(*c) != 0) c++; /* trim leading spaces */
  if (*c == '\"') 
  {
    c++;   /* if quoted, jump over initial quote */
    sslen = (int)strcspn(c,"\""); /* length up to the next \" */  
  }
  else
  {
    sslen = (int)strcspn(c,",\n"); /* length up to the next , or \n */  
  }
  while (sslen > 0 && isspace(c[sslen-1]) != 0) sslen--; /* Strip trailing spaces */
  if (sslen > bufsiz)  
  {
    /* sslen longer than expected; write log message and continue anyway */
    feclog("sub string \"%s\" too long (%d vs %d)",c,sslen,bufsiz);
    sslen = bufsiz;
  }
  strncpy(buf,c,sslen); /* Copy sub-string to buffer */
  if (sslen < bufsiz) buf[sslen] = 0; /* add \0 if there's room */

err:
  if (cc)
  {
    c = str == NULL ? "<null string>" : str;
    feclog("csv: column %d in %s -> %s",col,c,erlst[cc]);
  }
  return cc;
}

/**
 * \internal
 * \brief Extracts a partial string from a comma-separated line.
 * 
 * This function scans a comma-separated line and extracts the column
 * which is indexed by col.  The extracted string is copied into
 * a static buffer which is returned (not thread-safe!).
 * 
 * A column may me at most 127 bytes long or will be truncated.
 * 
 * \param col Column number. Indexing starts at 0.
 * \param str Comma-separated line.
 * 
 * \return Pointer to static buffer.
 * 
 */
char *colptr(int col,char *str)
{
  static char buf[128];
  buf[0] = 0;
  getColumnStr(col,str,buf,128);
  return buf;
}

/**
 * \internal
 * \brief Compares a column of a comma-separated line to a string.
 * 
 * This function scans a comma-separated line and extracts the column which is indexed by col. 
 * 
 * \param name String to compare.
 * \param str Comma-separated line.
 * \param col Column index (starting at 0)
 * \param limit Maximal length of the extracted column (length of result buffer).
 * 
 * \return like strcmp().
 * 
 */
int namcmp(char *name,char *str,int col,int limit)
{
  int n;
  char *scratch;
# if defined(VXWORKS)
  /*  VxWorks does not understand alloca() */
  char scratchBuffer[516];
  if (limit < 0 || limit > 512) return -1;
  scratch = scratchBuffer;
# else
  if (limit < 0) return -1;
  if ((scratch=(char *)alloca(limit)) == NULL) return -1; /* practically impossible */
# endif
  if (name == NULL && str != NULL) return -1; /* problem with arguments */
  if (str == NULL && name != NULL) return -1;
  if (name != NULL && col < 0) return -1; /* column not present */
  getColumnStr(col,str,scratch,limit); 
  n = MAX(limit,(int)strcspn(scratch,",\n"));
  return strncmp(name,scratch,n);
}

/**
 * \internal
 * \brief Standard Initializer used for most CSV database readers
 * 
 * Allocate internal buffer memory for a CSVDBStruct.
 * This function assumes that a CSVDBStruct is already partially initialized.
 * <ul>  
 * <li>It allocates memory for the buffer db->tgt if db->tgt_siz > 0. 
 * It uses the values of the 
 * fields db->siz, db->tgt_siz, and db->ncols to determine the buffer size.
 * </li>
 * <li> 
 * It also assumes that db->col is already initialized.
 * For each entry in db->col, it replaces the
 * pointer col->start with the previous value of col->ptr and
 * adds the pointer db->tgt as an offset to the pointer col->ptr.
 * </li>
 * 
 * <li> 
 * It sets db->cnt to 0. 
 * </li>
 * 
 * </ul>
 * 
 * \see CSVDBStruct
 * \todo Misleading name. Was this originally intended as (part of) an initializer ?
 * (This is THE standard initializer for most CSV files)
 * \todo What does the col pointer juggling mean (understood roughly)?
 * (if the csv contents are supposed to be filling a structure line by line, then the
 *  pointer has to be advanced).
 * \todo The concept behind CSVDBStruct and csv* functions is very hard to understand.
 * \bug Do not call csvStdInit twice (UNLESS csvStdExit is called in between) ... 
 */
EXPORT void csvStdInit(struct CSVDBStruct *db)
{
  int i;
  if (db == NULL || db->tgt_siz <= 0) return;
  db->cnt = 0;
  if (db->tgt != NULL) SystemFree(db->tgt);
  db->tgt = (BYTE *)SystemCalloc(db->siz,db->tgt_siz);
  for (i=0; i<db->ncols; i++) 
  {
    db->col[i].start = db->col[i].ptr; /* reset via csvStdExit() */
    /* set the ptr to the container addr plus the original offset */
    db->col[i].ptr = db->tgt + (size_t)db->col[i].start; 
  }
}
/**
 * \internal
 * \brief Standard Interator for most CSV database files
 *
 * Increments column index, increments column pointers.
 * \todo pointer juggling (see code)
 * \todo better: use real (separate) iterator instead of double meaning of CSVDBStruct
 *
 * \note the pointer advancement and size of the container have nothing to do with
 * column offsets and row indices.
 */
EXPORT int csvStdIterator(struct CSVDBStruct *db)
{
  int i;
  if (db == NULL || db->tgt_siz <= 0) return -1;  /* don't iterate */
  for (i=0; i<db->ncols; i++) if (db->col[i].fmt != CF_NULL) db->col[i].ptr += db->tgt_siz;
  db->cnt++;
  return 0;
}
/**
 * \internal
 * \brief Standard Finalizer for most CSV database files.
 * 
 * Restores column pointers to their values before memory allocation.
 * Has to be called before Memory-reallocation (csvStdInit()).
 * Does not de-allocate memory => Is not kind of a destructor
 * as the name suggests.
 * 
 * \todo misleading name (why?)
 * \todo pointer juggling (see csvStdIterator)  
 * \bug this function rather seems to be a work-around (see )
 * \see csvStdInit
 */
EXPORT void csvStdExit(struct CSVDBStruct *db)
{
  int i;
  if (db == NULL || db->tgt_siz <= 0) return;
  for (i=0; i<db->ncols; i++) db->col[i].ptr = db->col[i].start;
}
/**
 * \internal
 * \brief This function takes a string, converts it to a data type
 * specified by a format constant, and stores the result at a place
 * pointed to by a void pointer.
 */
EXPORT int csvAssignValue(void *ptr,int fmt,int siz,char *str)
{
  char *pstr = str;
  if (ptr == NULL) return 0;         /* intentionally not assigned */
  if (str == NULL) return argument_list_error;
  if (fmt != CF_TEXT && fmt != CF_NULL && strlen(str) == 0) pstr = "0";
  switch (fmt)
  {
    case CF_NULL: break;
    case CF_TEXT: if (siz == 0) *(char *)ptr = 0; else strncpy((char *)ptr,pstr,siz); break;
    case CF_BYTE: 
      if (siz > 1) HexToBytes(pstr,(BYTE *)ptr,siz); 
      else *(BYTE *)ptr = (BYTE)atoi(pstr);
      break;
    case CF_SHORT: *(short *)ptr = (short)strtol(pstr,NULL,0); break;
    case CF_LONG:  *(SINT32 *)ptr = strtol(pstr,NULL,0); break;
    case CF_FLOAT: *(float *)ptr = (float)strtod(pstr,NULL); break;
    case CF_DOUBLE: *(double *)ptr = strtod(pstr,NULL); break;
    default: return illegal_format;
  }
  return 0;
}
/**
 * \internal
 * \brief Reads the contents of a CSV file and places the entries in a container.
 * 
 * The contents of parameter target may be NULL or a valid pointer.
 * If target is not NULL, target will be freed and overwritten with 
 * a pointer to the internal memory of the database structure.
 * If db has no "ini", target will be overwritten
 * with a pointer to db's internal memory.
 * If db has an ini function, it reallocates its internal memory. 
 * \param path The directory where the file is
 * \param file The file to open
 * \param db The db structure 
 * \param target pointer which is freed an then set to db->tgt
 * \see findcol
 * \note The memory to which target points is freed before re-allocation.
 * \todo clarify weird allocation/deallocation
 * \bug db-tgt is not de-allocated when target != NULL
 *
 * \note To date, the only csv file which is occasionally re-read and 
 * actually needs the memory re-allocation is users.csv. Otherwise,
 * making the same csvReadFile() call a second time will have no effect
 * in the end as long as the standard ini, itr, and exi routines are used.
 *
 * \note some CSV File use different (non-standard) routines. e.g. csvExportIterator().
 *
 */
EXPORT int csvReadFile(char *path,char *file,CSVDB *db,void **target)
{
  FILE *fp=NULL;
  int i,cc=0,done=0;
  char *c,str[256],hdr[256],scratch[80],fn[128];

  if (file == NULL) return argument_list_error;
  if (path == NULL) strncpy(fn,file,128); else sprintf(fn,"%s%s",path,file);
  /* find database size => effectively the number of rows */
  if ((db->siz=GetDataBaseSize(fn)) < 0) ccerr(-db->siz);
  /* was a pre-allocated target passed ? */
  if (target != NULL) db->tgt = (BYTE *)*target;
  /* call init routine if there */
  if (db->ini != NULL) db->ini(db);
  /* either the target was pre-allocated or the initializer allocated it */
  if (target != NULL) *target = db->tgt;
  /* read through */
  if ((fp=fopen(fn,"r")) == NULL) ccerr(no_such_file);
  while (fgets(str,255,fp) != NULL)
  {
    strtrm(str);
    if (strlen(str) == 0) continue;               /* carraige return */
    if (strchr("%;#\n",str[0]) != NULL) continue; /* comment line */
    if (!done)
    {
      strncpy(hdr,str,255);
      for (i=0; i<db->ncols; i++) db->col[i].pos = findcol(hdr,db->col[i].nam);
      done = TRUE;
      continue;
    }
    for (i=0; i<db->ncols; i++)
    {
      if (db->col[i].pos < 0) 
      {
        c = db->col[i].def;
      }
      else
      {
        getColumnStr(db->col[i].pos,str,scratch,64); scratch[64] = 0;
        c = scratch;
      }
      /* assign value to structure location */
      cc = csvAssignValue(db->col[i].ptr,db->col[i].fmt,db->col[i].siz,c);
      if (cc)
      {
        sprintf(str,"column %s entry %d : %s\n",db->col[i].nam,i,erlst[cc]);
        feclog(str); ttyoutput(str);
        ccerr(cc);
      }
      /* call action function if there */
      if (db->col[i].fcn != NULL) if (db->col[i].fcn(c) != 0) break;    
    }
    /* call iterator if there */
    if (db->itr != NULL) if (db->itr(db) != 0) break;
  }
  /* call end routine if there */
  if (db->exi != NULL) db->exi(db);
err:
  if (fp) fclose(fp);
  if (cc) db->siz = 0;
  return cc;
}
/**
 * \internal
 * \brief Reads data from memory and converts it to a string.
 * \param ptr Pointer in memory
 * \param fmt TINE CF data type expected at pointer
 * \param siz Maximal string size
 * \param str Destination
 */
EXPORT int csvGetValue(void *ptr,int fmt,int siz,char *str)
{
  if (ptr == NULL) return 0;         /* intentionally not assigned */
  if (str == NULL) return argument_list_error;
  switch (fmt)
  {
    case CF_TEXT: strncpy(str,(char *)ptr,siz); break;
    case CF_BYTE: BytesToHex((BYTE *)ptr,str,siz); break;
    case CF_SHORT: sprintf(str,"%d",*(short *)ptr); break;
#   ifdef MSDOS
    case CF_LONG:  sprintf(str,"%ld",*(SINT32 *)ptr); break;
#   else
    case CF_LONG:  sprintf(str,"%d",*(SINT32 *)ptr); break;
#   endif
    case CF_FLOAT: sprintf(str,"%g",*(float *)ptr); break;
    case CF_DOUBLE: sprintf(str,"%g",*(double *)ptr); break;
    default: return illegal_format;
  }
  return 0;
}
/**
 * \internal
 * \brief Writes data in a CSVDB to a file.
 * \param path Directory
 * \param file File name
 * \param db A CSVDB
 */
EXPORT int csvWriteFile(char *path,char *file,CSVDB *db)
{
  FILE *fp=NULL;
  int i,j,cc=0;
  char str[256],scratch[64],fn[128];

  if (file == NULL) return argument_list_error;
  if (path == NULL) strncpy(fn,file,128); else sprintf(fn,"%s%s",path,file);
  if ((fp=fopen(fn,"w")) == NULL) ccerr(file_error);
  for (i=0,str[0]=0; i<db->ncols; i++)
  {
    strcat(str,db->col[i].nam);
    if (i<db->ncols-1) strcat(str,","); 
    if (db->tgt_siz > 0)  /* return pointers to start position */
    {
      db->col[i].ptr = db->tgt + (size_t)db->col[i].start;
    }
  }
  strcat(str,"\n");
  fwrite(str,strlen(str),1,fp);
  for (j=0; j<db->siz; j++)
  {
    str[0] = 0;
    for (i=0; i<db->ncols; i++)
    {
      csvGetValue(db->col[i].ptr,db->col[i].fmt,db->col[i].siz,scratch);
      strcat(str,scratch);
      if (i<db->ncols-1) strcat(str,","); 
    }
    strcat(str,"\n");
    fwrite(str,strlen(str),1,fp);
    /* call iterator if there */
    if (db->itr != NULL) if (db->itr(db) != 0) break;
  }
  /* call end routine if there */
  if (db->exi != NULL) db->exi(db);
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
#endif /* NO_DOXYGEN */
