/** \file history.c
 * \brief TINE Local History Server module plus API routines
 *
 * A TINE server can be instructed to maintain a local history of certain properties.  The
 * stored data can then be obtained with normal data acquition calls to the TINE Local History
 * Server.  By and large the instructions as to which properties are to be maintained in the
 * local history server and as to what filtering criteria are to be used in commiting the data
 * to disk are typically entered into a startup configuration file 'history.csv'.  As an 
 * alternative, see AppendHistoryInformation().  The local history server maintains two sets
 * of histories, a short-term history kept in main memory and a long-term history kept on disk.
 * Whereas the short-term history can be kept on a millisecond time raster, the long-term history
 * will not be written to disk any more often than once per second.
 *
 */
#ifndef NO_DOXYGEN
#include "prolog.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include "servdef.h"
#include "errors.h"
#include "swaplib.h"
#include "database.h"
#include "toolkit.h"
#include "history.h"
#ifdef WINNT
# include "direct.h"
#endif
#ifdef UNIX
# include <sys/stat.h>
# include <dirent.h>
# ifndef FREEBSD
# include <alloca.h>
# endif
#endif
#ifdef VAX_VMS
# include <file.h>
# include <unixio.h>
#endif

#if defined(MSDOS) || defined(WINNT)
# define O_mode O_RDONLY|O_BINARY
#else
# define O_mode O_RDONLY
#endif

extern char FecDBpath[80];
char ArchiveLocation[128];
HistoryRecord *historyRecords = NULL;
int nHistoryRecords = 0;
EXPORT int maxHistoryRecords = MAX_HISTORY_RECORDS;
EXPORT int minHistoryPollingRate = 1000;
EXPORT int useMonthlyHistoryFiles = FALSE;
EXPORT int useHighResolutionTimeStamp = FALSE;

int makeDirectoryPath(char *dname)
{
# if defined(UNIX) || defined(WINNT)
# if defined(UNIX)
  DIR *dp;
  char *tok = "/";
# elif defined(WINNT)
  char fn[128];
  char *tok = "\\";
  struct _finddata_t fp;
# endif
  char *c=dname,*cp=NULL,tmpname[128],dirname[128];

  strncpy(tmpname,dname,128); dirname[0] = 0;
  while (*c++ == *tok) strcat(dirname,tok); /* in case it begins with '/' or '\\' */
  for (c=strtok_r(tmpname,tok,&cp); c != NULL; c=strtok_r(NULL,tok,&cp))
  {
    if (strlen(c) == 0) break; /* terminating delimiter gives the last token empty */
    strcat(dirname,c);
    strcat(dirname,tok);
#   if defined(UNIX)
    if ((dp=opendir(dirname)) != NULL)
    {
      closedir(dp);
    }
    else
    {
      if (mkdir(dirname,S_IWUSR|S_IRUSR|S_IXUSR|S_IWGRP|S_IRGRP|S_IXGRP|S_IROTH)) return 0;
    }
#   elif defined(WINNT)
    sprintf(fn,"%s*.*",dirname);
    if (_findfirst(fn, &fp) == -1)
    {
      if (mkdir(dirname)) return 0;
    }
#   endif
  }
# endif
  return -1;
}

#define MakeTargetFileName(nm,yy,mm,dd,id) (\
  sprintf((nm),"%s%02d%02d%02d.%0x",\
  useHighResolutionTimeStamp ? "ha" : "ar",((yy)%100),(mm),\
  useMonthlyHistoryFiles ? 1 : (dd),historyRecords[(id)].recordIndex)\
)
int removeOutdatedLTSFile(char *fn)
{
  int filesRemoved = 0;
# ifndef NO_LTS
  int i;
  char fileName[128],lfn[32];
  static int stagger = -1;

  if (stagger < 0) stagger = (int)time(NULL); /* seed the start up */
  /* stagger file removal to avoid blocking the file system */
  stagger = (stagger + 1) % 100;
  if (stagger > 0) return -1; 

  for (i=1; i<32; i++)
  {
    strncpy(lfn,fn,6);
    sprintf(&lfn[6],"%02d",i);
    strcat(lfn,&fn[8]);
    sprintf(fileName,"%s%s",ArchiveLocation,lfn);
    if (unlink(fileName) == 0) filesRemoved++;
  }
# endif /* NO_LTS */
  if (filesRemoved > 0) feclog("HST: removed %d files from long term storage",filesRemoved);
  return filesRemoved;
}
int WriteRecordToLTS(int i)  /* write record #i to file */
{
# ifndef NO_LTS
  int hFile,ptr;
  HistoryRecord *hst;
  char fileName[128];
  BYTE *tmpHistoryDataBuffer;
  int timestampLength = useHighResolutionTimeStamp ? sizeof(double) : sizeof(UINT32);
  SINT32 lTimeStamp;

  if (i < 0 || i >= nHistoryRecords) return invalid_index;
  hst = &historyRecords[i];
  ptr = hst->dataPtr;
  sprintf(fileName,"%s%s",ArchiveLocation,hst->ArchiveFileName);

# if defined(MSDOS)
  if ( (hFile=open(fileName,O_RDWR|O_CREAT|O_APPEND|O_BINARY)) == -1 )
# elif defined(WINNT)
  if ( (hFile=open(fileName,O_RDWR|O_CREAT|O_APPEND|O_BINARY,_S_IREAD|_S_IWRITE)) == -1 )
# elif defined(VAX_VMS)
  if ( (hFile=open(fileName,O_RDWR|O_CREAT|O_APPEND)) == -1 )
# else
  if ( (hFile=open(fileName,O_RDWR|O_CREAT|O_APPEND,SH_DENYNONE)) == -1 )
# endif
  {
    if (NGdebug) perror(fileName);
    return file_error;
  }
  lseek(hFile,0L,SEEK_END);

  /* write the timestamp and record as one unit to avoid corruption */
# ifdef INTEL
  tmpHistoryDataBuffer = (BYTE *)alloca(historyRecords[i].recordLength+timestampLength);
# else
  tmpHistoryDataBuffer = tmpWorkArea;
# endif
  if (useHighResolutionTimeStamp)
  {
    memcpy(tmpHistoryDataBuffer,(char *)&hst->timestamp[ptr],sizeof(double));
  }
  else
  {
    lTimeStamp = (SINT32)hst->timestamp[ptr];
    memcpy(tmpHistoryDataBuffer,(char *)&lTimeStamp,sizeof(UINT32));
  }
  memcpy(&tmpHistoryDataBuffer[timestampLength],hst->data + ptr * hst->recordLength,hst->recordLength);

  write(hFile,(char *)tmpHistoryDataBuffer,hst->recordLength+timestampLength);
  
  close(hFile);
  hst->lastArchiveTime = hst->timestamp[ptr];
# endif /* NO_LTS */
  return 0;
}
SINT32 locateRecordFromLTS(char *fileName,int i,double dt)
{ /* p0,d0 < p1,d1 < p2,d2 */
# ifndef NO_LTS
  SINT32 t,t0,t1,t2;
  double dt0, dt1, dt2;
  SINT32 nrecs,p0,p1,p2;
  double f;
  int hFile,recsize,it=0;
  int timestampLength = useHighResolutionTimeStamp ? sizeof(double) : sizeof(SINT32);
  HistoryRecord *hst;

  if (i < 0 || i >= nHistoryRecords) return -invalid_index;
  hst = &historyRecords[i];
  recsize = hst->recordLength + timestampLength;

# ifdef VXWORKS
  if ((hFile=open(fileName,O_mode,SH_DENYNONE)) == -1) return -file_error;
# else
  if ((hFile=open(fileName,O_mode)) == -1) return -file_error;
# endif
  t = (SINT32)dt;
  nrecs = lseek(hFile,0,SEEK_END)/recsize;
  p0 = p1 = 0; p2 = (nrecs-1)*recsize;
  lseek(hFile,p0,SEEK_SET);
  if (useHighResolutionTimeStamp)
  {
    read(hFile,(char *)&dt0,timestampLength);
    t0 = (SINT32)dt0;
  }
  else
  {
    read(hFile,(char *)&t0,timestampLength);
    dt0 = (double)t0;
  }
  if (dt < dt0) goto out;
  lseek(hFile,p2,SEEK_SET); 
  if (useHighResolutionTimeStamp)
  {
    read(hFile,(char *)&dt2,timestampLength);
    t2 = (SINT32)dt2;
  }
  else
  {
    read(hFile,(char *)&t2,timestampLength);
    dt2 = (double)t2;
  }
  while (p2 > p0 + recsize)
  {
    if (dt > (dt1=dt2)) { p1 = p2; break; }
    while (dt1 > dt0)   /* seek <--  */
    {
      f = ((dt-dt0)/(dt1-dt0))*((double)(p2-p0)/recsize);
      if ((p1=p0+recsize*(SINT32)f) >= p2) break;
	    lseek(hFile,p1,SEEK_SET);
      if (useHighResolutionTimeStamp)
      {
        read(hFile,(char *)&dt1,timestampLength);
        t1 = (SINT32)dt1;
      }
      else
      {
        read(hFile,(char *)&t1,timestampLength);
        dt1 = (double)t1;
      }
      if (dt1 < dt) break;
      p2 = p1;
      t2 = t1;
      dt2 = dt1;
      it++;
    }
    dt0 = dt1; t0 = t1; p0 = p1;
    while (dt2 > dt1)  /* seek --> */
    {
      f = ((dt2-dt)/(dt2-dt1))*((double)(p2-p0)/recsize);
      if ((p1=p2-recsize*(SINT32)f) <= p0) break;
	    lseek(hFile,p1,SEEK_SET); 
      if (useHighResolutionTimeStamp)
      {
        read(hFile,(char *)&dt1,timestampLength);
        t1 = (SINT32)dt1;
      }
      else
      {
        read(hFile,(char *)&t1,timestampLength);
        dt1 = (double)t1;
      }
      if (dt1 > dt) break;
      p0 = p1; 
      t0 = t1;
      dt0 = dt1;
      it++;
    }
    p2 = p1; t2 = t1; dt2 = dt1;
  }
out: 
  close(hFile);
  return p1;
# else
  return 0;
# endif /* NO_LTS */
}
int needToArchiveRecord(int i)
{
  double t, ts = makeDataTimeStamp();
  BYTE *d0,*d1;
  HistoryRecord *hst;
  int p,ptr,delta_t,doarchive = 0;

  if (i < 0 || i >= nHistoryRecords) return FALSE;
  hst = &historyRecords[i];
  if (hst->archiveRate < 0 ||
      hst->heartbeat < 0   ||
      hst->depthLong < 0) return FALSE;/* no long term storage */
  ptr = hst->dataPtr;

  if (hst->timestamp[ptr] <= hst->lastArchiveTime) return FALSE;/* data timestamp hasn't changed */
  if (!useHighResolutionTimeStamp && (SINT32)hst->timestamp[ptr] == (SINT32)hst->lastArchiveTime) return FALSE;
  if (hst->lastArchiveTime == 0) hst->lastArchiveTime = ts - hst->archiveRate/1000.0;
  delta_t = (int)((ts - hst->lastArchiveTime)*1000);
  if (hst->isScheduled) delta_t = hst->archiveRate * 2;
  if (delta_t < hst->archiveRate) return FALSE;
  if (delta_t >= hst->heartbeat*1000) return TRUE; /* archive no matter what */

  d0 = hst->data + ptr * hst->recordLength;
  d1 = ptr > 0 ? hst->data + (ptr-1) * hst->recordLength :
                 hst->data + (hst->depthShort-1) * hst->recordLength;

  switch (LFMT(hst->c.EqpFormatOut))
  {
    case CF_DOUBLE:
         for (p=0; !doarchive && p<(int)hst->c.EqpSizeOut; p++)
         {
           t = hst->aTolerance +
               hst->pTolerance * fabs(((double *)d1)[p]);
           if ( fabs(((double *)d0)[p] - ((double *)d1)[p]) > t) doarchive = 1;
         }
         break;
    case CF_FLOAT:
         for (p=0; !doarchive && p<(int)hst->c.EqpSizeOut; p++)
         {
           t = hst->aTolerance +
               hst->pTolerance * fabs(((float *)d1)[p]);
           if ( fabs(((float *)d0)[p] - ((float *)d1)[p]) > t) doarchive = 1;
         }
         break;
    case CF_INTEGER:
         for (p=0; !doarchive && p<(int)hst->c.EqpSizeOut; p++)
         {
           t = hst->aTolerance +
               hst->pTolerance * abs(((short *)d1)[p]);
           if ( abs(((short *)d0)[p] - ((short *)d1)[p]) > t) doarchive = 1;
         }
         break;
      case CF_LONG:
         for (p=0; !doarchive && p<(int)hst->c.EqpSizeOut; p++)
         {
           t = hst->aTolerance +
               hst->pTolerance * labs(((SINT32 *)d1)[p]);
           if ( labs(((SINT32 *)d0)[p] - ((SINT32 *)d1)[p]) > t) doarchive = 1;
         }
         break;
      case CF_BYTE:
         for (p=0; !doarchive && p<(int)hst->c.EqpSizeOut; p++)
         {
           t = hst->aTolerance +
               hst->pTolerance * abs(((BYTE *)d1)[p]);
           if ( abs(((BYTE *)d0)[p] - ((BYTE *)d1)[p]) > t) doarchive = 1;
         }
         break;
      default:
         if (!doarchive && memcmp(d0,d1,hst->recordLength)) doarchive = 1;
         break;
  }
  return doarchive ? TRUE : FALSE;
}
int isInHistoryTable(HistoryRecord *hst,int nentries)
{
  int i;
  for (i=0; i<nentries; i++)
  {
    if (strncmp(historyRecords[i].c.EqpName,hst->c.EqpName,EQP_NAME_SIZE)) continue;
    if (strncmp(historyRecords[i].c.EqpProperty,hst->c.EqpProperty,EQP_PROP_SIZE)) continue;
    if (strncmp(historyRecords[i].c.EqpDeviceName,hst->c.EqpDeviceName,EQP_DEVICE_SIZE)) continue;
    if (historyRecords[i].c.EqpSizeOut != hst->c.EqpSizeOut) continue;
    if (historyRecords[i].c.EqpFormatOut != hst->c.EqpFormatOut) continue;
    feclog("HIST: ignoring double entry [%s] %s %s",hst->c.EqpName,hst->c.EqpDeviceName,hst->c.EqpProperty);
    return TRUE;
  }
  return FALSE; 
}
int LoadHistoryInformation(char *fname,char *expName,char *eqpName) /* read history record info from file */
{
  FILE *fp = NULL;
  char fn[128],str[256],scratch[48];
  int done = 0,i,n,cc = 0;
  int name_col=-1,property_col=-1,device_col=-1,datalen_col=-1,fmt_col=-1,
      heartbeat_col=-1,poll_col=-1,rate_col=-1,tol_col=-1,recidx_col=-1,
      sdepth_col=-1,mdepth_col=-1,exp_col,rhs_col=-1;
  time_t t = time(NULL);
  struct tm *lt=localtime(&t);

  sprintf(fn,"%s%s",FecDBpath,fname);
  if (historyRecords == NULL)
  {
    if ((i=GetDataBaseSize(fn)) < 0) ccerr(-i);
    if ((n=GetNumberEquipmentModules()) == 0) n = 1;
    if (i > maxHistoryRecords) maxHistoryRecords = i*n;
    if ((historyRecords=(HistoryRecord *)SystemCalloc(maxHistoryRecords,sizeof(HistoryRecord))) == NULL)
      ccerr(out_of_local_memory);
  }
# if defined(MSDOS) || defined(WINNT)
  if ((fp=fopen(fn,"rt")) == NULL) ccerr(file_error);
# else
  if ((fp=fopen(fn,"r")) == NULL) ccerr(file_error);
# endif
  for (i=nHistoryRecords; fgets(str,255,fp) != NULL; )
  {
    strtrm(str);
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      if ((recidx_col=findcol(str,"INDEX")) < 0) ccerr(no_such_line);
      if ((name_col=findcol(str,"LOCAL NAME")) < 0) ccerr(no_such_line);
      if ((property_col=findcol(str,"PROPERTY")) < 0) ccerr(no_such_line);
      if ((device_col=findcol(str,"DEVICE")) < 0) ccerr(no_such_line);
      if ((datalen_col=findcol(str,"DATA LENGTH")) < 0) ccerr(no_such_line);
      if ((fmt_col=findcol(str,"FORMAT")) < 0) ccerr(no_such_line);
      if ((heartbeat_col=findcol(str,"HEARTBEAT")) < 0) ccerr(no_such_line);
      if ((poll_col=findcol(str,"POLLING RATE")) < 0) ccerr(no_such_line);
      if ((rate_col=findcol(str,"ARCHIVE RATE")) < 0) ccerr(no_such_line);
      if ((tol_col=findcol(str,"TOLERANCE")) < 0) ccerr(no_such_line);
      if ((sdepth_col=findcol(str,"SHORT DEPTH")) < 0) ccerr(no_such_line);
      if ((mdepth_col=findcol(str,"LONG DEPTH")) < 0) ccerr(no_such_line);
      exp_col = findcol(str,"EXPORT NAME");
      rhs_col = findcol(str,"REMOTE SERVER");
      done = 1;
      continue;
    }
    getColumnStr(name_col,str,historyRecords[i].c.EqpName,EQP_NAME_SIZE); 
    if (eqpName != NULL && strncmp(historyRecords[i].c.EqpName,eqpName,EQP_NAME_SIZE)) continue;
    getColumnStr(property_col,str,historyRecords[i].c.EqpProperty,EQP_PROP_SIZE); 
    getColumnStr(device_col,str,historyRecords[i].c.EqpDeviceName,EQP_DEVICE_SIZE); 
    getColumnStr(datalen_col,str,scratch,16); 
    historyRecords[i].c.EqpSizeOut = atoi(scratch);
    getColumnStr(fmt_col,str,scratch,16); 
    historyRecords[i].c.EqpFormatOut = BFMT(ftoi(scratch));
    if (isInHistoryTable(&historyRecords[i],i)) continue;
    if (exp_col >= 0 && expName != NULL)
    { /* database has an export name column => find a match or skip ! */
      getColumnStr(exp_col,str,scratch,16);
      if (strncmp(scratch,expName,16)) continue; /* didn't match ! */
    }
    getColumnStr(recidx_col,str,scratch,16); 
    historyRecords[i].recordIndex = atoi(scratch);
    getColumnStr(heartbeat_col,str,scratch,16); 
    historyRecords[i].heartbeat = atoi(scratch);
    getColumnStr(poll_col,str,scratch,16); 
    historyRecords[i].pollingRate = atoi(scratch);
    if (historyRecords[i].pollingRate < 20) historyRecords[i].pollingRate *= 1000; /* was thinking in seconds ! */
    if (historyRecords[i].pollingRate < minHistoryPollingRate) minHistoryPollingRate = historyRecords[i].pollingRate;
    getColumnStr(rate_col,str,scratch,16); 
    historyRecords[i].archiveRate = atoi(scratch);
    if (historyRecords[i].archiveRate < 20) historyRecords[i].archiveRate *= 1000; /* was thinking in seconds ! */
    getColumnStr(tol_col,str,scratch,16); 
    if (strchr(scratch,'%')) historyRecords[i].pTolerance = (float)(atof(scratch) / 100.0);
    else historyRecords[i].aTolerance = (float)atof(scratch);
    getColumnStr(sdepth_col,str,scratch,16); 
    historyRecords[i].depthShort = atoi(scratch);
    getColumnStr(mdepth_col,str,scratch,16); 
    historyRecords[i].depthLong = atoi(scratch);
    historyRecords[i].recordLength =
      (int)historyRecords[i].c.EqpSizeOut * fmtsizeof(historyRecords[i].c.EqpFormatOut);
    MakeTargetFileName(historyRecords[i].ArchiveFileName,lt->tm_year,lt->tm_mon+1,lt->tm_mday,i);
    if (rhs_col >= 0)
    {
      getColumnStr(rhs_col,str,scratch,48); 
      if (strlen(scratch) > 0)
      {
        if ((historyRecords[i].prd=(PrpRedirBlk *)SystemCalloc(1,sizeof(PrpRedirBlk))) == NULL) 
          ccerr(out_of_server_memory);
        ParseRedirector(scratch,historyRecords[i].prd);
      }
    }
    if ((historyRecords[i].timestamp=(double *)SystemCalloc(historyRecords[i].depthShort,sizeof(double))) == NULL)
      ccerr(out_of_local_memory);
    if ((historyRecords[i].data=(BYTE *)SystemCalloc(historyRecords[i].depthShort,historyRecords[i].recordLength)) == NULL)
      ccerr(out_of_local_memory);
    feclog("HIST: %s %s %s (%d s %d mon)",
            historyRecords[i].c.EqpName,
            historyRecords[i].c.EqpProperty,
            historyRecords[i].c.EqpDeviceName,
            historyRecords[i].depthShort,
            historyRecords[i].depthLong);
    if (i++ >= maxHistoryRecords) break;
  }
  nHistoryRecords = i;
  if (MaxPollingRate > minHistoryPollingRate)
  {
    SysPoll = minHistoryPollingRate;
    MaxPollingRate = minHistoryPollingRate;
    if (minHistoryPollingRate < MinPollingRate) MinPollingRate = minHistoryPollingRate;
  }
err:
  if (fp != NULL) fclose(fp);
  if (cc && cc != no_such_file)
  {
    feclog("HIST: %s : %s",fn,erlst[cc]);
  }
  return cc;
}
void historyInit(char *expName,char *eqpName)
{
  static int history_home_done = FALSE;
  static char lastExpName[16] = "";
  static char lastEqpName[16] = "";
  char *ptr, hstfn[64];
# ifdef FS_DELIMITER
  int i;
# endif

  if ((expName == NULL || !strncmp(expName,lastExpName,16)) &&
      (eqpName == NULL || !strncmp(eqpName,lastEqpName,16)))
    return;
  if (expName == NULL) lastExpName[0] = 0; else strncpy(lastExpName,expName,16);
  if (eqpName == NULL) lastEqpName[0] = 0; else strncpy(lastEqpName,eqpName,16);
  if (!history_home_done)
  {
    if ((ptr=getenv("HISTORY_HOME")) == NULL && (ptr=getenv("HISTORYDB")) == NULL) 
    {
      ptr = FecDBpath;
    }
    strncpy(ArchiveLocation,ptr,80);
  # ifdef FS_DELIMITER
    if (((i=(int)strlen(ArchiveLocation)) > 0) &&
        (ArchiveLocation[i-1] != FS_DELIMITER))  /* no delimiter ! */
    {
      ArchiveLocation[i] = FS_DELIMITER; ArchiveLocation[i+1] = 0;
    }
  # endif
    makeDirectoryPath(ArchiveLocation);
    feclog("HISTORY HOME : %s",ArchiveLocation);
    history_home_done = TRUE;
  }
  strncpy(hstfn,HISTORY_FILENAME,64);
  if (strlen(gFecDBSubDir) > 0) sprintf(hstfn,"%s%s",gFecDBSubDir,HISTORY_FILENAME);
  LoadHistoryInformation(hstfn,expName,eqpName);
}
void historyCycle(void)
{
  int i,n,p,cc;
  ExportListStruct *el;
  char sDevName[16],rmvfn[16];
  struct tm *tm;
  int yy,mm,delta_t;
  static struct timeval lastCycleTime = {0, 0};
  static time_t lasttime = 0;
  static DTYPE din,dout;
  static int mday = -1;
  static int recordToRemove = 0;

  if (gEqpFcnBusySemaphore) return;
  if (lasttime == 0) memset(&din,0,sizeof(DTYPE));
  if (MSECS(thisCycleTime,lastCycleTime) < (int)(minHistoryPollingRate-gSystemTick)) return;
  lastCycleTime = thisCycleTime;
  lasttime = thisCycleTime.tv_sec;
  tm = localtime(&lasttime);
  for (i=0; i<nHistoryRecords; i++)
  {
    if ((el=getExportListItem(historyRecords[i].c.EqpName)) == NULL) continue;
    if (el->EqpFcn == NULL) continue;
    if (historyRecords[i].depthShort <= 0) continue;
    if (mday != tm->tm_mday)
    { /* it's the dawning of a new day (sort of) */
      MakeTargetFileName(historyRecords[i].ArchiveFileName,tm->tm_year,tm->tm_mon+1,tm->tm_mday,i);
      if (recordToRemove >= i && recordToRemove < nHistoryRecords)
      { /* careful of module arithmetic here. uses months from 0 to 11 */
        mm = ((tm->tm_mon+12) - ((historyRecords[recordToRemove].depthLong+1)%12))%12;
        yy = tm->tm_year - historyRecords[recordToRemove].depthLong/12;
        if (mm > tm->tm_mon) yy--;
        MakeTargetFileName(rmvfn,yy,mm+1,0,recordToRemove);
        if (removeOutdatedLTSFile(rmvfn) >= 0) recordToRemove++;
      }
    }
    gDataTimeStamp = putDataTimeStamp(gDataTimeStampOffset,thisCycleTime.tv_sec,(int)(thisCycleTime.tv_usec/1000));
    n = historyRecords[i].dataPtr;
    if (historyRecords[i].timestamp[n] == 0) 
      historyRecords[i].timestamp[n] = gDataTimeStamp - (double)historyRecords[i].pollingRate/1000.0;
    delta_t = (int)((gDataTimeStamp - historyRecords[i].timestamp[n])*1000);
    if (historyRecords[i].isScheduled) delta_t = (int)historyRecords[i].pollingRate * 2;
    if (delta_t < (int)(historyRecords[i].pollingRate-gSystemTick)) continue;
    n = (historyRecords[i].dataPtr + 1) % historyRecords[i].depthShort;
    dout.dArrayLength = historyRecords[i].c.EqpSizeOut;
    dout.dFormat = LFMT(historyRecords[i].c.EqpFormatOut);
    dout.data.vptr = historyRecords[i].data + n*historyRecords[i].recordLength;
    strncpy(sDevName,historyRecords[i].c.EqpDeviceName,EQP_DEVICE_SIZE);
    gEqpFcnBusySemaphore = 1;
    gEqpFcnAccessor = CA_HIST;
    cc = (*el->EqpFcn)(sDevName,historyRecords[i].c.EqpProperty,&dout,&din,CA_READ|CA_HIST);
    gEqpFcnAccessor = 0;
    gEqpFcnBusySemaphore = 0;
    if (cc != 0)
    {
      if (NGdebug)
      {
        dbglog("HIST: (%d) %s %s : <%d> %s",
                i,sDevName,historyRecords[i].c.EqpProperty,
                cc,cc > 0 && cc < numErr ? erlst[cc] : "undefined");
      }
      historyRecords[i].isScheduled = FALSE;
      continue;
    }
    p = historyRecords[i].dataPtr;
    delta_t = (int)((gDataTimeStamp - historyRecords[i].timestamp[p])*1000);
    if (historyRecords[i].isScheduled) delta_t = (int)historyRecords[i].pollingRate * 2;
    if (delta_t < (int)(historyRecords[i].pollingRate-gSystemTick)) continue;
    historyRecords[i].timestamp[n] = gDataTimeStamp;
    historyRecords[i].dataPtr = n;
    if (needToArchiveRecord(i)) WriteRecordToLTS(i);
    historyRecords[i].isScheduled = FALSE;
  }
  if (recordToRemove >= nHistoryRecords) 
  { /* finished removing old files */
    mday = tm->tm_mday;
    recordToRemove = 0;
  }
  return;
}
int getLastLTSData(int i,int index,void *datbuf,int datfmt,int bufsize)
{
# ifndef NO_LTS
  int hFile = -1,number,cc = 0;
  SINT32 pstart,rlen;
  char fn[16],fileName[128];
  DTYPE d;
  int fmtsize;
  int timestampLength = useHighResolutionTimeStamp ? sizeof(double) : sizeof(SINT32);
  time_t timestamp,starttime = time(NULL);
  double dtimestamp, datatime = 0;
  struct tm *lt=localtime((time_t *)&starttime);
  BYTE *tmpHistoryDataBuffer;

  if (bufsize < 1) return buffer_too_small;
  if (historyRecords == NULL) return un_allocated;
  if (i < 0 || i >= nHistoryRecords) return not_defined;
  if (historyRecords[i].depthLong < 0) return invalid_data;
  if (historyRecords[i].recordLength > WORK_SIZE) goto err;

  MakeTargetFileName(fn,lt->tm_year,lt->tm_mon+1,lt->tm_mday,i);
  sprintf(fileName,"%s%s",ArchiveLocation,fn);

  rlen = historyRecords[i].recordLength + timestampLength;
# ifdef VXWORKS
  if ((hFile=open(fileName,O_mode,SH_DENYNONE)) == -1) ccerr(file_error);
# else
  if ((hFile=open(fileName,O_mode)) == -1) ccerr(file_error);
# endif
  datatime = 0;
  pstart = lseek(hFile,0,SEEK_END);
  pstart -= rlen;

  if (lseek(hFile,pstart,SEEK_SET) < 0) ccerr(file_error);
# ifdef INTEL
  tmpHistoryDataBuffer = (BYTE *)alloca(historyRecords[i].recordLength);
# else
  tmpHistoryDataBuffer = tmpWorkArea;
# endif
  fmtsize = fmtsizeof(historyRecords[i].c.EqpFormatOut);
  d.dFormat = LFMT(historyRecords[i].c.EqpFormatOut);
  d.dArrayLength = 1;
  d.data.bptr = tmpHistoryDataBuffer + index*fmtsize;
  if (useHighResolutionTimeStamp)
  {
    if (read(hFile,(char *)&dtimestamp,timestampLength) != timestampLength) ccerr(file_error);
  }
  else
  {
    if (read(hFile,(char *)&timestamp,timestampLength) != timestampLength) ccerr(file_error);
    dtimestamp = (double)timestamp;
  }
  if (read(hFile,(char *)tmpHistoryDataBuffer,historyRecords[i].recordLength)
      != historyRecords[i].recordLength) ccerr(file_error);
  datatime = dtimestamp;

  switch (datfmt)
  {
    case CF_FLOAT:
      d.dArrayLength = MIN(bufsize,(int)historyRecords[i].c.EqpSizeOut);
      number = (int)d.dArrayLength;
      getValuesAsFloat(&d,(float *)datbuf,(short)number);
      break;
    case CF_LONG:
      d.dArrayLength = MIN(bufsize,(int)historyRecords[i].c.EqpSizeOut);
      number = (int)d.dArrayLength;
      getValuesAsLong(&d,(UINT32 *)datbuf,(short)number);
      break;
    case CF_DOUBLE:
      d.dArrayLength = MIN(bufsize,(int)historyRecords[i].c.EqpSizeOut);
      number = (int)d.dArrayLength;
      getValuesAsDouble(&d,(double *)datbuf,(short)number);
      break;
    default: goto err;
  }
err:
  if (hFile != -1) close(hFile);
  if (datatime > 0) setDataTimeStamp(datatime);
  return cc;
# else
  return 0;
# endif /* NO_LTS */
}
int loadLTSData(int i,int index,double dstarttime,double dendtime,
                 float lwrLimit, float uprLimit,
                 void *datbuf,int datfmt,int bufsize,int sample)
{
# ifndef NO_LTS
  int hFile = -1,nrecs,raster = 1,isSnapshot = 0;
  time_t starttime = (time_t)dstarttime;
  SINT32 pstart,pstop,cnt = 0,rlen;
  char fn[16],fileName[128];
  DTYPE d;
  int number = 0,fmtsize;
  int timestampLength = useHighResolutionTimeStamp ? sizeof(double) : sizeof(SINT32);
  time_t timestamp;
  double dtimestamp, datatime = 0, dtnow = makeDataTimeStamp();
  static double timebin = -1;
  struct tm *lt=localtime((time_t *)&starttime);
  BYTE *tmpHistoryDataBuffer, tsBuffer[8];
  UINT32 strsiz;

  if (timebin < 0) timebin = useHighResolutionTimeStamp ? 0.01 : 1.0;
  if (bufsize < 1) return 0;
  if (historyRecords == NULL) return FALSE;
  if (i < 0 || i >= nHistoryRecords) return FALSE;
  if (historyRecords[i].depthLong < 0) return FALSE;
  if ((dtnow - dstarttime) < historyRecords[i].depthShort) return FALSE;
  if (historyRecords[i].recordLength > WORK_SIZE) goto err;
  if (index < 0 || index >= (int)historyRecords[i].c.EqpSizeOut) goto err;
  if (((dendtime-dstarttime) < 0.01) && datfmt <= CF_FLOAT && bufsize > 1) isSnapshot = TRUE;
  if (isSnapshot) 
  {
    index = 0; /* snapshot requests always have index = 0 */
    datatime = dstarttime - timebin;
  }

  MakeTargetFileName(fn,lt->tm_year,lt->tm_mon+1,lt->tm_mday,i);
  sprintf(fileName,"%s%s",ArchiveLocation,fn);

  if (dendtime < dstarttime) dendtime = dtnow;
  if ((pstart=locateRecordFromLTS(fileName,i,dstarttime)) < 0) goto err;
  if ((pstop=locateRecordFromLTS(fileName,i,dendtime)) < 0) goto err;
  rlen = historyRecords[i].recordLength + timestampLength;
  nrecs = (int)((pstop-pstart)/rlen);
  if (bufsize == 1) 
  {
    setDataTimeStamp(dendtime);
    return MAX(nrecs,0);
  }
  if (bufsize < nrecs) raster += nrecs/(bufsize+1);
  if (sample > 0) raster = sample;

# ifdef VXWORKS
  if ((hFile=open(fileName,O_mode,SH_DENYNONE)) == -1) return 0;
# else
  if ((hFile=open(fileName,O_mode)) == -1) goto err;
# endif
  number = 0; cnt = 0; 
  lseek(hFile,pstart,SEEK_SET);

# ifdef INTEL
  tmpHistoryDataBuffer = (BYTE *)alloca(historyRecords[i].recordLength);
# else
  tmpHistoryDataBuffer = tmpWorkArea;
# endif
  fmtsize = fmtsizeof(historyRecords[i].c.EqpFormatOut);

  d.dFormat = LFMT(historyRecords[i].c.EqpFormatOut);
  d.dArrayLength = 1;
  d.data.bptr = tmpHistoryDataBuffer + index*fmtsize;

  while(read(hFile,(char *)tsBuffer,timestampLength) == timestampLength)   /* read through file */
  {
    if (useHighResolutionTimeStamp) 
    {
      memcpy(&dtimestamp,tsBuffer,timestampLength); 
      timestamp = (time_t)dtimestamp;
    }
    else 
    {
      memcpy(&timestamp,tsBuffer,timestampLength);
      dtimestamp = (double)timestamp;
    }
    cnt++;
    if (isSnapshot && dtimestamp > dendtime) dendtime = dtimestamp;
    if (dtimestamp > dendtime) goto err;
    if (dtimestamp < datatime + timebin)
    {
      if (lseek(hFile,historyRecords[i].recordLength,SEEK_CUR) < 0) goto err;
      continue;
    }
    if (read(hFile,(char *)tmpHistoryDataBuffer,historyRecords[i].recordLength)
        != historyRecords[i].recordLength) goto err;

    datatime = dtimestamp;

    switch (datfmt)
    {
      case CF_FLOAT:
        d.dArrayLength = MIN(bufsize,(int)historyRecords[i].c.EqpSizeOut);
        number = (int)d.dArrayLength;
        getValuesAsFloat(&d,(float *)datbuf,(short)number);
        goto err;
      case CF_FLTINT:
        ((FLTINT *)datbuf)[number].ival = (UINT32)timestamp;
        getValuesAsFloat(&d,&((FLTINT *)datbuf)[number].fval,1);
        if (sample < 0 && !isInRange((BYTE *)datbuf,datfmt,number,lwrLimit,uprLimit)) continue;
        break;
      case CF_INTFLTINT: /* CF_TDS */
        ((INTFLTINT *)datbuf)[number].i1val = (UINT32)timestamp;
        getValuesAsFloat(&d,&((INTFLTINT *)datbuf)[number].fval,1);
        ((INTFLTINT *)datbuf)[number].i2val = 0;
        if (sample < 0 && !isInRange((BYTE *)datbuf,datfmt,number,lwrLimit,uprLimit)) continue;
        break;
      case CF_FLTINTINT:
        getValuesAsFloat(&d,&((FLTINTINT *)datbuf)[number].fval,1);
        ((FLTINTINT *)datbuf)[number].i1val = (UINT32)timestamp;
        ((FLTINTINT *)datbuf)[number].i2val = 0;
        if (sample < 0 && !isInRange((BYTE *)datbuf,datfmt,number,lwrLimit,uprLimit)) continue;
        break;
      case CF_LONG:
        d.dArrayLength = MIN(bufsize,(int)historyRecords[i].c.EqpSizeOut);
        number = (int)d.dArrayLength;
        getValuesAsLong(&d,(UINT32 *)datbuf,(short)number);
        goto err;
      case CF_LNGINT:
        ((INTINT *)datbuf)[number].i2val = (UINT32)timestamp;
        getValuesAsLong(&d,&((INTINT *)datbuf)[number].i1val,1);
        if (sample < 0 && !isInRange((BYTE *)datbuf,datfmt,number,lwrLimit,uprLimit)) continue;
        break;
      case CF_DOUBLE:
        d.dArrayLength = MIN(bufsize,(int)historyRecords[i].c.EqpSizeOut);
        number = (int)d.dArrayLength;
        getValuesAsDouble(&d,(double *)datbuf,(short)number);
        goto err;
      case CF_DBLDBL:
        ((DBLDBL *)datbuf)[number].d2val = dtimestamp;
        getValuesAsDouble(&d,&((DBLDBL *)datbuf)[number].d1val,1);
        if (sample < 0 && !isInRange((BYTE *)datbuf,datfmt,number,lwrLimit,uprLimit)) continue;
        break;
      case CF_NAME16I:
        ((NAME16I *)datbuf)[number].ival = (UINT32)timestamp; strsiz = 16;
        getValuesAsString(&d,((NAME16I *)datbuf)[number].name,&strsiz);
        break;
      case CF_NAME32I:
        ((NAME32I *)datbuf)[number].ival = (UINT32)timestamp; strsiz = 32;
        getValuesAsString(&d,((NAME32I *)datbuf)[number].name,&strsiz);
        break;
      case CF_NAME48I:
        ((NAME48I *)datbuf)[number].ival = (UINT32)timestamp; strsiz = 48;
        getValuesAsString(&d,((NAME48I *)datbuf)[number].name,&strsiz);
        break;
      case CF_NAME64I:
        ((NAME64I *)datbuf)[number].ival = (UINT32)timestamp; strsiz = 64;
        getValuesAsString(&d,((NAME64I *)datbuf)[number].name,&strsiz);
        break;
      case CF_SHORT:
        d.dArrayLength = MIN(bufsize,(int)historyRecords[i].c.EqpSizeOut);
        number = (int)d.dArrayLength;
        getValuesAsShort(&d,(short *)datbuf,(short)number);
        goto err;
      default: goto err;
    }
    number++;
    if (number >= bufsize) goto err;
    if (raster > 1) /* move along through file */
    {
      if (lseek(hFile,rlen*(raster-1),SEEK_CUR) < 0) goto err;
    }
  }
err:
  if (hFile != -1) close(hFile);

  if (datatime > 0) setDataTimeStamp(datatime);
  return number;
# else
  return 0;
# endif /* NO_LTS */
}
int IsLiteralDevice(char *eqp,char *prp,char *dev)
{
  int i;
  for (i=0; i<nHistoryRecords; i++)
  {
    if (strncmp(historyRecords[i].c.EqpName,eqp,EQP_NAME_SIZE)) continue;
    if (strncmp(historyRecords[i].c.EqpProperty,prp,EQP_PROP_SIZE)) continue;
    if (strncmp(historyRecords[i].c.EqpDeviceName,dev,EQP_DEVICE_SIZE)) continue;
    break;
  }
  return nHistoryRecords > i ? TRUE: FALSE;
}
EXPORT int setSTSData(char *eqp,char *prp, char *dev,int index,BYTE *data,int fmt,int arraysize)
{
  int i,n,fmtsiz,sttptr=0,literalDevice;
  BYTE *dptr=NULL;

  if (nHistoryRecords <= 0 || historyRecords == NULL) return -un_allocated;
  literalDevice = IsLiteralDevice(eqp,prp,dev);
  for (i=0; i<nHistoryRecords; i++)
  {
    if (strncmp(historyRecords[i].c.EqpName,eqp,EQP_NAME_SIZE)) continue;
    if (strncmp(historyRecords[i].c.EqpProperty,prp,EQP_PROP_SIZE)) continue;
    if (strncmp(historyRecords[i].c.EqpDeviceName,dev,EQP_DEVICE_SIZE))
    {
      if (index == 0 && !literalDevice) index = GetDeviceNumberEx(eqp,dev,prp);
      else continue;
    }
    break;
  }
  if (i == nHistoryRecords) return -un_allocated;
  if (historyRecords[i].depthShort == 0) return -un_allocated;
  if (index < 0) return index;
  if (index >= (int)historyRecords[i].c.EqpSizeOut) return -out_of_range;
  if (arraysize < 1) return -buffer_too_small;
  fmtsiz = fmtsizeof(historyRecords[i].c.EqpFormatOut);
  if (arraysize > historyRecords[i].depthShort) 
  {
    sttptr = arraysize - historyRecords[i].depthShort;
    arraysize = historyRecords[i].depthShort;
  }

  for (n=0; n < arraysize && n <= historyRecords[i].depthShort; n++)
  {
    dptr = historyRecords[i].data + index*fmtsiz + n*historyRecords[i].recordLength;
    switch (fmt)
    {
      case CF_FLTINT: 
        if (LFMT(historyRecords[i].c.EqpFormatOut) != CF_FLOAT) return -illegal_format;
        *((float *)dptr) = ((FLTINT *)data)[n+sttptr].fval;
        historyRecords[i].timestamp[n] = (double)((FLTINT *)data)[n+sttptr].ival;
        break;
      case CF_DBLINT:
        if (LFMT(historyRecords[i].c.EqpFormatOut) != CF_DOUBLE) return -illegal_format;
        *((double *)dptr) = ((DBLINT *)data)[n+sttptr].dval;
        historyRecords[i].timestamp[n] = (double)((DBLINT *)data)[n+sttptr].ival;
        break;
      case CF_LNGINT: 
        if (LFMT(historyRecords[i].c.EqpFormatOut) != CF_LONG) return -illegal_format;
        *((UINT32 *)dptr) = ((LNGINT *)data)[n+sttptr].lval;
        historyRecords[i].timestamp[n] = (double)((LNGINT *)data)[n+sttptr].ival;
        break;
      case CF_DBLDBL:
        if (LFMT(historyRecords[i].c.EqpFormatOut) != CF_DOUBLE) return -illegal_format;
        *((double *)dptr) = ((DBLDBL *)data)[n+sttptr].d1val;
        historyRecords[i].timestamp[n] = ((DBLDBL *)data)[n+sttptr].d2val;
        break;
      case CF_NAME16I: 
        if (LFMT(historyRecords[i].c.EqpFormatOut) != CF_NAME16) return -illegal_format;
        strncpy(((NAME16 *)dptr)->name,((NAME16I *)data)[n+sttptr].name,16);
        historyRecords[i].timestamp[n] = (double)((LNGINT *)data)[n+sttptr].ival;
        break;
      case CF_NAME32I: 
        if (LFMT(historyRecords[i].c.EqpFormatOut) != CF_NAME32) return -illegal_format;
        strncpy(((NAME32 *)dptr)->name,((NAME32I *)data)[n+sttptr].name,32);
        historyRecords[i].timestamp[n] = (double)((LNGINT *)data)[n+sttptr].ival;
        break;
      case CF_NAME48I: 
        if (LFMT(historyRecords[i].c.EqpFormatOut) != CF_NAME48) return -illegal_format;
        strncpy(((NAME48 *)dptr)->name,((NAME48I *)data)[n+sttptr].name,48);
        historyRecords[i].timestamp[n] = (double)((LNGINT *)data)[n+sttptr].ival;
        break;
      case CF_NAME64I: 
        if (LFMT(historyRecords[i].c.EqpFormatOut) != CF_NAME64) return -illegal_format;
        strncpy(((NAME64 *)dptr)->name,((NAME64I *)data)[n+sttptr].name,64);
        historyRecords[i].timestamp[n] = (double)((LNGINT *)data)[n+sttptr].ival;
        break;
    }
  }    
  historyRecords[i].dataPtr = (arraysize-1) % historyRecords[i].depthShort;
  return n;
}
int getSTSData(char *eqp,char *prp, char *dev,int index,
		 double dstarttime,double dendtime,
     float lwrLimit, float uprLimit,
     BYTE *data,int fmt,int arraysize,int sample,int recsreadFlag)
{
  int i,n,number,rnumber=0,num,raster,cnt,fmtsiz,literalDevice;
  time_t timestamp;
  int isSnapshot=0;
  int sttidx = recsreadFlag ? 1 : 0;
  double dtimestamp,lastreadtimestamp=0;
  DTYPE d;
  UINT32 strsiz;

  if (nHistoryRecords <= 0 || historyRecords == NULL) return -un_allocated;
  literalDevice = IsLiteralDevice(eqp,prp,dev);
  for (i=0; i<nHistoryRecords; i++)
  {
    if (strncmp(historyRecords[i].c.EqpName,eqp,EQP_NAME_SIZE)) continue;
    if (strncmp(historyRecords[i].c.EqpProperty,prp,EQP_PROP_SIZE)) continue;
    if (strncmp(historyRecords[i].c.EqpDeviceName,dev,EQP_DEVICE_SIZE))
    {
      if (index == 0 && !literalDevice) index = GetDeviceNumberEx(eqp,dev,prp);
      else continue;
    }
    break;
  }
  if (i == nHistoryRecords) return -un_allocated;
  if (historyRecords[i].depthShort == 0) return -getRemoteHistoryServer(eqp,prp);
  if (index < 0) return index;
  if (index >= (int)historyRecords[i].c.EqpSizeOut) return -out_of_range;
  if (arraysize < 1) return -buffer_too_small;
  if (dstarttime < dendtime - (double)historyRecords[i].depthShort) 
    dstarttime = dendtime - (double)historyRecords[i].depthShort;

  d.dFormat = LFMT(historyRecords[i].c.EqpFormatOut);
  d.dArrayLength = 1;
  if ((raster=(int)((dendtime-dstarttime)/(double)arraysize)) < 1) raster = 1;
  if (arraysize == 1) 
  {
    raster = 1;
    sttidx = 0;
  }
  isSnapshot = ((dendtime-dstarttime) < .01) ? TRUE : FALSE;
  if (isSnapshot && arraysize > 1) index = 0; /* snapshot requests always have index = 0 */
  fmtsiz = fmtsizeof(d.dFormat);
  n = historyRecords[i].dataPtr;
  /* move raster + sttidx past the current position to start at the beginning */
  n = (n + raster + sttidx + historyRecords[i].depthShort) % historyRecords[i].depthShort;
  if (arraysize == (int)(dendtime-dstarttime)) /* caller wants the array filled */
  {
    if (arraysize <= (num=historyRecords[i].depthShort))
    {
      n = (n + (num-arraysize) + historyRecords[i].depthShort) % historyRecords[i].depthShort;
      dstarttime = historyRecords[i].timestamp[n];
      dendtime = makeDataTimeStamp();
    }
  }
  lastreadtimestamp = dstarttime;
  for (number=sttidx,cnt=0; number < arraysize &&
                            number <= historyRecords[i].depthShort &&
                            cnt < historyRecords[i].depthShort; cnt += raster)
  {
    d.data.vptr = historyRecords[i].data + index*fmtsiz + n*historyRecords[i].recordLength;
    dtimestamp = historyRecords[i].timestamp[n];
    n = (n + raster + historyRecords[i].depthShort) % historyRecords[i].depthShort;
    if (dtimestamp < lastreadtimestamp || dtimestamp > dendtime) continue;
    lastreadtimestamp = dtimestamp;
    timestamp = (time_t)dtimestamp;
    rnumber++;
    if (arraysize < 2) continue;
    switch (fmt)
    {
      case CF_FLOAT:
      	if (isSnapshot)
        {
          d.dArrayLength = MIN(arraysize,(int)historyRecords[i].c.EqpSizeOut);
          rnumber = number = (int)d.dArrayLength;
          getValuesAsFloat(&d,(float *)data,(short)number);
          goto err;
        }
        getValuesAsFloat(&d,&((float *)data)[number],1);
        break;
      case CF_FLTINT:
        ((FLTINT *)data)[number].ival = (UINT32)timestamp;
        getValuesAsFloat(&d,&((FLTINT *)data)[number].fval,1);
        if (sample < 0 && !isInRange(data,fmt,number,lwrLimit,uprLimit)) continue;
        break;
      case CF_INTFLTINT: /* CF_TDS */
        ((INTFLTINT *)data)[number].i1val = (UINT32)timestamp;
        getValuesAsFloat(&d,&((INTFLTINT *)data)[number].fval,1);
        ((INTFLTINT *)data)[number].i2val = 0;
        if (sample < 0 && !isInRange(data,fmt,number,lwrLimit,uprLimit)) continue;
        break;
      case CF_FLTINTINT: 
        getValuesAsFloat(&d,&((FLTINTINT *)data)[number].fval,1);
        ((FLTINTINT *)data)[number].i1val = (UINT32)timestamp;
        ((FLTINTINT *)data)[number].i2val = (UINT32)((dtimestamp-(double)timestamp)*1000);
        if (sample < 0 && !isInRange(data,fmt,number,lwrLimit,uprLimit)) continue;
        break;
      case CF_LONG:
      	if (isSnapshot)
        {
          d.dArrayLength = MIN(arraysize,(int)historyRecords[i].c.EqpSizeOut);
          rnumber = number = (int)d.dArrayLength;
          getValuesAsLong(&d,(UINT32 *)data,(short)number);
          goto err;
        }
        getValuesAsLong(&d,&((UINT32 *)data)[number],1);
        break;
      case CF_LNGINT:
        ((LNGINT *)data)[number].ival = (UINT32)timestamp;
        getValuesAsLong(&d,&((LNGINT *)data)[number].lval,1);
        if (sample < 0 && !isInRange(data,fmt,number,lwrLimit,uprLimit)) continue;
        break;
      case CF_DOUBLE:
      	if (isSnapshot)
        {
          d.dArrayLength = MIN(arraysize,(int)historyRecords[i].c.EqpSizeOut);
          rnumber = number = (int)d.dArrayLength;
          getValuesAsDouble(&d,(double *)data,(short)number);
          goto err;
        }
        getValuesAsDouble(&d,&((double *)data)[number],1);
        break;
      case CF_DBLDBL:
        ((DBLDBL *)data)[number].d2val = dtimestamp;
        getValuesAsDouble(&d,&((DBLDBL *)data)[number].d1val,1);
        if (sample < 0 && !isInRange(data,fmt,number,lwrLimit,uprLimit)) continue;
        break;
      case CF_NAME16I:
        ((NAME16I *)data)[number].ival = (UINT32)timestamp; strsiz = 16;
        getValuesAsString(&d,((NAME16I *)data)[number].name,&strsiz);
        break;
      case CF_NAME32I:
        ((NAME32I *)data)[number].ival = (UINT32)timestamp; strsiz = 32;
        getValuesAsString(&d,((NAME32I *)data)[number].name,&strsiz);
        break;
      case CF_NAME48I:
        ((NAME48I *)data)[number].ival = (UINT32)timestamp; strsiz = 48;
        getValuesAsString(&d,((NAME48I *)data)[number].name,&strsiz);
        break;
      case CF_NAME64I:
        ((NAME64I *)data)[number].ival = (UINT32)timestamp; strsiz = 64;
        getValuesAsString(&d,((NAME64I *)data)[number].name,&strsiz);
        break;
      case CF_SHORT:
      	if (isSnapshot)
        {
          d.dArrayLength = MIN(arraysize,(int)historyRecords[i].c.EqpSizeOut);
          rnumber = number = (int)d.dArrayLength;
          getValuesAsShort(&d,(short *)data,(short)number);
          goto err;
        }        
        return -illegal_format;
      default: goto err;
		}
    number++;
  }

  assignNumberOfPoints(data,fmt,arraysize,&number,&rnumber,recsreadFlag);

err:
  if (lastreadtimestamp > 0) setDataTimeStamp((double)lastreadtimestamp);
  return arraysize < 2 ? rnumber : number;
}
int isInRange(BYTE *data,int fmt,int number,float lwrLimit,float uprLimit)
{
  float fval;
  switch (fmt)
  {
    case CF_FLTINT:
      fval = ((FLTINT *)data)[number].fval;
      if (fval < lwrLimit || fval > uprLimit) 
      {
        ((FLTINT *)data)[number].ival = 0; /* flag the boundary */
        if (number > 0 && ((FLTINT *)data)[number-1].ival == 0) return 0;
      }
      return -1;
    case CF_FLTINTINT:
      fval = ((FLTINTINT *)data)[number].fval;
      if (fval < lwrLimit || fval > uprLimit) 
      {
        ((FLTINTINT *)data)[number].i1val = 0; /* flag the boundary */
        ((FLTINTINT *)data)[number].i2val = 0; /* flag the boundary */
        if (number > 0 && ((FLTINTINT *)data)[number-1].i1val == 0) return 0;
      }
      return -1;
    case CF_INTINT:
      fval = (float)((INTINT *)data)[number].i1val;
      if (fval < lwrLimit || fval > uprLimit) 
      {
        ((INTINT *)data)[number].i2val = 0; /* flag the boundary */
        if (number > 0 && ((INTINT *)data)[number-1].i2val == 0) return 0;
      }
      return -1;
    case CF_DBLDBL:
      fval = (float)((DBLDBL *)data)[number].d1val;
      if (fval < lwrLimit || fval > uprLimit) 
      {
        ((DBLDBL *)data)[number].d2val = 0; /* flag the boundary */
        if (number > 0 && ((DBLDBL *)data)[number-1].d2val == 0) return 0;
      }
      return -1;
    case CF_INTFLTINT:
      fval = (float)((INTFLTINT *)data)[number].fval;
      if (fval < lwrLimit || fval > uprLimit) 
      {
        ((INTFLTINT *)data)[number].i1val = 0; /* flag the boundary */
        if (number > 0 && ((INTFLTINT *)data)[number-1].i1val == 0) return 0;
      }
      return -1;
  }
  return -1;
}
void assignNumberOfPoints(BYTE *data,int fmt,int siz,int *npoints,int *nfound,int recsreadFlag)
{
  int padding = recsreadFlag ? 1 : 0;
  if (data == NULL || npoints == NULL || nfound == NULL) return;

  if (siz == 1) *npoints = *nfound + padding;
  else if (!recsreadFlag) return;

  switch (fmt)
  {
    case CF_FLTINT:
      if (padding) (*npoints)--;
      ((FLTINT *)data)[0].ival = (UINT32)(*npoints);
      ((FLTINT *)data)[0].fval = (float)(*npoints);
      break;
    case CF_FLTINTINT:
      if (padding) (*npoints)--;
      ((FLTINTINT *)data)[0].i1val = (UINT32)(*npoints);
      ((FLTINTINT *)data)[0].i2val = (UINT32)(*npoints);
      ((FLTINTINT *)data)[0].fval = (float)(*npoints);
      break;
    case CF_INTINT:
      if (padding) (*npoints)--;
      ((INTINT *)data)[0].i1val = (UINT32)(*npoints);
      ((INTINT *)data)[0].i2val = (UINT32)(*npoints);
      break;
    case CF_DBLDBL:
      if (padding) (*npoints)--;
      ((DBLDBL *)data)[0].d1val = (double)(*npoints);
      ((DBLDBL *)data)[0].d2val = (double)(*npoints);
      break;
    case CF_NAME16I:
      if (padding) (*npoints)--;
      ((NAME16I *)data)[0].ival = (UINT32)(*npoints);
      break;
    case CF_NAME32I:
      if (padding) (*npoints)--;
      ((NAME32I *)data)[0].ival = (UINT32)(*npoints);
      break;
    case CF_NAME48I:
      if (padding) (*npoints)--;
      ((NAME48I *)data)[0].ival = (UINT32)(*npoints);
      break;
    case CF_NAME64I:
      if (padding) (*npoints)--;
      ((NAME64I *)data)[0].ival = (UINT32)(*npoints);
      break;
    case CF_LONG:
      if (siz == 1) *(UINT32 *)data = (UINT32)(*nfound);
      break;
    case CF_FLOAT:
      if (siz == 1) *(float *)data = (float)(*nfound);
      break;
    case CF_INTFLTINT:
      if (siz == 1) 
      {
        ((INTFLTINT *)data)[0].i1val = (UINT32)(*nfound);
        ((INTFLTINT *)data)[0].i2val = (UINT32)(*nfound);
        ((INTFLTINT *)data)[0].fval = (float)(*nfound);
      }
      break;
    case CF_DOUBLE:
      if (siz == 1) *(double *)data = (double)(*nfound);
      break;
  }
}

int nDaysInMonthTable[2][12] = {
  {31,28,31,30,31,30,31,31,30,31,30,31},
  {31,29,31,30,31,30,31,31,30,31,30,31},
};
#define nDaysInMonth(yy,mm) (((yy) % 4) ? nDaysInMonthTable[0][mm] : nDaysInMonthTable[1][mm])

int nDaysInTimeSpan(struct tm *s,struct tm *e)
{
  int n = 0, nx, y, d, ms, me, i;

  d = nDaysInMonth(e->tm_year,e->tm_mon);
  nx = s->tm_mday + (d-e->tm_mday-1);
  for (y=s->tm_year; y<e->tm_year+1; y++)
  {
    ms = y == s->tm_year ? s->tm_mon : 0;
    me = y == e->tm_year ? e->tm_mon : 11;
    for (i=ms; i<me+1; i++) 
    {
      n += nDaysInMonth(y,i);
    }
  }
  return n - nx;
}
time_t getStartTimeFromDepth(char *eqp,char *prp, int depth)
{
  time_t t=time(NULL);
  int recsize,i,timestampLength = useHighResolutionTimeStamp ? sizeof(double) : sizeof(SINT32);
# ifndef NO_LTS
  int n,k,siz;
  struct tm *lt,it,itprev;
  char fn[16],fileName[128];
  FILE *fp = NULL;
# endif

  if (nHistoryRecords <= 0 || historyRecords == NULL) return -un_allocated;
  for (i=0; i<nHistoryRecords; i++)
  {
    if (strncmp(historyRecords[i].c.EqpName,eqp,EQP_NAME_SIZE)) continue;
    if (strncmp(historyRecords[i].c.EqpProperty,prp,EQP_PROP_SIZE)) continue;
    break;
  }
  if (i == nHistoryRecords) return -un_allocated;
  recsize = historyRecords[i].recordLength + timestampLength;
# ifndef NO_LTS
  /* if no long term storage then there's only what's in short term */
  if (historyRecords[i].depthLong == 0) return t - historyRecords[i].depthShort;
  /* start looking backwards from today to find a start point where at least 'depth' points are stored */
  lt = localtime(&t); 
  it = *lt; it.tm_sec = 0; it.tm_hour = 0; it.tm_min = 0; itprev = it;
  for (k=0,n=0; k<30 && n<depth; k++)
  {
    MakeTargetFileName(fn,it.tm_year,it.tm_mon+1,it.tm_mday,i);
    sprintf(fileName,"%s%s",ArchiveLocation,fn);
    if ((fp=fopen(fileName,"r")) != NULL)
    {
      fseek(fp,0,SEEK_END);
      siz = ftell(fp);
      fclose(fp);
      if (siz > 0) 
      {
        if (n + siz/recsize > depth) 
        {
          return n == 0 ? mktime(&it) : mktime(&itprev);
        }
        else n += siz/recsize;
      }
    }
    itprev = it;
    if (it.tm_mday > 1) it.tm_mday--;
    else 
    {
      if (it.tm_mon > 0)
      {
        it.tm_mon--;
      }
      else
      {
        it.tm_year--;
        it.tm_mon = 11;
      }
      it.tm_mday = nDaysInMonth(it.tm_year,it.tm_mon) + 1;
    }  
  }
  if (n > 0) return mktime(&itprev);
# endif
  return t - historyRecords[i].depthShort;
}
int getLTSData(char *eqp,char *prp, char *dev,int index,
               double dstarttime,double dendtime,
               float lwrLimit, float uprLimit,
               BYTE *data,int fmt,int arraysize,int sample,int recsreadFlag)
{
# ifndef NO_LTS
  SINT32 ps;
  int recsread=0,ptr=0,recsfound=0,rnumber=0,recsize,i,asize,daystotal,daysinmonth;
  int literalDevice,finalPass=FALSE;
  struct tm st,et,*lt,it;
  time_t t = time(NULL),itime = 0;
  time_t starttime = (time_t)dstarttime;
  time_t endtime = (time_t)dendtime;

  if (nHistoryRecords <= 0 || historyRecords == NULL) return -un_allocated;
  literalDevice = IsLiteralDevice(eqp,prp,dev);
  for (i=0; i<nHistoryRecords; i++)
  {
    if (strncmp(historyRecords[i].c.EqpName,eqp,EQP_NAME_SIZE)) continue;
    if (strncmp(historyRecords[i].c.EqpProperty,prp,EQP_PROP_SIZE)) continue;
    if (strncmp(historyRecords[i].c.EqpDeviceName,dev,EQP_DEVICE_SIZE))
    {
      if (index == 0 && !literalDevice) index = GetDeviceNumberEx(eqp,dev,prp);
      else continue;
    }
    break;
  }
  if (i == nHistoryRecords) return -un_allocated;
  if (historyRecords[i].depthShort == 0) return -getRemoteHistoryServer(eqp,prp);
  if (historyRecords[i].depthLong == 0) return 0;
  if (starttime >= t - historyRecords[i].depthShort * STS_DEADBAND) return 0;
  if (index < 0) return index;
  if (index > (int)historyRecords[i].c.EqpSizeOut) return -out_of_range;
  if (arraysize < 1) return -buffer_too_small;

  recsize = fmtsizeof(fmt);
  lt = localtime(&endtime); et = *lt;
  lt = localtime(&starttime); st = *lt;
  if (et.tm_mon == st.tm_mon && et.tm_mday == st.tm_mday)
  {
    asize = arraysize;
    finalPass = TRUE;
  }
  else if (arraysize == 1)   /* just a number query */
  {
    asize = arraysize;
  }
  else                      /* spans multiple files */
  {
    asize = 0;
  }

  daystotal = nDaysInTimeSpan(&st,&et);

  if (recsreadFlag && arraysize > 1)
  {
    recsread = 1;
    ptr = recsread * recsize;
  }

  while (recsread < arraysize)
  {
    if (asize < arraysize || arraysize == 1)
    {
      it = st;
      if (it.tm_year == et.tm_year && it.tm_mon == et.tm_mon && 
          (it.tm_mday == et.tm_mday || useMonthlyHistoryFiles))
      {
        asize = arraysize;
        finalPass = TRUE;
      }
      else  /* move along */
      {
        it.tm_mday = (it.tm_mday+1)%(nDaysInMonth(it.tm_year,it.tm_mon)+1);
        if (it.tm_mday == 0)
        {
          it.tm_mday++;
          it.tm_mon = (it.tm_mon+1)%12;
          if (it.tm_mon == 0) it.tm_year++;
        }
        it.tm_hour = 0; it.tm_min = 0; it.tm_sec = 0;
        itime = mktime(&it);
        daysinmonth = (int)MAX((itime-starttime)/(3600*24),1);
        if (arraysize > 1) 
        {
          if (fmt <= CF_FLOAT) 
          {
            asize = arraysize;
          }
          else 
          {
            asize += MAX((arraysize * daysinmonth)/daystotal,1);
          }
          if (asize >= arraysize) finalPass = TRUE; /* just in case */
        }
      }
    }
    if (asize > arraysize) asize = arraysize;
    ps = loadLTSData(i,index,dstarttime,dendtime,lwrLimit,uprLimit,&data[ptr],fmt,asize-recsread,sample);
    if (asize == 1) recsfound += (int)ps; else recsread += (int)ps;
    ptr = recsread * recsize;
    if (finalPass) break;
    starttime = itime;
    dstarttime = (double)starttime;
    lt = localtime(&starttime); st = *lt;
  }
  rnumber = recsfound > 0 ? recsfound : recsread;
  if (arraysize == 1) recsread = recsfound;

  assignNumberOfPoints(data,fmt,arraysize,&recsread,&recsfound,recsreadFlag);

  return rnumber;
# else
  return 0;
# endif /* NO_LTS */
}
extern char tRedirectedServer[64];
int getRemoteHistoryServer(char *eqp,char *prp)
{
  int i;
  if (nHistoryRecords <= 0 || historyRecords == NULL) return un_allocated;
  for (i=0; i<nHistoryRecords; i++)
  {
    if (strncmp(historyRecords[i].c.EqpName,eqp,EQP_NAME_SIZE)) continue;
    if (strncmp(historyRecords[i].c.EqpProperty,prp,EQP_PROP_SIZE)) continue;
    break;
  }
  if (i == nHistoryRecords) return un_allocated;
  if (historyRecords[i].prd == NULL) return not_defined;
  memcpy(tRedirectedServer,historyRecords[i].prd->prpRedirection[0].name,16*4);
  return server_redirection;
}
#endif /* NO_DOXYGEN */
#ifndef SMALL_TINE_LIB
EXPORT int AppendHistoryInformation(char *eqp,char *prp,char *dev,int len,int fmt,int idx,HistorySpecification *hspec)
{
  int i, cc = 0;
  time_t t = time(NULL);
  struct tm *lt=localtime(&t);
  static int init_done = FALSE;

  if (eqp == NULL || prp == NULL || dev == NULL || hspec == NULL) ccerr(argument_list_error);
  if (!init_done)
  {
    historyInit(GetExportedName(eqp),eqp);
    init_done = TRUE;
  }
  if (historyRecords == NULL)
  {
    if ((historyRecords=(HistoryRecord *)SystemCalloc(maxHistoryRecords,sizeof(HistoryRecord))) == NULL)
      ccerr(out_of_local_memory);
  }
  if ((i=nHistoryRecords) >= maxHistoryRecords) ccerr(resources_exhausted);
  historyRecords[i].recordIndex = idx;
  strncpy(historyRecords[i].c.EqpName,eqp,EQP_NAME_SIZE);
  strncpy(historyRecords[i].c.EqpProperty,prp,EQP_PROP_SIZE);
  strncpy(historyRecords[i].c.EqpDeviceName,dev,EQP_DEVICE_SIZE);
  historyRecords[i].c.EqpSizeOut = len;
  historyRecords[i].c.EqpFormatOut = BFMT(fmt);
  if (isInHistoryTable(&historyRecords[i],nHistoryRecords)) ccerr(0);
  historyRecords[i].heartbeat = hspec->heartbeat;
  historyRecords[i].pollingRate = hspec->pollingRate;
  if (historyRecords[i].pollingRate < 20) historyRecords[i].pollingRate *= 1000; /* was thinking in seconds ! */
  if (historyRecords[i].pollingRate < minHistoryPollingRate) minHistoryPollingRate = historyRecords[i].pollingRate;
  historyRecords[i].archiveRate = hspec->archiveRate;
  if (historyRecords[i].archiveRate < 20) historyRecords[i].archiveRate *= 1000; /* was thinking in seconds ! */
  historyRecords[i].pTolerance = hspec->pTolerance;
  if (historyRecords[i].pTolerance > 1.0) historyRecords[i].pTolerance /= 100;
  historyRecords[i].aTolerance = hspec->aTolerance;
  historyRecords[i].depthShort = hspec->depthShort;
  historyRecords[i].depthLong = hspec->depthLong;
  historyRecords[i].recordLength =
      (int)historyRecords[i].c.EqpSizeOut * fmtsizeof(historyRecords[i].c.EqpFormatOut);
# ifdef MSDOS
  if (historyRecords[i].recordLength > 0x3fff)
  {
    feclog("HIST: %s %s %s record too long (%d bytes)",
            historyRecords[i].c.EqpName,
            historyRecords[i].c.EqpProperty,
            historyRecords[i].c.EqpDeviceName,
            historyRecords[i].recordLength);
    ccerr(resources_exhausted);
  }
# endif
  MakeTargetFileName(historyRecords[i].ArchiveFileName,lt->tm_year,lt->tm_mon+1,lt->tm_mday,i);
  if (hspec->rhsServerName != NULL && strlen(hspec->rhsServerName) > 0)
  {
    if ((historyRecords[i].prd=(PrpRedirBlk *)SystemCalloc(1,sizeof(PrpRedirBlk))) == NULL) 
      ccerr(out_of_server_memory);

    strncpy(historyRecords[i].prd->prpRedirection[0].name,hspec->rhsServerName,16);
    if (hspec->rhsPropertyName != NULL) 
      strncpy(historyRecords[i].prd->prpRedirection[1].name,hspec->rhsPropertyName,16);
  }
  if ((historyRecords[i].timestamp=(double *)SystemCalloc(historyRecords[i].depthShort,sizeof(double))) == NULL)
      ccerr(out_of_local_memory);
  if ((historyRecords[i].data=(BYTE *)SystemCalloc(historyRecords[i].depthShort,historyRecords[i].recordLength)) == NULL)
      ccerr(out_of_local_memory);
  feclog("HIST: %s %s %s (%d s %d mon)",
          historyRecords[i].c.EqpName,
          historyRecords[i].c.EqpProperty,
          historyRecords[i].c.EqpDeviceName,
          historyRecords[i].depthShort,
          historyRecords[i].depthLong);
  nHistoryRecords++;
  if (MaxPollingRate > minHistoryPollingRate)
  {
    SysPoll = minHistoryPollingRate;
    MaxPollingRate = minHistoryPollingRate;
    if (minHistoryPollingRate < MinPollingRate) MinPollingRate = minHistoryPollingRate;
  }
err:
  return cc;
}
EXPORT int getLastStoredData(char *eqp,char *prp, char *dev,int index,BYTE *data,int fmt,int arraysize)
{
  int ptr=0,i,literalDevice;

  if (nHistoryRecords <= 0 || historyRecords == NULL) return un_allocated;
  literalDevice = IsLiteralDevice(eqp,prp,dev);
  for (i=0; i<nHistoryRecords; i++)
  {
    if (strncmp(historyRecords[i].c.EqpName,eqp,EQP_NAME_SIZE)) continue;
    if (strncmp(historyRecords[i].c.EqpProperty,prp,EQP_PROP_SIZE)) continue;
    if (strncmp(historyRecords[i].c.EqpDeviceName,dev,EQP_DEVICE_SIZE))
    {
      if (index == 0 && !literalDevice) index = GetDeviceNumberEx(eqp,dev,prp);
      else continue;
    }
    break;
  }
  if (i == nHistoryRecords) return un_allocated;
  if (historyRecords[i].depthLong == 0) return 0;
  if (index < 0) return -index;
  if (index > (int)historyRecords[i].c.EqpSizeOut) return out_of_range;
  if (arraysize < 1) return buffer_too_small;

  return getLastLTSData(i,index,&data[ptr],fmt,arraysize);
}
#endif /* SMALL_TINE_LIB */
