/** \file hsteqm.c
 * \brief TINE NETMEX archive server engine plus API routines
 *
 */
#include "prolog.h"
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include "errors.h"
#include "rpctypes.h"
#include "brdcast.h"
#include "servdef.h"
#include "toolkit.h"
#include "netmex.h"

unsigned long ElfHash(const unsigned char *key);
MEXDATA *MexData[PROPERTY_HASH_SIZE];
int numKeywords = 0;
int hsteqm(char *devName,char *keyword,DTYPE *dout,DTYPE *din,short access);
void hstini(void);

MEXDATA *getKeyword(char *keyword)
{
  int idx;
  MEXDATA *mexdata=NULL;

  idx = (int)(ElfHash((BYTE *)keyword) % PROPERTY_HASH_SIZE);
  for (mexdata=MexData[idx]; mexdata; mexdata=mexdata->next)
    if (!strcmp(keyword,mexdata->keyword)) break;
  return mexdata;
}

/* RegisterKeywordEx() */
/**
 * \brief The extended version of RegisterKeyword(). 
 *  Registers a Keyword with the TINE NETMEX engine
 * 
 * This is analogous to RegisterProperty() in a pure TINE server interface.
 * A server registers a Keyword, its size and data type, plus a background function
 * to be called upon to update the data associated with the keyword.  This call differs
 * from RegisterKeyword() in that it offers a 'decription' parameter allowing the caller
 * to provide a description of the NETMEX keyword.
 *
 * \param keyword is the NETMEX keyword to be registered.  Up to 32 characters are allowed,
 *  but it is preferable to stick to 16 or fewer.
 * \param description is the description of the keyword.
 * \param format is the TINE data type of the keyword
 * \param size is the number of elements associated with the keyword (i.e. the 'array size').
 * \param fcn is the background function which refreshes the data associated with the 
 *  keyword, to be called at regular intervals (see setKeywordRefreshRoutine()).
 *  This function must have the prototype fcn(float *data,unsigned short size,short index),
 *  which allows the keyword data set to be updated piecemeal, i.e. 'size' elements beginning
 *  at element number 'index' from the data object specified by 'data'.  If this data 
 *  object is not a 'float' it needs to be cast into the appropriate data set inside the
 *  refresh function.
 *  
 * \return 0 upon success, otherwise a TINE completion code.
 *
 * \see RegisterKeyword, RegisterKeywordDeviceName.
 *
 * \b Example:
 * 
 * \include eg_RegisterKeywordEx.c
 * 
 */
int RegisterKeywordEx(char *keyword,char *description,short format,unsigned short size,
                    short (*fcn)(float *,unsigned short,short))
{
  int idx;
  MEXDATA *mexdata;
  
  idx = (int)(ElfHash((BYTE *)keyword) % PROPERTY_HASH_SIZE);
  for (mexdata=MexData[idx]; mexdata; mexdata=mexdata->next)
    if (!strcmp(keyword,mexdata->keyword)) break;
  if (mexdata == NULL) 
  {
    numKeywords++;
    if ((mexdata=(MEXDATA *)calloc(1,sizeof(MEXDATA))) == NULL)
      return out_of_server_memory;
    strncpy(mexdata->keyword,keyword,16);
    if (description != NULL) strncpy(mexdata->description,description,96);
    mexdata->format = format;
    mexdata->size = size;
    mexdata->fcn = fcn;
    mexdata->next = MexData[idx];
    MexData[idx] = mexdata;
  }
  return 0;
}
/* RegisterKeyword() */
/**
 * \brief Registers a Keyword with the TINE NETMEX engine
 * 
 * This is analogous to RegisterProperty() in a pure TINE server interface.
 * A server registers a Keyword, its size and data type, plus a background function
 * to be called upon to update the data associated with the keyword.
 *
 * \param keyword is the NETMEX keyword to be registered.  Up to 32 characters are allowed,
 *  but it is preferable to stick to 16 or fewer.
 * \param format is the TINE data type of the keyword
 * \param size is the number of elements associated with the keyword (i.e. the 'array size').
 * \param fcn is the background function which refreshes the data associated with the 
 *  keyword, to be called at regular intervals (see setKeywordRefreshRoutine()).
 *  This function must have the prototype fcn(float *data,unsigned short size,short index),
 *  which allows the keyword data set to be updated piecemeal, i.e. 'size' elements beginning
 *  at element number 'index' from the data object specified by 'data'.  If this data 
 *  object is not a 'float' it needs to be cast into the appropriate data set inside the
 *  refresh function.
 *  
 * \return 0 upon success, otherwise a TINE completion code.
 *
 * \see RegisterKeywordEx, RegisterKeywordDeviceName.
 *
 * \b Example:
 * 
 * \include eg_RegisterKeyword.c
 */
int RegisterKeyword(char *keyword,short format,unsigned short size,
                    short (*fcn)(float *,unsigned short,short))
{
  return RegisterKeywordEx(keyword,NULL,format,size,fcn);
}
/* RegisterKeywordDeviceName() */
/**
 * \brief Registers a device name associated with the input Keyword.
 * 
 * This is analogous to RegisterDeviceName() in a pure TINE server interface.
 * A registered NETMEX Keyword might be an array of elements.  The individual 
 * elements can of course be addressed individually by number, but it is
 * 'polite' to associate the individual elements with device names where possible.
 *
 * \param keyword is the associated NETMEX keyword.
 * \param devicename is the device name to be bound to the index given.
 * \param index is the array element to be bound to the input devicename.  The index
 *  given must of course be smaller than the 'size' of the keyword, (counting from 0).
 *  
 * \return 0 upon success, otherwise a TINE completion code.
 *
 * \see RegisterKeywordEx, RegisterKeyword.
 *
 * \b Example:
 * 
 * \include eg_RegisterKeywordDeviceName.c
 */
int RegisterKeywordDeviceName(char *keyword,char *devicename,int index)
{
  int idx;
  MEXDATA *mexdata;

  if (strlen(keyword) > 27) return string_too_long;
  idx = (int)(ElfHash((BYTE *)keyword) % PROPERTY_HASH_SIZE);
  for (mexdata=MexData[idx]; mexdata; mexdata=mexdata->next)
    if (!strcmp(keyword,mexdata->keyword)) break;
  if (mexdata == NULL) return illegal_property;
  if (index < 0 || index >= mexdata->size) return illegal_equipment_number;
  if (mexdata->devicenames == NULL && 
      (mexdata->devicenames=(NAME16 *)calloc(mexdata->size,16)) == NULL)
    return out_of_server_memory;
  strncpy(mexdata->devicenames[index].name,devicename,16);
  return 0;
}
void hstini(void)
{
  int i;
  MEXDATA *mexdata=NULL;
  char prpqryname[32];
  char *kd, *defaultDescription = "machine parameter";
  
  for (i=0; i<PROPERTY_HASH_SIZE; i++)
    for (mexdata=MexData[i]; mexdata; mexdata=mexdata->next)
    {
      kd = strlen(mexdata->description) == 0 ? defaultDescription : mexdata->description;
      if (RegisterProperty("HSTEQM",mexdata->keyword,
          mexdata->size,mexdata->format,CA_READ,kd) < 0)
        printf("FILER: couldn't register %s for export\n>",mexdata->keyword);
      if (mexdata->devicenames != NULL)
      {
        sprintf(prpqryname,"%s.NAM",mexdata->keyword);
        if (RegisterProperty("HSTEQM",prpqryname,
            mexdata->size,CF_NAME16,CA_READ,kd) < 0)
          printf("FILER: couldn't register %s for export\n>",prpqryname);
      }
    }
}
/* initNetmexHistoryServer() */
/**
 * \brief Initializes the TINE NETMEX History engine and starts the server.
 * 
 * This call should follow the call to SystemInit() inside the main() routine.
 *  
 * \return none.
 *
 * \b Example:
 *
 * \include eg_mexsrv.c
 */
int initNetmexHistoryServer(void)
{
  static int initialized = 0;
  int cc;

  if (initialized) return 0;

  if ((cc=RegisterNetmexKeywords()) != 0)
  {
    printf("Error registering netmex keyword: %s\n",erlst[cc]);
    exit(1);
  }
  if ((cc=RegisterEquipmentModule(NULL,"HSTEQM",0,hsteqm,hstini,NULL,0,NULL)) != 0)
  {
    printf("Error registering equipment module: %s\n",erlst[cc]);
    exit(1);
  }
  initialized = 1;
  return 0;
}
int repackageSnapshotData(DTYPE *d,void *pdat,short fmt,int n)
{
  int len=d->dArrayLength;
  int fmtsiz;
  UINT32 len0;
  DTYPE d0;

  fmtsiz = fmtsizeof(fmt);
  d0.dArrayLength = n; 
  d0.dFormat = fmt;
  d0.data.vptr = pdat;
  switch (d->dFormat)
  {
    case CF_SHORT:
      getValuesAsShort(&d0,d->data.sptr,len);
      break;
    case CF_LONG:
      getValuesAsLong(&d0,d->data.lptr,len);
      break;
    case CF_FLOAT:
      getValuesAsFloat(&d0,d->data.fptr,len);
      break;
    case CF_DOUBLE:
      getValuesAsDouble(&d0,d->data.dptr,len);
      break;
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME48:
    case CF_NAME64:
      len0 = len*fmtsiz;
      getValuesAsString(&d0,(char *)d->data.vptr,&len0);
      break;
    default: 
      return -1;
  }
  return MIN(len,n);
}
int repackageData(DTYPE *d,void *pdat,short fmt,time_t *ptim,int n)
{
  int i,k,len=d->dArrayLength;
  int fmtsiz,raster=1;
  UINT32 len0;
  DTYPE d0;

  fmtsiz = fmtsizeof(fmt);
  if (len < n) raster += n/len;
  d0.dArrayLength = 1; 
  d0.dFormat = fmt;
  for (i=0,k=0; i<n && k<len; i+=raster)
  {
    d0.data.vptr = pdat + fmtsiz*i;
    len0 = fmtsiz;
    switch (d->dFormat)
    {
      case CF_LNGINT:
        ((LNGINT *)d->data.vptr)[k].ival = ptim[i];
        getValuesAsLong(&d0,&((LNGINT *)d->data.vptr)[k].lval,1);
        break;
      case CF_FLTINT:
        ((FLTINT *)d->data.vptr)[k].ival = ptim[i];
        getValuesAsFloat(&d0,&((FLTINT *)d->data.vptr)[k].fval,1);
        break;
      case CF_DBLINT:
        ((DBLINT *)d->data.vptr)[k].ival = ptim[i];
        getValuesAsDouble(&d0,&((DBLINT *)d->data.vptr)[k].dval,1);
        break;
      case CF_TDS:
        ((INTFLTINT *)d->data.vptr)[k].i1val = ptim[i];
        getValuesAsFloat(&d0,&((INTFLTINT *)d->data.vptr)[k].fval,1);
        ((INTFLTINT *)d->data.vptr)[k].i2val = 0;
        break;
      case CF_NAME16I:
        ((NAME16I *)d->data.vptr)[k].ival = ptim[i];
        getValuesAsString(&d0,((NAME16I *)d->data.vptr)[k].name,&len0);
        break;
      case CF_NAME32I:
        ((NAME32I *)d->data.vptr)[k].ival = ptim[i];
        getValuesAsString(&d0,((NAME32I *)d->data.vptr)[k].name,&len0);
        break;
      case CF_NAME48I:
        ((NAME48I *)d->data.vptr)[k].ival = ptim[i];
        getValuesAsString(&d0,((NAME48I *)d->data.vptr)[k].name,&len0);
        break;
      case CF_NAME64I:
        ((NAME64I *)d->data.vptr)[k].ival = ptim[i];
        getValuesAsString(&d0,((NAME64I *)d->data.vptr)[k].name,&len0);
        break;
      default:
        break;
    }
    k++;
  }
  return k;
}
int putLongValueInDoublet(DTYPE *d,int n)
{
  switch (d->dFormat)
  {
    case CF_SHTINT : ((SHTINT *)d->data.vptr)[0].ival = n; break;
    case CF_LNGINT : ((LNGINT *)d->data.vptr)[0].ival = n; break;
    case CF_FLTINT : ((FLTINT *)d->data.vptr)[0].ival = n; break;
    case CF_DBLINT : ((DBLINT *)d->data.vptr)[0].ival = n; break;
    case CF_TDS : ((INTFLTINT *)d->data.vptr)[0].i1val = n; break;
    case CF_NAME8I : ((NAME8I *)d->data.vptr)[0].ival = n; break;
    case CF_NAME16I : ((NAME16I *)d->data.vptr)[0].ival = n; break;
    case CF_NAME32I : ((NAME32I *)d->data.vptr)[0].ival = n; break;
    case CF_NAME48I : ((NAME48I *)d->data.vptr)[0].ival = n; break;
    case CF_NAME64I : ((NAME64I *)d->data.vptr)[0].ival = n; break;
    default : return illegal_format;
  }
  return 0;
}
int hsteqm(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access)
{
  int i,n,j,siz,fmt,isNameQuery=0,isSimpleFormat=0;
  int sample=0,idx=0;
  float upr=0,lwr=0;
  MEXDATA *mexdata=NULL;
  char *c, kwd[40];
  void *pdat=NULL;
  time_t *ptim=NULL,tnow=time(NULL),start,stop=tnow;
  
  strncpy(kwd,devProperty,32);
  if ((c=strstr(kwd,".NAM")) != NULL)
  {
    isNameQuery = TRUE;
    *c = 0;
  }
  if (!fmtsizeof(dout->dFormat)) return illegal_format;
  if (dout->dArrayLength < 1) return illegal_read_write;

   /* support a couple of useful queries: */
  if (!strcmp(kwd,"NTAGS"))
  {
    return putValuesFromLong(dout,(UINT32 *)&numKeywords,1);
  }
  if (!strcmp(kwd,"TAGS"))
  {
    switch (dout->dFormat)
    {
      case CF_NAME8:
      case CF_NAME16:
      case CF_NAME32: siz = fmtsizeof(dout->dFormat); break;
      default : return illegal_format;
    }
    for (i=0,j=0; i<dout->dArrayLength,j<PROPERTY_HASH_SIZE; j++)
      for (mexdata=MexData[j]; mexdata; mexdata=mexdata->next)
      { 
        strncpy(dout->data.cptr+i*siz,mexdata->keyword,siz);
        i++;
      }
    dout->dArrayLength = i;
    return 0;
  }

  /* must be a keyword from here on: */
  idx = GetDeviceNumberEx("HSTEQM",devName,kwd);
  if (idx < 0) return illegal_equipment_number;
  if ((mexdata=getKeyword(kwd)) == NULL) return non_existent;
  if (isNameQuery && mexdata->devicenames != NULL)
  {
    switch (dout->dFormat)
    {
      case CF_NAME16:
      case CF_NAME32:
      case CF_NAME48:
      case CF_NAME64:
        siz = fmtsizeof(dout->dFormat);
        for (i=0; i<dout->dArrayLength; i++)
        {
          if (mexdata->devicenames[i].name[0] == 0)
            sprintf(&dout->data.cptr[i*siz],"#%d",i);
          else
          strncpy(&dout->data.cptr[i*siz],mexdata->devicenames[i].name,16);
        }
        return 0;
      default:
        return illegal_format;
    }
  }
  /* the property is the keyword ! */
  fmt = getSimpleFormat(dout->dFormat);
  isSimpleFormat = fmt == -1 ? -1 : 0;
  if ((idx=GetDeviceNumberEx("HSTEQM",devName,kwd)) < 0) idx = 0;
  if (din->dArrayLength > 0)
  {
    switch (din->dFormat)
    {
      case CF_LONG:
        switch (din->dArrayLength)
        {
          default:
          case 4: sample = din->data.lptr[3];
          case 3: if (idx == 0) idx = din->data.lptr[2];
          case 2: stop  = din->data.lptr[1];
          case 1: start = din->data.lptr[0];
          case 0: break;
        }
        break;
      case CF_FWINDOW:
        start = ((FWINDOW *)din->data.vptr)[0].i1val;
        stop  = ((FWINDOW *)din->data.vptr)[0].i2val;
        lwr   = ((FWINDOW *)din->data.vptr)[0].f1val;
        upr   = ((FWINDOW *)din->data.vptr)[0].f2val;
        if (lwr < upr) sample  = -1;
        break;
      case CF_WINDOW:
        start = ((DWINDOW *)din->data.vptr)[0].t1;
        stop  = ((DWINDOW *)din->data.vptr)[0].t2;
        sample = ((DWINDOW *)din->data.vptr)[0].i2val;
        break;
      default:
        return illegal_format;
    }
  }
  else
  {
    start = tnow - dout->dArrayLength * 60;  /* assume depth is minutes */
  }
  if (stop > tnow) stop = tnow;
  if (!isSimpleFormat)
  {
    if (dout->dArrayLength == 1) /* just get the number of points in interval */
    {
      n = getNumberArchiveRecordsInInterval(kwd,start,stop);
      return putLongValueInDoublet(dout,n);
    }
    if ((n=getArchiveRecordsInInterval(kwd,idx,start,stop,&pdat,&ptim)) < 0)
      return invalid_data;
    if (pdat == NULL || ptim == NULL) return argument_list_error;
    n = repackageData(dout,pdat,mexdata->format,ptim,n);
  }
  else
  {
    if ((n=getArchiveRecord(kwd,idx,start,&pdat,&ptim)) < 0) return invalid_data;
    n = repackageSnapshotData(dout,pdat,mexdata->format,n);
    setDataTimeStamp((double)(*ptim));
  }
  if (n < dout->dArrayLength) dout->dArrayLength = n;
  return 0;
}

