/** \file keyeqm.c
 * \brief TINE NETMEX server engine plus API routines
 *
 */

#include "prolog.h"
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include "errors.h"
#include "rpctypes.h"
#include "brdcast.h"
#include "servdef.h"
#include "toolkit.h"
#include "netmex.h"

unsigned long ElfHash(const unsigned char *key);
MEXDATA *MexData[PROPERTY_HASH_SIZE];
int numKeywords = 0;
int keyeqm(char *devName,char *keyword,DTYPE *dout,DTYPE *din,short access);
void keyini(void);
void keytsk(void);

MEXDATA *getKeyword(char *keyword)
{
  int idx;
  MEXDATA *mexdata=NULL;

  idx = (int)(ElfHash((BYTE *)keyword) % PROPERTY_HASH_SIZE);
  for (mexdata=MexData[idx]; mexdata; mexdata=mexdata->next)
    if (!strcmp(keyword,mexdata->keyword)) break;
  return mexdata;
}

/* RegisterKeywordEx() */
/**
 * \brief The extended version of RegisterKeyword(). 
 *  Registers a Keyword with the TINE NETMEX engine
 * 
 * This is analogous to RegisterProperty() in a pure TINE server interface.
 * A server registers a Keyword, its size and data type, plus a background function
 * to be called upon to update the data associated with the keyword.  This call differs
 * from RegisterKeyword() in that it offers a 'decription' parameter allowing the caller
 * to provide a description of the NETMEX keyword.
 *
 * \param keyword is the NETMEX keyword to be registered.  Up to 32 characters are allowed,
 *  but it is preferable to stick to 16 or fewer.
 * \param description is the description of the keyword.
 * \param format is the TINE data type of the keyword
 * \param size is the number of elements associated with the keyword (i.e. the 'array size').
 * \param fcn is the background function which refreshes the data associated with the 
 *  keyword, to be called at regular intervals (see setKeywordRefreshRoutine()).
 *  This function must have the prototype fcn(float *data,unsigned short size,short index),
 *  which allows the keyword data set to be updated piecemeal, i.e. 'size' elements beginning
 *  at element number 'index' from the data object specified by 'data'.  If this data 
 *  object is not a 'float' it needs to be cast into the appropriate data set inside the
 *  refresh function.
 *  
 * \return 0 upon success, otherwise a TINE completion code.
 *
 * \see RegisterKeyword, RegisterKeywordDeviceName.
 *
 * \b Example:
 * 
 * \include eg_RegisterKeywordEx.c
 * 
 */

int RegisterKeywordEx(char *keyword,char *description,short format,unsigned short size,
                    short (*fcn)(float *,unsigned short,short))
{
  int idx;
  MEXDATA *mexdata;
  
  idx = (int)(ElfHash((BYTE *)keyword) % PROPERTY_HASH_SIZE);
  for (mexdata=MexData[idx]; mexdata; mexdata=mexdata->next)
    if (!strcmp(keyword,mexdata->keyword)) break;
  if (mexdata == NULL) 
  {
    numKeywords++;
    if ((mexdata=(MEXDATA *)calloc(1,sizeof(MEXDATA))) == NULL)
      return out_of_server_memory;
    strncpy(mexdata->keyword,keyword,16);
    if (description != NULL) strncpy(mexdata->description,description,96);
    mexdata->format = format;
    mexdata->size = size;
    mexdata->fcn = fcn;
    mexdata->next = MexData[idx];
    MexData[idx] = mexdata;
  }
  return 0;
}
/* RegisterKeyword() */

/**
 * \brief Registers a Keyword with the TINE NETMEX engine
 * 
 * This is analogous to RegisterProperty() in a pure TINE server interface.
 * A server registers a Keyword, its size and data type, plus a background function
 * to be called upon to update the data associated with the keyword.
 *
 * \param keyword is the NETMEX keyword to be registered.  Up to 32 characters are allowed,
 *  but it is preferable to stick to 16 or fewer.
 * \param format is the TINE data type of the keyword
 * \param size is the number of elements associated with the keyword (i.e. the 'array size').
 * \param fcn is the background function which refreshes the data associated with the 
 *  keyword, to be called at regular intervals (see setKeywordRefreshRoutine()).
 *  This function must have the prototype fcn(float *data,unsigned short size,short index),
 *  which allows the keyword data set to be updated piecemeal, i.e. 'size' elements beginning
 *  at element number 'index' from the data object specified by 'data'.  If this data 
 *  object is not a 'float' it needs to be cast into the appropriate data set inside the
 *  refresh function.
 *  
 * \return 0 upon success, otherwise a TINE completion code.
 *
 * \see RegisterKeywordEx, RegisterKeywordDeviceName.
 *
 * \b Example:
 * 
 * \include eg_RegisterKeyword.c
 */

int RegisterKeyword(char *keyword,short format,unsigned short size,
                    short (*fcn)(float *,unsigned short,short))
{
  return RegisterKeywordEx(keyword,NULL,format,size,fcn);
}
/* RegisterKeywordDeviceName() */

/**
 * \brief Registers a device name associated with the input Keyword.
 * 
 * This is analogous to RegisterDeviceName() in a pure TINE server interface.
 * A registered NETMEX Keyword might be an array of elements.  The individual 
 * elements can of course be addressed individually by number, but it is
 * 'polite' to associate the individual elements with device names where possible.
 *
 * \param keyword is the associated NETMEX keyword.
 * \param devicename is the device name to be bound to the index given.
 * \param index is the array element to be bound to the input devicename.  The index
 *  given must of course be smaller than the 'size' of the keyword, (counting from 0).
 *  
 * \return 0 upon success, otherwise a TINE completion code.
 *
 * \see RegisterKeywordEx, RegisterKeyword.
 *
 * \b Example:
 * 
 * \include eg_RegisterKeywordDeviceName.c
 */

int RegisterKeywordDeviceName(char *keyword,char *devicename,int index)
{
  int idx;
  MEXDATA *mexdata;

  if (strlen(keyword) > 27) return string_too_long;
  idx = (int)(ElfHash((BYTE *)keyword) % PROPERTY_HASH_SIZE);
  for (mexdata=MexData[idx]; mexdata; mexdata=mexdata->next)
    if (!strcmp(keyword,mexdata->keyword)) break;
  if (mexdata == NULL) return illegal_property;
  if (index < 0 || index >= mexdata->size) return illegal_equipment_number;
  if (mexdata->devicenames == NULL && 
      (mexdata->devicenames=(NAME16 *)calloc(mexdata->size,16)) == NULL)
    return out_of_server_memory;
  strncpy(mexdata->devicenames[index].name,devicename,16);
  return 0;
}
void keyini(void)
{
  int i;
  MEXDATA *mexdata=NULL;
  char prpqryname[32];
  char *kd, *defaultDescription = "machine parameter";
  
  for (i=0; i<PROPERTY_HASH_SIZE; i++)
    for (mexdata=MexData[i]; mexdata; mexdata=mexdata->next)
    {
      kd = strlen(mexdata->description) == 0 ? defaultDescription : mexdata->description;
      if (RegisterProperty("KEYWRD",mexdata->keyword,
          mexdata->size,mexdata->format,CA_READ,kd) < 0)
        printf("FILER: couldn't register %s for export\n>",mexdata->keyword);
      if (mexdata->devicenames != NULL)
      {
        sprintf(prpqryname,"%s.NAM",mexdata->keyword);
        if (RegisterProperty("KEYWRD",prpqryname,
            mexdata->size,CF_NAME16,CA_READ,kd) < 0)
          printf("FILER: couldn't register %s for export\n>",prpqryname);
	
      }
    }
}
void (*RefreshKeywords)();
int RefreshKeywordRate = 500; /* milliseconds */

/* setKeywordRefreshRoutine() */
/**
 * \brief Registers a background refresh routine to be called at the interval given.
 * 
 * This call will intruct the TINE NETMEX engine to call the routine provided at the
 * refresh rate interval provided.  This routine should be regarded as a timer whose
 * trigger should update the NETMEX keywords exported from the server.
 *
 * \param fcn is the background refresh routine to be called.  It must have the
 * prototype void fcn(void) and will be called by the TINE engine at the rate specified.
 * \param rate is the polling interval in milliseconds for which the background refresh
 * routine will be called.  
 *  
 * \return none.
 *
 * \b Example:
 *
 * \include eg_RegisterKeyword.c
 */

void setKeywordRefreshRoutine(void (*fcn)(void),int rate)
{
 RefreshKeywords = fcn;
 RefreshKeywordRate = rate;
}
void keytsk(void)
{
  if (RefreshKeywords != NULL) (*RefreshKeywords)();
}

/* initNetmexServer() */
/**
 * \brief Initializes the TINE NETMEX engine and starts the server.
 * 
 * This call should follow the call to SystemInit() inside the main() routine.
 *  
 * \return none.
 *
 * \b Example:
 *
 * \include eg_mexsrv.c
 */

int initNetmexServer(void)
{
  static int initialized = 0;
  int cc;

  if (initialized) return 0;

  if ((cc=RegisterNetmexKeywords()) != 0)
  {
    printf("Error registering netmex keyword: %s\n",erlst[cc]);
    exit(1);
  }
  
  if ((cc=RegisterEquipmentModule(NULL,"KEYWRD",0,keyeqm,keyini,keytsk,RefreshKeywordRate,NULL)) != 0)
  {
    printf("Error registering equipment module: %s\n",erlst[cc]);
    exit(1);
  }
  initialized = 1;
  return 0;
}
int keyeqm(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access)
{
  int i,j,siz,isNameQuery = 0;
  short index = 0;
  MEXDATA *mexdata=NULL;
  char *c, keyword[40];
  
  strncpy(keyword,devProperty,32);
  if ((c=strstr(keyword,".NAM")) != NULL)
  {
    isNameQuery = TRUE;
    *c = 0;
  }
  if (!fmtsizeof(dout->dFormat)) return illegal_format;
  if (dout->dArrayLength < 1) return illegal_read_write;
  index = GetDeviceNumberEx("KEYWRD",devName,keyword);
  if (index < 0) return illegal_equipment_number;

   /* support a couple of useful queries: */
  if (!strcmp(keyword,"NKEYWORD"))
  {
    return putValuesFromLong(dout,(UINT32 *)&numKeywords,1);
  }
  if (!strcmp(keyword,"KEYWORDS"))
  {
    switch (dout->dFormat)
    {
      case CF_STRING8:
      case CF_STRING16:
      case CF_STRING32: siz = fmtsizeof(dout->dFormat); break;
      default : return illegal_format;
    }
    for (i=0,j=0; i<dout->dArrayLength,j<PROPERTY_HASH_SIZE; j++)
      for (mexdata=MexData[j]; mexdata; mexdata=mexdata->next)
      { 
        strncpy(dout->data.cptr+i*siz,mexdata->keyword,siz);
        i++;
      }
    for (;i<dout->dArrayLength; i++) memset(dout->data.cptr+i*siz,' ',siz); /* leftovers */
    return 0;
  }

  /* must be a keyword from here on: */
  if ((mexdata=getKeyword(keyword)) == NULL) return non_existent;
  if (isNameQuery && mexdata->devicenames != NULL)
  {
    switch (dout->dFormat)
    {
      case CF_NAME16:
      case CF_NAME32:
      case CF_NAME48:
      case CF_NAME64:
        siz = fmtsizeof(dout->dFormat);
        for (i=0; i<dout->dArrayLength; i++)
        {
          if (mexdata->devicenames[i].name[0] == 0)
            sprintf(&dout->data.cptr[i*siz],"#%d",i);
          else
          strncpy(&dout->data.cptr[i*siz],mexdata->devicenames[i].name,16);
        }
        return 0;
      default:
        return illegal_format;
    }
  }
  if (mexdata->format != dout->dFormat) return illegal_format;

  /* let the mexdata->fcn() trap or wrap ! */
  /* if (dout->dArrayLength + index > mexdata->size) return dimension_error; */
  return mexdata->fcn(dout->data.fptr,dout->dArrayLength,index);
}
