/** \file listener.c
 * \brief TINE static listener documentation 
 *
 */
#ifndef NO_DOXYGEN
#include "prolog.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <sys/timeb.h>
#include "rpctypes.h"
#include "errors.h"
#include "database.h"
#include "brdcast.h"
#include "listener.h"

extern int gSystemInitialized;
#ifdef _USRDLL
void TineStartCycler(void);
#endif

mexListener **MexListenerTable = NULL;
int nListeners = 0;
int defaultPollingMode = CM_POLL;

void listenerCb(int id,int cc)
{
  int mlid = id - 1;
  mexListener *ml = MexListenerTable[mlid];
  if (ml == NULL) return;
  if (!ml->isGlobal)
  {
    if (ml->idleTimeInSeconds++ > ALLOWED_IDLE_TIME) 
    {
      stopListener(ml);
      return;
    }
    ml->timestamp = getCurrentDataTimeStamp(ml->lid);
  }
  else
  {
#   ifndef TINE32_DLL
    ml->timestamp = (double)NGlobalList[ml->lid]->time;
#   else
    ml->timestamp = (double)time(NULL);
#   endif
  }
  if (ml->nf != NULL) ml->nf(ml->nid);
  return;
}
mexListener *startListener(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,int pollrate,int flags)
{
  int i,lid,siz,mode=defaultPollingMode|CM_WAIT;
  mexListener *ml=NULL;

  if (MexListenerTable == NULL)
  {
    if ((MexListenerTable=(mexListener **)SystemCalloc(MAX_NUM_LISTENERS,sizeof(mexListener *))) == NULL)
    {
      printf("Cannot allocate listener table !\n");
      return NULL;
    }
  }
# ifdef USE_QUERY_INFO
  XPropertyQueryStruct xpqs;
  siz = 1;
  if (GetDevicePropertyInformationX(devName,devProperty,&xpqs,&siz) == 0)
  {
    if (dout->dArrayLength < xpqs.prpSize) dout->dArrayLength = xpqs.prpSize;
  }
# endif
  siz = dout->dArrayLength * fmtsizeof(dout->dFormat);
  if (dout->dFormat == CF_SPECTRUM) siz += 96;
  if ((ml=(mexListener *)calloc(1,sizeof(mexListener))) == NULL) return NULL;
  if ((ml->outbuf=(BYTE *)calloc(siz,1)) == NULL) 
  {
    free(ml);
    return NULL;
  }
  ml->dout.dArrayLength = dout->dArrayLength;
  ml->dout.dFormat = dout->dFormat;
  memcpy(ml->dout.dTag,dout->dTag,8);
  ml->dout.data.vptr = ml->outbuf;
  if (din != NULL && (siz=din->dArrayLength * fmtsizeof(din->dFormat)) > 0)
  {
    if (din->dFormat == CF_SPECTRUM) siz += 96;
    if ((ml->inbuf=(BYTE *)calloc(siz,1)) == NULL) 
    {
      if (ml->outbuf != NULL) free(ml->outbuf);
      free(ml);
      return NULL;
    }
    memcpy(ml->din.data.vptr,din->data.vptr,siz);
    ml->din.dArrayLength = din->dArrayLength;
    ml->din.dFormat = din->dFormat;
    memcpy(ml->din.dTag,din->dTag,8);
  }
  else
  {
    ml->din.dFormat = CF_NULL;
    ml->din.dArrayLength = 0;
    ml->din.data.vptr = NULL;
    memset(ml->din.dTag,0,8);
  }
  strncpy(ml->devName,devName,64);
  strncpy(ml->devProperty,devProperty,32);

  for (i=0; i<MAX_NUM_LISTENERS && MexListenerTable[i] != NULL; i++);
  if (i == MAX_NUM_LISTENERS) return NULL;
  ml->mlid = i;
  MexListenerTable[i] = ml;

  if (isGlobal(devName))
  {
    lid = recvNetGlobalEx(devProperty,&ml->dout,listenerCb,ml->mlid+1);
    ml->isGlobal = -1;
  }
  else
  {
    if (flags > 0) mode |= flags;
    lid = AttachLinkEx(devName,devProperty,&ml->dout,&ml->din,CA_READ,pollrate,listenerCb,mode,ml->mlid+1);
    ml->isGlobal = 0;
  }
  if (lid < 0)
  {
    if (ml->outbuf != NULL) free(ml->outbuf);
    if (ml->inbuf != NULL) free(ml->inbuf);
    free(ml);
    MexListenerTable[i] = NULL;
    return NULL;
  }
  if (ml->dout.dArrayLength != 0)
  {
    memcpy(dout->data.vptr,ml->outbuf,dout->dArrayLength*fmtsizeof(dout->dFormat));
  }
  ml->lid = lid;
  ml->pollingRate = pollrate;
  nMexListeners++;
  return ml;
}
int stopListener(mexListener *ml)
{
  int mlid;
  mexListener *m = NULL;
  if (ml == NULL) return non_existent_elem;
  Cancel(ml->lid);
  mlid = ml->mlid;
  if (MexListenerTable[mlid] == NULL) return not_running;
  if (ml->outbuf != NULL) free(ml->outbuf);
  if (ml->inbuf != NULL) free(ml->inbuf);
  free(ml);
  MexListenerTable[mlid] = NULL;
  nMexListeners--;
  return 0;
}
int nMexListeners = 0;
mexListener *getListener(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,int rate)
{
  mexListener *ml=NULL;
  UINT32 dsiz;
  int i;

  if (MexListenerTable == NULL) return NULL;
  for (i=0; i<nMexListeners; i++)
  {
    if ((ml=MexListenerTable[i]) == NULL) continue;
    if (strncmp(devName,ml->devName,64)) continue;
    if (strlen(devProperty) > 0 && strncmp(devProperty,ml->devProperty,32)) continue;
    if (dout != NULL)
    {
      if (dout->dFormat != ml->dout.dFormat) continue;
      if (dout->dArrayLength != ml->dout.dArrayLength) continue;
    }
    if (din != NULL && ml->din.dArrayLength != 0)
    {
      if (din->dFormat != ml->din.dFormat) continue;
      if (din->dArrayLength != ml->din.dArrayLength) continue;
      dsiz = ml->din.dArrayLength*fmtsizeof(ml->din.dFormat);
      if (din->dFormat == CF_SPECTRUM) dsiz += 96;
      if (memcmp(din->data.vptr,ml->inbuf,dsiz)) continue;
    }
    break;
  }
  if (i < nMexListeners)
  {
    if (ml != NULL && rate != 0 && rate < ml->pollingRate)
    {  /* force re-start of listener at the faster rate */
       stopListener(ml);
       return NULL;
    }
    return ml;
  }
  return NULL;  
}
int getDataFromListener(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,double *dTime,int rate,int flags)
{
  mexListener *ml;
  UINT32 siz;

  if (devName == NULL || devProperty == NULL || dout == NULL) return argument_list_error;
  if ( (ml=getListener(devName,devProperty,dout,din,rate)) == NULL && 
       (ml=startListener(devName,devProperty,dout,din,rate,flags)) == NULL )
  {
    return un_allocated;
  }
  if (dout->dFormat != ml->dout.dFormat) return illegal_format;
  ml->idleTimeInSeconds = 0;
  dout->dTimeStamp = ml->timestamp;
  siz = MIN(dout->dArrayLength,ml->dout.dArrayLength)*fmtsizeof(dout->dFormat);
  if (dout->dFormat == CF_SPECTRUM) siz += 96;
  if (dout->data.vptr != ml->outbuf) memcpy(dout->data.vptr,ml->outbuf,siz);
  if (dTime != NULL) *dTime = dout->dTimeStamp;
  return ml->status;
}
int getPropertyInfoFromListener(char *dev, char *prp, int *siz, int *fmt)
{
  mexListener *ml;

  if ((ml=getListener(dev,prp,NULL,NULL,0)) != NULL)
  {
    if (siz != NULL) *siz = ml->dout.dArrayLength;
    if (fmt != NULL) *fmt = ml->dout.dFormat;
    return 0;
  }
  return un_allocated;
}
int isGlobal(char *devName)
{
  if (strstr(devName,"GLOBQRY") != NULL) return -1;
  if (strstr(devName,GLOBALSRV) != NULL) return -1;
  return 0;
}
int isMetaCall(char *devName,char *devProperty)
{
  if (strstr(devName,"HISTORY") != NULL) return -1;
  if (strstr(devProperty,".HIST") != NULL) return -1;
  if (strstr(devProperty,".EGU") != NULL) return -1;
  if (strstr(devProperty,".NAM") != NULL) return -1;
  if (strstr(devProperty,".DESC") != NULL) return -1;
  return 0;
}
int isQueryProperty(char *devProperty)
{
  if (strstr(devProperty,"PROPS") != NULL) return -1;
  if (strstr(devProperty,"PROPERTIES") != NULL) return -1;
  if (strstr(devProperty,"DEVICES") != NULL) return -1;
  if (strstr(devProperty,"TAGS") != NULL) return -1;
  if (strstr(devProperty,"CONTEXTS") != NULL) return -1;
  if (strstr(devProperty,"SUBSYSTEMS") != NULL) return -1;
  if (strstr(devProperty,"USERS") != NULL) return -1;
  if (strstr(devProperty,"IPNETS") != NULL) return -1;
  if (strstr(devProperty,"CLIENTS") != NULL) return -1;
  if (strstr(devProperty,"CONTRACTS") != NULL) return -1;
  if (strstr(devProperty,"HISTORIES") != NULL) return -1;
  if (strstr(devProperty,"LOGFILE") != NULL) return -1;
  if (strstr(devProperty,"LOGDEPTH") != NULL) return -1;
  if (strstr(devProperty,"STRUCTFORMAT") != NULL) return -1;
  if (strncmp(devProperty,"SRV",3) == 0) return -1;
  if (strstr(devProperty,"APPDATE") != NULL) return -1;
  if (strstr(devProperty,"APPVERSION") != NULL) return -1;
  return 0;
}
#endif /* NO_DOXYGEN */
EXPORT_B int setDefaultListenerMode(int mode)
{
  if (mode > 0) defaultPollingMode = mode;
  return defaultPollingMode;
}

/* alsnExecLinkEx() */
/**
 * \brief Executes a synchronous call within an asynchronous static listener.
 * 
 * Asynchronous/Synchronous data exchange. For query calls (for instance to obtain a property
 * list), calls to the local history server, or calls with a timeout parameter less than 50 
 * milliseconds, the call reverts to ExecLinkEx(), i.e. synchronous data transfer.  Otherwise,
 * the initial call starts an asynchronous static listener for the selected device name and
 * property, polling at the rate determined by the timeout parameter. 
 * The initial call waits for the first data set to be returned before it completes.  Subsequent
 * calls retrieve data from the listener's buffer and complete immediately.  If no subsequent
 * calls are made within a 5 minute interval, the asynchronous link is cancelled.
 *
 * \param devName is the full device name (/\<Context\>/\<Server\>/\<Device\>) of the device to contact.
 * For example: "/HERA/BPM/WL167". 
 * \param devProperty is the device property requested, for example "ORBIT.X"
 * \param dout is a pointer to the output data set, that is, the data set to be returned from the server 
 * to the client.  This should refer to a DTYPE object giving the data size, format, and tag (if any)
 * and reference to the data of the output data set.
 * \param din is a pointer to the input data set, that is, the data set to be sent to the server from the 
 * client.  This should refer to a DTYPE object giving the data size, format, and tag (if any)
 * and reference to the data of the input data set.
 * \param access is the data access mode.  This can be any combination of access codes ORed together (CA_READ,
 * CA_WRITE, CA_CONNECT, etc.).
 * \param timeout the timeout period in milliseconds for the call to complete.  When it is known a priori 
 * that a call takes a long time to complete than you can set this value accordingly.  This parameter
 * has a particular meaning when used in the static listener routines.  Namely, it is also used to
 * determine the polling interval for the asynchronous listener which is attached to the call.  You
 * can force the call to execute synchronously without starting a listener by passing a value less
 * than 50 for this parameter.
 * \param dTimeStamp will be filled with the data time stamp corresponding to the requested data.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \see EXECLINK(), EXECLINKEx(), DTYPE
 */
EXPORT_B int alsnExecLinkEx(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access,long timeout,double *dTimeStamp)
{
  int cc,retry=2,len,accessflags=0;
  PropertyQueryStruct svrProp;
  DTYPE lcldin, *dinptr;

# ifdef _USRDLL
  if (!gSystemInitialized) TineStartCycler();
# endif
  if (din != NULL) 
  {
    dinptr = din;
  }
  else
  {
    lcldin.dArrayLength = 0;
    lcldin.dFormat = CF_NULL;
    dinptr = &lcldin;
  }
  if (isGlobal(devName))
  {
    if (dout == NULL) return argument_list_error;
    if (!strncmp(devProperty,"NPROPS",6))
    {
      dout->data.sptr[0] = 1;
      return 0;
    }
    if (!strncmp(devProperty,"PROPS",5))
    {
      if ((len=dinptr->dArrayLength) > 31) len = 31; 
      strncpy(svrProp.prpName,dinptr->data.cptr,len); svrProp.prpName[len] = 0;
      svrProp.prpSize = 1;
      if (strstr(dinptr->data.cptr,"OPTICS") || strstr(dinptr->data.cptr,"STATE"))
      {
        svrProp.prpFormat = BFMT(CF_NAME16);
      }
      else
      {
        svrProp.prpFormat = BFMT(CF_FLOAT);
      } 
      memcpy(dout->data.vptr,&svrProp,sizeof(PropertyQueryStruct));
      return 0;
    }
    /* a real global is handled within getDataFromListener */
  }
  if (!(access&CA_WRITE) && !isQueryProperty(devProperty) && 
      !isMetaCall(devName,devProperty) && timeout >= 50)
  {
    if (access & CA_NETWORK) accessflags = CM_NETWORK;
    cc = getDataFromListener(devName,devProperty,dout,dinptr,dTimeStamp,(int)timeout,accessflags);
    if (cc != use_stream_transport) return cc;
  }
  while (retry--)
  {
    if (timeout < 10) timeout *= 1000;
    cc = ExecLinkEx(devName,devProperty,dout,dinptr,access,(UINT16)timeout);
    if (dTimeStamp != NULL) *dTimeStamp = getDataTimeStamp();
    if (cc & CE_SENDDATA) cc = 0;
    if (cc == 0) break;
    if (timeout >= 1000) break;
  }
  return cc;
}
/* alsnExecLink() */
/**
 * \brief Executes a synchronous call within an asynchronous static listener.
 * 
 * Asynchronous/Synchronous data exchange. For query calls (for instance to obtain a property
 * list), calls to the local history server, or calls with a timeout parameter less than 50 
 * milliseconds, the call reverts to ExecLinkEx(), i.e. synchronous data transfer.  Otherwise,
 * the initial call starts an asynchronous static listener for the selected device name and
 * property, polling at the rate determined by the timeout parameter. 
 * The initial call waits for the first data set to be returned before it completes.  Subsequent
 * calls retrieve data from the listener's buffer and complete immediately.  If no subsequent
 * calls are made within a 5 minute interval, the asynchronous link is cancelled.
 *
 * \param devName is the full device name (/\<Context\>/\<Server\>/\<Device\>) of the device to contact.
 * For example: "/HERA/BPM/WL167". 
 * \param devProperty is the device property requested, for example "ORBIT.X"
 * \param dout is a pointer to the output data set, that is, the data set to be returned from the server 
 * to the client.  This should refer to a DTYPE object giving the data size, format, and tag (if any)
 * and reference to the data of the output data set.
 * \param din is a pointer to the input data set, that is, the data set to be sent to the server from the 
 * client.  This should refer to a DTYPE object giving the data size, format, and tag (if any)
 * and reference to the data of the input data set.
 * \param access is the data access mode.  This can be any combination of access codes ORed together (CA_READ,
 * CA_WRITE, CA_CONNECT, etc.).
 * \param timeout the timeout period in milliseconds for the call to complete.  When it is known a priori 
 * that a call takes a long time to complete than you can set this value accordingly.  This parameter
 * has a particular meaning when used in the static listener routines.  Namely, it is also used to
 * determine the polling interval for the asynchronous listener which is attached to the call.  You
 * can force the call to execute synchronously without starting a listener by passing a value less
 * than 50 for this parameter.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \see EXECLINK(), EXECLINKEx(), DTYPE
 */
EXPORT_B int alsnExecLink(char *devName,char *devProperty,
                          DTYPE *dout,DTYPE *din,short access,long timeout)
{
  return alsnExecLinkEx(devName,devProperty,dout,din,access,timeout,NULL);
}

EXPORT_B short tputgetArrayEx(char *fullDeviceNameAndProperty, short access,
                                                short putdatafmt,void *putdata,int putsiz,
                                                short getdatafmt, void *getdata, int getsiz,double *dTimeStamp)
{
  DTYPE dout, din;
  char *c;
  static char sbuf[256],dev[64],prp[48];
  int cc, i, rate = 1000;
  static int inside = FALSE;

  if (inside) return operation_busy;
  inside = TRUE;
  for (c=fullDeviceNameAndProperty; *c == ' ' && *c != 0; c++); // remove leading blanks
  strncpy(sbuf,c,256);
  if ((c=strchr(sbuf,'@')) != NULL)
  {
    *c++ = 0; // peel it off before we start tokenizing
    strupr(c);
    if (strstr(c,"NETWORK") != NULL || strstr(c,"MCAST") != NULL || strstr(c,"MULTICAST") != NULL)
      access |= CA_NETWORK;
    rate = atoi(c);
  }
  if ((c=strchr(sbuf,'[')) == NULL) return illegal_property;
  *c++ = 0; while (*c == ' ' && *c != 0) c++; // remove leading blanks
  strncpy(prp,c,32);  prp[32] = 0; // record property
  if ((c=strchr(prp,']')) != NULL) *c = 0;
  for (i=strlen(prp)-1; i > 0 && prp[i] == ' '; i--) prp[i] = 0; // remove trailing blanks
	strncpy(dev,sbuf,64);
  for (i=strlen(dev)-1; i > 0 && dev[i] == ' '; i--) dev[i] = 0; // remove trailing blanks
  
  din.dFormat = putdatafmt;
  din.dArrayLength = putsiz;
  memset(din.dTag,0,8);
  din.data.vptr = putdata;

  dout.dFormat = getdatafmt;
  dout.dArrayLength = getsiz;
  memset(dout.dTag,0,8);
  dout.data.vptr = getdata;

  cc = alsnExecLinkEx(dev,prp,&dout,&din,access,1,dTimeStamp);
  inside = FALSE;
  return cc;
}
EXPORT_B short tputgetArray(char *fullDeviceNameAndProperty, short access,
                                                short putdatafmt,void *putdata,int putsiz,
                                                short getdatafmt, void *getdata, int getsiz)
{
  return tputgetArrayEx(fullDeviceNameAndProperty,access,putdatafmt,putdata,putsiz,getdatafmt,getdata,getsiz,NULL);
}
EXPORT_B short tputString(char *fullDeviceNameAndProperty,char *putdata)
{
  char *c,sbuf[256];
  int len = 0;

  if (fullDeviceNameAndProperty == NULL || putdata == NULL) return argument_list_error;
  if ((len=strlen(putdata)) == 0) return argument_list_error;
  strncpy(sbuf,fullDeviceNameAndProperty,256);
  if ((c=strchr(sbuf,'@')) != NULL) *c = 0; // peel it off and replace it
  strcat(sbuf,"@1"); // make sure StringArray calls don't start a listener !
  return tputgetArray(sbuf,CA_WRITE|CA_READ,CF_TEXT,putdata,len,CF_NULL,NULL,0);
}
EXPORT_B short tputDoubleArray(char *fullDeviceNameAndProperty,double *putdata,int putsiz)
{
  void *outdata = NULL;
  int outsiz = 0;
  short outfmt = CF_NULL;

  return tputgetArray(fullDeviceNameAndProperty,CA_WRITE|CA_READ,CF_DOUBLE,putdata,putsiz,outfmt,outdata,outsiz);
}
EXPORT_B short tputSingleArray(char *fullDeviceNameAndProperty,float *putdata,int putsiz)
{
  void *outdata = NULL;
  int outsiz = 0;
  short outfmt = CF_NULL;

  return tputgetArray(fullDeviceNameAndProperty,CA_WRITE|CA_READ,CF_FLOAT,putdata,putsiz,outfmt,outdata,outsiz);
}
EXPORT_B short tputLongArray(char *fullDeviceNameAndProperty,int *putdata,int putsiz)
{
  void *outdata = NULL;
  int outsiz = 0;
  short outfmt = CF_NULL;

  return tputgetArray(fullDeviceNameAndProperty,CA_WRITE|CA_READ,CF_LONG,putdata,putsiz,outfmt,outdata,outsiz);
}
EXPORT_B short tputShortArray(char *fullDeviceNameAndProperty,short *putdata,int putsiz)
{
  void *outdata = NULL;
  int outsiz = 0;
  short outfmt = CF_NULL;

  return tputgetArray(fullDeviceNameAndProperty,CA_WRITE|CA_READ,CF_SHORT,putdata,putsiz,outfmt,outdata,outsiz);
}
EXPORT_B void tDetachNotifier(void (*nf)(int))
{
  mexListener *ml;
  int i;
  for (i=0; i<nMexListeners; i++)
  {
    if ((ml=MexListenerTable[i]) == NULL) continue;
    if (ml->nf == nf) ml->nf = NULL;
  }
}
EXPORT_B short tAttachNotifier(char *fullDeviceNameAndProperty, void (*nf)(int), int nid)
{
  char *c,sbuf[256],dev[64],prp[48];
  int i, rate = 1000;
  mexListener *ml=NULL;

  for (c=fullDeviceNameAndProperty; *c == ' ' && *c != 0; c++); // remove leading blanks
  strncpy(sbuf,c,256);
  if ((c=strchr(sbuf,'@')) != NULL)
  {
    *c++ = 0; // peel it off before we start tokenizing
    rate = atoi(c);
  }
  if ((c=strchr(sbuf,'[')) == NULL) return illegal_property;
  *c++ = 0; while (*c == ' ' && *c != 0) c++; // remove leading blanks
  strncpy(prp,c,32);  prp[32] = 0; // record property
  if ((c=strchr(prp,']')) != NULL) *c = 0;
  for (i=strlen(prp)-1; i > 0 && prp[i] == ' '; i--) prp[i] = 0; // remove trailing blanks
	strncpy(dev,sbuf,64);
  for (i=strlen(dev)-1; i > 0 && dev[i] == ' '; i--) dev[i] = 0; // remove trailing blanks

  for (i=0; i<nMexListeners; i++)
  {
    if ((ml=MexListenerTable[i]) == NULL) continue;
    if (strncmp(dev,ml->devName,64)) continue;
    if (strlen(prp) > 0 && strncmp(prp,ml->devProperty,32)) continue;
    break;
  }
  if (ml == NULL || i == nMexListeners) return link_not_open;
  ml->nf = nf;
  ml->nid = nid;
  return 0;
}
EXPORT_B short tgetArray(char *fullDeviceNameAndProperty,short datafmt,void *dataout,int buffersiz)
{
  DTYPE dout;
  char *c,sbuf[256],dev[64],prp[48];
  int i, rate = 1000;
  short access = CA_READ;

  for (c=fullDeviceNameAndProperty; *c == ' ' && *c != 0; c++); // remove leading blanks
  strncpy(sbuf,c,256);
  if ((c=strchr(sbuf,'@')) != NULL)
  {
    *c++ = 0; // peel it off before we start tokenizing
    strupr(c);
    if (strstr(c,"NETWORK") != NULL || strstr(c,"MCAST") != NULL || strstr(c,"MULTICAST") != NULL)
      access |= CA_NETWORK;
    rate = atoi(c);
  }
  if ((c=strchr(sbuf,'[')) == NULL) return illegal_property;
  *c++ = 0; while (*c == ' ' && *c != 0) c++; // remove leading blanks
  strncpy(prp,c,32);  prp[32] = 0; // record property
  if ((c=strchr(prp,']')) != NULL) *c = 0;
  for (i=strlen(prp)-1; i > 0 && prp[i] == ' '; i--) prp[i] = 0; // remove trailing blanks
	strncpy(dev,sbuf,64);
  for (i=strlen(dev)-1; i > 0 && dev[i] == ' '; i--) dev[i] = 0; // remove trailing blanks
  
  dout.dFormat = datafmt;
  dout.dArrayLength = buffersiz;
  memset(dout.dTag,0,8);
  dout.data.vptr = dataout;
  return alsnExecLink(dev,prp,&dout,NULL,access,rate);
}
/* tgetSingleArray() */
/**
 * \brief Executes a synchronous call within an asynchronous static listener.
 * 
 * Asynchronous/Synchronous data exchange. This is a convenience
 * routine which accesses alsnExecLink() explicitly for an array of Single (float) data type
 * without data input.
 *
 * \param fullDeviceNameAndProperty is the full device name 
 * appended with the requested property parsed according to 
 * '/\<Context\>/\<Server\>/\<Device\>[\<Property\>]@\<rate\>'.
 * For example: "/HERA/BPM/WL167[ORBIT.X]@500". If the '@\<rate\>' is omitted, a polling
 * rate of 1000 milliseconds is assumed.
 * \param dataout is a reference to the float array to receive the output data.
 * \param buffersiz is the size of (i.e. number of elements in) the float array.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \see alsnExecLink(), vBalsnExecLink().
 */
EXPORT_B short tgetSingleArray(char *fullDeviceNameAndProperty,float *dataout,int buffersiz)
{
  return tgetArray(fullDeviceNameAndProperty,CF_FLOAT,dataout,buffersiz);
}
EXPORT_B short tputgetSingleArray(char *fullDeviceNameAndProperty,float *datain,int insize,float *dataout,int outsize,double *datatime)
{
  return tputgetArrayEx(fullDeviceNameAndProperty,CA_READ,CF_FLOAT,datain,insize,CF_FLOAT,dataout,outsize,datatime);
}
/* tgetDoubleArray() */
/**
 * \brief Executes a synchronous call within an asynchronous static listener.
 * 
 * Asynchronous/Synchronous data exchange. This is a convenience
 * routine which accesses alsnExecLink() explicitly for an array of double data type
 * without data input.
 *
 * \param fullDeviceNameAndProperty is the full device name 
 * appended with the requested property parsed according to 
 * '/\<Context\>/\<Server\>/\<Device\>[\<Property\>]@\<rate\>'.
 * For example: "/HERA/BPM/WL167[ORBIT.X]@500". If the '@\<rate\>' is omitted, a polling
 * rate of 1000 milliseconds is assumed.
 * \param dataout is a reference to the double array to receive the output data.
 * \param buffersiz is the size of (i.e. number of elements in) the array.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \see alsnExecLink(), vBalsnExecLink().
 */
EXPORT_B short tgetDoubleArray(char *fullDeviceNameAndProperty,double *dataout,int buffersiz)
{
  return tgetArray(fullDeviceNameAndProperty,CF_DOUBLE,dataout,buffersiz);
}
EXPORT_B short tputgetDoubleArray(char *fullDeviceNameAndProperty,double *datain,int insize,double *dataout,int outsize,double *datatime)
{
  return tputgetArrayEx(fullDeviceNameAndProperty,CA_READ,CF_DOUBLE,datain,insize,CF_DOUBLE,dataout,outsize,datatime);
}
/* tgetLongArray() */
/**
 * \brief Executes a synchronous call within an asynchronous static listener.
 * 
 * Asynchronous/Synchronous data exchange. This is a convenience
 * routine which accesses alsnExecLink() explicitly for an array of long data type
 * without data input.
 *
 * \param fullDeviceNameAndProperty is the full device name 
 * appended with the requested property parsed according to 
 * '/\<Context\>/\<Server\>/\<Device\>[\<Property\>]@\<rate\>'.
 * For example: "/HERA/BPM/WL167[ORBIT.X]@500". If the '@\<rate\>' is omitted, a polling
 * rate of 1000 milliseconds is assumed.
 * \param dataout is a reference to the long integer array to receive the output data.
 * \param buffersiz is the size of (i.e. number of elements in) the array.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \see alsnExecLink(), vBalsnExecLink().
 */
EXPORT_B short tgetLongArray(char *fullDeviceNameAndProperty,long *dataout,int buffersiz)
{
  return tgetArray(fullDeviceNameAndProperty,CF_LONG,dataout,buffersiz);
}
EXPORT_B short tputgetLongArray(char *fullDeviceNameAndProperty,long *datain,int insize,long *dataout,int outsize,double *datatime)
{
  return tputgetArrayEx(fullDeviceNameAndProperty,CA_READ,CF_LONG,datain,insize,CF_LONG,dataout,outsize,datatime);
}
/* tgetShortArray() */
/**
 * \brief Executes a synchronous call within an asynchronous static listener.
 * 
 * Asynchronous/Synchronous data exchange. This is a convenience
 * routine which accesses alsnExecLink() explicitly for an array of short integer data type
 * without data input.
 *
 * \param fullDeviceNameAndProperty is the full device name 
 * appended with the requested property parsed according to 
 * '/\<Context\>/\<Server\>/\<Device\>[\<Property\>]@\<rate\>'.
 * For example: "/HERA/BPM/WL167[ORBIT.X]@500". If the '@\<rate\>' is omitted, a polling
 * rate of 1000 milliseconds is assumed.
 * \param dataout is a reference to the short integer array to receive the output data.
 * \param buffersiz is the size of (i.e. number of elements in) the array.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \see alsnExecLink(), vBalsnExecLink().
 */
EXPORT_B short tgetShortArray(char *fullDeviceNameAndProperty,short *dataout,int buffersiz)
{
  return tgetArray(fullDeviceNameAndProperty,CF_SHORT,dataout,buffersiz);
}
EXPORT_B short tputgetShortArray(char *fullDeviceNameAndProperty,short *datain,int insize,short *dataout,int outsize,double *datatime)
{
  return tputgetArrayEx(fullDeviceNameAndProperty,CA_READ,CF_SHORT,datain,insize,CF_SHORT,dataout,outsize,datatime);
}
/* tgetStringArray() */
/**
 * \brief Executes a synchronous call within an asynchronous static listener.
 * 
 * Synchronous data exchange. This is a convenience
 * routine which accesses ExecLink() explicitly for an array of NAME32 data type
 * without data input.  This is useful for retrieving a list of properties or devices from a 
 * server for instance.
 *
 * \param fullDeviceNameAndProperty is the full device name 
 * appended with the requested property parsed according to 
 * '/\<Context\>/\<Server\>/\<Device\>[\<Property\>]'.
 * For example: "/HERA/BPM/WL167[PROPERTIES]".
 * \param dataout is a reference to the NAME32 array to receive the output data.
 * \param buffersiz is the size of (i.e. number of elements in) the NAME32 array.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \see alsnExecLink(), vBalsnExecLink().
 */
EXPORT_B short tgetStringArray(char *fullDeviceNameAndProperty,char *dataout,int buffersiz)
{
  int cc, len, i, j;
  char *c,sbuf[256];

  if (buffersiz <=0 || fullDeviceNameAndProperty == NULL || dataout == NULL) return argument_list_error;

  strncpy(sbuf,fullDeviceNameAndProperty,256);
  if ((c=strchr(sbuf,'@')) != NULL) *c = 0; // peel it off and replace it
  strcat(sbuf,"@1"); // make sure StringArray calls don't start a listener !
  if ((c=alloca(buffersiz)) == NULL) return out_of_client_memory;
  if ((cc=tgetArray(sbuf,CF_NAME32,c,buffersiz/32)) != 0) return cc;

  for (j=0, i=0; i<buffersiz && j<buffersiz; j += 32)
  {
    len = MIN(strlen(&c[j]),32);
    strncpy(&dataout[i],&c[j],len);
    i += len;
    dataout[i++] = '\n';
  }
  return 0;
}
EXPORT_B short tgetString(char *fullDeviceNameAndProperty,char *dataout,int buffersiz)
{
  int cc;
  char *c,sbuf[256];

  if (buffersiz <=0 || fullDeviceNameAndProperty == NULL || dataout == NULL) return argument_list_error;

  strncpy(sbuf,fullDeviceNameAndProperty,256);
  if ((c=alloca(buffersiz)) == NULL) return out_of_client_memory;
  if ((cc=tgetArray(sbuf,CF_TEXT,c,buffersiz)) != 0) return cc;
  strncpy(dataout,c,buffersiz);
  return 0;
}
EXPORT_B short tputgetString(char *fullDeviceNameAndProperty,char *datain,char *dataout,int buffersiz)
{
  int cc, inputlength = 0;
  char *c,sbuf[256],*cin = NULL;

  if (buffersiz <=0 || fullDeviceNameAndProperty == NULL || dataout == NULL) return argument_list_error;
  if (datain != NULL) 
  {
    cin = datain;
    inputlength = strlen(cin);
  }
  strncpy(sbuf,fullDeviceNameAndProperty,256);
  if ((c=alloca(buffersiz)) == NULL) return out_of_client_memory;
  if ((cc=tputgetArray(sbuf,CA_READ,CF_TEXT,cin,inputlength,CF_TEXT,c,buffersiz)) != 0) return cc;
  strncpy(dataout,c,buffersiz);
  return 0;
}
EXPORT_B short tputgetStringEx(char *fullDeviceNameAndProperty,char *datain,char *dataout,int buffersiz,double *datatime)
{
  int cc, inputlength = 0;
  char *c,sbuf[256],*cin = NULL;

  if (buffersiz <=0 || fullDeviceNameAndProperty == NULL || dataout == NULL) return argument_list_error;
  if (datain != NULL) 
  {
    cin = datain;
    inputlength = strlen(cin);
  }
  strncpy(sbuf,fullDeviceNameAndProperty,256);
  if ((c=alloca(buffersiz)) == NULL) return out_of_client_memory;
  if ((cc=tputgetArrayEx(sbuf,CA_READ,CF_TEXT,cin,inputlength,CF_TEXT,c,buffersiz,datatime)) != 0) return cc;
  strncpy(dataout,c,buffersiz);
  return 0;
}
