/** \file mexlib.c
 * \brief TINE NETMEX library routines
 *
 */
#include "cprolog.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "rpctypes.h"
#include "servdef.h"
#include "brdcast.h"
#include "database.h"
#include "srvip.h"
#include "netmex.h"            /* NextMex only */

#ifndef NextCount
int NextCount = 0;             /* # of outstanding requests */
#endif

extern int syncCount;
extern void SyncAsync(int id,int cc);
unsigned long ElfHash(const unsigned char *key);
int NextMex(NETKEY* key)
{
  int cc;
  DTYPE dout;
  char devname[32];

  sprintf(devname,"%s/#%d",key->Eqpname,key->EqNo);  
  dout.dArrayLength = key->no_el;
  dout.dFormat = key->format;
  dout.data.vptr = key->data;

  cc = ExecLink(devname,key->Property,&dout,NULL,CA_READ);
  if (!cc) key->TimeStamp = (int) dout.dTimeStamp;
  return cc;
}
/* NetmexGetEx() */
/**
 * \brief The extended version of NetmexGet(). Synchronous.
 *  Retrieves a NETMEX keyword.
 * 
 * This call differs from NetmexGet() in that it provides parameters for specifying
 * input information when the Keyword in question requires input (seldem the case).
 * This is a synchronous call to obtain the NETMEX Keyword information for the input
 * Keyword.
 *
 * \param key (input/output) is the NETMEX key for which the information is to be
 *  acquired.  The parameter type NETKEY has usually already been afixed in the appropriate
 *  netmex header file.  A successful call to NetmexGetEx() will then refresh the data
 *  contents of the keyword syncrhonously.
 * \param inbuf is a pointer to a buffer holding the input data required by the keyword.
 * \param inbuffmt is the input TINE data type.
 * \param inbuflen is the size of the input buffer (number of elements).
 *  
 * \return 0 upon success, otherwise a TINE completion code.
 *
 * \see NetmexGet, LinkNetmex, LinkNetmexEx
 */
int NetmexGetEx(NETKEY* key,void *inbuf,int inbuffmt,int inbuflen)
{
  int cc;
  DTYPE dout,din,*dinptr = NULL;
  char devname[52];

  if (strlen(key->DevName))
    sprintf(devname,"%s/%s",key->Eqpname,key->DevName);
  else
    sprintf(devname,"%s/#%d",key->Eqpname,key->EqNo);

  dout.dArrayLength = key->no_el;
  dout.dFormat = key->format;
  dout.data.vptr = key->data;

  if (inbuf && inbuflen > 0)
  {
    din.dArrayLength = inbuflen;
    din.dFormat = inbuffmt;
    din.data.vptr = inbuf;
    dinptr = &din;
  }
  return ExecLink(devname,key->Property,&dout,dinptr,CA_READ);
}
/* NetmexGet() */
/**
 * \brief The extended version of NetmexGet(). Synchronous.
 *  Retrieves a NETMEX keyword.
 * 
 * This is a synchronous call to obtain the NETMEX Keyword information for the input
 * Keyword.
 *
 * \param key (input/output) is the NETMEX key for which the information is to be
 *  acquired.  The parameter type NETKEY has usually already been afixed in the appropriate
 *  netmex header file.  A successful call to NetmexGet() will then refresh the data
 *  contents of the keyword syncrhonously.
 *  
 * \return 0 upon success, otherwise a TINE completion code.
 *
 * \see NetmexGetEx, LinkNetmex, LinkNetmexEx
 */
int NetmexGet(NETKEY* key)
{
  return NetmexGetEx(key,NULL,0,0);
}
NETKEY *NextList[MAX_ENTRIES];
void NextAsync(int id,int cc)
{
  NETKEY *key;
  if (NGdebug > 1) printf("Async CallBack %d\n",id);

  if ((key=NextList[id]) == NULL) return;
  key->Ret_Code = cc;
  if (!cc) key->TimeStamp = (int) connTable[id]->dtimestamp;
  if (NextList[id]->fcn != NULL)
    NextList[id]->fcn(NextList[id]->Property,cc);
}
int LinkNextMex(NETKEY *key,int PollRate,void (*function)(char *,int),short mode)
{
  DTYPE dout;
  int id;
  char devname[32];

  sprintf(devname,"%s/#%d",key->Eqpname,key->EqNo);  
  dout.dArrayLength = key->no_el;
  dout.dFormat = key->format;
  dout.data.vptr = key->data;

  if ((id=AttachLink(devname,key->Property,&dout,NULL,CA_READ,PollRate,NextAsync,REGISTER)) < 0)
    return -id;
  connTable[id]->mode = mode;
  connTable[id]->heartbeat = SUB_PENDING;

  key->fcn = function;
  NextList[id] = key;

  return 0;
}
/* LinkNetmexEx() */
/**
 * \brief Extended version of LinkNetmex(). Asynchronous NETMEX data acquisition. 
 *  Refreshes the input keyword and calls the input callback function.
 *  
 * This call differs from LinkNetmex() in that it provides parameters for specifying
 * input information when the Keyword in question requires input (seldem the case).
 * This is an asynchronous mechanism to obtain the NETMEX Keyword information for the input
 * Keyword, and is the preferred method of obtaining NETMEX parameters.
 *
 * \param key (input/output) is the NETMEX key for which the information is to be
 *  acquired.  The parameter type NETKEY has usually already been afixed in the appropriate
 *  netmex header file.  A successful call to NetmexGet() will then refresh the data
 *  contents of the keyword syncrhonously.
 * \param PollRate is the polling interval in milliseconds passed along the the corresponding
 *  server.
 * \param function is the callback function to be called when new data arrive.  It must have
 *  the prototype function(char *keywordName,int completionCode). 
 * \param mode is the TINE access mode to be used in the data acquisition, typically one of
 *  CM_POLL or CM_REFRESH.
 * \param inbuf is a pointer to a buffer holding the input data required by the keyword.
 * \param inbuffmt is the input TINE data type.
 * \param inbuflen is the size of the input buffer (number of elements).
 *  
 * \return 0 upon success, otherwise a TINE completion code.
 *
 * \see LinkNetmex, NetmexGet, NetmexGetEx
 */
int LinkNetmexEx(NETKEY *key,int PollRate,void (*function)(char *,int),
               short mode,void *inbuf,int inbuffmt,int inbuflen)
{
  DTYPE dout,din,*dinptr=NULL;
  int id;
  char devname[32];

  if (strlen(key->DevName))
    sprintf(devname,"%s/%s",key->Eqpname,key->DevName);
  else
    sprintf(devname,"%s/#%d",key->Eqpname,key->EqNo);
    
  dout.dArrayLength = key->no_el;
  dout.dFormat = key->format;
  dout.data.vptr = key->data;

  if (inbuf && inbuflen > 0)
  {
    din.dArrayLength = inbuflen;
    din.dFormat = inbuffmt;
    din.data.vptr = inbuf;
    dinptr = &din;
  }
  if ((id=AttachLink(devname,key->Property,&dout,dinptr,CA_READ,PollRate,NextAsync,REGISTER)) < 0)
    return -id;
  connTable[id]->mode = mode;
  connTable[id]->heartbeat = SUB_PENDING;

  key->fcn = function;
  NextList[id] = key;

  return 0;
}
/* LinkNetmex() */
/**
 * \brief Asynchronous NETMEX data acquisition. Refreshes the input keyword and calls
 *  the input callback function.
 *  
 * This is an asynchronous mechanism to obtain the NETMEX Keyword information for the input
 * Keyword, and is the preferred method of obtaining NETMEX parameters.
 *
 * \param key (input/output) is the NETMEX key for which the information is to be
 *  acquired.  The parameter type NETKEY has usually already been afixed in the appropriate
 *  netmex header file.  A successful call to NetmexGet() will then refresh the data
 *  contents of the keyword syncrhonously.
 * \param PollRate is the polling interval in milliseconds passed along the the corresponding
 *  server.
 * \param function is the callback function to be called when new data arrive.  It must have
 *  the prototype function(char *keywordName,int completionCode). 
 * \param mode is the TINE access mode to be used in the data acquisition, typically one of
 *  CM_POLL or CM_REFRESH.
 *  
 * \return 0 upon success, otherwise a TINE completion code.
 *
 * \see LinkNetmexEx, NetmexGet, NetmexGetEx
 *
 * \b Example
 *
 * \include eg_LinkNetmex.c
 *
 */
int LinkNetmex(NETKEY *key,int PollRate,void (*function)(char *,int),short mode)
{
  return LinkNetmexEx(key,PollRate,function,mode,NULL,0,0);
}

/*==========================LEGACY NextMex Packing with UDP subscriptions======================*/

#define NMAX_PACKS 20
int nmpacks =0;

typedef struct {
   char Name[16];
   int size;
   NETKEY* key_id[50];
   int connTab[50];
}  NMPACKLIST;

NMPACKLIST nmpackList[NMAX_PACKS];

int InitPack(char *Name, NETKEY* key)   {   /* UDP version 28.02.95 */
  int i,id,n ;

  for (n=0; n<nmpacks && n<NMAX_PACKS; n++) if (!strcmp(Name,nmpackList[n].Name)) break;
  if (n==NMAX_PACKS) return -out_of_client_memory;
  if (n == nmpacks) {
     strcpy(nmpackList[nmpacks].Name,Name);    /* wasn't in packing list.  New entry! */
     nmpackList[nmpacks].size = 0;
  }
  for (i=0; i<nmpackList[n].size; i++) if (key==nmpackList[n].key_id[i]) {           /* test for duplicates */
    if (NGdebug) printf("duplicate %s  %s\n",key->Eqpname,key->Property);
    return -1;
 }
  nmpackList[n].key_id[nmpackList[n].size] = key;
  nmpackList[n].size++;
  if (n == nmpacks) nmpacks++;
  return 0;
}

int NextPack(char *Name) {      /* UDP version 28.02.95 */
  int id,n,packid= -1;
  NETKEY *key;
  DTYPE dout;
  char devname[32];

  NextCount=0;   /* ~global */

  for (n=0; n<nmpacks && n<NMAX_PACKS; n++) if (!strcmp(Name,nmpackList[n].Name)) packid=n;
  if (packid<0) return -1;
  for (n=0; n<nmpackList[packid].size; n++) 
  {
     key = nmpackList[packid].key_id[n];

     sprintf(devname,"%s/#%d",key->Eqpname,key->EqNo);  
     dout.dArrayLength = key->no_el;
     dout.dFormat = key->format;
     dout.data.vptr = key->data;

     if ((id=AttachLink(devname,key->Property,&dout,NULL,CA_READ,1000,NextAsync,SINGLE)) >= 0)
     {
       connTable[id]->dataOut = (char*) key->data;
       nmpackList[packid].connTab[n] = id;
       NextCount++;
     }
     else 
     {
       nmpackList[packid].connTab[n] = id;
       if (NGdebug) printf("Sub Fails %s %s %d\n",key->Eqpname,key->Property,id);
     }
  }
  for (syncCount=NextCount; syncCount>0;) WaitForSyncCount(syncCount);

  for (n=0; n<nmpackList[packid].size; n++) 
  {
    key = nmpackList[packid].key_id[n];
    id=nmpackList[packid].connTab[n];
    if (id<0) {  key->Ret_Code = -id; }
    else {
      key->Ret_Code = connTable[id]->dataFlg;
      key->TimeStamp = (int)connTable[id]->dtimestamp;
      key->Status = 0;     key->Version = 0;
    }
  }
  return 0;
}
