#ifndef _RPCDEFS_H_
#define _RPCDEFS_H_

#include "project.def"
#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif

#ifndef EXPORT
# define EXPORT
#endif

#ifndef CNTXT_PREFIX
#define CNTXT_PREFIX "\\\\"
#endif

#ifndef DELIMITER
#define DELIMITER "/\\"
#endif

#ifndef RTOL
/** 
 * \brief Direction: Remote to Local flag
 */
#define RTOL    0
/**
 * \brief  Direction: Local to Remote flag
 */  
#define LTOR    1  
#endif


#ifndef MAX
/** 
 * \brief Returns the greater of two variables.
 */
#define MAX(a,b)  ((a) > (b) ? a : b)
#endif

#ifndef MIN
/**
 * \brief Returns the lesser of two variables 
 */
#define MIN(a,b)  ((a) < (b) ? a : b) 
#endif

#ifndef BYTE
#define BYTE unsigned char
#endif
#ifndef UINT8
#define UINT8 unsigned char
#endif
#ifndef UINT16
#define UINT16 unsigned short
#endif
#ifndef UINT32
#define UINT32 unsigned int
#endif
#ifndef SINT32
#define SINT32 int
#endif
#ifndef WINNT
#ifndef WORD
#define WORD UINT16
#endif
#ifndef DWORD
#define DWORD UINT32
#endif
#endif

#ifndef TRUE
/**
 * \brief Definition of logical True 
 */
#define TRUE -1
#endif

#ifndef FALSE
/**
 * \brief Definition of logical False 
 */
#define FALSE 0
#endif

/**
 * \page tinetypes TINE Data Formats and Representations
 *
 * TINE offers a wide range of system recognized data types as shown in the \ref DataTypes "list"
 * below.
 *
 * \section tineCompoundDataTypes Compound Data Types
 *
 * In addition to the fundemental data types (e.g. short, float, double, etc.) there are a number
 * of compound format types.  For instance, the fundemental types can be paired with a long integer
 * given rise to data formats such as FLTINT (a float value plus a long integer value).  Such data
 * types are important when transfering archived data for example, where the returned data consists
 * of a value plus a UTC timestamp.  There are also a set of fix-length string format types, such
 * as NAME32, which deliver (an array of) ascii character strings of the given length.  There
 * are also a range of data triplets and quadruplets (such as INTFLTINT or NAME16FI or DADDRESS)
 * which are useful in certain circumstances.  The most complex format type offered systematically
 * is DUSTRING, offering a quadruplet of data types plus an 80-character string, and a
 * DSPECTRUM, offering a DUSTRING plus an array of up to 1024 float values.
 *
 * \section tineTaggedStructures User-defined Data Types
 *
 * In addition to the data type offered, TINE allows user-defined data structures.
 *
 * The issue here is the 
 * recognition of data sets by client and server over the network. If the client asks for 
 * N of something, both sides must know how big (how many bytes) that something is.  Furthermore 
 * both sides must know how to make sure all the bytes end up in the right order for the 
 * platform in question.  This is guaranteed to work if you stick to data of a standard format 
 * or one of the control system types.  If you want to define your own data structure type and 
 * use it, things are a bit more complicated but can also work just fine if you pay attention 
 * on both the client and server ends.
 *
 * The first step on each end is to register the structure.  You do this at initialization time 
 * by declaring a structure tag name and filling in the sizes and locations of the individual 
 * elements.  Then when a client sends or asks for a structure and it sends the structure tag 
 * along with the request, the server knows all about the structure and how to handle it.  As 
 * an example, well look at another snippet from the sinsrv.c routine.
 * 
 * Using tagged structures also has the side benefit of allowing queries from the client-side 
 * to determine the structure of the structure.  Essentially a client can discover all it needs
 * to know as to how to handle the structure.
 *
 * \include eg_taggedStruct.c
 * 
 * The registerStructs() function above is to be called in an initialization routine on both 
 * the client and server end.  The client can then simply pass a pointer to (an array of ) the 
 * structure in question.  The server will then see (an array of ) the structure in question.  
 *
 * The server equipment function might do something like this:
 *
 * \include eg_taggedStruct2.c
 *
 * \anchor DataTypes
 * List of format types
 * - \b CF_TEXT (\e alias \b CF_CHAR)
 *   - typedef \e c: \b char  \e vb: \b String  \e java: \b char
 *   - size: 1 byte
 *   - In Visual Basic beware of unicode character representation!
 * - \b CF_BYTE (\e alias \b CF_UINT8)
 *   - typedef \e c: \b BYTE  \e vb: \b Byte \e java: \b byte
 *   - size: 1 byte
 * - \b CF_BIT 
 *   - typedef \e c: \b BYTE  \e vb: \b Byte \e java: \b byte
 *   - size: 1 bit
 *   - BIT format returns a 0 or 1 and is carried as a 4-byte integer
 * - \b CF_SHORT (\e alias \b CF_INTEGER, \b CF_UINT16)
 *   - typedef \e c: \b short  \e vb: \b Integer  \e java: \b short
 *   - size: 2 bytes
 * - \b CF_LONG (\e alias \b CF_INT4, \b CF_UINT32)
 *   - typedef \e c: \b long  \e vb: \b Long  \e java: \b int
 *   - size: 4 bytes
 * - \b CF_BOOLEAN
 *   - typedef \e c: \b long  \e vb: \b Long  \e java: \b int
 *   - size: 4 bytes
 * - \b CF_FLOAT   
 *   - typedef \e c: \b float  \e vb: \b Single  \e java: \b float
 *   - size: 4 bytes
 * - \b CF_DOUBLE   
 *   - typedef \e c: \b double  \e vb: \b Double  \e java: \b double
 *   - size: 8 bytes
 * - \b CF_STRUCT 
 *   - typedef \e c: \b struct  \e vb: \b Type  \e java: \b class
 *   - size: 1 byte
 *   - If 'tagged structures' are used then the size is the size of the structure in bytes.
 * - \b CF_NAME8 (\e alias \b CF_CHAR8, \b CF_STRING8)
 *   - typedef \b NAME8
 *   - size: 8 bytes
 * - \b CF_NAME16 (\e alias \b CF_CHAR16, \b CF_STRING16)
 *   - typedef \b NAME16
 *   - size: 16 bytes
 * - \b CF_NAME32 (\e alias \b CF_CHAR32, \b CF_STRING32)
 *   - typedef \b NAME32
 *   - size: 32 bytes
 * - \b CF_NAME48 (\e alias \b CF_CHAR48, \b CF_STRING48)
 *   - typedef \b NAME48
 *   - size: 48 bytes
 * - \b CF_NAME64 (\e alias \b CF_CHAR64, \b CF_STRING64)
 *   - typedef \b NAME64
 *   - size: 64 bytes
 * - \b CF_FLTFLT (\e alias \b CF_FF, \b CF_XY, \b CF_POINT)
 *   - typedef \b FLTFLT
 *   - size: 8 bytes (1 float + 1 float)
 * - \b CF_FLTINT (\e alias \b CF_FI)
 *   - typedef \b FLTINT
 *   - size: 8 bytes (1 float + 1 long integer)
 * - \b CF_LNGINT (\e alias \b INTINT, \b CF_II, \b CF_LI, \b CF_LL)
 *   - typedef \b LNGINT
 *   - size: 8 bytes (1 long interger + 1 long integer)
 * - \b CF_DBLDBL
 *   - typedef \b DBLDBL
 *   - size: 16 bytes (1 double + 1 double)
 * - \b CF_NAME8I (\e alias \b CF_CHAR8I)
 *   - typedef \b NAME8I
 *   - size: 12 bytes (8 chars + 1 long integer)
 * - \b CF_NAME16I (\e alias \b CF_CHAR16I)
 *   - typedef \b NAME16I
 *   - size: 20 bytes (16 chars + 1 long integer)
 * - \b CF_NAME32I (\e alias \b CF_CHAR32I)
 *   - typedef \b NAME32I
 *   - size: 36 bytes (32 chars + 1 long integer)
 * - \b CF_NAME48I (\e alias \b CF_CHAR48I)
 *   - typedef \b NAME48I
 *   - size: 52 bytes (48 chars + 1 long integer)
 * - \b CF_NAME64I (\e alias \b CF_CHAR64I)
 *   - typedef \b NAME64I
 *   - size: 68 bytes (64 chars + 1 long integer)
 * - \b CF_NAME16FI (\e alias \b CF_NAMEFI, \b CF_CHAR16FI, \b CF_STRING16FI)
 *   - typedef \b NAME16FI
 *   - size: 24 bytes (16 chars + 1 float + 1 long integer)
 * - \b CF_NAME16II (\e alias \b CF_NAMEII, \b CF_CHAR16II, \b CF_STRING16II)
 *   - typedef \b NAME16II
 *   - size: 24 bytes (16 chars + 1 long integer + 1 long integer)
 * - \b CF_NAME16DBLDBL (\e alias \b CF_NAMEDBLDBL)
 *   - typedef \b NAME16DBLDBL
 *   - size: 32 bytes (16 chars + 1 double + 1 double)
 * - \b CF_NAME32DBLDBL (\e alias \b CF_CHAR32DBLDBL, \b CF_STRING32DBLDBL)
 *   - typedef \b NAME32DBLDBL
 *   - size: 48 bytes (32 chars + 1 double + 1 double)
 * - \b CF_NAME64DBLDBL (\e alias \b CF_CHAR64DBLDBL, \b CF_STRING64DBLDBL)
 *   - typedef \b NAME64DBLDBL
 *   - size: 80 bytes (64 chars + 1 double + 1 double)
 * - \b CF_INTFLTINT (\e alias \b CF_TDS, \b CF_IFI)
 *   - typedef \b INTFLTINT
 *   - size: 12 bytes (1 long integer + 1 float + 1 long integer)
 * - \b CF_FLTFLTINT (\e alias \b CF_XYS, \b CF_FFI)
 *   - typedef \b FLTFLTINT
 *   - size: 12 bytes (1 float + 1 float + 1 long integer)
 * - \b CF_FLTINTINT (\e alias \b CF_FII, \b CF_FIS)
 *   - typedef \b FLTINTINT
 *   - size: 12 bytes (1 float + 1 long integer + 1 long integer)
 * - \b CF_INTINTINT (\e alias \b CF_III)
 *   - typedef \b INTINTINT
 *   - size: 12 bytes (1 long integer + 1 long integer + 1 long integer)
 * - \b CF_FILTER (\e alias \b CF_IFFF, \b CF_INTFLTFLTFLT)
 *   - typedef \b DFILTER
 *   - size: 16 bytes (1 long integer + 1 float + 1 float + 1 float)
 * - \b CF_ADDRESS (\e alias \b CF_IIII, \b CF_INTINTINTINT)
 *   - typedef \b DADDRESS
 *   - size: 16 bytes (1 long integer + 1 long integer + 1 long integer + 1 long integer)
 * - \b CF_WINDOW (\e alias \b CF_TTII)
 *   - typedef \b DWINDOW
 *   - size: 16 bytes (1 long integer + 1 long integer + 1 long integer + 1 long integer)
 * - \b CF_FWINDOW (\e alias \b CF_IIFF \b CF_INTINTFLTFLT)
 *   - typedef \b FWINDOW
 *   - size: 16 bytes (1 long integer + 1 long integer + 1 float + 1 float)
 * - \b CF_FLTINTFLTINT (\e alias \b CF_FIFI)
 *   - typedef \b FLTINTFLTINT
 *   - size: 16 bytes (1 float + 1 long integer + 1 float + 1 long integer)
 * - \b CF_USTRING 
 *   - typedef \b DUSTRING
 *   - size: 96 bytes (80 chars + 1 long integer + 1 float + 1 float + 1 long integer)
 * - \b CF_SPECTRUM
 *   - typedef \b DSPECTRUM
 *   - size: 1120 bytes (DUSTRING + 1024 floats)
 * - \b CF_UNAME (\e alias \b CF_IFFFNAME \b CF_IFFFNAME16)
 *   - typedef \b DUNAME
 *   - size: 32 bytes (1 long integer + 1 float + 1 float + 1 float + 1 16 char name)
 * - \b CF_XML 
 *   - typedef \e c: \b char  \e vb: \b String  \e java: \b char
 *   - size: 1 byte
 */
#undef CF_TEXT
#define CF_DOUBLE       512
#define CF_UINT16       513
#define CF_SHORT        513
#define CF_INTEGER      513
#define CF_UINT8        514
#define CF_BYTE         514
#define CF_UINT32       515
#define CF_INT4         515
#define CF_LONG         515
#define CF_TEXT         516
#define CF_CHAR         516
#define CF_FLOAT        517
#define CF_POCALBLK     518
#define CF_STRUCT       519
#define CF_CHAR8        520
#define CF_NAME8        520
#define CF_STRING8      520
#define CF_CHAR16       521
#define CF_NAME16       521
#define CF_STRING16     521
#define CF_CHAR16FI     522
#define CF_NAME16FI     522
#define CF_BIT          523
#define CF_HITTS        524
#define CF_CHAR32       525
#define CF_NAME32       525
#define CF_STRING32     525
#define CF_FI           526
#define CF_FLTINT       526
#define CF_II           527
#define CF_LI           527
#define CF_LL           527
#define CF_INTINT       527
#define CF_LNGINT       527
#define CF_DI           528
#define CF_DBLINT       528
#define CF_SI           529
#define CF_SHTINT       529
#define CF_TDS          530
#define CF_IFI          530
#define CF_INTFLTINT    530
#define CF_CHAR48       531
#define CF_NAME48       531
#define CF_STRING48     531
#define CF_CHAR8I       532
#define CF_NAME8I       532
#define CF_CHAR16I      533
#define CF_NAME16I      533
#define CF_CHAR32I      534
#define CF_NAME32I      534
#define CF_CHAR48I      535
#define CF_NAME48I      535

#define CF_USTRING      536
#define CF_XYS          537
#define CF_FFI          537
#define CF_FLTFLTINT    537
#define CF_FII          538
#define CF_FLTINTINT    538
#define CF_FIS          538
#define CF_FILTER       539
#define CF_IFFF         539
#define CF_INTFLTFLTFLT 539
#define CF_SPECTRUM     540
#define CF_ADDRESS      541
#define CF_IIII         541
#define CF_INTINTINTINT 541
#define CF_WINDOW       542
#define CF_TTII         542
#define CF_POINT        543
#define CF_FF           543
#define CF_XY           543
#define CF_FLTFLT       543
#define CF_TDSI         544
#define CF_INTFLTSHTSHT 544
#define CF_BOOLEAN      545
#define CF_IIFF         546
#define CF_INTINTFLTFLT 546
#define CF_FWINDOW      546
#define CF_FIFI         547
#define CF_FLTINTFLTINT 547
#define CF_CHAR64       548
#define CF_NAME64       548
#define CF_STRING64     548
#define CF_CHAR64I      549
#define CF_NAME64I      549
#define CF_INTDBL       550
#define CF_LNGDBL       550
#define CF_FLTDBL       551
#define CF_DBLDBL       552
#define CF_NAMEDBLDBL   553
#define CF_NAME16DBLDBL 553
#define CF_NAME16INTINT 554
#define CF_CHAR16INTINT 554
#define CF_NAME16II     554
#define CF_CHAR16II     554
#define CF_STRING16II   554
#define CF_NAMEII       554
#define CF_INTINTINT    555
#define CF_III          555
#define CF_NAME32DBLDBL 556
#define CF_CHAR32DBLDBL 556
#define CF_STRING32DBLDBL 556
#define CF_NAME64DBLDBL 557
#define CF_CHAR64DBLDBL 557
#define CF_STRING64DBLDBL 557
#define CF_IFFFNAME     558
#define CF_IFFFNAME16   558
#define CF_INTFLTFLTFLTNAME16 558
#define CF_UNAME        558
#define CF_XML          559

#define CF_DEFAULT      766
#undef CF_NULL
#define CF_NULL     767

#define _getFormatSize_(fmt)\
 ( fmt == CF_DOUBLE  ? 8 :\
   fmt == CF_SHORT   ? 2 :\
   fmt == CF_BYTE    ? 1 :\
   fmt == CF_LONG    ? 4 :\
   fmt == CF_FLOAT   ? 4 :\
   fmt == CF_STRUCT  ? 1 :\
   fmt == CF_POCALBLK? 1 :\
   fmt == CF_CHAR8   ? 8 :\
   fmt == CF_CHAR16  ? 16:\
   fmt == CF_CHAR32  ? 32:\
   fmt == CF_CHAR48  ? 48:\
   fmt == CF_CHAR64  ? 64:\
   fmt == CF_CHAR16FI? 24:\
   fmt == CF_SI      ? 6 :\
   fmt == CF_FI      ? 8 :\
   fmt == CF_II      ? 8 :\
   fmt == CF_DI      ? 12:\
   fmt == CF_TDS     ? 12:\
   fmt == CF_HITTS   ? 6 :\
   fmt == CF_CHAR8I  ? 12:\
   fmt == CF_CHAR16I ? 20:\
   fmt == CF_CHAR32I ? 36:\
   fmt == CF_CHAR48I ? 52:\
   fmt == CF_CHAR64I ? 68:\
   fmt == CF_USTRING ? 96:\
   fmt == CF_FFI     ? 12:\
   fmt == CF_FII     ? 12:\
   fmt == CF_IFFF    ? 16:\
   fmt == CF_IIII    ? 16:\
   fmt == CF_TTII    ? 16:\
   fmt == CF_IIFF    ? 16:\
   fmt == CF_FIFI    ? 16:\
   fmt == CF_XY      ?  8:\
   fmt == CF_TDSI    ? 12:\
   fmt == CF_SPECTRUM?  4:\
   fmt == CF_TEXT    ?  1:\
   fmt == CF_BOOLEAN ?  4:\
   fmt == CF_BIT     ?  4:\
   fmt == CF_LNGDBL  ?  12:\
   fmt == CF_FLTDBL  ?  12:\
   fmt == CF_DBLDBL  ?  16:\
   fmt == CF_NAMEDBLDBL ?  32:\
   fmt == CF_NAME16II ?  24:\
   fmt == CF_III     ?  12:\
   fmt == CF_NAME32DBLDBL ?  48:\
   fmt == CF_NAME64DBLDBL ?  80:\
   fmt == CF_IFFFNAME ? 32:\
   fmt == CF_XML     ?  1:\
   fmt == CF_DEFAULT ?  1: 0 )

EXPORT int getFormatSize(int fmt);
#define fmtsizeof(fmt) getFormatSize(((fmt)%256)+CF_DOUBLE)

#define LFMT(fmt) ((fmt)+CF_DOUBLE)
#define BFMT(fmt) ((fmt)-CF_DOUBLE)

#define isSimpleFormat(fmt) (\
  fmt <= CF_NAME16 ? TRUE :\
  fmt == CF_NAME32 ? TRUE :\
  fmt == CF_NAME48 ? TRUE:\
  fmt == CF_NAME64 ? TRUE: FALSE )

#define getSimpleFormatType(txt) (\
  !strcmp(txt,"BYTE")     ? CF_BYTE :\
  !strcmp(txt,"SHORT")    ? CF_SHORT :\
  !strcmp(txt,"INT")      ? CF_SHORT :\
  !strcmp(txt,"INTEGER")  ? CF_SHORT :\
  !strcmp(txt,"LONG")     ? CF_LONG :\
  !strcmp(txt,"FLOAT")    ? CF_FLOAT :\
  !strcmp(txt,"SINGLE")   ? CF_FLOAT :\
  !strcmp(txt,"DOUBLE")   ? CF_DOUBLE :\
  !strcmp(txt,"STRUCT")   ? CF_STRUCT :\
  !strcmp(txt,"TEXT")     ? CF_TEXT :\
  !strcmp(txt,"CHAR")     ? CF_TEXT :\
  !strcmp(txt,"STRING")   ? CF_TEXT :\
  !strcmp(txt,"NAME8")    ? CF_CHAR8 :\
  !strcmp(txt,"CHAR8")    ? CF_CHAR8 :\
  !strcmp(txt,"NAME16")   ? CF_CHAR16 :\
  !strcmp(txt,"CHAR16")   ? CF_CHAR16 :\
  !strcmp(txt,"CHAR32")   ? CF_CHAR32 :\
  !strcmp(txt,"NAME32")   ? CF_CHAR32 :\
  !strcmp(txt,"CHAR48")   ? CF_CHAR48 :\
  !strcmp(txt,"NAME48")   ? CF_CHAR48 :\
  !strcmp(txt,"CHAR64")   ? CF_CHAR64 :\
  !strcmp(txt,"NAME64")   ? CF_CHAR64 :\
  !strcmp(txt,"BIT")      ? CF_BIT :\
  !strcmp(txt,"XML")      ? CF_XML :\
  CF_NULL )

#define getCompoundFormatType(txt) (\
  !strcmp(txt,"NAME16FLTINT") ? CF_CHAR16FI :\
  !strcmp(txt,"NAME16FI") ? CF_CHAR16FI :\
  !strcmp(txt,"CHAR16FI") ? CF_CHAR16FI :\
  !strcmp(txt,"NAME8I")   ? CF_CHAR8I :\
  !strcmp(txt,"CHAR8I")   ? CF_CHAR8I :\
  !strcmp(txt,"NAME16I")  ? CF_CHAR16I :\
  !strcmp(txt,"CHAR16I")  ? CF_CHAR16I :\
  !strcmp(txt,"NAME32I")  ? CF_CHAR32I :\
  !strcmp(txt,"CHAR32I")  ? CF_CHAR32I :\
  !strcmp(txt,"NAME48I")  ? CF_CHAR48I :\
  !strcmp(txt,"CHAR48I")  ? CF_CHAR48I :\
  !strcmp(txt,"NAME64I")  ? CF_CHAR64I :\
  !strcmp(txt,"CHAR64I")  ? CF_CHAR64I :\
  !strcmp(txt,"HITTS")    ? CF_HITTS :\
  !strcmp(txt,"FI")       ? CF_FI :\
  !strcmp(txt,"FLTINT")   ? CF_FI :\
  !strcmp(txt,"FLTDBL")   ? CF_FLTDBL :\
  !strcmp(txt,"II")       ? CF_II :\
  !strcmp(txt,"INTINT")   ? CF_II :\
  !strcmp(txt,"LI")       ? CF_II :\
  !strcmp(txt,"LNGINT")   ? CF_II :\
  !strcmp(txt,"LNGDBL")   ? CF_LNGDBL :\
  !strcmp(txt,"INTDBL")   ? CF_LNGDBL :\
  !strcmp(txt,"DBLDBL")   ? CF_DBLDBL :\
  !strcmp(txt,"NAMEDBLDBL") ? CF_NAMEDBLDBL :\
  !strcmp(txt,"NAME16DBLDBL") ? CF_NAMEDBLDBL :\
  !strcmp(txt,"NAME32DBLDBL") ? CF_NAME32DBLDBL :\
  !strcmp(txt,"NAME64DBLDBL") ? CF_NAME64DBLDBL :\
  !strcmp(txt,"NAME16II") ? CF_NAME16II :\
  !strcmp(txt,"CHAR16II") ? CF_NAME16II :\
  !strcmp(txt,"NAME16INTINT") ? CF_NAME16II :\
  !strcmp(txt,"TDS")      ? CF_TDS :\
  !strcmp(txt,"IFI")      ? CF_TDS :\
  !strcmp(txt,"INTFLTINT")? CF_TDS :\
  !strcmp(txt,"USTRING")  ? CF_USTRING :\
  !strcmp(txt,"XYS")      ? CF_XYS :\
  !strcmp(txt,"FFI")      ? CF_XYS :\
  !strcmp(txt,"FLTFLTINT")? CF_XYS :\
  !strcmp(txt,"FLTINTINT")? CF_FII :\
  !strcmp(txt,"FIS")      ? CF_FII :\
  !strcmp(txt,"FILTER")   ? CF_FILTER :\
  !strcmp(txt,"IFFF")     ? CF_IFFF :\
  !strcmp(txt,"INTFLTFLTFLT")? CF_IFFF :\
  !strcmp(txt,"SPECTRUM") ? CF_SPECTRUM :\
  !strcmp(txt,"ADDRESS")  ? CF_ADDRESS  :\
  !strcmp(txt,"INTINTINTINT")? CF_ADDRESS  :\
  !strcmp(txt,"IIII")     ? CF_ADDRESS  :\
  !strcmp(txt,"INTINTINT")? CF_III  :\
  !strcmp(txt,"III")      ? CF_III  :\
  !strcmp(txt,"WINDOW")   ? CF_WINDOW  :\
  !strcmp(txt,"TTII")     ? CF_WINDOW  :\
  !strcmp(txt,"POINT")    ? CF_POINT  :\
  !strcmp(txt,"XY")       ? CF_POINT  :\
  !strcmp(txt,"FLTFLT")   ? CF_POINT  :\
  !strcmp(txt,"TDSI")     ? CF_TDSI  :\
  !strcmp(txt,"INTFLTSHTSHT")? CF_TDSI  :\
  !strcmp(txt,"BOOLEAN")? CF_BOOLEAN  :\
  !strcmp(txt,"DEFAULT")  ? CF_DEFAULT :\
  !strcmp(txt,"IIFF")     ? CF_IIFF :\
  !strcmp(txt,"INTINTFLTFLT")? CF_IIFF :\
  !strcmp(txt,"FIFI")     ? CF_FIFI :\
  !strcmp(txt,"FLTINTFLTINT")? CF_FIFI :\
  !strcmp(txt,"FWINDOW")   ? CF_FWINDOW  :\
  !strcmp(txt,"IFFFNAME")   ? CF_IFFFNAME  :\
  !strcmp(txt,"INTFLTFLTFLTNAME") ? CF_IFFFNAME  :\
  !strcmp(txt,"UNAME") ? CF_IFFFNAME  :\
  CF_NULL )

/**
 * \page tinemodes TINE Data Transfer Modes and Flags
 *
 * Below is a \ref TransferModes "list" of the available data transfer modes
 * 
 * TINE allows both synchronous and asynchronous data exchange. Synchronous data exchange 
 * is easier to understand and easier to program, but is generally much less efficient than 
 * asynchronous data exchange. With synchronous data exchange, a client would simply make a 
 * call to ExecLink().  When the call completes, the data-exchange activity is over:  The 
 * data have either been acquired or an error code has been set.  However, the process calling 
 * ExecLink() is completely blocked until the call completes.  This is a very annoying 
 * side-effect if a client is talking to several servers, and one of them is down.  From the 
 * server side, the request for data-exchange is treated as an isolated case. That is, no 
 * persistent record of the client or the contract is kept, and if several clients in fact want 
 * the same thing, a server must deal with each one individually.  This can be an annoying burden 
 * on the server if the number of synchronous clients wanting the same thing is large. On the 
 * other hand, a client could set up an asynchronous data-exchange link via a call to 
 * AttachLink(). When programming asynchronously, one inevitably deals with callbacks. Although it 
 * is possible to avoid them in TINE, it is strongly suggested that you use them. In other words 
 * a call to AttachLink() is guaranteed to complete immediately (unless the CM_WAIT access bit 
 * is specified). When the requested data transfer completes, the data set registered in the 
 * call to AttachLink() will be updated and a user-defined callback function will be called 
 * (see the discussion concerning AttachLink()). In principle, one could ignore the callback 
 * and simply assume that the data set always represents the current server data. However this 
 * is dangerous, as the callback function will also be called upon any error condition 
 * (including timeout), giving a client the chance to take appropriate action.
 *
 * Following in the asynchronous vein, TINE permits "Network Links" as well, where a call to 
 * AttachLink() can request that the data be sent to the entire network and not just the caller.
 * To the developer, the callback logic is identical to normal asynchronous links. To specify a
 * "Network Link", the transfer mode only needs to turn on the CM_BCAST bit. In most cases, 
 * client programs do not run often enough on the same network to warrant this action.  Occasionally 
 * though, there are a sufficient number of clients requesting a large amount of data, so that 
 * this does turn out to make sense.  If a client program demands 10 Kbytes from a server, and 
 * the client program is so popular that it is run on 10 stations, then the server must deliver 
 * the same 10 Kbytes to 10 different customers.  One can reduce the demand for bandwidth and 
 * at the same time relieve the server of a bit of work by writing the client program so that 
 * it requests a broadcast to the network instead. Then, the server sends the 10 Kbytes only 
 * once, and all 10 customers get it nonetheless.
 *
 * A server will accept the request to send the contract as a broadcast only if the network 
 * from which the request was made, is contained in its ipbcast.csv (or ipxbcast.csv) startup 
 * file.
 *
 * For example (ipbcast.csv):
 *
 * \include eg_ipbcast.csv
 * 
 * If the network is not so registered, then the client request will be processed as a 
 * normal subscription.
 * 
 * TINE uses connectionless communication be default (i.e. datagrams). Under some circumstances,
 * it could be desired to run data communication over connected sockets (TCP vs UDP, for 
 * instance). This is allowed, in that one applies the CM_CONNECT bit to the requested access 
 * mode. In such as cause, all communication peer-to-peer (from this client to that server) 
 * will go via a connected socket, regardless of whether the CM_CONNECT flag has been set for 
 * other links to the same front end or not. Note that at the server side, each connected 
 * client takes up extra resources, and a front can specify how many connected clients it 
 * will allow, via the maxTCPConnections variable (default in most cases is 32).  It is usually 
 * only under special circumstances that a client should request a connection. In almost all 
 * cases, the connectionless communication has been made reliable enough, to the extent that 
 * "commands" sent from the client require an acknowledgment and are retried following timeout.  
 * A potential problem arises when the initial command arrived at the server and was executed, 
 * whereas the acknowledgement was lost in transmission.
 * As long as 'gRetardSingleContractRemoval' is set to TRUE (the default) such retries are
 * flagged with CM_RETRY and a server accepting the retry will re-transmit the results of the
 * first call.  If CM_RETRY is FALSE, the server cannot re-transmit the results of the previous
 * call since it has disgarded them.  In such cases, the command could be 
 * issued twice. In TCP, retransmission of missing packets occurs at the network layer.  
 * TCP connections are also necessary for JAVA applets, which are not allowed to open a UDP 
 * socket unless the local policies allow it.
 * 
 * \anchor TransferModes 
 * List of data acquisition modes
 * - \b CM_CANCEL (\e alias \b CM_NULL)
 *    - Cancels an active data link
 * - \b CM_SINGLE 
 *    - One-time non-persistent data acquisition
 * - \b CM_REFRESH
 *    - Persistent data acquisition with data transfer on change of data
 *    - 'change' of data is determined by zero-tolerance data comparison of the data payload
 * - \b CM_POLL
 *    - persistent data acquisition at the specified polling rate
 * - \b CM_REGISTER
 *    - Registers the data acquisition request in a client's connection table.
 *    - As this does not send a data acquisition request to a server it is used primarily internally
 * - \b CM_NODUPLICATES
 *    - Instructs the client engines to suppress event (i.e. callback) notification if the
 *      incoming data do not change within the given tolerance (see: SetNotificationTolerance()).
 * - \b CM_BCAST
 *    - Apply this flag to signal the server to send the data as a broadcast on the 'control' network
 *    - The server's definition of the 'control' network is contained in its ipbcast.csv 
 *     configuration file (or ipxbcast.csv file for ipx-speaking servers).
 * - \b CM_MCAST (\e alias \b CM_NETWORK)
 *    - Apply this flag to signal the server to send the data as a multicast
 * - \b CM_USE_ON_ERROR (\e alias \b CM_USEONERROR)
 *    - Apply this flag to signal the client engine to replace faulty data (non-zero completion code)
 *      with initial (i.e. pre-link) contents of the output data buffer.
 * - \b CM_RETRY
 *    - Used internally only to notify a server that the request is being retried
 * - \b CM_GROUPED
 *    - Apply this flag to signal the client engine to call the callback routine only once when
 *      the data for all links with the same callback function has arrived.
 *    - If one link in the group times out, then the callback is delayed until the timeout
 *      is signalled.  Any errors that occur on any of the links belonging to the group will 
 *      establish the completion code of the group.  Thus is the grouped callback is not successful
 *      you will have to check the invidual return codes one by one for the individual links.
 *      Using CM_USE_ON_ERROR to flag faulty data can go a long way in identifying which links
 *      where down!
 *    - You should avoid using CM_REFRESH mode with the CM_GROUPED flag. Since the callback
 *      notification does not occur until all links in the group have arrived, any link whose
 *      data do not change will delay notification until the heartbeat kicks in.
 * - \b CM_CONNECT
 *    - Apply this flag to signal that the data acquisition should occur via a connected
 *      socket, i.e. using TCP/IP instead of UDP datagrams.
 *    - This flag is incompatible with CM_BCAST or CM_MCAST.
 * - \b CM_WAIT
 *    - Apply this flag to signal the client code execution to wait for the first data
 *      acquisition to complete before continuing.  For instance, the application requires a
 *      persistent data link but needs to know the values contained in the initial link during
 *      initialization.
 * - \b CM_SYNCNOTIFY
 *    - Apply this flag to signal the client engine should call its callback routines as the
 *      data come in. Normally, the callback nofitication is deferred until all incoming
 *      data is processed.
 *    - If this flag is used, you CANNOT use synchronous links (e.g. ExecLinkEx()) 
 *      inside a callback routine.
 */
#define CM_NULL         0x00
#define CM_CANCEL       0x00
#define CM_SINGLE       0x01
#define CM_REFRESH      0x02
#define CM_POLL         0x03
#define CM_REGISTER     0x04
#define CM_RECEIVE      0x05
#define CM_NODUPLICATES 0x10
#define CM_SEEKPATTERN  0x20
#define CM_MISSPATTERN  0x40
#define CM_BCAST        0x100
#define CM_NETWORK      0x200
#define CM_MCAST        0x200
#define CM_USE_ON_ERROR 0x400
#define CM_USEONERROR   0x400
#define CM_RETRY        0x800
#define CM_GROUPED      0x1000
#define CM_CONNECT      0x2000
#define CM_WAIT         0x4000
#define CM_SYNCNOTIFY   0x8000

#define BASEMODE(m)  ((m) & 0x07)
#define TARGETMODE(m)  ((m) & 0x0fff)
#define CM_REMOVEVOLATILE (CM_NODUPLICATES|CM_SEEKPATTERN|CM_MISSPATTERN|CM_USEONERROR|CM_GROUPED|CM_WAIT)

#define getAccessMode(txt) (\
	strstr(txt,"READ")    ? CM_SINGLE :\
	strstr(txt,"WRITE")   ? CM_SINGLE :\
	strstr(txt,"SINGLE")  ? CM_SINGLE :\
	strstr(txt,"REFRESH") ? CM_REFRESH :\
	strstr(txt,"POLL")    ? CM_POLL :\
	strstr(txt,"REGISTER")? CM_REGISTER :\
	CM_CANCEL )

/**
 * \page tineaccess TINE Data Access and Flags
 * 
 * Below is a \ref DataAccess "list" of the available data access flags
 * 
 * A TINE data request is accompanied by an access code. At the client-side, the request 
 * might be made by specifying any combination of CA_READ access and CA_WRITE access. 
 * Note that by default, any data exchange will automatically carry the CA_READ access bit.  
 * TINE servers are also open to CA_READ access from any client on the network.  Data requests 
 * specifying CA_WRITE access must pass through the TINE security mechanism (see the discussion 
 * concerning users.csv, ipnets.csv, and ipxnets.csv).  It is generally suggested that those 
 * data-exchange requests which involve commands to change hardware settings or otherwise directly 
 * access front-end hardware always require CA_WRITE access. The server equipment module can 
 * then examine the access code in the data request. At the server-side, the following access 
 * bits can be checked against the data access \ref DataAccess "list" shown below.
 *
 * The scope bits CA_FIRST and CA_LAST can be quite useful, as the initial data access for a 
 * particular contract (which is being polled) might trigger some activity which is not needed 
 * in subsequent requests.  Likewise, the final data request (when the contract is being 
 * cancelled) might also be used to terminate related background activities. Of course, single 
 * commands will always carry both the CA_FIRST and the CA_LAST bits, as the request is merely 
 * transient and is coming into scope and going out of scope at the same time.
 *
 * The access bit CA_CONNECT is entirely analogous to the transfer mode bit CM_CONNECT 
 * (please see the discussion there). It is offered as an access bit in the case of synchronous 
 * calls using ExecLink(), in which case there is no possibility to otherwise modify the 
 * transfer mode (de-facto CM_SINGLE).
 *
 * \anchor DataAccess
 * List of data acquisition access flags
 * - \b CA_NULL
 *    - Used internally
 * - \b CA_READ
 *    - The call is requesting read access.  
 *    - This access flag will be applied by default.
 * - \b CA_WRITE
 *    - The call is requesting write access. Calls which change settings should always 
 *    request write access and servers which accept such requests should always demand that
 *    write access is applied to the call before allowing a setting to be changed.
 *    - Write access is always checked against a server's allowed users list (users.csv)
 *    and against a server's allowed networks list (ipnets.csv and/or ipxnets.csv).  If no
 *    users.csv file exists, then all users have write access.  If no ipnets.csv file exists,
 *    the all ip addresses and subnets have write access.
 * - \b CA_ASYNC
 *    - Set by the server prior to calling the equipment module.
 *    - The call is being made on behalf of a peristent data link.
 * - \b CA_SYNC
 *    - Set by the server prior to calling the equipment module.
 *    - The call is being made on behalf of a non-peristent synchronous data link.
 * - \b CA_ALARM
 *    - Set by the server prior to calling the equipment module.
 *    - The call is being made on behalf of the local alarm server (due to the Alarm Watch Table).
 * - \b CA_HIST
 *    - Set by the server prior to calling the equipment module.
 *    - The call is being made on behalf of the local history server.
 * - \b CA_FIRST
 *    - Set by the server prior to calling the equipment module.
 *    - The call is coming into scope.  For instance a persistent data link for a particular
 *      contract has just been established.
 * - \b CA_LAST
 *    - Set by the server prior to calling the equipment module.
 *    - The call is going out of scope.  For instance a persistent data link for a particular
 *      contract has just been canceled.
 *    - Synchronous calls will set the \b CA_FIRST bit along with the \b CA_LAST bit as well as
 *      the \b CA_SYNC bit.
 * - \b CA_RETRY
 *    - Used in conjunction with access mode CM_SINGLE only.  It instructs the subsystem to
 *      retry the data acquistion in case of connection timeout or link error.
 * - \b CA_CONNECT
 *    - Has the same effect as the mode flag \b CM_CONNECT.  Is useful for API interfaces such
 *      as ExecLinkEx() which does not take a 'mode' parameter.
 * - \b CA_MUTABLE
 *    - Applies only to synchronous calls ExecLink() or ExecLinkEx().  This allows the subsystem
 *      to adjust the output data object to account for the amount of data actually returned
 *      by the server.  This is the only thread-safe way of obtaining this information with a
 *      sychronous call.  Note that calling GetCompletionDataSize(-1) is not thread-safe.
 * - \b CA_SYNCNOTIFY
 *    - Has the same effect as the mode flag \b CM_SYNCNOTIFY.
 */
#define CA_NULL         0x00
#define CA_READ         0x01
#define CA_WRITE        0x02
#define CA_ASYNC        0x04
#define CA_SYNC         0x08
#define CA_ALARM        0x10
#define CA_HIST         0x20
#define CA_FIRST        0x40
#define CA_LAST         0x80
#define CA_RETRY        0x100
#define CA_NETWORK      0x200
#define CA_DISCARD      0x1000
#define CA_CONNECT      0x2000
#define CA_MUTABLE      0x4000
#define CA_SYNCNOTIFY   0x8000

#define getAccessType(txt) ( strstr(txt,"WRITE") ? CA_WRITE : CA_READ )

short ftoi(char *fmt);
#define getFormatType(f) ftoi(f)
char *itof(int fmt);
short mtoi(char *mod);
char *itom(short accessMode);

#define GT_LIN  0
#define GT_LOG  1
#define GT_SLOG 2
#define GT_TD   3
#define GT_BIT  4
#define GT_TEXT 5
#define GT_ABS  6
#define GT_HEX  7

#define getGraphType(txt) (\
  txt == NULL           ? GT_LIN:\
  strstr(txt,"LIN")     ? GT_LIN :\
  strstr(txt,"LOG")     ? GT_LOG :\
  strstr(txt,"SLOG")    ? GT_SLOG :\
  strstr(txt,"TD")      ? GT_TD :\
  strstr(txt,"BIT")     ? GT_BIT :\
  strstr(txt,"TEXT")    ? GT_TEXT :\
  strstr(txt,"ABS")     ? GT_ABS :\
  strstr(txt,"HEX")     ? GT_HEX :\
  GT_LIN )

/**
 * \page tinearraytype TINE Data Array Types 
 * 
 * Below is a \ref DataArrayType "list" of the available data array type codes
 * 
 * A TINE query for extended property information returns amoung other information
 * a code specifying the type of data array returned.  If the property in question
 * does not return an array, then 'AT_SCALAR' is returned if data are returned at all.
 * 'AT_UNKNOWN' is returned in the case of 'trigger' properties which do not involve
 * data transfer at all (in which case the property's data size is 0).  
 *
 * The codes below are bit-wise unique and can appear in combination.
 * 
 * \anchor DataArrayType
 * List of data array type flags
 * - \b AT_UNKNOWN
 *    - The array type is either unspecified or the property is question does not
 *    involve a transfer of data.
 * - \b AT_SCALAR
 *    - The property does not return an array but instead a single value.
 * - \b AT_SINGLE
 *    - The property returns a single dimensioned array (i.e. a vector).
 * - \b AT_DOUBLE
 *    - The property returns a double dimensioned array (a matrix).
 * - \b AT_BOOLEAN
 *    - The property returns a boolean value (TRUE or FALSE), or it's value
 *    should be interpretted in a boolean way, i.e. displayed as "TRUE" or
 *    "FALSE".
 * - \b AT_CHANNEL
 *    - The property returns an array and the array reprsents a multi-channel
 *    array, where the array elements represent the associated property values
 *    of corresponding device modules.  Typically the property is overloaded
 *    to return the group of devices beginning with and specified by the 
 *    device name used in the call and to fill in the requested array 
 *    up to the length specified.  Hence the entire group can be obtained, or
 *    a single device, or a subset of the array property so registered.
 *    Such arrays are best displayed as a historgram.
 * - \b AT_SPECTRUM (alias \b AT_TRACE, \b AT_WAVEFORM)
 *    - The property returns an array the the array represents a trace or 
 *    waveform.  Such arrays are best displayed as a line trace.  The horizontal
 *    display information concerning the range and units can likewise be obtained with
 *    a call for extended property query information.
 * - \b AT_STRUCTURED (alias \b AT_COLLECTION)
 *    - The property returns an array whose elements represent non-plottable
 *    information unrelated to a common set of engineering units.  For instance
 *    an array of long integers where the first gives an number, the second a timestamp,
 *    the third a target value, and so on.
 * - \b AT_UNSTRUCTURED (alias \b AT_BLOB, \b AT_IMAGE)
 *    - The property returns an array whose elements are unstructured and 
 *    are to be manipulated (and understood) by the caller.  For instance an image
 *    is usually sent and a unstructured byte-blob which needs to be decompressed
 *    by the client, before being displayed.
 * - \b AT_JPEG 
 *    - The property returns an array whose elements are to be intpretted as a
 *    byte blob representing a JPEG image.
 * - \b AT_MPEG 
 *    - The property returns an array whose elements are to be intpretted as a
 *    byte blob representing a MPEG image.
 * - \b AT_BMP 
 *    - The property returns an array whose elements are to be intpretted as a
 *    byte blob representing a bitmap image.
 * - \b AT_GIF 
 *    - The property returns an array whose elements are to be intpretted as a
 *    byte blob representing a GIF image.
 * - \b AT_WAV 
 *    - The property returns an array whose elements are to be intpretted as a
 *    byte blob representing a WAV audio stream.
 * - \b AT_MP3 
 *    - The property returns an array whose elements are to be intpretted as a
 *    byte blob representing a MP3 audio stream.
 */
#define AT_UNKNOWN      0x00
#define AT_SCALAR       0x01
#define AT_SINGLE       0x02
#define AT_DOUBLE       0x04
#define AT_BOOLEAN      0x08
#define AT_CHANNEL      0x10
#define AT_SPECTRUM     0x20
#define AT_TRACE        0x20
#define AT_WAVEFORM     0x20
#define AT_STRUCTURED   0x40
#define AT_COLLECTION   0x40
#define AT_UNSTRUCTURED 0x80
#define AT_BLOB         0x80
#define AT_IMAGE        0x80
#define AT_JPEG        0x100
#define AT_MPEG        0x200
#define AT_BMP         0x400
#define AT_GIF         0x800
#define AT_WAV        0x1000
#define AT_MP3        0x2000

short getArrayType(char *fmt);
short getArrayRowLength(char *fmt);

#define CSHOSTS       "ENSEQM"
#define CSHOSTS_FILE  "cshosts.csv"
#define FECID_FILE    "fecid.csv"
#define EXPORT_FILE   "exports.csv"
#define EQPDB_FILE    "eqpdbase.csv"
#define FECDB_FILE    "fecaddr.csv"
#define USERS_FILE    "users.csv"
#define DENYUSER_FILE "denyuser.csv"
#define IPNETS_FILE   "ipnets.csv"
#define IPBCASTNETS_FILE "ipbcast.csv"
#define IPXNETS_FILE  "ipxnets.csv"
#define IPXBCASTNETS_FILE "ipxbcast.csv"
#define ALIAS_FILE    "alias.csv"

#define MAJOR_VERSION 3
#define MINOR_VERSION 31
#define VERSION_REVISION 23

#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif

#endif /* _RPCDEFS_H_ */
