/** \file servdef.h
 *
 * server-side (and client-side) transport definitions and prototypes.
 *
 */
#ifndef SERVDEF_H_
#define SERVDEF_H_

#include "project.def"
#include "prolog.h"
#include "rpctypes.h"
#include "errors.h"
#include "almsrv.h"

#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif

/**
 * \internal
 * Assign a completion code and jump to a 'cleanup' label where
 * files can be fclose()ed, memory free()ed. etc. if need be ...
 *
 * A very useful macro when you're using straight "c".
 */
#define ccerr(e) {cc = e; goto err; }

/**
 * \internal
 * \brief Upper bound on the system polling rate.
 *
 * Absolute maximum allowed wait time on a select().  Typically the system polling
 * rate is on the order of 10 to 20 milliseconds by default (a clock tick).
 */
#define MAXALLOWED_POLLRATE 65535L

/**
 * \internal
 * \brief difference in milliseonds between two struct timeval's.
 */
#define MSECS(a,b) ((a).tv_sec-(b).tv_sec > MAXALLOWED_POLLRATE ? MAXALLOWED_POLLRATE :\
		    ((a).tv_sec-(b).tv_sec)*1000+((a).tv_usec-(b).tv_usec)/1000)

/**
 * \internal
 * \brief Maximal length of a FEC name.
 */
#define FEC_NAME_SIZE     16

/**
 * \internal
 * \brief maximal length of an export name.
 */
#define ELEM_NAME_SIZE    16

/**
 * \internal
 * \brief Maximal length of a local module name.
 */
#define EQP_NAME_SIZE      6

/**
 * \internal
 * \brief Maximal length of a property name.
 */
#define EQP_PROP_SIZE     32

/**
 * \internal
 * \brief Maximal length of a device name.
 */
#define EQP_DEVICE_SIZE   16

/**
 * \internal
 * \brief Maximal length of a context name.
 */
#define EQP_CONTEXT_SIZE  16

#define EQP_SUBSYSTEM_SIZE 16
#define EQP_SUBSYSTEM_SHORTTAG_SIZE 5
/**
 * \internal
 * \brief Default value for MaxNumConnections.
 */
#ifndef MAX_ENTRIES
# define MAX_ENTRIES     100 /* connection table entries  */
#endif

/**
 * \internal
 * \brief Communication protocol macro definitions ...
 */
#define IPX             0x01
#define SPX             0x02
#define TCP             0x04
#define UDP             0x08
#define PIPE            0x10
#define MMF             0x20

#define getProtocolAsString(i) (\
  (i) == IPX ? "IPX" :\
  (i) == SPX ? "SPX" :\
  (i) == TCP ? "TCP" :\
  (i) == UDP ? "UDP" :\
  (i) == PIPE ? "LCL" :\
  (i) == MMF ? "LCL" :\
  "---"\
)

#define CLN_REQ_ 0x01
#define SRV_RSP_ 0x02

/**
 * \internal
 * \brief Tine Engine operation mode
 *
 * Single process OSes (MSDOS, NIOS) use POLLING, where in some sense, TINE becomes the OS.  
 * Multi-process OSes (UNIX, WINDOWS, VxWorks, VMS, etc.) use SELECTING, where the
 * TINE engines only wakes up if 'select()' fires (incoming data) or if the 
 * SystemPolling rate so dictates.
 */
#define POLLING         0x01
#define SELECTING       0x02

/**
 * \internal
 * \brief The default value for MaxPollingRate.
 *
 * Client contracts asking to poll properties at an interval greater than MaxPollingRate will
 * be automatically converted to MaxPollingRate.
 * This in effect sets the upper limit for the timeout parameter for select() when SELECTING
 * mode is in effect.  Note that 'Rate' is actually a misnomer as the value of this parameter
 * actually gives an interval and not a rate.
 *
 */
#ifndef DEFAULT_POLLRATE
#  ifndef SYSPOLL
#    ifdef MULTITHREADED
#      define DEFAULT_POLLRATE 20
#    else
#      define DEFAULT_POLLRATE  1000
#    endif
#  else
#    define DEFAULT_POLLRATE SYSPOLL
#  endif
#endif

/**
 * \internal
 * \brief The default value for MinPollingRate.
 *
 * Client contracts asking to poll properties at an interval smaller than MinPollingRate will
 * be automatically converted to MinPollingRate.  Thus a renegade client cannot force a server
 * to poll a property at smaller intervals.
 *
 */
#ifndef MIN_POLLRATE
#  ifndef SYSPOLL
#    define MIN_POLLRATE 20
#  else
#    define MIN_POLLRATE SYSPOLL
#  endif
#endif

/**
 * \internal
 * \brief The default timeout used when ExecLink() is called.
 *
 * ExecLink() calls ExecLinkEx() with DEFAULT_CLIENT_POLLRATE as the timeout parameter.
 *
 */
#define DEFAULT_CLIENT_POLLRATE 1000

/**
 * \internal
 * \brief The default value for MaxRPCTransportSize
 *
 * Buffer space is reserved at initialization according to this value (64 KBytes).  
 * Servers which need to support larger data transfers should set this value
 * accordingly.
 *
 * \todo Make this gracefully dynamic if possible. (Watch out for VxWorks, NIOS).
 *
 */
#define MAX_TRANSPORTSIZE 0xffff

 /**
 * \internal
 * \brief The default value of gBurstLimit.
 *
 * This corresponds to the number of data packets sent without a break.
 * If UDP is in use, there is no flow control, so data transfers are made assuming that
 * all clients can handle the network speed of the server.  
 * This may or may not be true. The default value is a generally reasonable choice, but
 * Video servers sending large blocks of data will probably want to set gBurstLimit to a 
 * high value on initialization.
 *
 */
#ifndef BURST_LIMIT
# define BURST_LIMIT 20
#endif
extern int gBurstLimit;

 /**
 * \internal
 * \brief The default value of gCycleDelay.
 *
 * This corresponds to the number of milliseonds to delay following a burst transmission.
 * If UDP is in use, there is no flow control, so data transfers are made assuming that
 * all clients can handle the network speed of the server.  
 * This may or may not be true. The default value is a generally reasonable choice, but
 * Video servers sending large blocks of data will probably want to set gCycleDelay to a 
 * smaller value (or 0) on initialization.
 *
 */
#ifndef CYCLE_DELAY
# define CYCLE_DELAY   20
#endif
extern int gCycleDelay;

/**
 * \internal
 * \brief The default value of gSystemTick.
 *
 * Used internally within the TINE engine scheduler.
 *
 */
#ifndef SYSTEM_TICK
# define SYSTEM_TICK 10
#endif
extern UINT32 gSystemTick;

/**
 * \internal
 * \brief The default value of gSynchronizeContracts.
 *
 * Determines whether the TINE engine scheduler should attempt to
 * force contracts with polling intervals of (modulo) 1000 msec
 * should be called consecutively within the same cycle.  This
 * greatly improves the 'packing' of contracts being returned to
 * a given client.
 *
 */
#ifndef CONTRACT_SYNCHRONIZE
# define CONTRACT_SYNCHRONIZE TRUE
#endif
extern UINT32 gSynchronizeContracts;

/**
 * \internal
 * \brief Used as a flag in lieu of the startup time field in a contract.
 *
 * A server servicing a NETWORK subscription will set the startup time 
 * field in the contract header to BCAST_ID (for either a multicast or
 * broadcast request) as a special signature.
 *
 */
#define BCAST_ID 255

/**
 * \internal
 * \brief Special scheduling signals for the TINE scheduler.
 *
 */
#define PRP_SCHEDULE_SIGNAL 0
#define PRP_RETRY_SIGNAL 1
#define PRP_CANCEL_SIGNAL 2

/**
 * \internal
 * \brief Special reserved system names ...
 *
 */
#define SRVEQM_NAME "_SRV__"
#define SRVFEC_NAME "NETWORK"
#define SRVEXP_NAME "NETWORK"

/**
 * \internal
 * \brief Determine whether this request is coming from a NETWORK request
 * for services (e.g. ENS, CAS, etc.).
 *
 */
#define isServiceRequest(pctr) (strcmp((pctr)->EqpName,SRVEQM_NAME) == 0 ? TRUE : FALSE)

/**
 * \internal
 * \brief The default value for gRespondToServiceRequests.
 *
 * Determine whether this server can respond to Service requests.
 * Typically only the ENS will have this features turned on.  (Or possibly the CAS) 
 *
 */
#ifndef RESPOND_TO_SERVICE_REQUESTS
# define RESPOND_TO_SERVICE_REQUESTS FALSE
#endif
extern EXPORT int gRespondToServiceRequests;

#define REGISTERED_ENS  0x01
#define REGISTERED_GENS 0x02
#define REGISTERED_ALL  0x03

typedef struct EQPTask
{
  void (*tskfcn)(void); /* background task function */
  int weight;           /* background task weight factor */
  struct EQPTask *next;
} EQPTASK;
extern EQPTASK *EQPTaskList;

#define EQPBKG void (*)(void)
#define EQPINI void (*)(void)
#define EQPEXI void (*)(void)

/**
 * \internal
 * \brief The top level header which defines the incoming data packet.
 * 
 * \todo split TotalSize away from the header.  This will involve a good look
 * at the routines in swaplib.c.  The advantage of doing this is primarily to 
 * deal with a HeaderStruct with fields on a 4-byte boundary.  The first two
 * bytes of what has come in doesn't need to be a part of this header. BUT,
 * splitting this off from the header has low priority.
 *
 */
typedef struct       /* struct is passed over the net */
{

  /** The total size in bytes (as a consistency check,
   * must match the bytes read) of everthing in this packet.
   */
  UINT16 TotalSize;

  /** The Client name sending the contract. */
  char UserName[16]; 

  /** The contract header protocol level (levels 1 and 2
   * are no longer supported; 4 is probably no longer in use; 3 is
   * used by globals contracts; 5 is currently modern).
   */
  short Protocol;
  
   /**  A vestige field which nominally should give the number
    * of independent contract headers to follow. (Not used in protocol level 5).
    */
  short EqpSize;

} HeaderStruct;
#define HEADERSTRUCT_SIZE  22
#define SUB_PROT_LOCATION 18

/**
 * \internal
 * \brief Legacy header for contract protocol 1
 */
typedef struct       /* struct is passed over the net */
{
  char ElemName[16];
  char EqpModName[8];
  short EqpModNo;
  char Property[12];
  char Cycle[22];
  short Format;
  short DataSize;
} EqpInfoStruct;
#define EQPINFOSTRUCT_SIZE 64

/**
 * \internal
 * \brief Linked List structure used by clients to manage redirected calls.
 */
typedef struct RedirectedLinkStruct
{
  char srcContext[16];
  char srcServerName[16];
  char srcProperty[32];
  char srcDeviceName[32];
  char dstServerName[16];
  char dstProperty[32];
  struct RedirectedLinkStruct *prv;
  struct RedirectedLinkStruct *nxt;
} REDIRECTED_LINK;
extern REDIRECTED_LINK *RedirectedLinkList;

/**
 * \internal
 * \brief Linked List structure used by connection table if multiple links to same contract
 */
typedef struct DependentLinkStruct
{
  int linkId;
  struct DependentLinkStruct *prv;
  struct DependentLinkStruct *nxt;
} DEPENDENT_LINK;

typedef struct
{
  float absolute;
  float percent;
} DataToleranceFactor;

/**
 * \internal
 * \brief Linked List structure used by clients to manage grouped calls.
 */
typedef struct grpStruct
{
  short numPending;     /* num of group events pending */
  short numInGrp;       /* number of elements in group */
  size_t GrpID;         /* group identification */
  short  GrpCc;         /* group completion */
  struct grpStruct *next;
} GroupTableEntry;

#ifdef MSDOS
# define WILDCARD_LINKID(i) ((i) & 0xff)
# define WILDCARD_LISTID(i) ((i) >> 8)
# define MAKE_WILDCARD_CALLBACKID(wcid,lstid) ((wcid) + ((lstid)<<8))
#else
# define WILDCARD_LINKID(i) ((i) & 0xffff)
# define WILDCARD_LISTID(i) ((i) >> 16)
# define MAKE_WILDCARD_CALLBACKID(wcid,lstid) ((wcid) + ((lstid)<<16))
#endif

/**
 * \internal
 * \brief Table entry structure used by clients to manage wildcard calls
 */
typedef struct wildCardTableStruct
{
  int linkId;  /* xref into connTableEntry */
  int callbackId; /* user's callback id */
  void (*callback)(int,int); /* user's callback */
  int length;  /* length of device list */
  int format;  /* original caller format */
  int numPending; /* number waiting to notify */
  int numActive;  /* number of active links in this set */
  NAME16 *list;/* list of device names */
  BYTE *data;  /* either floats or longs */
  int *status; /* list of individual comp codes */
  int *listId; /* list of individual link ids */
} wildCardTableEntry;

extern wildCardTableEntry **wcTable;
extern int nWildCardTableEntries;

#define UNASSIGNED_CALLBACKID 0xeeeeeeee
/**
 * \internal
 * \brief Connection table structure used by clients.
 */
typedef struct          /* local */
{              
  char name[16];        /* eqp function name */
  char prop[32];        /* eqp property */
  char deviceName[32];  /* eqp device name */
  char context[16];     /* eqp context */
  short EqNo;           /* eqp number */
  short access;         /* data access */
  short formatIn;       /* eqp data format down */
  short formatOut;      /* eqp data format up */
  UINT32 sizeIn;        /* data array size down */
  UINT32 sizeOut;       /* data array size up */
  char strTagIn[8];     /* eqp structure tag down */
  char strTagOut[8];    /* eqp structure tag up */
  short dataFlg;        /* signal new arrival (negative completion code) */
  short tocounter;      /* link timeout counter */
  int  RPCindex;        /* index into RPC database */
  int  FECindex;        /* index into FEC database */
  UINT32 pollRate;      /* polling rate for connectioned clients */
  UINT32 heartbeat;     /* used by subscriptions */
  UINT16 mode;          /* used by subscriptions */
  UINT16 RpcID;         /* connection ID */
  int RPCprotocol;      /* RPC header type */
  int protocol;         /* transport protocol (IPX,SPX,UDP,TCP) */
  UINT32 time;          /* last polling request */
  UINT32 starttime;     /* entry into table (used by subscriptions) */
  UINT32 dtimestamp;    /* arriving RPC timestamp */
  int dtimestampMSEC;   /* arriving RPC timestamp (msec fraction) */
  UINT32 lastDataSize;  /* arriving RPC data size */
  int  TCPsocket;
  int  counter;         /* current subscription counter */
  char *dataIn;         /* pointer to associated data down */
  char *dataOut;        /* pointer to associated data up */
  char *bufOut;         /* pointer to raw data up buffer if necessary */
  BYTE *bufOutBlks;     /* pointer to data-up buffer block progress if necessary */
  char *bufCmp;         /* pointer to comparison buffer if supressHearbeatNotification == TRUE */
  DataToleranceFactor *dtf; /* pointer to comparison tolerance if applied */
  short allocated;      /* is dataOut allocated memory ? */
  short useErrValue;    /* flag  */
  UINT16 numblks;       /* progress counters for incoming data */
  UINT16 blksin;        /* progress counters for incoming data */
  UINT32 bytesin;       /* progress counters for incoming data  */
  UINT32 nextDataSize;  /* next outgoing data size in bytes */
  UINT32 nextDataPtr;   /* next outgoing data array pointer */
  UINT32 BytesRemaining;/* number of bytes data remaining */
  UINT32 BytesTotal;    /* total number of bytes data */
  UINT16 datamtu;       /* outgoing data minimum transport unit */
  UINT16 blkidIn;       /* block Id of data to send */
  UINT16 numblksIn;     /* num blks input data to send */
  UINT16 blknumIn;      /* this blk num of input data */
  UINT32 callbackID;    /* pivate callback ID */
  void (*fcn)(int id,int cc); /* call back function */
  UINT8 grouped;        /* flag to indicate if element is grouped */
  UINT8 pending;        /* flag to indicate if element is pending */
  UINT16 srvID;         /* server subscription ID (applicable in MCAST) */
  UINT32 redirectID;    /* if non-null gives child connection ID for redirected calls */
  wildCardTableEntry *wcLink; /* wildcard entry if relevant */
  char redirectName[16];/* redirected eqp function name */
  char redirectProp[32];/* redirected eqp property */
  short linkStatus;     /* current link status */
  short needsNotification; /* flag to call NotifyClient() or not */
  short hasUserCallbackId; /* flag : TRUE if set by caller, otherwise false */
  short RPCLastErrorSource; /* LCL or RMT */
  char RPCLastErrorString[32]; /* temporary storage for return status */
  BYTE  errValue[64];   /* buffer for assigned error value */
  REDIRECTED_LINK *redirectedLink; /* pointer into the redirection table */
  int cancelPending;    /* delay CANCEL because of dependencies */
  int boundToId;        /* id of parent link if this one is bound (>0) */
  DEPENDENT_LINK *dependentLink; /* pointer to dependency list (if this link is a parent) */
} ConnTableEntry;

/**
 * \internal
 * \brief Structure used to managed global data.
 */
typedef struct         /* net */
{
  char  name[16];      /* global element name  (16 bytes) */
  short size;          /* number of elements gcast under "name" (2 bytes) */
  short format;        /* data format (as per RPC definitions)  (2 bytes) */
  SINT32 time;         /* timestamp of gcast data (4 bytes) */
} GlobalDataHdr;
#define GLOBALDATA_HEADER_SIZE (16 + 2 + 2 + 4)

#define EMPTY_PROP    0
#define TARGET_PROP   1
#define WILDCARD_PROP 2
#define STOCK_PROP    3

#ifndef PROPERTY_HASH_SIZE
# define PROPERTY_HASH_SIZE 211
#endif
#define PROPERTY_HASH_INDEX(i) ((i) & 0xff)
#define PROPERTY_NUM_INDEX(i) (((i) >> 8) & 0xffff)
#define MAKE_PROPERTY_ID(i,indx) (((i) << 8) | (indx))

/**
 * \internal
 * \brief Structure used to hold engineering units, ranges, etc. for properties
 */
typedef struct PropertyEGU
{
  char units[16];
  float min;
  float max;
  BYTE graph;
  BYTE reserved[3];
} PropertyEGUStruct;

/**
 * \internal
 * \brief Structure used to hold property redirection information.
 */
typedef struct PrpRedirBlkStruct
{
  NAME16 prpRedirection[4];  
  struct PrpRedirBlkStruct *next;
} PrpRedirBlk;

/**
 * \internal
 * \brief Structure used to hold exported property information.
 */
typedef struct ExportPropertyList
{
  char  prpName[32];       /* property name */
  char  prpAlias[32];      /* property alias name */
  SINT32 prpId;            /* property id */
  UINT32 prpSize;          /* maximum supported array size */
  UINT32 prpSizeIn;        /* maximum supported input array size */
  short prpFormat;         /* format output data */
  short prpFormatIn;       /* format input data */
  short prpAccessMode;     /* bit 0: read access, bit 1: write access */
  short prpArrayType;      /* property array type if array */
  short numRows;           /* num rows if array */
  short rowSize;           /* row size if array */
  NAME8 prpFormatTag;      /* output data structure tag if non zero-length */
  NAME8 prpFormatTagIn;    /* input data structure tag if non zero-length */
  PrpRedirBlk *prd;        /* re-director if not NULL */
  NAME32 *devNames;        /* device list if not NULL */
  int numDevices;          /* number of devices in device list */
  char  prpDescription[64];/* short description of property */
  PropertyEGUStruct egu;   /* EGU information for display */
  PropertyEGUStruct xegu;  /* x-axis EGU information for spectrum display */
  struct ExportPropertyList *next;
} ExportPropertyListStruct;

#ifndef MODULE_HASH_SIZE
# define MODULE_HASH_SIZE 211
#endif
#define MODULE_HASH_PAYOFF (MODULE_HASH_SIZE/10) /* don't hash if fewer modules than this */

typedef struct HashXRef
{
  int index;
  struct HashXRef *next;
} HashXRefStruct;

#ifndef STOCKPROP_HASH_SIZE
# define STOCKPROP_HASH_SIZE 103
#endif

extern HashXRefStruct **StockPropXRefTable;

typedef struct 
{
  NAME32 *prpList;
  int prpListSize;
  char filename[64];
} EqpModulePropertyList;
/**
 * \internal
 * \brief Linked list structure used to hold equipment module information.
 *
 * \todo Does anybody STILL need the ancient NODAL prototype?
 */
typedef struct ExportListTag /* local */
{
  char   Context[16];       /* exported equipment name context */
  char   tagName[16];       /* exported equipment name tag */
  char   EqpSubSystem[16];  /* equipment module sub system tag */
  char   EqpName[8];        /* local equipment name tag */
  char   EqpAlias[8];       /* local equipment alias name tag */
  char   group[16];         /* group name, if applicable */
  int    groupIndex;        /* group index if applicable */
  int    exportId;          /* local ID */
  ExportPropertyListStruct  /* pointer to chained equipment property hash list */
         *EqpProperty[PROPERTY_HASH_SIZE+1]; /* Add room for Wild-card list */
  int    (*EqpPropertyQueryFcn)(char *devName,PropertyQueryStruct **pqs);
  int    (*XEqpPropertyQueryFcn)(char *devName,XPropertyQueryStruct **xpqs);
  int    hasPropertyDeviceNames;
  NAME32 *EqpModuleName;    /* pointer to device name list */
  NAME32 *EqpModuleRdr;     /* pointer to device redirection list */
  EqpModulePropertyList *EqpModulePrpList;/* pointer to array of device-property lists */
  HashXRefStruct **EqpModuleXRef; /* ModuleName cross-reference */
  short  EqpNumProperties;  /* total number of properties supported */
  short  EqpNumModules;     /* total number of devices supported */
  short  EqpNumModulesSpace;/* total amount of space initially allocated */
  short  registered;        /* registered with name server */
  ADS    *adsTable;         /* alarm definition structure table */
  short  alarmSystem;       /* alarm system flag */
  ALARM  **almLst;          /* associated alarm list */
  short  nalarms;           /* number of active alarms */
  UINT32 almMRT;            /* alarm Most Recent Timestamp */
  UINT32 almMRCT;           /* alarm Most Recent Complete Timestamp */
  short  almHSV;            /* alarm highest Severity */
  short  nAlmMRT;           /* number of alarms at MRT */
  short  nAlmMRCT;          /* number of alarms at MRCT */
  short  nAlmHSV;           /* number of alarms at HSV */
  short  nclients;          /* number of active clients */
  short  ncontracts;        /* number of active contracts */
# ifdef NODAL
  int  (*NdlFcn)(double *,int,int,char *); /* associated Nodal equipment function */
# else
  int  (*NdlFcn)(double *,short,short,short,char *);
# endif
  int  (*EqpFcn)(char *,char *,DTYPE *,DTYPE *,short);/* associated equipment function */
  void (*eqpBkg)(void);     /* associated background task */
  int hEqpBkg;              /* handle to background task (in case of threads) */
  SINT32 rateBkg;           /* background polling rate msec */
  short bkgbusy;            /* background busy semaphore */
  short inidone;            /* init task done flag */
  struct timeval lastBkg;   /* background task last polled */
  void (*eqpIni)(void);     /* associated initialization  */
  void (*eqpExi)(void);     /* associated cleanup */
  struct ExportListTag *next;  /* Linked list */
} ExportListStruct;

extern int gEqpFcnBusySemaphore;
extern int gEqpFcnAccessor;
extern ExportPropertyListStruct *ExportPropertyList;
extern ExportListStruct *ExportList;
extern int NrOfEqpTypes;

/**
 * \internal
 * \brief A Structure which is used to map a Context/ExportName to a FecName/EqpName.
 * 
 * The client uses Context and Export name to address the FEC. 
 * TINE internally uses the FecName and the Equipment name.
 * 
 * The TINE library uses this structure for internal caches.
 */
typedef struct                      /* struct is passed over the net */
{
  char  FecName[FEC_NAME_SIZE];     /**< FEC Name */
  char  SubSystem[5];               /* short sub-system name */
  char  ExportName[ELEM_NAME_SIZE];	/**< Export Name */
  char  EqName[EQP_NAME_SIZE+1];	  /**< Equipment name */
  char  Context[EQP_CONTEXT_SIZE];	/**< Context */
  short alarmSystem;
} RPCDataStruct;
#define RPCDATASTRUCT_SIZE (FEC_NAME_SIZE+5+ELEM_NAME_SIZE+EQP_NAME_SIZE+1+2+16)

extern UINT32 IncomingNetworkProtocol;
typedef struct                      /* struct is passed over the net */
{
  char Name[FEC_NAME_SIZE];
  BYTE reserved[3];
  BYTE Port[2];
  BYTE Network[4];
  BYTE Node[6];
  BYTE ImmediateAddress[6];
  char IP[16];
  BYTE PortOffset;
  char IPh_addr[4];
  short  Prot;
} RPCFecStruct;
#define RPCFECSTRUCT_SIZE (FEC_NAME_SIZE+3+2+16+21+2)

typedef struct
{
  char os[12];
  char desc[64];
  char loc[32];
  char hdw[32];
  char sub[16];
  char resp[32];
} RPCFecExStruct;

typedef struct
{
  char FecName[FEC_NAME_SIZE];
  char reserved[4];
  UINT32 localtime;
  UINT32 starttime;
  UINT32 systemPollingRate;
  short  numBkgTsks;
  short  numTotalContracts;
  short  numTotalClients;
  short  numTargetContracts;
  short  numTargetClients;
  short  numConnections;
  UINT32 numConnectionTimeouts;
  UINT32 numConnectionArrivals;
  UINT32 numUDPpkts;
  UINT32 numTCPpkts;
  UINT32 numIPXpkts;
  UINT32 numSPXpkts;
} ActivityQueryStruct;
#define ACTIVITYQUERYSTRUCT_SIZE (FEC_NAME_SIZE+4+4*3+2*6+4*6)

typedef struct
{
  char desc[64];
  BYTE dataFmt;
  BYTE dataSiz;
  BYTE data[64];
  UINT32 timeEntered;
  UINT32 codeNow;
  UINT32 codePrv;
  UINT32 codeNxt;
} ServerStateStruct;
#define SERVERSTATESTRUCT_SIZE (64+2+64+4*4)

typedef struct
{
  UINT32 AveBusyTime;
  UINT32 CycleCounts;
  UINT32 MaxCycleCounts;
  UINT32 SingleLinkCount;
  UINT32 ClientMisses;
  UINT32 ClientReconnects;
  UINT32 ClientRetries;
  UINT32 ContractMisses;
  UINT32 ContractDelays;
  UINT32 BurstLimitReachedCount;
  UINT32 DataTimeStampOffset;
} ServerStatsStruct;
#define SERVERSTATSSTRUCT_SIZE (11*4)

/**
 * \internal
 * \brief Determines the default value of MaxNumClients
 *
 * The places an upper limit on the number of simultaneous clients a server is willing
 * to manage.  If it is known that a server needs to service more clients than this
 * default, then the value of MaxNumClients should be set accordingly at initialization time.
 */
#ifndef NCONSUMER
# define NCONSUMER     100
#endif

/**
 * \internal
 * \brief Determines the default value of MaxNumContracts
 *
 * The places an upper limit on the number of simultaneous contracts a server is willing
 * to manage.  If it is known that a server needs to service more contracts than this
 * default, then the value of MaxNumContracts should be set accordingly at initialization time.
 */
#ifndef NCONTRACT
# define NCONTRACT     1000
#endif

#define RENEWAL       60
#define REMINDER      5
#define ACK_REQUEST   (2*REMINDER)
#define HEARTBEAT     60
#define REPLY_PENDING 3
#define SUB_PENDING   0x3fff
#define ACK_PENDING   0x7fff

/**
 * \internal
 * \brief The payload size of an IPX Datagram.
 *
 */
#define IPX_DATA_SIZE   512
#define IPX_BUFFER_SIZE 1466

/**
 * \internal
 * \brief The payload size of a UDP Datagram.
 *
 * The value here works almost everywhere except NIOS (where #define UDP_BUFFER_SIZE is
 * set to 1024 in project.def).
 *
 * \note TINE performs its own packet reassembly for incoming datagrams.  For many OSes 
 * this is not necessary, but as packet reassembly cannot be assumed to be present or 
 * might have an upper limit relating to the number of recv buffers, it is a fundemental
 * feature of TINE.  
 */
#ifndef UDP_BUFFER_SIZE
# define UDP_BUFFER_SIZE 1472
#endif
/**
 * \internal
 * \brief The default value for one of srvWorkAreaSize, WorkAreaSize, and/or tmpWorkAreaSize.
 *
 * TINE uses a work area buffer for managing contracts calls.  On some (single-threaded) platforms 
 * all work areas point to the same buffer.
 */
#ifndef WORK_SIZE
# define WORK_SIZE  0x10000 /* 64 KB */
#endif

/**
 * \internal
 * \brief The default value for one of clnWorkAreaSize.
 *
 * TINE uses a work area buffer for managing client calls.  On some (single-threaded) platforms 
 * this points to the same area as the srvWorkAreaSize.
 */
#ifndef MIN_WORK_SIZE
# define MIN_WORK_SIZE 2000
#endif

/**
 * \internal
 * \brief The maximum payload size for incoming network globals.
 *
 */
#define BRDCST_BUFFER_SIZE IPX_BUFFER_SIZE

/**
 * \internal
 * \brief The buffer size for a console command
 *
 */
#ifndef CMDSIZE
# define CMDSIZE 256
#endif

extern char *SubType[];
extern int nconsumer;
extern int ncontract;
extern int StaleData;
extern int nExpiredContracts;

/**
 * \internal
 * \brief Legacy contract input header for protocol level 4.
 *
 */
typedef struct             /* struct is passed over the net  */
{
  char  EqpName[8];
  char  Property[12];
  short EqpSize;
  short EqpFormat;
  short EqpNumber;
#ifdef NODAL
  int  (*fcn)(double *,int,int,char *);
#else
  int  (*fcn)(double *,short,short,short,char *);
#endif
} CONTRACT4;
#define CONTRACT4_SIZE 30

/**
 * \internal
 * \brief Contract input header for protocol level 5.
 *
 */
typedef struct             /* struct is passed over the net  */
{
  char   EqpProperty[32];
  char   EqpDeviceName[16];
  char   EqpName[8];
  UINT32 EqpSizeIn;
  UINT32 EqpSizeOut;
  BYTE   hEqpName;
  BYTE   EqpAccess;
  BYTE   EqpFormatIn;
  BYTE   EqpFormatOut;
  char   strTagIn[8];
  char   strTagOut[8];
} CONTRACT;
#define CONTRACT_SIZE (32+16+8+4*2+4*1+2*8)

typedef struct
{
  CONTRACT contract;
  SINT32 pollrate;
  SINT32 nconsumers;
} ContractQueryStruct;
#define CONTRACT_QUERY_SIZE (CONTRACT_SIZE + 4 + 4)

/**
 * \internal
 * \brief Client struct for protocol level 5.
 *
 */
typedef struct                    /* struct is passed over the net (up to the transport field) */
{
  char               userName[16];
  IPXAddress         IPXaddress;  /* if 0 then IP transfer is assumed */
  BYTE               node[6];     /* IPX routing node (Immediate Address) */
  struct sockaddr_in IPaddress;
  short              count;       /* number of registrations */
  short              protocol;    /* target client protocol level */
  short              transport;   /* transport protocol */
  int                sck;         /* communication socket */
} CLN;
#define CLIENT_SIZE (16 + 12 + 6 + 16 + 2 + 2)

/**
 * \internal
 * \brief Client query struct for protocol level 5.
 *
 */
typedef struct                    /* struct is passed over the net */
{
  char               userName[16];
  IPXAddress         IPXaddress;  /* if 0 then IP transfer is assumed */
  BYTE               node[6];     /* IPX routing node (Immediate Address) */
  struct sockaddr_in IPaddress;
  short              count;       /* number of registrations */
  short              protocol;    /* target client protocol level */
} CLNQS;

typedef struct
{
  CLNQS cln;
} ClientQueryStruct;
#define CLIENT_QUERY_SIZE (16 + 12 + 6 + 16 + 2 + 2)

/**
 * \internal
 * \brief Definitions relevant to an Access Lock
 *
 */
#define LOCK_UNLOCKED 0
#define LOCK_PREEMPTIVE 1
#define LOCK_PERSISTENT 2
#define LOCK_ABORT 3
typedef struct
{
  CLN cln;
  int lockType;
  int lockDuration;
  UINT32 lockToken;
  UINT32 lockStart;
} AccessLockStruct;

/**
 * \internal
 * \brief Client connection status struct used by TINE Engine.
 *
 */
typedef struct                /* local */
{
  short  id;                  /* contract ID as seen by consumer */
  short  mode;                /* SINGLE,REFRESH,POLL,CANCEL */
  SINT32 PollingRate;         /* requested polling rate */
  struct timeval lasttime;    /* time of most recent update */
  UINT32 starttime;           /* start of initial subscription */
  BYTE   counter;             /* counts remaining in subscription(s) */
  BYTE   Stale;               /* 1 = send new data to consumer */
  UINT16 numblks;             /* total number of message blocks */
  UINT16 blknum;              /* block number of this message */
  short  statusCode;          /* interim completion */
  UINT16 mtu;                 /* maximal transport unit */
  UINT32 nextDataSize;        /* next outgoing data size in bytes */
  UINT32 nextDataPtr;         /* next outgoing data array pointer */
  UINT32 BytesRemaining;      /* number of bytes data remaining */
  UINT32 BytesTotal;          /* total number of bytes data */
  UINT32 misses;              /* data return after expiration */
  UINT32 reconnects;          /* client unexpectedly re-establishes link */
  int    isRetry;             /* the contract is being resubmitted by the client */
} ClientSTATUS;

#define MAXLONGGRAMS 5
typedef struct
{
  UINT32  haddr;               /* identify client by IP address */
  UINT16 SubscriptionID;      /* supplied by the consumer upon registration */
  UINT16 blksin;              /* # blocks arrived from long datagram */
  UINT16 blkid;               /* block id of long datagram */
  UINT8  busy,
	 reserved;
  UINT8  *data;
  UINT32 blktime;
  UINT32 SubsStarttime;
} ContractProgress;

typedef struct ConStruct
{
  CLN              *client;
  ClientSTATUS     status;
  struct ConStruct *next;
} CONSUMER;

/**
 * \internal
 * \brief Legacy Contract List Table for protocol level 4.
 *
 */
typedef struct                  /* local */
{
  CONSUMER *consumer;           /* assciated consumer list */
  int      nconsumer;           /* number of consumers for this contract */
  CONTRACT4 contract;
  SINT32   PollingRate;         /* in seconds, smallest number wins ... */
  struct   timeval lasttime;    /* most recent data polled */
  struct   timeval lastmarked;  /* last time the data were marked STALE */
  UINT32   dtimestamp;        /* data timestamp */
  BYTE     *Data;               /* data to send */
} ContractListStruct4;

#define CONTRACT_ACTIVE    0x00
#define CONTRACT_EXPIRED   0x01
#define CONTRACT_AVAILABLE 0x02
/**
 * \internal
 * \brief Contract List Table.
 *
 */
typedef struct                  /* local */
{
  CONSUMER *consumer;           /* assciated consumer list */
  short    nconsumer;           /* number of consumers for this contract */
  short    contractID;          /* local contract ID */
  CONTRACT contract;            /* see above */
  char     *lngDeviceName;      /* long device name buffer */
  char     *lngPropertyName;    /* long property name buffer */
  SINT32   PollingRate;         /* in mseconds, smallest number wins ... */
  struct   timeval lasttime;    /* most recent time the data polled */
  struct   timeval lastmarked;  /* last time the data were marked STALE */
  struct   timeval starttime;   /* start time of the current call */
  UINT32   dtimestamp;          /* data timestamp */
  int      dtimestampMSEC;      /* data timestamp (msec fraction) */
  BYTE     *DataOut;            /* data to send  */
  BYTE     *DataIn;             /* data received */
  BYTE     *DataBuf;            /* data working area */
  DTYPE    *dout;               /* prepared output DTYPE */
  DTYPE    *din;                /* prepared input DTYPE */
  UINT32   dataSizeOut;         /* output data size < contract.dataSize */
  int      pending;             /* pending flag for DeliverData() vs ProduceData() */
  int      expired;             /* expired flag for contract removal */
  int      compStatus;          /* completion status */
  int      returnCode;          /* contract completion return code */
  UINT32   delays;              /* delayed scheduling due to pending */
  UINT32   misses;              /* missed scheduling (i.e. > 2x polling rate) */
  char     compString[64];      /* completion string (redirection ?) */
  ExportListStruct *exp;        /* associated export element */
  int      listId;              /* cross-reference to this entry */
} ContractListStruct;

/**
 * \internal
 * \brief Legacy Contract returned header struct for protocol level 4.
 *
 */
typedef struct                  /* struct is passed over the net : Protocol 4 (out) */
{
  short msgsize;
  short SubscriptionID;         /* supplied by the consumer upon registration */
  short CompletionCode;
  short counter;                /* tell consumer how much is left */
  UINT32 timestamp;             /* RPCtimestamp */
  UINT32 starttime;             /* supplied by the consumer upon registration */
} ProducerHeaderStruct4;
#define PRODUCER4_SIZE 16

/**
 * \internal
 * \brief Contract returned header struct
 *
 */
typedef struct                  /* struct is passed over the net : Protocol 5 (out) */
{
  UINT16 msgsize;               /* message size in bytes */
  UINT16 SubscriptionID;        /* supplied by the consumer upon registration */
  UINT16 CompletionCode;        /* completion or interim status */
  UINT16 numblks;               /* total number of message blocks */
  UINT16 blknum;                /* block number of this message */
  UINT16 mtu;                   /* maximum transport unit */
  UINT16 timestampMSEC;         /* msec fraction of data timestamp */
  BYTE   EqpFormat;             /* Eqp Format of returned data */
  BYTE   counter;               /* tell consumer how much is left */
  UINT32 timestamp;             /* data timestamp */
  UINT32 ClnStarttime;          /* supplied by the consumer upon registration */
} ProducerHeaderStruct;
#define PRODUCER_SIZE (7*2+2*1+2*4)
#define PMTU(s)  (((s) ? IPX_BUFFER_SIZE : UDP_BUFFER_SIZE) - PRODUCER_SIZE - sizeof(UINT16))

/**
 * \internal
 * \brief Legacy Contract header struct for protocol level 4
 *
 */
struct subscription4            /* struct is passed over the net : Protocol 4 (in) */
{
  short    msgsize;
  UINT16   PollingRate;
  short    id;
  short    mode;
  UINT32   starttime;
  CONTRACT4 contract;
};
typedef struct subscription4 SUBSCRIPTION4;
#define SUBSCRIPTION4_SIZE (12+CONTRACT4_SIZE)

/**
 * \internal
 * \brief Contract header struct
 *
 */
struct subscription         /* struct is passed over the net : Protocol 5 (in) */
{
  UINT16    msgsize;
  UINT16    mtu;            /* maximum transport unit for incoming data */
  UINT8     numblks;        /* total number of message blocks           */
  UINT8     blknum;         /* block number of this message             */
  UINT16    id;
  short     mode;
  short     blkid;          /* (incremented) block id */
  UINT32    PollingRate;
  UINT32    starttime;
  CONTRACT contract;
};
typedef struct subscription SUBSCRIPTION;
#define SUBSCRIPTION_SIZE (6*2+2*4+CONTRACT_SIZE)
#define SUB_MODE_LOCATION (HEADERSTRUCT_SIZE + 8)

extern ContractListStruct **ContractList;
extern CLN **Consumers;         /* registered consumer list */

extern int clnSocketNumber;     /* SPX Client Socket                 */
extern int ProducerSocket;      /* UDP Producer Socket               */
extern int ConsumerSocket;      /* UDP Consumer Socket               */
extern int MCastConsumerSocket; /* UDP MCast/BCast Consumer Socket   */
extern int NetServiceSocket;    /* UDP MCast Net Services Socket     */
extern int GlobalSocket;        /* UDP Global listening Socket       */
extern int SyncSocket;          /* UDP trivial RPC Socket            */
extern struct sockaddr_in prod_name;
extern AccessLockStruct currentAccessLock;

void doConsumer(void);

typedef struct
{
  char prpName[32];
  UINT16 prpSize;          /* maximum supported array size */
  short prpFormat;         /* native format */
  short prpAccessMode;     /* bit 0: read access, bit 1: write access */
  char  prpDescription[32];/* short description of property */
  int  (*fcn)(CONTRACT *,char *,char *);
} StockPropertyStruct;

typedef struct
{
  char prpName[32];
  UINT32 prpSize;          /* maximum supported array size (output) */
  UINT32 prpSizeIn;        /* maximum supported array size (input)  */
  short prpFormat;         /* format (output) */
  short prpFormatIn;       /* format (input) */
  short prpAccessMode;     /* bit 0: read access, bit 1: write access */
  short prpArrayType;      /* output array type */
  char  prpStructTag[8];   /* output data structure tag */
  char  prpDescription[64];/* short description of property */
  int  (*fcn)(CONTRACT *,char *,char *);
} StockPropertyStructEx;

#ifndef _StockPropertyStructEx 
# define _StockPropertyStructEx StockPropertyStructEx
#endif
extern int NrStockProperties;
extern _StockPropertyStructEx StockProperty[];
#define GetStockPropertyId(str) \
  getTableIndex(str,StockPropXRefTable,STOCKPROP_HASH_SIZE,(char *)StockProperty,NrStockProperties,OFFSETIN(_StockPropertyStructEx,prpName),sizeof(_StockPropertyStructEx))

typedef struct
{
  char modName[16];
  int  (*fcn)(CONTRACT *,char *,char *);
} StockModifierStruct;
#ifndef _StockModifierStruct
# define _StockModifierStruct StockModifierStruct
#endif
extern int NrStockModifiers;
extern _StockModifierStruct StockModifier[];
short IsMetaProperty(char *Property,char **baseProperty,char **metaProperty);

#ifndef ALIAS_HASH_SIZE
# define ALIAS_HASH_SIZE 103
#endif
#define ALIAS_HASH_PAYOFF (ALIAS_HASH_SIZE/10)
typedef struct
{
  char name[32];
  char alias[32];
} ALIAS_TABLE;
extern ALIAS_TABLE *gAliasTable;
extern int nAliasTableEntries;
extern HashXRefStruct **AliasXRefTable;

int makeAliasXRefTable(void);
int GetAliasTable(void);
int getTableIndex(char *str,HashXRefStruct **xrf,int hshsiz,char *tbl,int tblsiz,int off,int reclen);
#define getAliasTableIndex(str) \
   getTableIndex(str,AliasXRefTable,ALIAS_HASH_SIZE,(char *)gAliasTable,nAliasTableEntries,OFFSETIN(ALIAS_TABLE,alias),sizeof(ALIAS_TABLE))
#define getAliasTableIndexFromName(str) \
   getTableIndex(str,NULL,0,(char *)gAliasTable,nAliasTableEntries,OFFSETIN(ALIAS_TABLE,name),sizeof(ALIAS_TABLE))

extern EXPORT void (*cdiSystemInitHook)(void);
extern EXPORT int (*cdiCloseLinkHook)(int);
extern EXPORT char *(*getCdiLinkNameHook)(char *,char *);
extern EXPORT int (*cdiExecLinkHook)(char *,char *,DTYPE *,DTYPE *,short,UINT16);
extern EXPORT int (*cdiAttachLinkHook)(char *,char *,DTYPE *,DTYPE *,short,int,void (*)(int,int),int,UINT32);
extern EXPORT void (*cdiSystemFreeHook)(void);

#ifdef SERVER_POLLING_MODE
#define ServerCycleMode    POLLING
#else
#define ServerCycleMode    SELECTING
#endif
extern int RequireAcknowledgments;
extern int AveBusyTime;
extern struct timeval thisCycleTime;
extern int broadcast, nglobals,TCPdebug,TCPerrcnt;
extern BYTE *BrdcstBuffer;
extern UINT32 TCPcumcnt,TCPinitcnt;
extern time_t startuptime;
extern short RPCverion[2];
extern int RPCprotocol;
extern int hConn, hControl;
extern WORD ISwap(WORD w);
extern int connectionTableID,TCPloaded,IPXloaded,numErr;
extern char gUserName[20];      /* login name of PC user */
extern EXPORT char gFecName[20];
extern BYTE *clientbuf;
extern BYTE *WorkArea;
extern BYTE *tmpWorkArea;
extern UINT32 WorkAreaSize;
extern UINT32 tmpWorkAreaSize;
extern CONSUMER *EQPconsumerlist;   /* global reference  */
extern int EQPnconsumers;
extern CONTRACT *EQPcontract;

extern char fecaddr[30],clientaddr[30],fecprot[4];
extern char *SRVDataBuffer;    /* global Buffer pointers for data exchange */
extern int ListenForMCasts;
extern int foreground;
extern int foregroundTTY;
extern UINT32 CMDcount;

extern ContractListStruct *ContractListItem;
extern ConnTableEntry **connTable;
extern GroupTableEntry *grpTable;
extern int ReturnToClientCycle;
extern int ReturnToServerCycle;
extern int nConnectionTableEntries;
extern int doConsumerSemaphore;
extern int ExecLinkSemaphore;
extern int AttachLinkSemaphore;
extern int recvNetGlobalSemaphore;
extern struct sockaddr_in srv_name,brdcst_name;
extern int TCPParent,TCPProducerSocket,UDPSocket;
extern fd_set TCPrset, rwset, conset, UDPset;
extern struct timeval TCPto, fec_to;
extern int NrOfUsers,WRITEaccess;
extern BYTE rpc_socket[2];
extern const BYTE brdcst_socket[2];
extern BYTE reg_socket[2];
extern const BYTE mcast_socket[2];
extern char PATH[];
extern RPCFecStruct *RPCFec, *NameServer;
extern RPCDataStruct *RPCData;
extern RPCFecExStruct RPCFecEx;
extern int numNameServers,curNameServer;
extern int RPCNumRecsRead,RPCNumFecsRead,FECallocNum;
extern UINT32 TCPcount,UDPcount,SPXcount,IPXcount,BRDcount;
extern UINT32 nConnectionTimeouts,nArrivals;
extern UINT32 nPartialTransfers;
extern int NrOfEqpTypes;
extern short NrEQPBckTsks;
extern int NumNodalFunctions;
extern time_t *tnull;
extern EXPORT char gDeviceContext[20];
extern EXPORT char gDeviceSubSystem[20];
extern UINT32 gContractMisses;
extern UINT32 gContractDelays;
extern UINT32 gClientMisses;
extern UINT32 gClientReconnects;
extern UINT32 gClientRetries;
extern UINT32 gBurstLimitReachedCount;
extern UINT32 gSingleLinkCount;
extern int gCycleCounts, gMaxCycleCounts;
extern char gLastWriteAccessUsr[32];
extern char gLastWriteAccessAddr[32];
extern char gLastWriteAccessPrp[36];
extern char gLastWriteAccessDev[36];
extern double gLastWriteAccessTime;
extern int gLastMinuteInitializationDone;
extern int max_rcvbuf;
extern int max_sndbuf;
extern EXPORT int gRetardSingleContractRemoval;
extern int ServerExitCondition;
extern EXPORT int ServerInitialized;
extern int acquireDeviceNamesFromProperties;
extern int BroadcastServer;
extern char *ltzname[2];
extern char tExecLinkReturnedCompletion[36];
extern int tExecLinkReturnedDataSize;

NAME32 *GetEqpModuleNames(char *eqname,char *prpname,int num,NAME32 **rdrnames,EqpModulePropertyList **prplist);
short IsIllegalProperty(char *eqpname,char *prop,char *dev,CLN *cln,short access,ExportListStruct **listEntry);
int doContract(ContractListStruct *cl);
void doGCast(struct timeval *thisTime);
int GetPrpId(char *eqm,char *prpName);
void devNamesInitialization(void);
void ExitEQPModules(void);
void RegisterEQPtasks(char *name,void (*ini)(void),void (*tsk)(void),int rate,void (*exi)(void));
void checkTasklist(void);
int GetDeviceListFromProperty(char *eqm,char *prp,NAME16 *devlst,UINT32 *devlen);
ExportPropertyListStruct *AssignDeviceListFromProperty(char *eqm,char *prp);
int PrepSubBuffer(BYTE *data);
int SendSubBuffer(int FECindex,int sck,int transport,BYTE *data);
unsigned short execRequest(CLN *cln);
int ProduceData(struct timeval *thisTime);
int DeliverData(struct timeval *thisTIme);
int SendToPeer(CLN *c,BYTE *payload,int msgtype);
int doRequest(int num, int *i);
int getConsumerRequest(CLN *cln,BYTE *buf,int packetsize);
int getConnectionEntry(DTYPE *dout,DTYPE *din,char *devcontxt,char *devgrp,char *devname,char *property);
void displayData(int format, int fec, int dir);
int dumpConsumerList(void);
int dumpContractList(void);
int dumpConnectionList(void);
void dumpRegisteredUsers(void);
int dumpSemaphores(void);
void consumeData(BYTE *buf);
int addFecToList(char   *name);
int initFecTable(void);
int datacmp(BYTE *d1,BYTE *d2,int siz,int fmt,double t);
int initBroadcastList(void);
void getGlobalUpdate(BYTE *buf);
void quitServer(void);
void hardDelay(int msec);
int IsMemberLockSet(CLN *cln);
CLN *getNetworkClientFromList(void);
CLN *LocateConsumerInList(struct sockaddr_in *ip,IPXAddress *nwAddr,BYTE *immAddr);
EXPORT UINT32 ElfHash(const unsigned char *key);
void makeEqpModuleHashTable(ExportListStruct *el);
int makeStockPropXRefTable(void);
#define GATE_BOOLEAN 0x01
#define GATE_FILTER 0x02
short IsGatedModifier(CONTRACT *con,char *baseProperty,UINT32 *gate);
short IsStockProperty(char *Property);
int lookupPropertyName(ExportListStruct *el,char *prpName,char *devName);
char *lookupPropertyId(ExportListStruct *el,short id);
int sndRegisteredExports(struct timeval *tv);
void ParsePropertyDescription(char *str,ExportPropertyListStruct *prp);
void ParseRedirector(char *str,PrpRedirBlk *prd);
ExportListStruct *getExportListItem(char *eqm);
ExportListStruct *getExportListItemFromId(int id);
int SystemStartGlobalSynchronization(void);
EXPORT int reformat(BYTE *dstData,int dstFmt,BYTE *srcData,int srcFmt,int num);
int getSrvAddress(DTYPE *sa,char *eqm,char *devsrv);

int IsInForeground(void);
void checkIPXNets(struct timeval *tv);
int GetSPXStream(WORD ID,CLN *cln);
int initIPX(void);
void CleanupIPX(struct timeval *tv);
void closeIPXServer(void);
void AcceptIPX(struct timeval *tv);
void AcceptSPXStream(void);
void CloseSPXStream(WORD ID);
int SendIPXData(CLN *cln, BYTE *data);
int SendIPXBroadcast(BYTE *buf,int size);
void fulfillSPXStreamRequest(WORD ID);
CLN *getConsumerIPXAddress(void);
char *ipxnet_ntoa(BYTE *node);
void dumpIPXNets(void);

void NotifyClient(int i,int cc);
void NotifyGlobalClient(int i,int cc);
int doTCPClient(int fec);
int listenUDP(int *s,unsigned short port);
int sendUDPBuffer(struct sockaddr_in *client, int msgsize);
int connectTCP(int port,int i);
int initTCP(void);
int initClient(void);
int GetTCPStream(int socket,CLN *cln);
void CloseTCPStream(int socket);
int SendTCPData(int socket,BYTE *buf,SINT32 buflen);
int SendIPData(CLN *cln,BYTE *data);
int SendIPBroadcast(BYTE *buf,int size);
void fulfillTCPStreamRequest(int socket);
CLN *getConsumerIPAddress(int sck,int prot);
void AcceptTCPStream(void);
void dumpIPNets(void);
void AcceptIP(struct timeval *tv);
void CleanupIP(struct timeval *tv);
void closeIPServer(void);
int IsMemberIPNets(int num,struct sockaddr_in *lst,struct sockaddr_in *ipaddr);
int tUDPCall(int fecindex,BYTE *buf,int timout);
int tIPXCall(int fecindex,BYTE *buf,int timout);
int tRPC(int fecindex,char *data,char *name,short EqNo,char *prop,
           short size,short format,short timout);
int tExecLink(int fecindex,char *data,char *name,char *dev,char *prop,
           short size,short format,short timout);
int tExecLinkEx(int fecindex,char *eqm,char *dev,char *prp,
	   short siz,short fmt,char *dat,short sizIn,short fmtIn,char *datIn,
     short access,short timout);
void checkClientNotification(void);
int readSelfTest(char *eqm,char *textbuf,int sizeInBytes);
int readFecLog(char *textbuf,int sizeInBytes);
int readFecLogFile(char *fname,char *textbuf,int sizeInBytes);
int writeFecLogFile(char *fname,char *textbuf,int sizeInBytes);
int scanFecLogFiles(char *path,char *subpath,NAME64 *namebuf,int *numfiles,int numoffset);
void applyFecLogPath(char *path,int pathlen);
UINT16 GetPortOffset(RPCFecStruct *f);
void checkAccessLocks(time_t t);
int isDeniedUser(char *usr);
EXPORT void haltServer(void);
void freeServerMemory(void);
EXPORT void FreeEQPModules(void);
EXPORT int dbglog(char *text,...);
int dbgoutput(char *str,time_t *ts);
int _datcmp(char *d1,char *d2,int dsz,short fmt,DataToleranceFactor *dtf);
int isSpectrumArray(char *eqm,char *prpName);
int isChannelArray(char *eqm,char *prpName);
void getWildcardTokens(char *wildcardString,char *beginToken,char *endToken,char *midToken);
int wildcardMatches(char *devName,char *beginToken,int beginLen,char *endToken,int endLen,char *midToken,int midLen);
int isContiguous(char *eqm,char *devName,NAME32 *devList,int *start,int *len);
int isLocal(char *eqm,char *devName,char *prpName);
ExportPropertyListStruct *GetProperyListStruct(char *eqm,char *prpName);

EXPORT void resetLastMinuteInitialization(void);

/* FindServerOnNetwork() */
/**
 * \fn FindServerOnNetwork
 * \brief Issues a multicast (or broadcast) to which configured TINE central servers respond
 * 
 * Diskless (or otherwise 'in-a-box') servers which need to find central servers such as the
 * Equipment Name Server, Central Alarm Server, or Post-Mortem Server, etc. can issue this
 * call to ascertain the address.
 *  
 * \param context (optional) the Context of the desired central server
 * \param eqmName the local equipment module name of the desired central server (e.g. "ENSEQM")
 * \param exportName the exported device server name of the desired central server. Note either
 * exportName or eqmName, but not both are optional input parameters.
 * \param fec (optional) is the returned FEC address structure of the server responding to the
 * call
 * \param srv (optional) is the returned Device Server Data address structure of the server responding
 * to the call.
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 */
EXPORT int FindServerOnNetwork(char *context,char *eqmName,char *exportName,RPCFecStruct *fec,RPCDataStruct *srv);

/* GetRegisteredPropertyListStruct() */
/**
 * \brief Returns the ExportPropertyListStruct structure for the registered property
 * specified by the first two arguments.
 * 
 * A server can retieve the ExportPropertyListStruct structure of the input registered 
 * property by making this.  This will give the full information pertaining to the 
 * property given.
 *
 * \param eqm (input) is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param prpName (input) is the name of the property for which the ExportPropertyListStruct
 * structure is desired.
 * \param prpLstStruct (output) contains the output structure when the call completes.
 *
 * \return '0' upon success, othereise a tine return code.
 *
 * \see GetRegisteredPropertyList()
 *  
 */
EXPORT int GetRegisteredPropertyListStruct(char *eqm,char *prpName,ExportPropertyListStruct *prpLstStruct);

/* getLastContractId() */
/**
 * \brief Returns the last contract identifier.
 * 
 * A server can retieve the last contract identifier at any time.
 * 
 * \note A contract table identifier is a dynamic thing, to the extent that if a contract
 * comes and goes (is not persistent) it can have a different ID when it comes again.
 *
 * \return a valid contract ID or -1 if the contract table is empty.
 *
 */
EXPORT int getLastContractId(void);

/* getLastContractTableId() */
/**
 * \brief Returns the last contract table identifier.
 * 
 * A server can retieve the last contract table identifier at any time.  This 
 * routine is most usefull however when used within a server's equipment module.
 * 
 * \note A contract table identifier is a dynamic thing, to the extent that if a contract
 * comes and goes (is not persistent) it can have a different ID when it comes again.
 *
 * \return a valid contract ID or -1 if the contract table is empty.
 *
 */
EXPORT int getLastContractTableId(void);

/* getLastContractTableId() */
/**
 * \brief Returns the access rate (interval) associated with the given contract
 * 
 * A server can retieve the access rate of the specified contract.  If a negative
 * number is passed, then the most recently invoked contract is assumed.
 * This routine is most usefull however when used within a server's equipment module.
 * 
 * \param id is the contract table indentifier whose access rate is desired
 * 
 * \note A contract table identifier is a dynamic thing, to the extent that if a contract
 * comes and goes (is not persistent) it can have a different ID when it comes again.
 *
 * \return a valid contract ID or -1 if the contract table is empty.
 *
 */
EXPORT int getContractAccessRate(int id);

EXPORT void UseMultiThreadedCycler(int value);

#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif

#endif  /* SERVDEF_H_ */
