/**
 * \internal
 * \file swaplib.c
 * \brief Byte-Swapping and Structures
 * 
 * This module contains
 * <ul>
 * <li>functions for byte-swapping</li>
 * <li>functions which handle structure descriptions and tagged structures</li>
 * <li>an initializer for globally used structure definitions</li>
 * </ul>
 * 
 * 	 swaplib.c is also to be used when machines either do not use
 * 	 INTEL byte order or use INTEL byte order but have non-BYTE
 * 	 alignment (such as OSF)
*/
#include "prolog.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "servdef.h"
#include "swaplib.h"
#include "rpctypes.h"
#include "errors.h"

/*----------------------------------------------------------------------------*
 *  Switches between Intel and Motorola byte order conventions 15.09.93 swh 
 *  Special KLUGE for VMS - deliver Mot order for float,double 10.03.94 swh
 *  Let VMS compile using IEEE format pass thru SwapBytes      10.09.2006 pdd
 *----------------------------------------------------------------------------*/

#if !defined(INTEL) && !defined(NONALIGNED) && !defined(VAX_VMS_IEEE)
int SwapBytes(char *dat,int size,int fmt,int dir)
{
  register int i,tmp;
# ifdef VAX_VMS
  int j, a = (dir==RTOL?0:1),b = (a+1)%2;
  char c;
# endif
  hdrSPECTRUM shs;

  switch(fmt)
  {
    case CF_NULL: 
      return 0;
    case CF_TEXT:
    case CF_BYTE:
    case CF_XML:
    case CF_STRING8:
    case CF_STRING16:
    case CF_STRING32:
    case CF_STRING48:
    case CF_STRUCT:
      return 0;
    case CF_SHORT:
#     ifndef VAX_VMS
      for (i=0; i<size; i++) _ISWAP(((short *)dat)[i]);
#     endif
	    return 0;
    case CF_BOOLEAN:
    case CF_LONG:
#     ifdef VAX_VMS
      return 0;
#     endif
    case CF_FLOAT:
#     ifndef VAX_VMS
      for (i=0; i<size*4; i+=4) _LSWAP(dat[i]);
#     else
      for (i=0; i<size; i++) if (((int *)dat)[i]) for (j=0; j<2; j++)
      {
        c = dat[i*4+j+2];
        dat[i*4+j+2] = dat[i*4+j] - a*j;
        dat[i*4+j] = c + b*j;
      }
#     endif
      return 0;
    case CF_DOUBLE:
#     ifndef VAX_VMS
      for (i=0; i<size*8; i+=8) _DSWAP(dat[i]);
#     else
      for (i=0; i<size; i+=8) for (j=0; j<4; j++)
      {
        c= *(dat+i+8+j-2-(j/2));
        *(dat+i+8+j-2-(j/2))= *(dat+i+j);         /*not for VMS serv yet */
        *(dat+i+j)=c + b*(j ? 0 : 0x20);
      }
#     endif
      return 0;
#   ifndef VAX_VMS
    case CF_DBLDBL:
      for (i=0; i<size*16; i+=16) {_DSWAP(dat[i]); _DSWAP(dat[i+8]);}
      return 0;
    case CF_NAMEDBLDBL:
      for (i=0; i<size*32; i+=32) {_DSWAP(dat[16+i]); _DSWAP(dat[16+i+8]);}
      return 0;
    case CF_NAME32DBLDBL:
      for (i=0; i<size*48; i+=48) {_DSWAP(dat[32+i]); _DSWAP(dat[32+i+8]);}
      return 0;
    case CF_NAME64DBLDBL:
      for (i=0; i<size*80; i+=80) {_DSWAP(dat[64+i]); _DSWAP(dat[64+i+8]);}
      return 0;
    case CF_NAME16FI:
    case CF_NAME16II:
      for (i=0; i<size*24; i+=24) {_LSWAP(dat[i+16]); _LSWAP(dat[i+20]);}
      return 0;
    case CF_NAME8I:
      for (i=0; i<size*12; i+=12) {_LSWAP(dat[i+8]);}
      return 0;
    case CF_NAME16I:
      for (i=0; i<size*20; i+=20) {_LSWAP(dat[i+16]);}
      return 0;
    case CF_NAME32I:
      for (i=0; i<size*36; i+=36) {_LSWAP(dat[i+32]);}
      return 0;
    case CF_NAME48I:
      for (i=0; i<size*52; i+=52) {_LSWAP(dat[i+48]);}
      return 0;
    case CF_NAME64I:
      for (i=0; i<size*68; i+=68) {_LSWAP(dat[i+64]);}
      return 0;
    case CF_FF:
    case CF_FI:
    case CF_II:
      for (i=0; i<size*8; i+=8) {_LSWAP(dat[i]); _LSWAP(dat[i+4]);}
      return 0;
    case CF_III:
    case CF_FII:
    case CF_FFI:
    case CF_TDS:
      for (i=0; i<size*12; i+=12) {_LSWAP(dat[i]); _LSWAP(dat[i+4]); _LSWAP(dat[i+8]);}
      return 0;
    case CF_USTRING:
      for (i=0; i<size*96; i+=96)
      {
        _LSWAP(dat[i]); _LSWAP(dat[i+4]); _LSWAP(dat[i+8]); _LSWAP(dat[i+12]);
      }
      return 0;
    case CF_UNAME:
      for (i=0; i<size*32; i+=32)
      {
        _LSWAP(dat[i]); _LSWAP(dat[i+4]); _LSWAP(dat[i+8]); _LSWAP(dat[i+12]);
      }
      return 0;
    case CF_TTII:
    case CF_IIII:
    case CF_IIFF:
    case CF_FIFI:
    case CF_IFFF:
      for (i=0; i<size*16; i+=16)
      {
        _LSWAP(dat[i]); _LSWAP(dat[i+4]); _LSWAP(dat[i+8]); _LSWAP(dat[i+12]);
      }
      return 0;
    case CF_SPECTRUM:
      memcpy(&shs,dat,sizeof(hdrSPECTRUM));
      _StructSwap(SpectrumHeaderStruct,(unsigned char *)dat,(unsigned char *)&shs,dir);
      for (i=96; i<size*4; i+=4) _LSWAP(dat[i]);
      return 0;
#   endif
    default: if (NGdebug > 1) dbglog("SwapBytes : bad format size");
  }
  return illegal_format;
}
#endif  /* !INTEL && !NONALIGNED && !VAX_VMS_IEEE */

struct structStruct *structList = NULL;

struct structStruct *getStructFromTag(char *tag)
{
  struct structStruct *s;

  if (tag == NULL || strlen(tag) == 0) return NULL;
  for (s=structList; s && strncmp(s->name,tag,8); s=s->next);
  if (s != NULL && (s->rawlen == 0 || s->size == 0))
  {
    if (NGdebug) dbglog("getStructFromTag : struct %s not properly terminated",tag);
    return NULL;
  }
  return s;
}

/**
 * \internal
 * \brief Finds or allocates a taged structure.
 */
struct structStruct *findstruct(char *tag)
{
  struct structStruct *s;

  for (s=structList; s; s=s->next)
    if (!strncmp(s->name,tag,8)) return s;
  if ((s=(struct structStruct *)SystemCalloc(1,sizeof(struct structStruct)))
       != NULL)
  {
    strncpy(s->name,tag,8);
    s->rawlen = 0;
    s->next = structList;
    structList = s;
    if (NGdebug) dbglog("findstruct : New structure tag registered: %s",tag);
  }
  
  return s;
}


int queryStructFormat(char *tag,INTINT *sfq,int *sfqsiz)
{
  int i,n,sz=0,nq,cc=0;
  struct structStruct *s;
  struct structFormat *sf;
  
  if (sfqsiz == NULL) ccerr(argument_list_error);
  nq = *sfqsiz;
  if ((s=getStructFromTag(tag)) == NULL) ccerr(invalid_structure_tag);
  for (sf=s->s,n=0; sf != NULL; sf=sf->nxt,n++); /* count them first (reverse ordered!) */
  for (i=0,sf=s->s; i<nq && i<n && sf != NULL; i++, sf=sf->nxt)
  {
    sfq[n-i-1].i1val = sf->siz;
    sfq[n-i-1].i2val = sf->fmt;
    sz += sf->siz * fmtsizeof(sf->fmt);
  }
  if (i == nq) i--;
  sfq[i].i1val = sz;
  sfq[i].i2val = CF_NULL;
  *sfqsiz = n+1;
err:
  return cc;
}

#ifdef MSDOS
#pragma argsused
#endif
/**
 * \brief Prepend a structFormat to a structFomat list.
 * 
 * This method creates a new structFormat and sets it in
 * the first element of a linked list of structFormats.
 * A structFormat may describe a scalar or an array.
 * 
 * 
 * \param target void pointer which is interpreted as a double pointer to a structFormat. Insert the new struct format before this.
 * \param addr Address offset of this field in the structure. 
 * \param size Array length if the described field. (not the size in bytes).
 * \param fmt TINE CF_ format of the new struct element
 * 
 * \todo target should be structFormat double pointer.
 */ 
int addtostruct(void *target,int addr,int size, int fmt)
{
  struct structFormat *tmp,**s = (struct structFormat **)target;
  int offset = (*s) ? (*s)->off + (*s)->siz*getFormatSize((*s)->fmt) : 0;

  tmp = (struct structFormat *)SystemCalloc(1,sizeof(struct structFormat));
  if (tmp == NULL) return out_of_local_memory;
  tmp->off = offset;
  tmp->fmt = fmt;
  tmp->siz = size;
  tmp->addr= addr;
  
  /* Add tmp to list */
  tmp->nxt = *s;  
  *s = tmp;
  return 0;
}

/**
 * \internal
 * \brief Frees a structFormat list
 * \param sf Pointer to the first list element
 */
void freeStruct(struct structFormat *sf)
{
  struct structFormat *tmp,*tmp2;
  for (tmp=sf; tmp != NULL; tmp=tmp2)
  {
    tmp2 = tmp->nxt;
    SystemFree(tmp); 
  } 
}

/**
 * \brief Adds a field descritpion to a tagged structure.
 * 
 * This function has two meanings:
 * <ol>
 * <li>Adds a field to a tagged structure</li>
 * <li>Sets the size and the array size of a tagged structure</li>
 * </ol>
 * 
 * \param tag Tag name
 * \param addr <ol><li>Address offset of this field in the structure.</li>
 *  <li>Size of the element in bytes</li></ol>
 * \param size <ol><li>Array length if the described field. (not the size in bytes).</li>
 *  <li>Number of elements</li></ol>
 * \param fmt TINE CF_ format of the new struct element
 * 
 * 
 * \return 0 or out_of_local_memory if no struct could be found/allocated for tag.
 * 
 * \see StructStruct
 * \see findstruct
 * \todo Better use two separate functions than two meanings in one function.
 */ 
EXPORT int fillstruct(char *tag,int addr,int size,int fmt)
{
  struct structStruct *s;

  if ((s=findstruct(tag)) == NULL) return out_of_local_memory;
  if (fmt == CF_NULL)
  {
    s->num = size;
    s->size = addr;
#   ifndef MSDOS
    if ((s->sdata=(BYTE *)SystemCalloc(size,s->size)) == NULL) return out_of_local_memory;
#   else
    s->rawlen = s->size;
#   endif
    return 0;
  }
  s->rawlen += size*getFormatSize(fmt);
  return addtostruct(&s->s,addr,size,fmt);
}

EXPORT int GetStructSize(char *tag)
{
  struct structStruct *s=findstruct(tag);
  return s == NULL ? 0 : s->rawlen;
}
int _StructSwap(struct structFormat *s,unsigned char *dat,unsigned char *struc,int dir)
{
  int n,dsiz;

  if (dir == RTOL)  /* in-coming */
  {
    for (n=0; s != NULL && s->siz; n += dsiz)
    {
      SwapBytes((char *)&dat[s->off],s->siz,s->fmt,RTOL);
      memcpy(&struc[s->addr],&dat[s->off],dsiz=s->siz*getFormatSize(s->fmt));
      s = s->nxt;
    }
  }
  else             /* out-going  */
  {
    for (n=0; s != NULL && s->siz; n += dsiz)
    {
      memcpy(&dat[s->off],&struc[s->addr],dsiz=s->siz*getFormatSize(s->fmt));
      SwapBytes((char *)&dat[s->off],s->siz,s->fmt,LTOR);
      s = s->nxt;
    }
  }
  return n;  /* number of bytes swapped */
}

int StructSwap(char *tag,unsigned char *dat,unsigned char *struc,int dir)
{
  struct structStruct *ss=findstruct(tag);
# ifdef MSDOS
  if (dir == RTOL) memcpy(struc,dat,ss->rawlen);
  else memcpy(dat,struc,ss->rawlen);
  return ss->rawlen;
# else
  return _StructSwap(ss->s,dat,struc,dir);
# endif
}

int StructSwapEx(char *tag,unsigned char *dat,unsigned char *struc,int dir,int num,int structsiz)
{
# ifndef MSDOS
  int i,n=0;
# endif
  struct structStruct *ss=findstruct(tag);

  if (ss->rawlen == 0) return -un_allocated;
  if (num <1 || structsiz < 1) return -out_of_range;
# ifdef MSDOS
  if (dir == RTOL) memcpy(struc,dat,ss->rawlen*num);
  else memcpy(dat,struc,ss->rawlen*num);
  return ss->rawlen*num;
# else
  for (i=0; i<num; i++)
    n += _StructSwap(ss->s,&dat[i*ss->rawlen],&struc[i*structsiz],dir);
  return n;
# endif
}

int _StructSwapEx(struct structFormat *ss,unsigned char *dat,unsigned char *struc,int dir,int num,int structsiz)
{
# ifndef MSDOS
  int i,n;
# endif
  int len = 0;
  struct structFormat *s = ss;

  if (num <1 || structsiz < 1) return -out_of_range;
  while ( s != NULL && s->siz)
  {
    len += s->siz*getFormatSize(s->fmt);
    s = s->nxt;
  }
# ifdef MSDOS
  if (dir == RTOL) memcpy(struc,dat,len*num);
  else memcpy(dat,struc,len*num);
  return len*num;
# else
  for (i=0,n=0; i<num; i++)
    n += _StructSwap(ss,&dat[i*len],&struc[i*structsiz],dir);
  return n;
# endif
}

/**
 * \internal
 * \brief Structure description for HeaderStruct
 * \see HeaderStruct
 */
struct structFormat *HeaderStructStruct;

/**
 * \internal
 * \brief Structure description for EqpInfoStruct
 * \see EqpInfoStruct
 */
struct structFormat *EqpInfoStructStruct;

/**
 * \internal
 * \brief Structure description for CONTRACT4
 * \see CONTRACT4
 */
struct structFormat *CONTRACT4Struct;

/**
 * \internal
 * \brief Structure description for CONTRACT
 * \see CONTRACT
 */
struct structFormat *CONTRACTStruct;

/**
 * \internal
 * \brief Structure description for ProducerHeaderStruct4
 * \see ProducerHeaderStruct4
 */
struct structFormat *ProducerHeaderStruct4Struct;

/**
 * \internal
 * \brief Structure description for ProducerHeaderStruct
 * \see ProducerHeaderStruct
 */
struct structFormat *ProducerHeaderStructStruct;

/**
\internal
\brief Structure description for SUBSCRIPTION4
\see SUBSCRIPTION4
*/
struct structFormat *SUBSCRIPTION4Struct;

/**
 * \internal
 * \brief Structure description for SUBSCRIPTION
 * \see SUBSCRIPTION
 */
struct structFormat *SUBSCRIPTIONStruct;

/**
 * \internal
 * \brief Structure description for RPCDataStruct
 * \see RPCDataStruct
 */
struct structFormat *RPCDataStructStruct;

/**
 * \internal
 * \brief Structure description for RPCFecStruct
 * \see RPCFecStruct
 */
struct structFormat *RPCFecStructStruct;

/**
 * \internal
 * \brief Structure description for AMSStruct
 * \see AMSStruct
 */
struct structFormat *AMSStructStruct;

/**
 * \internal
 * \brief Structure description for ADSStruct
 * \see ADSStruct
 */
struct structFormat *ADSStructStruct;

/**
 * \internal
 * \brief Structure description for PropertyQueryStruct
 * \see PropertyQueryStruct
 */
struct structFormat *PropertyQueryStructStruct;

/**
 * \internal
 * \brief Structure description for PropertyQueryStructEx
 * \see PropertyQueryStructEx
 */
struct structFormat *PropertyQueryStructExStruct;

/**
\internal
\brief Structure description for XPropertyQueryStruct
\see XPropertyQueryStruct
*/
struct structFormat *XPropertyQueryStructStruct;

/**
 * \internal
 * \brief Structure description for ActivityQueryStruct
 * \see ActivityQueryStruct
 */
struct structFormat *ActivityQueryStructStruct;

/**
 * \internal
 * \brief Structure description for ServerStateStruct
 * \see ServerStateStruct
 */
struct structFormat *ServerStateStructStruct;

/**
 * \internal
 * \brief Structure description for GlobalDataHdr
 * \see GlobalDataHdr
 */
struct structFormat *GlobalDataHdrStruct;

/**
 * \internal
 * \brief Structure description for ClientQueryStruct
 * \see ClientQueryStruct
*/
struct structFormat *ClientQueryStructStruct;

/**
 * \internal
 * \brief Structure description for SpectrumHeader
 * \see SpectrumHeader
 */
struct structFormat *SpectrumHeaderStruct;

/**
 * \internal
 * \brief Initializes and registers pre-defined tagged structures.
 * 
 * \see HeaderStructStruct 
 * \see EqpInfoStructStruct
 * \see CONTRACT4Struct
 * \see SUBSCRIPTIONStruct
 * \see SUBSCRIPTION4Struct
 * \see ProducerHeaderStructStruct
 * \see ProducerHeaderStruct4Struct
 * \see ServerStateStructStruct
 * \see RPCDataStructStruct
 * \see RPCFecStructStruct
 * \see ServerStateStructStruct
 * \see GlobalDataHdrStruct
 * \see SpectrumHeaderStruct
 * \see PropertyQueryStruct
 * \see PropertyQueryStructEx
 * \see XPropertyQueryStruct
 * \see ActivityQueryStruct
 * \see ContractQueryStruct
 * \see ClientQueryStruct
 * \see AMSStructStruct
 * \see ADSStructStruct
 */
int initControlStructs(void)
{
  static int initialized = 0;
  struct structStruct *ss = NULL;
  int cc;

  if (initialized) return 0;
  initialized = 1;

# ifndef MSDOS
  if ((cc=addtostruct(&HeaderStructStruct,
		       OFFSETIN(HeaderStruct,TotalSize),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=addtostruct(&HeaderStructStruct,
		       OFFSETIN(HeaderStruct,UserName),
		       16,CF_TEXT)) != 0) return cc;
  if ((cc=addtostruct(&HeaderStructStruct,
		       OFFSETIN(HeaderStruct,Protocol),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=addtostruct(&HeaderStructStruct,
		       OFFSETIN(HeaderStruct,EqpSize),
		       1,CF_SHORT)) != 0) return cc;

  if ((cc=addtostruct(&EqpInfoStructStruct,
		       OFFSETIN(EqpInfoStruct,ElemName),
		       16,CF_TEXT)) != 0) return cc;
  if ((cc=addtostruct(&EqpInfoStructStruct,
		       OFFSETIN(EqpInfoStruct,EqpModName),
		       8,CF_TEXT)) != 0) return cc;
  if ((cc=addtostruct(&EqpInfoStructStruct,
		       OFFSETIN(EqpInfoStruct,EqpModNo),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=addtostruct(&EqpInfoStructStruct,
		       OFFSETIN(EqpInfoStruct,Property),
		       12,CF_TEXT)) != 0) return cc;
  if ((cc=addtostruct(&EqpInfoStructStruct,
		       OFFSETIN(EqpInfoStruct,Cycle),
		       22,CF_TEXT)) != 0) return cc;
  if ((cc=addtostruct(&EqpInfoStructStruct,
		       OFFSETIN(EqpInfoStruct,Format),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=addtostruct(&EqpInfoStructStruct,
		       OFFSETIN(EqpInfoStruct,DataSize),
		       1,CF_SHORT)) != 0) return cc;

  if ((cc=addtostruct(&CONTRACT4Struct,
		       OFFSETIN(CONTRACT4,EqpName),
		       8,CF_TEXT)) != 0) return cc;
  if ((cc=addtostruct(&CONTRACT4Struct,
		       OFFSETIN(CONTRACT4,Property),
		       12,CF_TEXT)) != 0) return cc;
  if ((cc=addtostruct(&CONTRACT4Struct,
		       OFFSETIN(CONTRACT4,EqpSize),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=addtostruct(&CONTRACT4Struct,
		       OFFSETIN(CONTRACT4,EqpFormat),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=addtostruct(&CONTRACT4Struct,
		       OFFSETIN(CONTRACT4,EqpNumber),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=addtostruct(&CONTRACT4Struct,
		       OFFSETIN(CONTRACT4,fcn),
		       4,CF_BYTE)) != 0) return cc;

  if ((cc=addtostruct(&SUBSCRIPTIONStruct,
		       OFFSETIN(SUBSCRIPTION,msgsize),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTIONStruct,
		       OFFSETIN(SUBSCRIPTION,mtu),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTIONStruct,
		       OFFSETIN(SUBSCRIPTION,numblks),
		       1,CF_BYTE)) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTIONStruct,
		       OFFSETIN(SUBSCRIPTION,blknum),
		       1,CF_BYTE)) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTIONStruct,
		       OFFSETIN(SUBSCRIPTION,id),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTIONStruct,
		       OFFSETIN(SUBSCRIPTION,mode),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTIONStruct,
		       OFFSETIN(SUBSCRIPTION,blkid),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTIONStruct,
		       OFFSETIN(SUBSCRIPTION,PollingRate),
		       1,CF_LONG)) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTIONStruct,
		       OFFSETIN(SUBSCRIPTION,starttime),
		       1,CF_LONG)) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTIONStruct,
		       OFFSETIN(SUBSCRIPTION,contract.EqpProperty),
		       32,CF_TEXT)) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTIONStruct,
		       OFFSETIN(SUBSCRIPTION,contract.EqpDeviceName),
		       16,CF_TEXT)) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTIONStruct,
		       OFFSETIN(SUBSCRIPTION,contract.EqpName),
		       8,CF_TEXT)) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTIONStruct,
		       OFFSETIN(SUBSCRIPTION,contract.EqpSizeIn),
		       1,CF_LONG)) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTIONStruct,
		       OFFSETIN(SUBSCRIPTION,contract.EqpSizeOut),
		       1,CF_LONG)) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTIONStruct,
		       OFFSETIN(SUBSCRIPTION,contract.hEqpName),
		       1,CF_BYTE)) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTIONStruct,
		       OFFSETIN(SUBSCRIPTION,contract.EqpAccess),
		       1,CF_BYTE)) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTIONStruct,
		       OFFSETIN(SUBSCRIPTION,contract.EqpFormatIn),
		       1,CF_BYTE)) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTIONStruct,
		       OFFSETIN(SUBSCRIPTION,contract.EqpFormatOut),
		       1,CF_BYTE)) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTIONStruct,
		       OFFSETIN(SUBSCRIPTION,contract.strTagIn),
		       8,CF_TEXT)) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTIONStruct,
		       OFFSETIN(SUBSCRIPTION,contract.strTagOut),
		       8,CF_TEXT)) != 0) return cc;

  if ((cc=addtostruct(&SUBSCRIPTION4Struct,
		       OFFSETIN(SUBSCRIPTION4,msgsize),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTION4Struct,
		       OFFSETIN(SUBSCRIPTION4,PollingRate),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTION4Struct,
		       OFFSETIN(SUBSCRIPTION4,id),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTION4Struct,
		       OFFSETIN(SUBSCRIPTION4,mode),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTION4Struct,
		       OFFSETIN(SUBSCRIPTION4,starttime),
		       1,CF_LONG)) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTION4Struct,
		       OFFSETIN(SUBSCRIPTION4,contract.EqpName),
		       8,CF_TEXT)) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTION4Struct,
		       OFFSETIN(SUBSCRIPTION4,contract.Property),
		       12,CF_TEXT)) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTION4Struct,
		       OFFSETIN(SUBSCRIPTION4,contract.EqpSize),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTION4Struct,
		       OFFSETIN(SUBSCRIPTION4,contract.EqpFormat),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTION4Struct,
		       OFFSETIN(SUBSCRIPTION4,contract.EqpNumber),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTION4Struct,
		       OFFSETIN(SUBSCRIPTION4,contract.fcn),
		       4,CF_BYTE)) != 0) return cc;

  if ((cc=addtostruct(&ProducerHeaderStructStruct,
		       OFFSETIN(ProducerHeaderStruct,msgsize),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=addtostruct(&ProducerHeaderStructStruct,
		       OFFSETIN(ProducerHeaderStruct,SubscriptionID),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=addtostruct(&ProducerHeaderStructStruct,
		       OFFSETIN(ProducerHeaderStruct,CompletionCode),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=addtostruct(&ProducerHeaderStructStruct,
		       OFFSETIN(ProducerHeaderStruct,numblks),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=addtostruct(&ProducerHeaderStructStruct,
		       OFFSETIN(ProducerHeaderStruct,blknum),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=addtostruct(&ProducerHeaderStructStruct,
		       OFFSETIN(ProducerHeaderStruct,mtu),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=addtostruct(&ProducerHeaderStructStruct,
		       OFFSETIN(ProducerHeaderStruct,timestampMSEC),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=addtostruct(&ProducerHeaderStructStruct,
		       OFFSETIN(ProducerHeaderStruct,EqpFormat),
		       1,CF_BYTE)) != 0) return cc;
  if ((cc=addtostruct(&ProducerHeaderStructStruct,
		       OFFSETIN(ProducerHeaderStruct,counter),
		       1,CF_BYTE)) != 0) return cc;
  if ((cc=addtostruct(&ProducerHeaderStructStruct,
		       OFFSETIN(ProducerHeaderStruct,timestamp),
		       1,CF_LONG)) != 0) return cc;
  if ((cc=addtostruct(&ProducerHeaderStructStruct,
		       OFFSETIN(ProducerHeaderStruct,ClnStarttime),
		       1,CF_LONG)) != 0) return cc;

  if ((cc=addtostruct(&ProducerHeaderStruct4Struct,
		       OFFSETIN(ProducerHeaderStruct4,msgsize),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=addtostruct(&ProducerHeaderStruct4Struct,
		       OFFSETIN(ProducerHeaderStruct4,SubscriptionID),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=addtostruct(&ProducerHeaderStruct4Struct,
		       OFFSETIN(ProducerHeaderStruct4,CompletionCode),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=addtostruct(&ProducerHeaderStruct4Struct,
		       OFFSETIN(ProducerHeaderStruct4,counter),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=addtostruct(&ProducerHeaderStruct4Struct,
		       OFFSETIN(ProducerHeaderStruct4,timestamp),
		       1,CF_LONG)) != 0) return cc;
  if ((cc=addtostruct(&ProducerHeaderStruct4Struct,
		       OFFSETIN(ProducerHeaderStruct4,starttime),
		       1,CF_LONG)) != 0) return cc;

  if ((cc=addtostruct(&RPCDataStructStruct,
		       OFFSETIN(RPCDataStruct,FecName),
		       FEC_NAME_SIZE,CF_TEXT)) != 0) return cc;
  if ((cc=addtostruct(&RPCDataStructStruct,
		       OFFSETIN(RPCDataStruct,SubSystem),
		       5,CF_TEXT)) != 0) return cc;
  if ((cc=addtostruct(&RPCDataStructStruct,
		       OFFSETIN(RPCDataStruct,ExportName),
		       ELEM_NAME_SIZE,CF_TEXT)) != 0) return cc;
  if ((cc=addtostruct(&RPCDataStructStruct,
		       OFFSETIN(RPCDataStruct,EqName),
		       EQP_NAME_SIZE+1,CF_TEXT)) != 0) return cc;
  if ((cc=addtostruct(&RPCDataStructStruct,
		       OFFSETIN(RPCDataStruct,Context),
		       16,CF_TEXT)) != 0) return cc;
  if ((cc=addtostruct(&RPCDataStructStruct,
		       OFFSETIN(RPCDataStruct,alarmSystem),
		       1,CF_SHORT)) != 0) return cc;

  if ((cc=addtostruct(&RPCFecStructStruct,
		       OFFSETIN(RPCFecStruct,Name),
		       FEC_NAME_SIZE,CF_TEXT)) != 0) return cc;
  if ((cc=addtostruct(&RPCFecStructStruct,
		       OFFSETIN(RPCFecStruct,reserved),
		       5,CF_BYTE)) != 0) return cc;
  if ((cc=addtostruct(&RPCFecStructStruct,
		       OFFSETIN(RPCFecStruct,Network),
		       4,CF_BYTE)) != 0) return cc;
  if ((cc=addtostruct(&RPCFecStructStruct,
		       OFFSETIN(RPCFecStruct,Node),
		       6,CF_BYTE)) != 0) return cc;
  if ((cc=addtostruct(&RPCFecStructStruct,
		       OFFSETIN(RPCFecStruct,ImmediateAddress),
		       6,CF_BYTE)) != 0) return cc;
  if ((cc=addtostruct(&RPCFecStructStruct,
		       OFFSETIN(RPCFecStruct,IP),
		       16,CF_TEXT)) != 0) return cc;
  if ((cc=addtostruct(&RPCFecStructStruct,
		       OFFSETIN(RPCFecStruct,PortOffset),
		       1,CF_BYTE)) != 0) return cc;
  if ((cc=addtostruct(&RPCFecStructStruct,
		       OFFSETIN(RPCFecStruct,IPh_addr),
		       4,CF_BYTE)) != 0) return cc;
  if ((cc=addtostruct(&RPCFecStructStruct,
		       OFFSETIN(RPCFecStruct,Prot),
		       2,CF_BYTE)) != 0) return cc;

  if ((cc=addtostruct(&ServerStateStructStruct,
		       OFFSETIN(ServerStateStruct,desc),
		       64,CF_TEXT)) != 0) return cc;
  if ((cc=addtostruct(&ServerStateStructStruct,
		       OFFSETIN(ServerStateStruct,dataFmt),
		       1,CF_BYTE)) != 0) return cc;
  if ((cc=addtostruct(&ServerStateStructStruct,
		       OFFSETIN(ServerStateStruct,dataSiz),
		       1,CF_BYTE)) != 0) return cc;
  if ((cc=addtostruct(&ServerStateStructStruct,
		       OFFSETIN(ServerStateStruct,data),
		       64,CF_BYTE)) != 0) return cc;
  if ((cc=addtostruct(&ServerStateStructStruct,
		       OFFSETIN(ServerStateStruct,timeEntered),
		       1,CF_LONG)) != 0) return cc;
  if ((cc=addtostruct(&ServerStateStructStruct,
		       OFFSETIN(ServerStateStruct,codeNow),
		       1,CF_LONG)) != 0) return cc;
  if ((cc=addtostruct(&ServerStateStructStruct,
		       OFFSETIN(ServerStateStruct,codePrv),
		       1,CF_LONG)) != 0) return cc;
  if ((cc=addtostruct(&ServerStateStructStruct,
		       OFFSETIN(ServerStateStruct,codeNxt),
		       1,CF_LONG)) != 0) return cc;

  if ((cc=addtostruct(&GlobalDataHdrStruct,
		       OFFSETIN(GlobalDataHdr,name),
		       16,CF_TEXT)) != 0) return cc;
  if ((cc=addtostruct(&GlobalDataHdrStruct,
		       OFFSETIN(GlobalDataHdr,size),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=addtostruct(&GlobalDataHdrStruct,
		       OFFSETIN(GlobalDataHdr,format),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=addtostruct(&GlobalDataHdrStruct,
		       OFFSETIN(GlobalDataHdr,time),
		       1,CF_LONG)) != 0) return cc;

  if ((cc=addtostruct(&SpectrumHeaderStruct,
		       OFFSETIN(hdrSPECTRUM,comment),
		       80,CF_TEXT)) != 0) return cc;
  if ((cc=addtostruct(&SpectrumHeaderStruct,
		       OFFSETIN(hdrSPECTRUM,tm),
		       1,CF_LONG)) != 0) return cc;
  if ((cc=addtostruct(&SpectrumHeaderStruct,
		       OFFSETIN(hdrSPECTRUM,s_start),
		       1,CF_FLOAT)) != 0) return cc;
  if ((cc=addtostruct(&SpectrumHeaderStruct,
		       OFFSETIN(hdrSPECTRUM,s_inc),
		       1,CF_FLOAT)) != 0) return cc;
  if ((cc=addtostruct(&SpectrumHeaderStruct,
		       OFFSETIN(hdrSPECTRUM,status),
		       1,CF_LONG)) != 0) return cc;
# endif

  if ((cc=fillstruct("PQS",
		       OFFSETIN(PropertyQueryStruct,prpName),
		       32,CF_TEXT)) != 0) return cc;
  if ((cc=fillstruct("PQS",
		       OFFSETIN(PropertyQueryStruct,prpDescription),
		       32,CF_TEXT)) != 0) return cc;
  if ((cc=fillstruct("PQS",
		       OFFSETIN(PropertyQueryStruct,prpSize),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=fillstruct("PQS",
		       OFFSETIN(PropertyQueryStruct,prpFormat),
		       1,CF_BYTE)) != 0) return cc;
  if ((cc=fillstruct("PQS",
		       OFFSETIN(PropertyQueryStruct,prpAccess),
		       1,CF_BYTE)) != 0) return cc;
  if ((cc=fillstruct("PQS",
		       sizeof(PropertyQueryStruct),
		       100,CF_NULL)) != 0) return cc;

  ss = findstruct("PQS");
  PropertyQueryStructStruct = ss->s;

  if ((cc=fillstruct("PQSX",
		       OFFSETIN(PropertyQueryStructEx,prpName),
		       32,CF_TEXT)) != 0) return cc;
  if ((cc=fillstruct("PQSX",
		       OFFSETIN(PropertyQueryStructEx,prpDescription),
		       64,CF_TEXT)) != 0) return cc;
  if ((cc=fillstruct("PQSX",
		       OFFSETIN(PropertyQueryStructEx,prpRedirection),
		       32,CF_TEXT)) != 0) return cc;
  if ((cc=fillstruct("PQSX",
		       OFFSETIN(PropertyQueryStructEx,prpTag),
		       16,CF_TEXT)) != 0) return cc;
  if ((cc=fillstruct("PQSX",
		       OFFSETIN(PropertyQueryStructEx,prpUnits),
		       16,CF_TEXT)) != 0) return cc;
  if ((cc=fillstruct("PQSX",
		       OFFSETIN(PropertyQueryStructEx,prpMinValue),
		       2,CF_FLOAT)) != 0) return cc;
  if ((cc=fillstruct("PQSX",
		       OFFSETIN(PropertyQueryStructEx,prpSize),
		       3,CF_LONG)) != 0) return cc;
  if ((cc=fillstruct("PQSX",
		       OFFSETIN(PropertyQueryStructEx,prpHistoryDepthLong),
		       2,CF_SHORT)) != 0) return cc;
  if ((cc=fillstruct("PQSX",
		       OFFSETIN(PropertyQueryStructEx,prpFormat),
		       4,CF_BYTE)) != 0) return cc;
  if ((cc=fillstruct("PQSX",
		       sizeof(PropertyQueryStructEx),
		       100,CF_NULL)) != 0) return cc;

  ss = findstruct("PQSX");
  PropertyQueryStructExStruct = ss->s;

  if ((cc=fillstruct("XPQS",
		       OFFSETIN(XPropertyQueryStruct,prpName),
		       32,CF_TEXT)) != 0) return cc;
  if ((cc=fillstruct("XPQS",
		       OFFSETIN(XPropertyQueryStruct,prpDescription),
		       64,CF_TEXT)) != 0) return cc;
  if ((cc=fillstruct("XPQS",
		       OFFSETIN(XPropertyQueryStruct,prpRedirection),
		       32,CF_TEXT)) != 0) return cc;
  if ((cc=fillstruct("XPQS",
		       OFFSETIN(XPropertyQueryStruct,prpTag),
		       16,CF_TEXT)) != 0) return cc;
  if ((cc=fillstruct("XPQS",
		       OFFSETIN(XPropertyQueryStruct,prpUnits),
		       16,CF_TEXT)) != 0) return cc;
  if ((cc=fillstruct("XPQS",
		       OFFSETIN(XPropertyQueryStruct,prpMinValue),
		       2,CF_FLOAT)) != 0) return cc;
  if ((cc=fillstruct("XPQS",
		       OFFSETIN(XPropertyQueryStruct,prpSize),
		       2,CF_LONG)) != 0) return cc;
  if ((cc=fillstruct("XPQS",
		       OFFSETIN(XPropertyQueryStruct,prpNumOverloads),
		       1,CF_LONG)) != 0) return cc;
  if ((cc=fillstruct("XPQS",
		       OFFSETIN(XPropertyQueryStruct,prpHistoryDepthShort),
		       2,CF_SHORT)) != 0) return cc;
  if ((cc=fillstruct("XPQS",
		       OFFSETIN(XPropertyQueryStruct,prpFormat),
		       4,CF_BYTE)) != 0) return cc;
  if ((cc=fillstruct("XPQS",
		       OFFSETIN(XPropertyQueryStruct,rngUnits),
		       16,CF_TEXT)) != 0) return cc;
  if ((cc=fillstruct("XPQS",
		       OFFSETIN(XPropertyQueryStruct,rngMinValue),
		       2,CF_FLOAT)) != 0) return cc;
  if ((cc=fillstruct("XPQS",
		       OFFSETIN(XPropertyQueryStruct,numRows),
		       6,CF_SHORT)) != 0) return cc;
  if ((cc=fillstruct("XPQS",
		       sizeof(XPropertyQueryStruct),
		       10,CF_NULL)) != 0) return cc;

  ss = findstruct("XPQS");
  XPropertyQueryStructStruct = ss->s;

  if ((cc=fillstruct("AQS",
		       OFFSETIN(ActivityQueryStruct,FecName),
		       FEC_NAME_SIZE,CF_TEXT)) != 0) return cc;
  if ((cc=fillstruct("AQS",
		       OFFSETIN(ActivityQueryStruct,reserved),
		       4,CF_TEXT)) != 0) return cc;
  if ((cc=fillstruct("AQS",
		       OFFSETIN(ActivityQueryStruct,localtime),
		       1,CF_LONG)) != 0) return cc;
  if ((cc=fillstruct("AQS",
		       OFFSETIN(ActivityQueryStruct,starttime),
		       1,CF_LONG)) != 0) return cc;
  if ((cc=fillstruct("AQS",
		       OFFSETIN(ActivityQueryStruct,systemPollingRate),
		       1,CF_LONG)) != 0) return cc;
  if ((cc=fillstruct("AQS",
		       OFFSETIN(ActivityQueryStruct,numBkgTsks),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=fillstruct("AQS",
		       OFFSETIN(ActivityQueryStruct,numTotalContracts),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=fillstruct("AQS",
		       OFFSETIN(ActivityQueryStruct,numTotalClients),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=fillstruct("AQS",
		       OFFSETIN(ActivityQueryStruct,numTargetContracts),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=fillstruct("AQS",
		       OFFSETIN(ActivityQueryStruct,numTargetClients),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=fillstruct("AQS",
		       OFFSETIN(ActivityQueryStruct,numConnections),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=fillstruct("AQS",
		       OFFSETIN(ActivityQueryStruct,numConnectionTimeouts),
		       1,CF_LONG)) != 0) return cc;
  if ((cc=fillstruct("AQS",
		       OFFSETIN(ActivityQueryStruct,numConnectionArrivals),
		       1,CF_LONG)) != 0) return cc;
  if ((cc=fillstruct("AQS",
		       OFFSETIN(ActivityQueryStruct,numUDPpkts),
		       1,CF_LONG)) != 0) return cc;
  if ((cc=fillstruct("AQS",
		       OFFSETIN(ActivityQueryStruct,numTCPpkts),
		       1,CF_LONG)) != 0) return cc;
  if ((cc=fillstruct("AQS",
		       OFFSETIN(ActivityQueryStruct,numIPXpkts),
		       1,CF_LONG)) != 0) return cc;
  if ((cc=fillstruct("AQS",
		       OFFSETIN(ActivityQueryStruct,numSPXpkts),
		       1,CF_LONG)) != 0) return cc;
  if ((cc=fillstruct("AQS",
		       sizeof(ActivityQueryStruct),
		       1,CF_NULL)) != 0) return cc;

  ss = findstruct("AQS");
  ActivityQueryStructStruct = ss->s;

  if ((cc=fillstruct("CONQS",
		       OFFSETIN(ContractQueryStruct,contract.EqpProperty),
		       32,CF_TEXT)) != 0) return cc;
  if ((cc=fillstruct("CONQS",
		       OFFSETIN(ContractQueryStruct,contract.EqpDeviceName),
		       16,CF_TEXT)) != 0) return cc;
  if ((cc=fillstruct("CONQS",
		       OFFSETIN(ContractQueryStruct,contract.EqpName),
		       8,CF_TEXT)) != 0) return cc;
  if ((cc=fillstruct("CONQS",
		       OFFSETIN(ContractQueryStruct,contract.EqpSizeIn),
		       1,CF_LONG)) != 0) return cc;
  if ((cc=fillstruct("CONQS",
		       OFFSETIN(ContractQueryStruct,contract.EqpSizeOut),
		       1,CF_LONG)) != 0) return cc;
  if ((cc=fillstruct("CONQS",
		       OFFSETIN(ContractQueryStruct,contract.hEqpName),
		       1,CF_BYTE)) != 0) return cc;
  if ((cc=fillstruct("CONQS",
		       OFFSETIN(ContractQueryStruct,contract.EqpAccess),
		       1,CF_BYTE)) != 0) return cc;
  if ((cc=fillstruct("CONQS",
		       OFFSETIN(ContractQueryStruct,contract.EqpFormatIn),
		       1,CF_BYTE)) != 0) return cc;
  if ((cc=fillstruct("CONQS",
		       OFFSETIN(ContractQueryStruct,contract.EqpFormatOut),
		       1,CF_BYTE)) != 0) return cc;
  if ((cc=fillstruct("CONQS",
		       OFFSETIN(ContractQueryStruct,contract.strTagIn),
		       8,CF_TEXT)) != 0) return cc;
  if ((cc=fillstruct("CONQS",
		       OFFSETIN(ContractQueryStruct,contract.strTagOut),
		       8,CF_TEXT)) != 0) return cc;
  if ((cc=fillstruct("CONQS",
		       OFFSETIN(ContractQueryStruct,pollrate),
		       1,CF_LONG)) != 0) return cc;
  if ((cc=fillstruct("CONQS",
		       OFFSETIN(ContractQueryStruct,nconsumers),
		       1,CF_LONG)) != 0) return cc;
  if ((cc=fillstruct("CONQS",
		       sizeof(ContractQueryStruct),
		       1,CF_NULL)) != 0) return cc;

  ss = findstruct("CONQS");
  CONTRACTStruct = ss->s;

  if ((cc=fillstruct("CLNQS",
		       OFFSETIN(ClientQueryStruct,cln.userName),
		       16,CF_TEXT)) != 0) return cc;
  if ((cc=fillstruct("CLNQS",
		       OFFSETIN(ClientQueryStruct,cln.IPXaddress),
		       12,CF_BYTE)) != 0) return cc;
  if ((cc=fillstruct("CLNQS",
		       OFFSETIN(ClientQueryStruct,cln.node),
		       6,CF_BYTE)) != 0) return cc;
  if ((cc=fillstruct("CLNQS",
		       OFFSETIN(ClientQueryStruct,cln.IPaddress.sin_family),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=fillstruct("CLNQS",
		       OFFSETIN(ClientQueryStruct,cln.IPaddress.sin_port),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=fillstruct("CLNQS",
		       OFFSETIN(ClientQueryStruct,cln.IPaddress.sin_addr.s_addr),
		       4,CF_BYTE)) != 0) return cc;
  if ((cc=fillstruct("CLNQS",
		       OFFSETIN(ClientQueryStruct,cln.IPaddress.sin_zero),
		       8,CF_BYTE)) != 0) return cc;
  if ((cc=fillstruct("CLNQS",
		       OFFSETIN(ClientQueryStruct,cln.count),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=fillstruct("CLNQS",
		       OFFSETIN(ClientQueryStruct,cln.protocol),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=fillstruct("CLNQS",
		       sizeof(ClientQueryStruct),
		       1,CF_NULL)) != 0) return cc;

  ss = findstruct("CLNQS");
  ClientQueryStructStruct = ss->s;

  if ((cc=fillstruct("AMS",
		       OFFSETIN(AMS,tag),
		       16,CF_TEXT)) != 0) return cc;
  if ((cc=fillstruct("AMS",
		       OFFSETIN(AMS,device),
		       16,CF_TEXT)) != 0) return cc;
  if ((cc=fillstruct("AMS",
		       OFFSETIN(AMS,alarmTag),
		       16,CF_TEXT)) != 0) return cc;
  if ((cc=fillstruct("AMS",
		       OFFSETIN(AMS,alarmCode),
		       1,CF_LONG)) != 0) return cc;
  if ((cc=fillstruct("AMS",
		       OFFSETIN(AMS,timestamp),
		       1,CF_LONG)) != 0) return cc;
  if ((cc=fillstruct("AMS",
		       OFFSETIN(AMS,alarmMask),
		       1,CF_LONG)) != 0) return cc;
  if ((cc=fillstruct("AMS",
		       OFFSETIN(AMS,alarmData),
		       6,CF_BYTE)) != 0) return cc;
  if ((cc=fillstruct("AMS",
		       OFFSETIN(AMS,alarmDataFormat),
		       1,CF_BYTE)) != 0) return cc;
  if ((cc=fillstruct("AMS",
		       OFFSETIN(AMS,alarmDataArraySize),
		       1,CF_BYTE)) != 0) return cc;
  if ((cc=fillstruct("AMS",
		       OFFSETIN(AMS,severity),
		       1,CF_BYTE)) != 0) return cc;
  if ((cc=fillstruct("AMS",
		       OFFSETIN(AMS,descriptor),
		       1,CF_BYTE)) != 0) return cc;
  if ((cc=fillstruct("AMS",
		       OFFSETIN(AMS,alarmSystem),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=fillstruct("AMS",
		       sizeof(AMS),
		       1000,CF_NULL)) != 0) return cc;

  ss = findstruct("AMS");
  AMSStructStruct = ss->s;
    
  if ((cc=fillstruct("ADS",
		       OFFSETIN(ADS,alarmTag),
		       16,CF_TEXT)) != 0) return cc;
  if ((cc=fillstruct("ADS",
		       OFFSETIN(ADS,alarmCode),
		       1,CF_LONG)) != 0) return cc;
  if ((cc=fillstruct("ADS",
		       OFFSETIN(ADS,alarmMask),
		       1,CF_LONG)) != 0) return cc;
  if ((cc=fillstruct("ADS",
		       OFFSETIN(ADS,severity),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=fillstruct("ADS",
		       OFFSETIN(ADS,alarmDataFormat),
		       1,CF_BYTE)) != 0) return cc;
  if ((cc=fillstruct("ADS",
		       OFFSETIN(ADS,alarmDataArraySize),
		       1,CF_BYTE)) != 0) return cc;
  if ((cc=fillstruct("ADS",
		       OFFSETIN(ADS,timeRes),
		       1,CF_LONG)) != 0) return cc;
  if ((cc=fillstruct("ADS",
		       OFFSETIN(ADS,hitsChannel),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=fillstruct("ADS",
		       OFFSETIN(ADS,hitsModule),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=fillstruct("ADS",
		       OFFSETIN(ADS,alarmText),
		       40,CF_TEXT)) != 0) return cc;
  if ((cc=fillstruct("ADS",
		       OFFSETIN(ADS,deviceText),
		       38,CF_TEXT)) != 0) return cc;
  if ((cc=fillstruct("ADS",
		       OFFSETIN(ADS,alarmSystem),
		       1,CF_SHORT)) != 0) return cc;
  if ((cc=fillstruct("ADS",
		       OFFSETIN(ADS,dataText),
		       40,CF_TEXT)) != 0) return cc;
  if ((cc=fillstruct("ADS",
		       OFFSETIN(ADS,url),
		       40,CF_TEXT)) != 0) return cc;
  if ((cc=fillstruct("ADS",
		       OFFSETIN(ADS,ndefs),
		       1,CF_LONG)) != 0) return cc;
  if ((cc=fillstruct("ADS",
		       sizeof(ADS),
		       100,CF_NULL)) != 0) return cc;

  ss = findstruct("ADS");
  ADSStructStruct = ss->s;
  
  feclog("Control Structures Swap information registered");
  return 0;
}
