/** \file toolkit.c
 * \brief TINE API tools
 *
 * Many optional routines are offered to make programming easier.  Below are API routines which
 * are useful both for client-side programming and server-side programming.  Some query calls
 * such as GetDeviceNames(), etc. are clearly suitable for information gathering.  The set of
 * of data manipulation routines such at getValuesAsFloat() or putValuesFromLongEx() are 
 * designed for easy integration into server code, where format conversion, array-wrapping, 
 * array segmenting, etc. are automatically provided.  Helper routines such as sendPMTrigger()
 * are also provided.
 */
#ifndef NO_DOXYGEN
#include "prolog.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "rpcdefs.h"
#include "servdef.h"
#include "errors.h"
#include "swaplib.h"
#include "database.h"
#include "toolkit.h"

char feclogfile[80] = "";

static char *months[12] = {"Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"};
EXPORT time_t parseCtimeStr(char *cTimeStr)
{
  struct tm it;
  char *c = NULL;
  int i;

  for (i=0; i<12; i++) if ((c=strstr(cTimeStr,months[i])) != NULL) break;
  if (c == NULL) return (time_t)0;
  it.tm_mon = i; c += 3;
  it.tm_mday = atoi(c); c += 4;
  it.tm_hour = atoi(c); c += 3;
  it.tm_min = atoi(c);  c += 3;
  it.tm_sec = atoi(c);
  /* if this string comes from 'date' then there's a TZ field in front of the year */
  c = &cTimeStr[strlen(cTimeStr)-5];
  it.tm_year = atoi(c) - 1900;
  it.tm_isdst = -1;
  return mktime(&it);
}
EXPORT time_t parseCDateAndTimeStr(char *cDateStr,char *cTimeStr)
{
  struct tm it;
  char *c = NULL;
  int i;

  for (i=0; i<12; i++) if ((c=strstr(cDateStr,months[i])) != NULL) break;
  if (c == NULL) return (time_t)0;
  it.tm_mon = i; c += 3;
  it.tm_mday = atoi(c); c += 4;
  it.tm_year = atoi(c) - 1900;
  if ((c=cTimeStr) == NULL) return (time_t)0;
  it.tm_hour = atoi(c); c += 3;
  it.tm_min = atoi(c);  c += 3;
  it.tm_sec = atoi(c);  c += 3;
  it.tm_isdst = -1;
  return mktime(&it);
}
#ifndef SMALL_TINE_LIB
EXPORT int PropertyDataExchange(void *data,short fmt,short siz,
                        DTYPE *dout, DTYPE *din,short access)
{
  BYTE *dst=NULL,*src=NULL;
  int nbytes=0;
  if (din->dArrayLength > 0 && data != NULL)
  {
    if (!(access & CA_WRITE)) return illegal_read_write;
    switch(fmt)
    {
      case CF_SHORT: 
        return getValuesAsShort(din,(short *)data,siz);
      case CF_LONG:
        return getValuesAsLong(din,(UINT32 *)data,siz);
      case CF_FLOAT:
        return getValuesAsFloat(din,(float *)data,siz);
      case CF_DOUBLE:
        return getValuesAsDouble(din,(double *)data,siz);
      default:
        if (din->dFormat != fmt) return illegal_format;
        dst = (BYTE *)data; src = din->data.bptr; nbytes = (int)(din->dArrayLength*fmtsizeof(fmt));
        goto err;
    }
  }
  if (dout->dArrayLength > 0 && data != NULL)
  {
    switch(fmt)
    {
      case CF_SHORT: 
        return putValuesFromShort(dout,(short *)data,siz);
      case CF_LONG:
        return putValuesFromLong(dout,(UINT32 *)data,siz);
      case CF_FLOAT:
        return putValuesFromFloat(dout,(float *)data,siz);
      case CF_DOUBLE:
        return putValuesFromDouble(dout,(double *)data,siz);
      default:
        if (dout->dFormat != fmt) return illegal_format;
        dst = dout->data.bptr; src = (BYTE *)data; nbytes = (int)(dout->dArrayLength*fmtsizeof(fmt));
        goto err;
    }
  }
err:
  if (dst != NULL && src != NULL && nbytes > 0) memcpy(dst,src,nbytes);
  return 0;
}
#endif /* SMALL_TINE_LIB */
#if !defined(MSDOS) && !defined(SMALL_TINE_LIB)
int isAbsolutePath(char *fname);
extern char gFeclogPath[80];
time_t getTimeStampFromLogEntry(char *str)
{
  struct tm tms;
  tms.tm_mday = atoi(str);
  tms.tm_mon = atoi(&str[3]) - 1;
  tms.tm_year = 100 + atoi(&str[6]);
  tms.tm_hour = atoi(&str[9]);
  tms.tm_min = atoi(&str[12]);
  tms.tm_sec = atoi(&str[15]);
  tms.tm_isdst = memcmp(&str[22],ltzname[1],3) == 0 ? 1 : 0;
  return mktime(&tms);
}
int findLogFileEntry(FILE *fp,int fpos,time_t timestamp)
{
  int pos=fpos;
  time_t t=0;
  char str[256];

  fseek(fp,pos,SEEK_SET);
  while (fgets(str,255,fp) != NULL)
  {
    t = getTimeStampFromLogEntry(str);
    if (t > timestamp) return pos;
    pos += (int)strlen(str);
  }
  return pos;
}
int readFecLogFileEx(char *fname,char *textbuf,time_t start,time_t stop,int sizeInBytes)
{
  int cc=0,done=0,useAbsolutePath=0;
  FILE *fp = NULL, *fpb = NULL;
  char str[256],fn[128],bakname[96], *c;
  int sz=0,pos=0;
  int pos1=0,pos2=0;
  static char *fmod = "r";

  if (nofeclog) return 0;
  if (fname == NULL) return no_such_file;
  if (stop <= start) stop = time(NULL);
  useAbsolutePath = isAbsolutePath(fname);
  if (useAbsolutePath) strncpy(fn,fname,128); else sprintf(fn,"%s%s",gFeclogPath,fname);
  if ((fp=fopen(fn,fmod)) == NULL) goto checkbak;
  pos1 = findLogFileEntry(fp,0,start);
  pos2 = findLogFileEntry(fp,pos1,stop);
  pos = 0;
checkbak:
  if (pos1 == 0)  /* look for back file */
  {
    strncpy(bakname,fname,95); bakname[95] = 0;
    if ((c=strstr(bakname,".log")) != NULL) *c = 0;
    strcat(bakname,".bak");
    if (useAbsolutePath) strncpy(fn,bakname,128); else sprintf(fn,"%s%s",gFeclogPath,bakname);
    if ((fpb=fopen(fn,fmod)) == NULL) goto filesread;
    pos1 = findLogFileEntry(fpb,0,start);
    pos2 = findLogFileEntry(fpb,pos1,stop);
    fseek(fpb,pos1,SEEK_SET);
    while (fgets(str,255,fp) != NULL)
    {
      if ((sz=(int)strlen(str)) + pos > sizeInBytes) 
      {
        done = TRUE;
        break;
      }
      strncpy(&textbuf[pos],str,sz);
      if ((pos += sz) > pos2) 
      {
        done = TRUE;
        break;
      }
    }
    fclose(fpb);
    pos1 = 0;
  }
filesread:
  if (fp)
  {
    pos2 = findLogFileEntry(fp,pos1,stop);
    fseek(fp,pos1,SEEK_SET);
    while (fgets(str,255,fp) != NULL && !done)
    {
      if ((sz=(int)strlen(str)) + pos > sizeInBytes) 
      {
        done = TRUE;
        break;
      }
      strncpy(&textbuf[pos],str,sz);
      if ((pos += sz) > pos2) 
      {
        done = TRUE;
        break;
      }
    }
    fclose(fp);
  }
  textbuf[pos] = 0;
  if (sz == 0) cc = no_such_file;
  if (cc) dbglog("%s : %s",fn,erlst[cc]);
  return cc;
}
int GetTagList(char *context,char *tagtype,NAME16 *taglist,int *num)
{
  int cc = 0;
  short n = 0;
  char *c,prop[32],srv[64],*sub = NULL;
  DTYPE dout,din, *dinptr = NULL;

  if (num == NULL || *num <= 0) return buffer_too_small;
  strncpy(srv,CSHOSTS,16);
  /* is a context given ? */
  if (context != NULL && strlen(context) > 0) 
  {
    /* is a subsystem given ? */
    if ((c=strchr(context,'/')) != NULL)
    {
      /* has the form <context>/<subsystem> */
      *c++ = 0; sub = c;
    }
    if (strlen(context) > 0) /* might have been given as /<subsystem> ! */
    {
      strcat(srv,"/");
      strncat(srv,context,16);
    }
  }
  if (sub != NULL)
  {
    din.dFormat = CF_TEXT;
    din.dArrayLength = (UINT32)strlen(sub);
    din.data.cptr = sub;
    dinptr = &din;
  }
  dout.dFormat = CF_SHORT;
  dout.dArrayLength = 1;
  dout.data.vptr = &n;
  sprintf(prop,"N%s",tagtype);
  if ((cc=ExecLinkEx(srv,prop,&dout,dinptr,CA_READ|CA_RETRY,200)) != 0) return cc;
  if (n > *num) n = *num;

  if (taglist != NULL)
  { /* Get Tags from name server (synchronous call) */
    dout.dFormat = CF_STRING16;
    dout.dArrayLength = n;
    dout.data.vptr = taglist;
    if ((cc=ExecLinkEx(srv,tagtype,&dout,dinptr,CA_READ|CA_RETRY,200)) != 0) return cc;
  }
  *num = n;
  return 0;
}
int GetDevicePropertyInformationEx(char *srv,char *prp,PropertyQueryStructEx *srvProps,int *num)
{
  DTYPE dout, din, *dinptr = NULL;
  int cc = 0;
  short n;
  BYTE *data = (BYTE *)srvProps;

# ifndef INTEL
  data = tmpWorkArea;
# endif
  if (num == NULL || (n = *num) <= 0) return buffer_too_small;
  if (prp != NULL)
  {
    din.dArrayLength = 1;
    din.dFormat = CF_NAME32;
    din.data.cptr = prp;
    memset(din.dTag,0,8);
    dinptr = &din;
  }
  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = n*PROPERTYQUERYSTRUCTEX_SIZE;
  dout.data.vptr = data;
  if ((cc=ExecLinkEx(srv,"PROPS",&dout,dinptr,CA_READ|CA_RETRY,800)) != 0) return cc;
  *num = GetCompletionDataSize(-1);
# ifndef INTEL
  if (*num > 0) PrepIncomingPQSX(srvProps,data,*num);
# endif
  return 0;
}
#endif /* not MSDOS or SMALL_TINE_LIB */
#endif /* NO_DOXYGEN */

int fecLogFileDepth = 500;
#ifndef MSDOS
EXPORT int GetMyServerAddress(char *eqmName,char *expName,char *ctxName,char *fecName,int *port)
{
  int i,cc=0;
  short sizIn=0,fmtIn=CF_NULL;
  NAME16 n16[5];
  char *c = NULL;

  if (eqmName != NULL && strlen(eqmName) > 0)
  {
    strcpy(n16[0].name,eqmName);
    c = n16[0].name;
    sizIn = 1;
    fmtIn = CF_NAME16;
  }
  for (i=0,cc=link_not_open; cc != 0 && i <5; i++)
  {
    cc = tExecLinkEx(0,CSHOSTS,"#0","WHOAMI",5,CF_NAME16,n16[0].name,sizIn,fmtIn,c,CA_READ|CA_RETRY,200);
  }
  if (cc == 0)
  {
    if (expName != NULL) strncpy(expName,n16[0].name,ELEM_NAME_SIZE);
    if (fecName != NULL) strncpy(fecName,n16[2].name,FEC_NAME_SIZE);
    if (ctxName != NULL) strncpy(ctxName,n16[3].name,EQP_CONTEXT_SIZE);
    if (port != NULL) *port = atoi(n16[4].name);
  }
  return cc;
}
#endif /* MSDOS */
#if !defined(MSDOS) && !defined(SMALL_TINE_LIB)
EXPORT void AppendHistoryLog(char *text,...)
{
# ifndef FS_RDONLY
  double ts = makeDataTimeStamp();
  FILE *fp = NULL, *fpb = NULL;
  SINT32 flen;
  int l;
  char s[256],tmpfilename[128];
  va_list args;

  if (strlen(feclogfile) == 0) sprintf(feclogfile,"%s.log",gFecName);
  if ((fp=fopen(feclogfile,"a")) == NULL)
  {
    perror("LOG: fopen");
    goto error;
  }
  va_start(args,text);
  sprintf(s,"%s[%s] ",getDataTimeString(ts,0),gFecName);
  vsprintf(&s[strlen(s)],text,args);
  va_end(args);

  if (s[strlen(s)-1] != '\n') strcat(s,"\n");
  fprintf(fp,s);

  fseek(fp, 0L, SEEK_END);
  flen = ftell(fp);

  if (flen > 2*fecLogFileDepth*80)
  {
    fclose(fp);
    strncpy(tmpfilename,feclogfile,79); tmpfilename[79] = 0;
    strncpy(&tmpfilename[strlen(tmpfilename)-3],"BAK",3);
    rename(feclogfile,tmpfilename);
    if ((fp=fopen(feclogfile,"w")) == NULL) goto error;
    if ((fpb=fopen(tmpfilename,"r")) == NULL) goto error;
    for (l=0; fgets(s,255,fpb) != NULL; l++)
    {
      if (l < fecLogFileDepth) continue;
      fwrite(s,strlen(s),1,fp);
    }

  }
error:
  if (fp) fclose(fp);
  if (fpb) { fclose(fpb); remove(tmpfilename); }
# endif
  return;
}
EXPORT int GetDeviceContexts(NAME16 *clist,int *num)
{
  return GetTagList(NULL,"CONTEXTS",clist,num);
}
EXPORT int GetDeviceServersEx(char *context,NAME16 *dslist,int *num)
{
  return GetTagList(context,"TAGS",dslist,num);
}
EXPORT int GetDeviceServers(NAME16 *dslist,int *num)
{
  return GetTagList(NULL,"TAGS",dslist,num);
}
EXPORT int GetServersEx(char *context,NAME16 *slist,int *num)
{
  return GetTagList(context,"FECS",slist,num);
}
EXPORT int GetServers(NAME16 *slist,int *num)
{
  return GetTagList(NULL,"FECS",slist,num);
}
EXPORT int GetDevicePropertyEGU(char *srv,char *prp,float *max,float *min,char *egu)
{
  DTYPE dout;
  DUSTRING ustr;
  int cc = 0;
  char uprp[36];

  dout.dFormat = CF_USTRING;
  dout.dArrayLength = 1;
  dout.data.vptr = &ustr;
  strncpy(uprp,prp,28); uprp[28] = 0; strcat(uprp,".EGU");
  if ((cc=ExecLinkEx(srv,uprp,&dout,NULL,CA_READ|CA_RETRY,800)) != 0) return cc;
  if (max != NULL) *max = ustr.f1val;
  if (min != NULL) *min = ustr.f2val;
  if (egu != NULL) strncpy(egu,ustr.str,80);
  return 0;
}
EXPORT int GetTargetPropertyInformation(char *srv,char *prp,int *fmt,int *siz,char *dsc)
{
  DTYPE dout,din;
  int cc = 0;
  PropertyQueryStruct srvProps;
  NAME32 pname;

  if (prp != NULL) strncpy(pname.name,prp,32);
  din.dFormat = CF_NAME32;
  din.dArrayLength = 1;
  din.data.vptr = pname.name;
  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = PROPERTYQUERYSTRUCT_SIZE;
  dout.data.vptr = &srvProps;
  if ((cc=ExecLinkEx(srv,"PROPS",&dout,&din,CA_READ|CA_RETRY,800)) != 0) return cc;
  if (fmt != NULL) *fmt = srvProps.prpFormat;
  if (siz != NULL) *siz = ISWAP(srvProps.prpSize);
  if (dsc != NULL) strncpy(dsc,srvProps.prpDescription,32);
  return 0;
}
EXPORT int GetDevicePropertyInformation(char *srv,PropertyQueryStruct *srvProps,int *num)
{
  DTYPE dout;
  int cc = 0;
  short n, i;

  if (num == NULL || (n = *num) <= 0) return buffer_too_small;
  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = n*PROPERTYQUERYSTRUCT_SIZE;
  dout.data.vptr = srvProps;
  if ((cc=ExecLinkEx(srv,"PROPS",&dout,NULL,CA_READ|CA_RETRY,800)) != 0) return cc;
  for (i=0; i<n; i++) srvProps[i].prpSize = ISWAP(srvProps[i].prpSize);
  *num = GetCompletionDataSize(-1);
  return 0;
}
EXPORT int GetDevicePropertyInformationX(char *srv,char *prp,XPropertyQueryStruct *srvProps,int *num)
{
  DTYPE dout, din, *dinptr = NULL;
  int cc = 0;
  short n;
  BYTE *data = (BYTE *)srvProps;

# ifndef INTEL
  data = tmpWorkArea;
# endif
  if (num == NULL || (n = *num) <= 0) return buffer_too_small;
  if (prp != NULL)
  {
    din.dArrayLength = 1;
    din.dFormat = CF_NAME32;
    din.data.cptr = prp;
    memset(din.dTag,0,8);
    dinptr = &din;
  }
  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = n;
  strncpy(dout.dTag,"XPQS",8);
  dout.data.vptr = data;
  if ((cc=ExecLinkEx(srv,"PROPS",&dout,dinptr,CA_READ|CA_RETRY,800)) != 0) return cc;
  *num = GetCompletionDataSize(-1);
  return 0;
}
EXPORT int savePropertyValues(char *name,void *values,short format,int size)
{
  int i, cc = 0, len = 0;
  FILE *fp = NULL;
  char s[128], fn[128];

  if (name == NULL || values == NULL || size <= 0) ccerr(argument_list_error);
  sprintf(fn,"%s-setting.csv",name);
  if ((fp=fopen(fn,"w")) == NULL) ccerr(file_error);
  fprintf(fp,"VALUES\n");
  /* just take simple types in Release 3.XX */
  switch (format)
  {
    case CF_BYTE:
      for (i=0; i<size; i++) fprintf(fp,"%d\n",((BYTE *)values)[i]);
      break;
    case CF_SHORT:
      for (i=0; i<size; i++) fprintf(fp,"%d\n",((short *)values)[i]);
      break;
    case CF_LONG:
      for (i=0; i<size; i++) fprintf(fp,"%d\n",((UINT32 *)values)[i]);
      break;
    case CF_FLOAT:
      for (i=0; i<size; i++) fprintf(fp,"%g\n",((float *)values)[i]);
      break;
    case CF_DOUBLE:
      for (i=0; i<size; i++) fprintf(fp,"%g\n",((double *)values)[i]);
      break;
    case CF_NAME8:
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME48:
    case CF_NAME64:
      len = fmtsizeof(format);
      for (i=0; i<size; i++) 
      {
        strncpy(s,&((char *)values)[i*len],len);  s[len] = 0;
        fprintf(fp,"%s\n",s);
      }
      break;
    default: 
      ccerr(illegal_format);
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
EXPORT int restorePropertyValues(char *name,void *values,short format,int size)
{
  int i = 0, cc = 0, done = 0, col = -1, len = fmtsizeof(format);
  FILE *fp = NULL;
  char scratch[128], str[256], fn[128];

  if (name == NULL || values == NULL || size <= 0 || len <= 0) ccerr(argument_list_error);
  sprintf(fn,"%s-setting.csv",name);
  if ((fp=fopen(fn,"r")) == NULL) ccerr(no_such_file);
  memset(values,0,size*len);
  while (fgets(str,255,fp) != NULL)
  {
    if (i > size) break;
    strtrm(str);
    if (strlen(str) == 0) continue;               /* carraige return */
    if (strchr("%;#\n",str[0]) != NULL) continue; /* comment line */
    if (!done)
    {
      col = findcol(str,"VALUES");
      done = TRUE;
      continue;
    }
    if (col >= 0) getColumnStr(col,str,scratch,64); else strcpy(str,"0");
    scratch[64] = 0;
    switch (format)
    {
      case CF_BYTE:
        ((BYTE *)values)[i] = atoi(scratch);
        break;
      case CF_SHORT:
        ((short *)values)[i] = atoi(scratch);
        break;
      case CF_LONG:
        ((UINT32 *)values)[i] = atoi(scratch);
        break;
      case CF_FLOAT:
        ((float *)values)[i] = (float)strtod(scratch,NULL);
        break;
      case CF_DOUBLE:
        ((double *)values)[i] = strtod(scratch,NULL);
        break;
      case CF_NAME8:
      case CF_NAME16:
      case CF_NAME32:
      case CF_NAME48:
      case CF_NAME64:
        strncpy(&((char *)values)[i*len],scratch,len);
        break;
      default: 
        ccerr(illegal_format);
    }
    i++;
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}

#endif /* not MSDOS or SMALL_TINE_LIB */

EXPORT int GetDeviceProperties(char *srv,NAME32 *props,int *num)
{
  DTYPE dout;
  int cc = 0;
  short n;

  if ((n = *num) <= 0) return buffer_too_small;
  dout.dFormat = CF_NAME32;
  dout.dArrayLength = n;
  dout.data.cptr = (char *)props;
  if ((cc=ExecLinkEx(srv,"PROPS",&dout,NULL,CA_READ|CA_RETRY,800)) != 0 && !(cc & CE_SENDDATA)) return cc;
  *num = GetCompletionDataSize(-1);
  return 0;
}
EXPORT int GetDeviceNamesEx(char *srv,char *prp, NAME16 *devs,int *num)
{
  DTYPE dout;
  int i,cc = 0;
  short n;
  char tmpprp[64];
  ExportListStruct *el = getExportListItem(srv);
  NAME32 *lcldevlst;

  if ((n = *num) <= 0) return buffer_too_small;
  if (prp != NULL && strlen(prp) < 28)
  {
    sprintf(tmpprp,"%s.NAM",prp);
  }
  else
  {
    strcpy(tmpprp,"DEVICES");
  }
  if (el != NULL) /* local eqm name given */
  {
    lcldevlst = GetExportedDeviceList(srv);
    for (i=0; i<el->EqpNumModules && i<n; i++) 
      strncpy(devs[i].name,lcldevlst[i].name,16);
    *num = el->EqpNumModules;
  }
  else
  {
    dout.dFormat = CF_NAME16;
    dout.dArrayLength = n;
    dout.data.vptr = devs;
    if ((cc=ExecLinkEx(srv,tmpprp,&dout,NULL,CA_READ|CA_RETRY,800)) != 0 && !(cc & CE_SENDDATA)) return cc;
    *num = GetCompletionDataSize(-1);
  }
  return 0;
}
EXPORT int GetDeviceNames(char *srv,NAME16 *devs,int *num)
{
  return GetDeviceNamesEx(srv,NULL,devs,num);
}
#if !defined(MSDOS) && !defined(SMALL_TINE_LIB)
int parseFullDeviceName(char *fullname,char *dev,char *prp,int isHistoryCall)
{
  char *c,tmp[96];
  int isBracketed = FALSE;

  strncpy(tmp,fullname,96);
  if ((c=strchr(tmp,'[')) != 0)
  {
    *c++ = 0;
    strncpy(prp,c,32);
    if ((c=strchr(prp,']')) != 0) *c = 0;
    isBracketed = TRUE;
  }
  if (!isBracketed)
  {
    strrev(tmp);
    if ((c=strchr(tmp,'/')) == 0) return illegal_delimiter;
    *c++ = 0;
    strncpy(prp,tmp,32); strrev(prp);
    strncpy(dev,c,64); strrev(dev);
  }
  else
  {
    strncpy(dev,tmp,64);
  }
  if (isHistoryCall)
  {
    if (strstr(dev,"HISTORY") == NULL && /* not a central archive call */
      strstr(prp,".HIST") == NULL && strstr(prp,".HST") == NULL) /* meta property not yet appended */
    {
      if (strlen(prp) < 28) strcat(prp,".HST");
    }
  }
  return 0;
}
EXPORT int GetArchivedData(char *devsrv,int index,time_t start,time_t stop,FLTINT *fiDataArray,short *num)
{
  DTYPE din,dout;
  UINT32 startstopArray[3];
  char aServer[64],prp[32];
  int cc;

  if (num == NULL || *num <= 0) return buffer_too_small;
  if ((cc=parseFullDeviceName(devsrv,aServer,prp,TRUE)) != 0) return cc;

  startstopArray[0] = (UINT32)start;
  startstopArray[1] = (UINT32)stop;
  startstopArray[2] = index;
  dout.dArrayLength = *num;
  dout.dFormat = CF_FLTINT;
  dout.data.vptr = fiDataArray;
  din.dArrayLength = 3;
  din.dFormat = CF_LONG;
  din.data.lptr = startstopArray;

  if ((cc=ExecLinkEx(aServer,prp,&dout,&din,CA_READ|CA_RETRY,200)) != 0) *num = 0;
  else *num = (int)GetCompletionDataSize(-1);

  return cc;
}
EXPORT int GetArchivedTraceDataAsFloat(char *devsrv,time_t start,time_t stop,float *fDataArray,int num)
{
  DTYPE din,dout;
  UINT32 startstopArray[3];
  char aServer[64],prp[32];
  int cc;

  if (num <= 0) return buffer_too_small;
  if ((cc=parseFullDeviceName(devsrv,aServer,prp,TRUE)) != 0) return cc;
  startstopArray[0] = (UINT32)start;
  startstopArray[1] = (UINT32)stop;
  dout.dArrayLength = num;
  dout.dFormat = CF_FLOAT;
  dout.data.vptr = fDataArray;
  din.dArrayLength = 2;
  din.dFormat = CF_LONG;
  din.data.lptr = startstopArray;

  return ExecLinkEx(aServer,prp,&dout,&din,CA_READ|CA_RETRY,200);
}
EXPORT int GetArchivedDataAsSnapshot(char *devsrv,time_t *target,float *fDataArray,int size)
{
  DTYPE din,dout;
  UINT32 startstopArray[3];
  char aServer[64],prp[32];
  int cc;

  if (size <= 0) return buffer_too_small;
  if (target == NULL) return argument_list_error;
  if ((cc=parseFullDeviceName(devsrv,aServer,prp,TRUE)) != 0) return cc;
  startstopArray[0] = (UINT32) *target;
  startstopArray[1] = (UINT32) *target;
  dout.dArrayLength = size;
  dout.dFormat = CF_FLOAT;
  dout.data.vptr = fDataArray;
  din.dArrayLength = 2;
  din.dFormat = CF_LONG;
  din.data.lptr = startstopArray;

  if ((cc=ExecLinkEx(aServer,prp,&dout,&din,CA_READ|CA_RETRY,200)) != 0) *target = 0;
  else *target = (time_t)gDataTimeStamp;

  return cc;
}
EXPORT int GetArchivedDataAsFloat(char *devsrv,time_t start,time_t stop,FLTINT *fiDataArray,int *num)
{
  DTYPE din,dout;
  UINT32 startstopArray[3];
  char aServer[64],prp[32];
  int cc;

  if (num == NULL || *num <= 0) return buffer_too_small;
  if ((cc=parseFullDeviceName(devsrv,aServer,prp,TRUE)) != 0) return cc;
  startstopArray[0] = (UINT32)start;
  startstopArray[1] = (UINT32)stop;
  dout.dArrayLength = *num;
  dout.dFormat = CF_FLTINT;
  dout.data.vptr = fiDataArray;
  din.dArrayLength = 2;
  din.dFormat = CF_LONG;
  din.data.lptr = startstopArray;

  if ((cc=ExecLinkEx(aServer,prp,&dout,&din,CA_READ|CA_RETRY,200)) != 0) *num = 0;
  else *num = (int)GetCompletionDataSize(-1);

  return cc;
}
EXPORT int GetArchivedDataAsText(char *devsrv,time_t start,time_t stop,NAME32I *niDataArray,int *num)
{
  DTYPE din,dout;
  UINT32 startstopArray[3];
  char aServer[64],prp[32];
  int cc;

  if (num == NULL || *num <= 0) return buffer_too_small;
  if ((cc=parseFullDeviceName(devsrv,aServer,prp,TRUE)) != 0) return cc;
  startstopArray[0] = (UINT32)start;
  startstopArray[1] = (UINT32)stop;
  dout.dArrayLength = *num;
  dout.dFormat = CF_NAME32I;
  dout.data.vptr = niDataArray;
  din.dArrayLength = 2;
  din.dFormat = CF_LONG;
  din.data.lptr = startstopArray;

  if ((cc=ExecLinkEx(aServer,prp,&dout,&din,CA_READ|CA_RETRY,200)) != 0) *num = 0;
  else *num = (int)GetCompletionDataSize(-1);

  return cc;
}
#endif /* not MSDOS or SMALL_TINE_LIB */

EXPORT int getValuesAsString(DTYPE *d,char *str,UINT32 *dsiz)
{
  int i,n,len,siz,cc=0;
  int valueOnly = FALSE;
  time_t t;

  if (d == NULL) return argument_list_error;
  if (str == NULL || dsiz == NULL) return un_allocated;
  if ((siz = (int)(*dsiz)) < 16) return dimension_error;
  if (siz <= 64) valueOnly = TRUE;

  str[0] = 0; len = 0;
  n = (int)d->dArrayLength;
  switch (d->dFormat)
  {
    case CF_TEXT:
      len = MIN(siz,n);
      strncpy(str,d->data.cptr,len);
      if (n < siz) str[n] = 0;
      ccerr(0);
    case CF_BYTE:   
      for (i=0; i<n; i++) 
      {
        if (valueOnly) { sprintf(str,"%d",d->data.bptr[i]); goto err; }
        sprintf(&str[len],"%d\n",d->data.bptr[i]);
        if ((len=(int)strlen(str)) > siz + 16) break;
      }
      ccerr(0);
    case CF_SHORT:  
      for (i=0; i<n; i++) 
      {
        if (valueOnly) { sprintf(str,"%d",d->data.sptr[i]); goto err; }
        sprintf(&str[len],"%d\n",d->data.sptr[i]);
        if ((len=(int)strlen(str)) > siz + 16) break;
      }
      ccerr(0);
    case CF_LONG:
      for (i=0; i<n; i++) 
      {
#       ifdef MSDOS
        if (valueOnly) { sprintf(str,"%ld",d->data.lptr[i]); goto err; }
        sprintf(&str[len],"%ld\n",d->data.lptr[i]);
#       else
        if (valueOnly) { sprintf(str,"%d",d->data.lptr[i]); goto err; }
        sprintf(&str[len],"%d\n",d->data.lptr[i]);
#       endif
        if ((len=(int)strlen(str)) > siz + 16) break;
      }
      ccerr(0);
    case CF_FLOAT:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) { sprintf(str,"%g",d->data.fptr[i]); goto err; }
        sprintf(&str[len],"%g\n",d->data.fptr[i]);
        if ((len=(int)strlen(str)) > siz + 16) break;
      }
      ccerr(0);
    case CF_DOUBLE:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) { sprintf(str,"%g",d->data.dptr[i]); goto err; }
        sprintf(&str[len],"%g\n",d->data.dptr[i]);
        if ((len=(int)strlen(str)) > siz + 16) break;
      }
      ccerr(0);
    case CF_NAME8:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) { sprintf(str,"%s",((NAME8 *)d->data.vptr)[i].name); goto err; }
        sprintf(&str[len],"%s\n",((NAME8 *)d->data.vptr)[i].name);
        if ((len=(int)strlen(str)) > siz + 16) break;
      }
      ccerr(0);
    case CF_NAME16:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) { sprintf(str,"%s",((NAME16 *)d->data.vptr)[i].name); goto err; }
        sprintf(&str[len],"%s\n",((NAME16 *)d->data.vptr)[i].name);
        if ((len=(int)strlen(str)) > siz + 32) break;
      }
      ccerr(0);
    case CF_NAME32:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) { sprintf(str,"%s",((NAME32 *)d->data.vptr)[i].name); goto err; }
        sprintf(&str[len],"%s\n",((NAME32 *)d->data.vptr)[i].name);
        if ((len=(int)strlen(str)) > siz + 48) break;
      }
      ccerr(0);
    case CF_NAME48:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) { sprintf(str,"%s",((NAME48 *)d->data.vptr)[i].name); goto err; }
        sprintf(&str[len],"%s\n",((NAME48 *)d->data.vptr)[i].name);
        if ((len=(int)strlen(str)) > siz + 64) break;
      }
      ccerr(0);
    case CF_NAME64:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) { sprintf(str,"%s",((NAME64 *)d->data.vptr)[i].name); goto err; }
        sprintf(&str[len],"%s\n",((NAME64 *)d->data.vptr)[i].name);
        if ((len=(int)strlen(str)) > siz + 80) break;
      }
      ccerr(0);
#if !defined(MSDOS) && !defined(SMALL_TINE_LIB)
    case CF_NAME16I:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) { sprintf(str,"%s",((NAME16I *)d->data.vptr)[i].name); goto err; }
        sprintf(&str[len],"%s %d\n",((NAME16I *)d->data.vptr)[i].name,((NAME16I *)d->data.vptr)[i].ival);
        if ((len=(int)strlen(str)) > siz + 64) break;
      }
      ccerr(0);
    case CF_NAME32I:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) { sprintf(str,"%s",((NAME32I *)d->data.vptr)[i].name); goto err; }
        sprintf(&str[len],"%s %d\n",((NAME32I *)d->data.vptr)[i].name,((NAME32I *)d->data.vptr)[i].ival);
        if ((len=(int)strlen(str)) > siz + 64) break;
      }
      ccerr(0);
    case CF_NAME48I:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) { sprintf(str,"%s",((NAME48I *)d->data.vptr)[i].name); goto err; }
        sprintf(&str[len],"%s %d\n",((NAME48I *)d->data.vptr)[i].name,((NAME48I *)d->data.vptr)[i].ival);
        if ((len=(int)strlen(str)) > siz + 64) break;
      }
      ccerr(0);
    case CF_NAME64I:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) { sprintf(str,"%s",((NAME64I *)d->data.vptr)[i].name); goto err; }
        sprintf(&str[len],"%s %d\n",((NAME64I *)d->data.vptr)[i].name,((NAME64I *)d->data.vptr)[i].ival);
        if ((len=(int)strlen(str)) > siz + 80) break;
      }
      ccerr(0);
    case CF_FLTINT:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) goto err;
        sprintf(&str[len],"%g %d\n",((FLTINT *)d->data.vptr)[i].fval,((FLTINT *)d->data.vptr)[i].ival);
        if ((len=(int)strlen(str)) > siz + 32) break;
      }
      ccerr(0);
    case CF_LNGINT:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) goto err;
        sprintf(&str[len],"%d %d\n",((LNGINT *)d->data.vptr)[i].lval,((LNGINT *)d->data.vptr)[i].ival);
        if ((len=(int)strlen(str)) > siz + 32) break;
      }
      ccerr(0);
    case CF_DBLINT:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) goto err;
        sprintf(&str[len],"%g %d\n",((DBLINT *)d->data.vptr)[i].dval,((DBLINT *)d->data.vptr)[i].ival);
        if ((len=(int)strlen(str)) > siz + 32) break;
      }
      ccerr(0);
    case CF_INTFLTINT:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) goto err;
        sprintf(&str[len],"%d %g %d\n",((INTFLTINT *)d->data.vptr)[i].i1val,
          ((INTFLTINT *)d->data.vptr)[i].fval,((INTFLTINT *)d->data.vptr)[i].i2val);
        if ((len=(int)strlen(str)) > siz + 32) break;
      }
      ccerr(0);
    case CF_USTRING:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) goto err;
        t = (time_t)((DUSTRING *)d->data.vptr)[i].tm;
        sprintf(&str[len],"%g %g %d %s %s",((DUSTRING *)d->data.vptr)[i].f1val,
          ((DUSTRING *)d->data.vptr)[i].f2val,((DUSTRING *)d->data.vptr)[i].ival,
          ((DUSTRING *)d->data.vptr)[i].str,ctime(&t));
        if ((len=(int)strlen(str)) > siz + 128) break;
      }
      ccerr(0);
#   endif /* not MSDOS or SMALL_TINE_LIB */
    default:
      return illegal_format;
  }
err:
  *dsiz = len;
  return cc;
}
EXPORT int getValuesAsShort(DTYPE *d,short *sval,int num)
{
  int i,n;

  if (d == NULL) return argument_list_error;
  if (d->dArrayLength < 1 || num < 1) return dimension_error;
  n = MIN((int)d->dArrayLength,num);
  switch (d->dFormat)
  {
    case CF_BYTE:   for (i=n-1; i>=0; i--) sval[i] = (short)d->data.bptr[i];
                    return 0;
    case CF_SHORT:  if (sval == (short *)d->data.sptr) return 0;
		                memcpy(sval,d->data.sptr,n*sizeof(UINT16));
                    return 0;
    case CF_LONG:   for (i=0; i<n; i++) sval[i] = (short)((SINT32 *)d->data.lptr)[i];
                    return 0;
    case CF_FLOAT:  for (i=0; i<n; i++) sval[i] = (short)d->data.fptr[i];
                    return 0;
    case CF_DOUBLE: for (i=0; i<n; i++) sval[i] = (short)d->data.dptr[i];
                    return 0;
    default:        return illegal_format;
  }
}
EXPORT int getValuesAsFloat(DTYPE *d,float *fval,int num)
{
  int i,n;

  if (d == NULL) return argument_list_error;
  if (d->dArrayLength < 1 || num < 1) return dimension_error;
  n = MIN((int)d->dArrayLength,num);
  switch (d->dFormat)
  {
    case CF_BYTE:   for (i=n-1; i>=0; i--) fval[i] = (float)d->data.bptr[i];
                    return 0;
    case CF_SHORT:  for (i=n-1; i>=0; i--) fval[i] = (float)((short *)d->data.sptr)[i];
                    return 0;
    case CF_LONG:   for (i=0; i<n; i++) fval[i] = (float)((SINT32 *)d->data.lptr)[i];
                    return 0;
    case CF_FLOAT:  if (fval == d->data.fptr) return 0;
		                memcpy(fval,d->data.fptr,n*sizeof(float));
                    return 0;
    case CF_DOUBLE: for (i=0; i<n; i++) fval[i] = (float)d->data.dptr[i];
                    return 0;
    default:        return illegal_format;
  }
}
EXPORT int getValuesAsLong(DTYPE *d,UINT32 *lval,int num)
{
  int i,n;

  if (d == NULL) return argument_list_error;
  if (d->dArrayLength < 1 || num < 1) return dimension_error;
  n = MIN((int)d->dArrayLength,num);
  switch (d->dFormat)
  {
    case CF_BYTE:   for (i=n-1; i>=0; i--) lval[i] = (UINT32)d->data.bptr[i];
                    return 0;
    case CF_SHORT:  for (i=n-1; i>=0; i--) lval[i] = (UINT32)d->data.sptr[i];
                    return 0;
    case CF_LONG:   if (lval == d->data.lptr) return 0;
		                memcpy(lval,d->data.lptr,n*sizeof(UINT32));
                    return 0;
    case CF_FLOAT:  for (i=0; i<n; i++) lval[i] = (UINT32)d->data.fptr[i];
                    return 0;
    case CF_DOUBLE: for (i=0; i<n; i++) lval[i] = (UINT32)d->data.dptr[i];
                    return 0;
    default:        return illegal_format;
  }
}
EXPORT int getValuesAsDouble(DTYPE *d,double *dval,int num)
{
  int i,n;

  if (d == NULL) return argument_list_error;
  if (d->dArrayLength < 1 || num < 1) return dimension_error;
  n = MIN((int)d->dArrayLength,num);
  switch (d->dFormat)
  {
    case CF_BYTE:   for (i=n-1; i>=0; i--) dval[i] = (double)d->data.bptr[i];
                    return 0;
    case CF_SHORT:  for (i=n-1; i>=0; i--) dval[i] = (double)((short *)d->data.sptr)[i];
                    return 0;
    case CF_LONG:   for (i=n-1; i>=0; i--) dval[i] = (double)((SINT32 *)d->data.lptr)[i];
                    return 0;
    case CF_FLOAT:  for (i=n-1; i>=0; i--) dval[i] = (double)d->data.fptr[i];
                    return 0;
    case CF_DOUBLE: if (dval == d->data.dptr) return 0;
		                memcpy(dval,d->data.dptr,n*sizeof(double));
                    return 0;
    default:        return illegal_format;
  }
}
EXPORT int putValuesFromDoubleEx(DTYPE *d,double *dval,int num,int offset)
{
  int i,n,n1,n2;
  DTYPE dl;

  if (d == NULL) return argument_list_error;
  if ((d->dFormat != CF_TEXT) && (d->dArrayLength < 1 || num < 1)) return dimension_error;
  if (offset >= num) return dimension_error;
  n = MIN((int)d->dArrayLength,num);
  switch (d->dFormat)
  {
    case CF_BYTE:   for (i=0; i<n; i++) d->data.bptr[i] = (BYTE)dval[(i+offset)%num];
                    return 0;
    case CF_SHORT:  for (i=0; i<n; i++) d->data.sptr[i] = (short)dval[(i+offset)%num];
                    return 0;
    case CF_LONG:   for (i=0; i<n; i++) d->data.lptr[i] = (SINT32)dval[(i+offset)%num];
                    return 0;
    case CF_FLOAT:  for (i=0; i<n; i++) d->data.fptr[i] = (float)dval[(i+offset)%num];
                    return 0;
    case CF_DOUBLE: n1 = MIN(num-offset,n); n2 = n - n1;
                    if (n1 > 0) memcpy(d->data.dptr,&dval[offset],n1*sizeof(double));
                    if (n2 > 0) memcpy(&d->data.dptr[n1],dval,n2*sizeof(double));
                    return 0;
    case CF_TEXT:   dl.dArrayLength = num; dl.dFormat = CF_DOUBLE; dl.data.dptr = dval;
                    return getValuesAsString(&dl,d->data.cptr,&d->dArrayLength);
    default:        return illegal_format;
  }
}
EXPORT int putValuesFromFloatEx(DTYPE *d,float *fval,int num,int offset)
{
  int i,n,n1,n2;
  DTYPE dl;

  if (d == NULL) return argument_list_error;
  if ((d->dFormat != CF_TEXT) && (d->dArrayLength < 1 || num < 1)) return dimension_error;
  if (offset >= num) return dimension_error;
  n = MIN((int)d->dArrayLength,num);
  switch (d->dFormat)
  {
    case CF_BYTE:   for (i=0; i<n; i++) d->data.bptr[i] = (BYTE)fval[(i+offset)%num];
                    return 0;
    case CF_SHORT:  for (i=0; i<n; i++) d->data.sptr[i] = (short)fval[(i+offset)%num];
                    return 0;
    case CF_LONG:   for (i=0; i<n; i++) d->data.lptr[i] = (SINT32)fval[(i+offset)%num];
                    return 0;
    case CF_FLOAT:  n1 = MIN(num-offset,n); n2 = n - n1;
                    if (n1 > 0) memcpy(d->data.fptr,&fval[offset],n1*sizeof(float));
                    if (n2 > 0) memcpy(&d->data.fptr[n1],fval,n2*sizeof(float));
                    return 0;
    case CF_DOUBLE: for (i=0; i<n; i++) d->data.dptr[i] = (double)fval[(i+offset)%num];
                    return 0;
    case CF_TEXT:   dl.dArrayLength = num; dl.dFormat = CF_FLOAT; dl.data.fptr = fval;
                    return getValuesAsString(&dl,d->data.cptr,&d->dArrayLength);
    default:        return illegal_format;
  }
}
EXPORT int putValuesFromLongEx(DTYPE *d,UINT32 *lval,int num,int offset)
{
  int i,n,n1,n2;
  DTYPE dl;

  if (d == NULL) return argument_list_error;
  if ((d->dFormat != CF_TEXT) && (d->dArrayLength < 1 || num < 1)) return dimension_error;
  if (offset >= num) return dimension_error;
  n = MIN((int)d->dArrayLength,num);
  switch (d->dFormat)
  {
    case CF_BYTE:   for (i=0; i<n; i++) d->data.bptr[i] = (BYTE)lval[(i+offset)%num];
                    return 0;
    case CF_SHORT:  for (i=0; i<n; i++) d->data.sptr[i] = (short)lval[(i+offset)%num];
                    return 0;
    case CF_LONG:   n1 = MIN(num-offset,n); n2 = n - n1;
                    if (n1 > 0) memcpy(d->data.lptr,&lval[offset],n1*sizeof(UINT32));
                    if (n2 > 0) memcpy(&d->data.lptr[n1],lval,n2*sizeof(UINT32));
                    return 0;
    case CF_FLOAT:  for (i=0; i<n; i++) d->data.fptr[i] = (float)lval[(i+offset)%num];
                    return 0;
    case CF_DOUBLE: for (i=0; i<n; i++) d->data.dptr[i] = (double)lval[(i+offset)%num];
                    return 0;
    case CF_TEXT:   dl.dArrayLength = num; dl.dFormat = CF_LONG; dl.data.lptr = lval;
                    return getValuesAsString(&dl,d->data.cptr,&d->dArrayLength);
    default:        return illegal_format;
  }
}
EXPORT int putValuesFromShortEx(DTYPE *d,short *sval,int num,int offset)
{
  int i,n,n1,n2;
  DTYPE dl;

  if (d ==NULL) return argument_list_error;
  if ((d->dFormat != CF_TEXT) && (d->dArrayLength < 1 || num < 1)) return dimension_error;
  if (offset >= num) return dimension_error;
  n = MIN((int)d->dArrayLength,num);
  switch (d->dFormat)
  {
    case CF_BYTE:   for (i=0; i<n; i++) d->data.bptr[i] = (BYTE)sval[(i+offset)%num];
                    return 0;
    case CF_SHORT:  n1 = MIN(num-offset,n); n2 = n - n1;
                    if (n1 > 0) memcpy(d->data.sptr,&sval[offset],n1*sizeof(UINT16));
                    if (n2 > 0) memcpy(&d->data.sptr[n1],sval,n2*sizeof(UINT16));
                    return 0;
    case CF_LONG:   for (i=0; i<n; i++) d->data.lptr[i] = (SINT32)sval[(i+offset)%num];
                    return 0;
    case CF_FLOAT:  for (i=0; i<n; i++) d->data.fptr[i] = (float)sval[(i+offset)%num];
                    return 0;
    case CF_DOUBLE: for (i=0; i<n; i++) d->data.dptr[i] = (double)sval[(i+offset)%num];
                    return 0;
    case CF_TEXT:   dl.dArrayLength = num; dl.dFormat = CF_SHORT; dl.data.sptr = (UINT16 *)sval;
                    return getValuesAsString(&dl,d->data.cptr,&d->dArrayLength);
    default:        return illegal_format;
  }
}
EXPORT int putValuesFromUnsignedShortEx(DTYPE *d,UINT16 *sval,int num,int offset)
{
  int i,n,n1,n2;
  DTYPE dl;

  if (d ==NULL) return argument_list_error;
  if ((d->dFormat != CF_TEXT) && (d->dArrayLength < 1 || num < 1)) return dimension_error;
  if (offset >= num) return dimension_error;
  n = MIN((int)d->dArrayLength,num);
  switch (d->dFormat)
  {
    case CF_BYTE:   for (i=0; i<n; i++) d->data.bptr[i] = (BYTE)sval[(i+offset)%num];
                    return 0;
    case CF_SHORT:  n1 = MIN(num-offset,n); n2 = n - n1;
                    if (n1 > 0) memcpy(d->data.sptr,&sval[offset],n1*sizeof(UINT16));
                    if (n2 > 0) memcpy(&d->data.sptr[n1],sval,n2*sizeof(UINT16));
    case CF_LONG:   for (i=0; i<n; i++) d->data.lptr[i] = (SINT32)sval[(i+offset)%num];
                    return 0;
    case CF_FLOAT:  for (i=0; i<n; i++) d->data.fptr[i] = (float)sval[(i+offset)%num];
                    return 0;
    case CF_DOUBLE: for (i=0; i<n; i++) d->data.dptr[i] = (double)sval[(i+offset)%num];
                    return 0;
    case CF_TEXT:   dl.dArrayLength = num; dl.dFormat = CF_SHORT; dl.data.sptr = sval;
                    return getValuesAsString(&dl,d->data.cptr,&d->dArrayLength);
    default:        return illegal_format;
  }
}
EXPORT int sendPMTrigger(char *dev,char *cmt,short triggerLevel)
{
  DTYPE din;
  char triggerDevName[64];
  char cmtbuf[256];
  int cc;

  if (dev == NULL || strlen(dev) == 0) return illegal_equipment_name;
  if (strchr(dev,'/') == 0)
  {
    if (strlen(dev) > 16) return illegal_equipment_name;
    sprintf(triggerDevName,"/%s/STORAGE/%s",gDeviceContext,dev);
  }
  else strncpy(triggerDevName,dev,64);

  din.dFormat = CF_SHORT;
  din.dArrayLength = 1;
  din.data.sptr = (UINT16 *)&triggerLevel;
  if ((cc=ExecLinkEx(triggerDevName,"TRIGGER",NULL,&din,CA_WRITE,200)) != 0) return cc;

  /* send comment buffer */
  strncpy(cmtbuf,cmt,255);
  din.dFormat = CF_TEXT;
  din.data.cptr = cmtbuf;
  sprintf(&cmtbuf[strlen(cmtbuf)],"@%s",getDataTimeString((double)time(NULL),FALSE));
  din.dArrayLength = (UINT32)strlen(cmtbuf);
  if ((cc=ExecLinkEx(triggerDevName,"ANNOTATE",NULL,&din,CA_WRITE,200)) != 0) return cc;
  return 0;
}
