#ifndef _TOOLKIT_H_
#define _TOOLKIT_H_

#include "project.def"

#include "errors.h"
#include "rpctypes.h"

#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif

/* fecLogFileDepth */
/**
 * \brief Sets the depth in lines of a server's secondary log file (if utilized).
 * 
 * For most cases the system log file "fec.log" is sufficient for logging server 
 * activities.  Furthermore, "fec.log" is retrievable via a systematic TINE call
 * so that it can be viewed from a remote location without requiring a mount to a 
 * file system.  However, for more detailing logging you may want to make use of a
 * secondary log file whose name is given by <fecname>.log.  This log file resides
 * only on the local file system.  You can fix the depth of the this log file by
 * setting this parameter.
 *
 * \see AppendHistoryLog()
 *
 */
extern int fecLogFileDepth;

/* getValuesAsShort() */
/**
 * \brief Retrieves incoming data as an array of short integers
 * 
 * This routine will convert incoming data if possible to an array of short integers.  If
 * it is not possible to convert to an array of short integers an error is returned.
 *
 * \param d is a TINE DTYPE data object containing the incoming data.
 * \param sval is a pointer to the short integer array to receive the converted input data
 * \num is the maximum number of array elements which the short integer array can hold
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \see getValuesAsLong(), getValuesAsFloat(), getValuesAsDouble()
 *
 * \b Alias: GetValuesAsShort
 *
 * \b Example:
 *
 * \include eg_getValuesAsShort.c
 */
EXPORT int getValuesAsShort(DTYPE *d,short *sval,int num);
#define GetValuesAsShort(d,s,n) getValuesAsShort(d,s,n)

/* putValuesFromShortEx() */
/**
 * \brief Submits outgoing data as an array of short integers
 * 
 * This routine will convert outgoing data if possible to an array of short integers.  If
 * it is not possible to convert to an array of short integers an error is returned.
 *
 * \param d is a TINE DTYPE data object into which the outgoing data is to be copied.
 * \param sval is a pointer to the short integer array containing the source data
 * \num is the number of array elements contained in the source array
 * \offset is the starting point in the source data array
 *
 * \note This is the extended version of putValuesFromShort(), the difference being that
 * putValuesFromShortEx() takes a pointer to the entire source array, will begin copying
 * at the offset given, and will wrap the source array if necessary.  This feature allows a
 * caller to receive an entire array of data (an orbit for instance) but have the initial
 * element begin at a location of his choosing.
 *
 * \note The macro putValuesFromShort is a call to putValuesFromShortEx() with offset 0
 * 
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \see putValuesFromUnsignedShortEx(), putValuesFromLongEx(), putValuesFromFloatEx(), putValuesFromDoubleEx()
 *
 * \b Alias: PutValuesFromShortEx
 *
 * \b Example:
 *
 * \include eg_putValuesFromShortEx.c
 */
EXPORT int putValuesFromShortEx(DTYPE *d,short *sval,int num,int offset);
#define putValuesFromShort(d,s,n) putValuesFromShortEx(d,s,n,0)
#define PutValuesFromShort(d,s,n) putValuesFromShortEx(d,s,n,0)
#define PutValuesFromShortEx(d,s,n,o) putValuesFromShortEx(d,s,n,o)

/* getValuesAsFloat() */
/**
 * \brief Retrieves incoming data as an array of floats
 * 
 * This routine will convert incoming data if possible to an array of floats.  If
 * it is not possible to convert to an array of floats an error is returned.
 *
 * \param d is a TINE DTYPE data object containing the incoming data.
 * \param fval is a pointer to the float array to receive the converted input data
 * \num is the maximum number of array elements which the float array can hold
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \see getValuesAsShort(), getValuesAsLong(), getValuesAsDouble()
 *
 * \b Alias: GetValuesAsFloat()
 *
 * \b Example:
 * (see getValuesAsShort() for example)
 */
EXPORT int getValuesAsFloat(DTYPE *d,float *fval,int num);
#define GetValuesAsFloat(d,s,n) getValuesAsFloat(d,s,n)

/* putValuesFromFloatEx() */
/**
 * \brief Submits outgoing data as an array of floats
 * 
 * This routine will convert outgoing data if possible to an array of floats.  If
 * it is not possible to convert to an array of floats an error is returned.
 *
 * \param d is a TINE DTYPE data object into which the outgoing data is to be copied.
 * \param fval is a pointer to the float array containing the source data
 * \num is the number of array elements contained in the source array
 * \offset is the starting point in the source data array
 *
 * \note This is the extended version of putValuesFromFloat(), the difference being that
 * putValuesFromFloatEx() takes a pointer to the entire source array, will begin copying
 * at the offset given, and will wrap the source array if necessary.  This feature allows a
 * caller to receive an entire array of data (an orbit for instance) but have the initial
 * element begin at a location of his choosing.
 *
 * \note The macro putValuesFromFloat is a call to putValuesFromFloatEx() with offset 0
 * 
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \see putValuesFromUnsignedShortEx(), putValuesFromShortEx(), putValuesFromLongEx(), putValuesFromDoubleEx()
 *
 * \b Alias: PutValuesFromFloatEx
 *
 * \b Example:
 *
 * \include eg_putValuesFromFloatEx.c
 */
EXPORT int putValuesFromFloatEx(DTYPE *d,float *fval,int num,int offset);
#define putValuesFromFloat(d,s,n) putValuesFromFloatEx(d,s,n,0)
#define PutValuesFromFloat(d,s,n) putValuesFromFloatEx(d,s,n,0)
#define PutValuesFromFloatEx(d,s,n,o) putValuesFromFloatEx(d,s,n,o)

/* getValuesAsDouble() */
/**
 * \brief Retrieves incoming data as an array of doubles
 * 
 * This routine will convert incoming data if possible to an array of doubles.  If
 * it is not possible to convert to an array of doubles an error is returned.
 *
 * \param d is a TINE DTYPE data object containing the incoming data.
 * \param dval is a pointer to the double array to receive the converted input data
 * \num is the maximum number of array elements which the double array can hold
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \see getValuesAsShort(), getValuesAsLong(), getValuesAsFloat()
 *
 * \b Alias: GetValuesAsDouble()
 *
 * \b Example:
 * (see getValuesAsShort() for example)
 */
EXPORT int getValuesAsDouble(DTYPE *d,double *dval,int num);
#define GetValuesAsDouble(d,s,n) getValuesAsDouble(d,s,n)

/* putValuesFromDoubleEx() */
/**
 * \brief Submits outgoing data as an array of doubles
 * 
 * This routine will convert outgoing data if possible to an array of doubles.  If
 * it is not possible to convert to an array of doubles an error is returned.
 *
 * \param d is a TINE DTYPE data object into which the outgoing data is to be copied.
 * \param fval is a pointer to the double array containing the source data
 * \num is the number of array elements contained in the source array
 * \offset is the starting point in the source data array
 *
 * \note This is the extended version of putValuesFromDouble(), the difference being that
 * putValuesFromDoubleEx() takes a pointer to the entire source array, will begin copying
 * at the offset given, and will wrap the source array if necessary.  This feature allows a
 * caller to receive an entire array of data (an orbit for instance) but have the initial
 * element begin at a location of his choosing.
 *
 * \note The macro putValuesFromDouble is a call to putValuesFromDoubleEx() with offset 0
 * 
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \see putValuesFromShortEx(), putValuesFromLongEx(), putValuesFromFloatEx()
 *
 * \b Alias: PutValuesFromDoubleEx
 *
 * \b Example:
 * (see putValuesFromFloatEx() for example)
 */
EXPORT int putValuesFromDoubleEx(DTYPE *d,double *dval,int num,int offset);
#define putValuesFromDouble(d,s,n) putValuesFromDoubleEx(d,s,n,0)
#define PutValuesFromDouble(d,s,n) putValuesFromDoubleEx(d,s,n,0)
#define PutValuesFromDoubleEx(d,s,n,o) putValuesFromDoubleEx(d,s,n,o)

/* getValuesAsLong() */
/**
 * \brief Retrieves incoming data as an array of long integers
 * 
 * This routine will convert incoming data if possible to an array of long integers.  If
 * it is not possible to convert to an array of long integers an error is returned.
 *
 * \param d is a TINE DTYPE data object containing the incoming data.
 * \param lval is a pointer to the long integer array to receive the converted input data
 * \num is the maximum number of array elements which the long integer array can hold
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \see getValuesAsShort(), getValuesAsFloat(), getValuesAsDouble()
 *
 * \b Alias: GetValuesAsLong()
 *
 * \b Example:
 * (see getValuesAsShort() for example)
 */
EXPORT int getValuesAsLong(DTYPE *d,UINT32 *lval,int num);
#define GetValuesAsLong(d,s,n) getValuesAsLong(d,s,n)

/* putValuesFromLongEx() */
/**
 * \brief Submits outgoing data as an array of long integers
 * 
 * This routine will convert outgoing data if possible to an array of long integers.  If
 * it is not possible to convert to an array of long integers an error is returned.
 *
 * \param d is a TINE DTYPE data object into which the outgoing data is to be copied.
 * \param fval is a pointer to the long integer array containing the source data
 * \num is the number of array elements contained in the source array
 * \offset is the starting point in the source data array
 *
 * \note This is the extended version of putValuesFromLong(), the difference being that
 * putValuesFromLongEx() takes a pointer to the entire source array, will begin copying
 * at the offset given, and will wrap the source array if necessary.  This feature allows a
 * caller to receive an entire array of data (an orbit for instance) but have the initial
 * element begin at a location of his choosing.
 *
 * \note The macro putValuesFromLong is a call to putValuesFromLongEx() with offset 0
 * 
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \see putValuesFromUnsignedShortEx(), putValuesFromShortEx(), putValuesFromFloatEx(), putValuesFromDoubleEx()
 *
 * \b Alias: PutValuesFromLongEx
 *
 * \b Example:
 * (see putValuesFromFloatEx() for example)
 */
EXPORT int putValuesFromLongEx(DTYPE *d,UINT32 *lval,int num,int offset);
#define putValuesFromLong(d,s,n) putValuesFromLongEx(d,s,n,0)
#define PutValuesFromLong(d,s,n) putValuesFromLongEx(d,s,n,0)
#define PutValuesFromLongEx(d,s,n,o) putValuesFromLongEx(d,s,n,o)

/* putValuesFromUnsignedShortEx() */
/**
 * \brief Submits outgoing data as an array of unsigned short integers
 * 
 * This routine will convert outgoing data if possible to an array of unsigned short integers.  If
 * it is not possible to convert to an array of unsigned short integers an error is returned.
 *
 * \param d is a TINE DTYPE data object into which the outgoing data is to be copied.
 * \param fval is a pointer to the unsigned short integer array containing the source data
 * \num is the number of array elements contained in the source array
 * \offset is the starting point in the source data array
 *
 * \note This is the extended version of putValuesFromUnsignedShort(), the difference being that
 * putValuesFromUnsignedShortEx() takes a pointer to the entire source array, will begin copying
 * at the offset given, and will wrap the source array if necessary.  This feature allows a
 * caller to receive an entire array of data (an orbit for instance) but have the initial
 * element begin at a location of his choosing.
 *
 * \note The macro putValuesFromUnsignedShort is a call to putValuesFromUnsignedShortEx() with offset 0
 * 
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \see putValuesFromShortEx(), putValuesFromFloatEx(), putValuesFromDoubleEx()
 *
 * \b Alias: PutValuesFromUnsignedShortEx
 *
 * \b Example:
 * (see putValuesFromFloatEx() for example)
 */
EXPORT int putValuesFromUnsignedShortEx(DTYPE *d,UINT16 *sval,int num,int offset);
#define putValuesFromUnsignedShort(d,s,n) putValuesFromUnsignedShortEx(d,s,n,0)
#define PutValuesFromUnsignedShort(d,s,n) putValuesFromUnsignedShortEx(d,s,n,0)
#define PutValuesFromUnsignedShortEx(d,s,n,o) putValuesFromUnsignedShortEx(d,s,n,o)

/* getValuesAsString() */
/**
 * \brief Retrieves incoming data as a string buffer
 * 
 * This routine will convert incoming data if possible to a string representation.  If
 * it is not possible to convert to a string an error is returned.
 *
 * \param d is a TINE DTYPE data object containing the incoming data.
 * \param str is a pointer to the string buffer to receive the converted input data
 * \num is the maximum size of the string buffer
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \see getValuesAsShort(), getValuesAsLong(), getValuesAsFloat(), getValuesAsDouble()
 *
 * \b Alias: GetValuesAsString()
 */
EXPORT int getValuesAsString(DTYPE *d,char *str,UINT32 *dsiz);
#define GetValuesAsString(d,s,n) getValuesAsString(d,s,n)

/* AppendHistoryLog() */
/**
 * \brief Appends a line of text to a server's secondary log file
 * 
 * For most cases the system log file "fec.log" is sufficient for logging server 
 * activities.  Furthermore, "fec.log" is retrievable via a systematic TINE call
 * so that it can be viewed from a remote location without requiring a mount to a 
 * file system.  However, for more detailing logging you may want to make use of a
 * secondary log file whose name is given by <fecname>.log.  This log file resides
 * only on the local file system.  
 *
 * \param str is a 'printf'-like variable argument list string containing the text to be appended
 *
 * \return none
 *
 * \see fecLogFileDepth
 *
 */
EXPORT void AppendHistoryLog(char *str,...);

/* sendPMTrigger() */
/**
 * \brief Sends a Post-Mortem (i.e. event) Trigger to the designated Post-Mortem Server
 * 
 * By making this call, a configured Post-Mortem Server can be sent a trigger initializing 
 * the collection of data for the data set specified.
 *
 * \param devname is the trigger-appended full device name of the post mortem server to 
 * receive the trigger.  For instance if a post-mortem server "PMSERVER" in context "SERVICE"
 * reacts to a trigger "p-loss", then specifying "/SERVICE/PMSERVER/p-loss" as this paramter
 * will trigger data collection.
 * \param cmt is any accompanying comment text which should be associated with the post-mortem
 * event.  This is typically the reason for the event archive (which modules signalled the 
 * alarm, etc.).
 * \param triggerLevel is the level at which the post-mortem event is initialized.  Typically
 * a value of "1" should be used to signify beginning at the first step.  If a value of "-1" is
 * passed, the post-mortem event will forward the trigger to all other post-mortem servers 
 * indentifying itself as the initialized.  Other post-mortem servers which may or may not have
 * missed the event can then either react or not according to some pre-coded event critiera.
 * A value of "0" has no effect, and values greater than "1" will step the data recording 
 * sequence at a higher (usually unknown) level.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \see GetDeviceServersEx()
 *
 */
EXPORT int sendPMTrigger(char *devname,char *cmt,short triggerLevel);
int GetTagList(char *context,char *tagtype,NAME16 *taglist,int *num);

/* GetDeviceServers() */
/**
 * \brief Retrieves a list of servers via query to the Equipment Name Server
 * 
 * This call simply retrieves a list of all devices servers irrespective of server
 * context.  For a context specific list, please use the extended call GetDeviceServersEx().
 *
 * \param dslist is a pointer to a NAME16 buffer to receive the name list.
 * \param num is the size of the buffer pointed to by the first parameter.  Note that this
 * buffer space must be pre-allocated, otherwise a system crash will ensue.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \see GetDeviceServersEx()
 *
 */
EXPORT int GetDeviceServers(NAME16 *dslist,int *num);

/* GetDeviceServersEx() */
/**
 * \brief Retrieves a list of servers via query to the Equipment Name Server
 * 
 * This call retrieves a list of all devices servers associated with the
 * context given.  
 *
 * \param context is the context for which the query is to be made
 * \param dslist is a pointer to a NAME16 buffer to receive the name list.
 * \param num is the size of the buffer pointed to by the first parameter.  Note that this
 * buffer space must be pre-allocated, otherwise a system crash will ensue.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \see GetDeviceServers()
 *
 */
EXPORT int GetDeviceServersEx(char *context,NAME16 *dslist,int *num);

/* GetDeviceContexts() */
/**
 * \brief Retrieves a list of server contexts via query to the Equipment Name Server
 * 
 * This call retrieves a list of all server context managed by the Equipment Name Server
 *
 * \param clist is a pointer to a NAME16 buffer to receive the name list.
 * \param num is the size of the buffer pointed to by the first parameter.  Note that this
 * buffer space must be pre-allocated, otherwise a system crash will ensue.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 */
EXPORT int GetDeviceContexts(NAME16 *clist,int *num);

/* GetServers() */
/**
 * \brief Retrieves a list of servers via query to the Equipment Name Server
 * 
 * This call simply retrieves a list of all Front End Computer Names (FECs) irrespective of server
 * context.  For a context specific list, please use the extended call GetServersEx().
 *
 * \param slist is a pointer to a NAME16 buffer to receive the name list.
 * \param num is the size of the buffer pointed to by the first parameter.  Note that this
 * buffer space must be pre-allocated, otherwise a system crash will ensue.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \see GetServersEx()
 *
 */
EXPORT int GetServers(NAME16 *slist,int *num);

/* GetServersEx() */
/**
 * \brief Retrieves a list of FECs via query to the Equipment Name Server
 * 
 * This call retrieves a list of all Front End Computer Names (FECs) associated with the server
 * context given.
 *
 * \param context is the context for which the query is to be made
 * \param slist is a pointer to a NAME16 buffer to receive the name list.
 * \param num is the size of the buffer pointed to by the first parameter.  Note that this
 * buffer space must be pre-allocated, otherwise a system crash will ensue.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \see GetServers()
 *
 */
EXPORT int GetServersEx(char *context,NAME16 *slist,int *num);

/* GetDeviceProperties() */
/**
 * \brief Retrieves a list of properties via query to the server given
 * 
 * This call retrieves a list of all properties registered on the server associated with 
 * full device server name specified.  A classic device server will export its behavior through its properties,
 * and export its "instances" through its "device names", according to the motto "All 
 * devices found on this server have the same set of properties, and all properties apply to
 * any device found on this server".  In this case, simply specifying the device server name
 * as the first input parameter will return the set of properties valid for all devices registered
 * on the server.  Under some circumstances, however, the set of properties associated with a
 * particular device might be different from device to device. If this is the case, 
 * the full device server name should be used as the first parameter, where the target device
 * name is also sent as part of the query.
 *
 * \param srv is the full device server name for which the property query is to be made
 * \param plist is a pointer to a NAME32 buffer to receive the name list.
 * \param num is the size of the buffer pointed to by the first parameter.  Note that this
 * buffer space must be pre-allocated, otherwise a system crash will ensue.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \see GetDevicePropertyInformation(), GetTargetPropertyInformation(), GetDevicePropertyEGU()
 *
 */
EXPORT int GetDeviceProperties(char *srv,NAME32 *plist,int *num);

/* GetTargetPropertyInformation() */
/**
 * \brief Retrieves the default set of property information parameters for the property specified
 * 
 * This call retrieves the default set of property information parameters for the property specified.
 *
 * \param srv [in] is the full device server name for which the property query is to be made
 * \param prp [in] is the property for which the information is desired.
 * \param fmt [out] is a pointer to receive the TINE format data type (if non-NULL).
 * \param siz [out] is a pointer to receive the data size (if non-NULL)
 * \param dsc [out] is a string buffer to receive the property description (if non-NULL).
 * Note that buffer space must be pre-allocated, otherwise a system crash will ensue.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \note In cases where property behavior has been overloaded, this call will only retrieve
 * the properties "default" behavior, i.e. the information pertaining to the last registered 
 * property.
 *
 * \see GetDevicePropertyInformation(), GetDeviceProperties(), GetDevicePropertyEGU()
 *
 */
EXPORT int GetTargetPropertyInformation(char *srv,char *prp,int *fmt,int *siz,char *dsc);

/* GetDevicePropertyInformation() */
/**
 * \brief Retrieves a list of property query structures for the device server specified
 * 
 * This call retrieves a list of property information parameters in the form of 
 * PropertyQueryStruct structures for the device server specified.
 *
 * \param srv [in] is the full device server name for which the property query is to be made
 * \param srvProps [out] is a pointer to an array of PropertyQueryStruct objects.
 * Note that buffer space must be pre-allocated, otherwise a system crash will ensue.
 * \param num [in/out] as input is a pointer to the maximum number of PropertyQueryStruct 
 * objects which srvProps can hold. As output it contains the total number of properties.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \see GetTargetPropertyInformation(), GetDeviceProperties(), GetDevicePropertyInformationX()
 *
 */
EXPORT int GetDevicePropertyInformation(char *srv,PropertyQueryStruct *srvProps,int *num);
int GetDevicePropertyInformationEx(char *srv,char *prp,PropertyQueryStructEx *srvProps,int *num);

/* GetDevicePropertyInformationX() */
/**
 * \brief Retrieves a list of extended property query structures for the device server and target 
 * property specified
 * 
 * This call retrieves a list of all relevant property information parameters in the form of 
 * extended XPropertyQueryStruct structures for the device server and target property specified.
 *
 * \param srv [in] is the full device server name for which the property query is to be made
 * \param prp [in] is the property name for which the extended property query is to be made. 
 * Passing a NULL for this parameter will result in the query returning information for all
 * properties until the buffer space (given by 'num') is exhausted.  Passing a wildcard 
 * character '*' will return information for those properties matching the pattern given.
 * \param srvProps [out] is a pointer to an array of XPropertyQueryStruct objects.  The
 * extended property query objects contain all detailed information concerning the queried
 * property.
 * Note that buffer space must be pre-allocated, otherwise a system crash will ensue.
 * \param num [in/out] as input is a pointer to the maximum number of XPropertyQueryStruct 
 * objects which srvProps can hold. As output it contains the total number of properties.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \see GetTargetPropertyInformation(), GetDeviceProperties(), GetDevicePropertyInformation()
 *
 */
EXPORT int GetDevicePropertyInformationX(char *srv,char *prp,XPropertyQueryStruct *srvProps,int *num);

/* GetDeviceNames() */
/**
 * \brief Retrieves a list of device names via query to the server given
 * 
 * This call retrieves a list of all device names registered on the server.
 * A classic device server will export its behavior through its properties,
 * and export its "instances" through its "device names", according to the motto "All 
 * devices found on this server have the same set of properties, and all properties apply to
 * any device found on this server".  Under some circumstances this is not true, in which case
 * a more generalized query GetDeviceNamesEx() is required to retrieve a device list associated with a 
 * given property.
 *
 * \param srv is the full device server name for which the property query is to be made
 * \param dlist is a pointer to a NAME16 buffer to receive the name list.
 * \param num is the size of the buffer pointed to by the first parameter.  Note that this
 * buffer space must be pre-allocated, otherwise a system crash will ensue.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \see GetDeviceNamesEx()
 *
 */
EXPORT int GetDeviceNames(char *srv,NAME16 *dlist,int *num);

/* GetDeviceNamesEx() */
/**
 * \brief Retrieves a list of device names associated with a given property via query to the server specified.
 * 
 * This call retrieves a list of all device names registered on the server associated with the property
 * specified in the call.
 * A classic device server will export its behavior through its properties,
 * and export its "instances" through its "device names", according to the motto "All 
 * devices found on this server have the same set of properties, and all properties apply to
 * any device found on this server".  For certain classes of server (in particular middle layer
 * servers) the properties exported might refer to 'apples' and 'oranges' simply because the
 * server is archiving them.  Or there might be other reasons why the device list associated
 * with a particular property has little or nothing to do with that associated with some other
 * property.  Use this call to obtain the device list associated with a particular property.
 *
 * \param srv is the full device server name for which the property query is to be made
 * \param prp is the property for which the device list is desired.
 * \param dlist is a pointer to a NAME16 buffer to receive the name list.
 * \param num is the size of the buffer pointed to by the first parameter.  Note that this
 * buffer space must be pre-allocated, otherwise a system crash will ensue.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \see GetDeviceNames()
 *
 */
EXPORT int GetDeviceNamesEx(char *srv,char *prp, NAME16 *dlist,int *num);

/* GetDevicePropertyEGU() */
/**
 * \brief Retrieves the maximum, minimum values and engineering units for the property specified
 * 
 * This call retrieves the maximum, minimum values and engineering units for the property specified.
 *
 * \param srv [in] is the full device server name for which the property query is to be made
 * \param prp [in] is the property for which the information is desired.
 * \param max [out] is a pointer to receive maximum property value (if non-NULL) as registered.
 * \param min [out] is a pointer to receive minimum property value (if non-NULL) as registered.
 * \param egu [out] is a string buffer to receive the property engineering unit (if non-NULL).
 * Note that buffer space must be pre-allocated, otherwise a system crash will ensue.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \note It is always assumed thar regardless of property format overloading, the maximum and minimum
 * values as well as the engineering units remain the same.
 *
 * \see GetDevicePropertyInformation(), GetDeviceProperties(), GetTargetPropertyInformation()
 *
 */
EXPORT int GetDevicePropertyEGU(char *srv,char *prp,float *max,float *min,char *egu);

/* GetArchivedData() */
/**
 * \brief Retrieves archive data from the 'Central Archiver'.
 * 
 * This call retrieves archive data from either the default 'Central Archive' or the archiver
 * requested in the call.
 *
 * \param keyword [in] is either the keyword-appended full device server name for which the 
 *  archive data is desired or simply the keyword, in which case the default central archiver is
 *  called (registered with export name "HISTORY").
 * \param keyindex [in] is an optional index (if non-zero) into a keyword array (for instance if
 *  the keyword is a trace.
 * \param start [in] is the start time input (expressed as a UNIX timestamp) for which the
 *  archive data are desired.
 * \param stop [in] is the end time input (expressed as a UNIX timestamp) for which the
 *  archive data are desired.
 * \param fiDataArray [out] is a pointer to array of FLTINT objects to receive the archive data.
 * \param num [in/out] is a pointer to an integer giving (as input) the size of the FLTINT buffer 
 *  which is to receive the archive data, and (as ouput) which contains the amount of archive data
 *  actually returned by the call.
 *
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \note It is in general best to use the keyword-appended full device name in this call. For 
 * instance, if the horizontal orbit has been archived as keyword "ORBIT.X", a call to get the 
 * orbit position for a BPM with device name "MONITOR3" could be made by specifying 
 * "ORBIT.X" as the keyword and "3" as the keyindex and assume that this keyword is available 
 * at a central archive server named "HISTORY".  Otherwise, if the name of the relevant archive
 * server is known (either "HISTORY" or something else), then the call can be made by specifying
 * "/<context>/<archive server>/MONITOR3/ORBIT.X" as the keyword and "0" as the keyindex.  The
 * latter although longer is a much more general way to obtain archive data via this call.
 *
 * \note The data in this call are returned as an array of FLTINT objects, containing a data-point 
 * + timestamp pair.
 * 
 * \note If you desire the an entire trace or spectrum of data at a particular timestamp, you
 * should make a call to GetArchivedTraceDataAsFloat() in which case an array of floats is returned at
 * the first timestamp encountered while scanning from start to stop.
 * 
 * \see GetArchivedDataAsText(), GetArchivedDataAsFloat(), GetArchivedTraceDataAsFloat()
 *
 */
EXPORT int GetArchivedData(char *keyword,int keyindex,time_t start,time_t stop,FLTINT *fiDataArray,short *num);

/* GetArchivedDataAsText() */
/**
 * \brief Retrieves archive data as text from the 'Central Archiver'.
 * 
 * This call retrieves archive data as text from the archiver
 * requested in the call.
 *
 * \param devsrv [in] is the keyword-appended full device server name for which the 
 *  archive data is desired.
 * \param start [in] is the start time input (expressed as a UNIX timestamp) for which the
 *  archive data are desired.
 * \param stop [in] is the end time input (expressed as a UNIX timestamp) for which the
 *  archive data are desired.
 * \param niDataArray [out] is a pointer to array of NAME32I objects to receive the archive data.
 * \param num [in/out] is a pointer to an integer giving (as input) the size of the FLTINT buffer 
 *  which is to receive the archive data, and (as ouput) which contains the amount of archive data
 *  actually returned by the call.
 *
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \note This calls uses only the keyword-appended full device name in this call.
 * If the name of the relevant archive
 * server is known (either "HISTORY" or something else), then the call can be made by specifying
 * "/<context>/<archive server>/MONITOR3/ORBIT.X" as the keyword. The
 * archived data will be converted to a text string and entered into the "NAME32" part of the 
 * array of NAME32I doublets returned.  This call must be used in order to retrieve data 
 * stored as strings, such as machine states, etc.
 *
 * \see GetArchivedData(), GetArchivedDataAsFloat(), GetArchivedTraceDataAsFloat()
 *
 */
EXPORT int GetArchivedDataAsText(char *devsrv,time_t start,time_t stop,NAME32I *niDataArray,int *num);

/* GetArchivedDataAsFloat() */
/**
 * \brief Retrieves archive data from the 'Central Archiver' requested in the call.
 * 
 * This call retrieves archive data from the archiver requested in the call.
 * This call supercedes GetArchivedData(), which still allows a simple
 * call into the 'default central archiver'.
 *
 * \param devsrv [in] must be the keyword-appended full device server name for which the 
 *  archive data is desired.
 * \param start [in] is the start time input (expressed as a UNIX timestamp) for which the
 *  archive data are desired.
 * \param stop [in] is the end time input (expressed as a UNIX timestamp) for which the
 *  archive data are desired.
 * \param fiDataArray [out] is a pointer to array of FLTINT objects to receive the archive data.
 * \param num [in/out] is a pointer to an integer giving (as input) the size of the FLTINT buffer 
 *  which is to receive the archive data, and (as ouput) which contains the amount of archive data
 *  actually returned by the call.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \note You must use the keyword-appended full device name in this call. For 
 * instance, if the horizontal orbit has been archived as keyword "ORBIT.X", a call to get the 
 * orbit position for a BPM with device name "MONITOR3" is to be made by specifying
 * "/<context>/<archive server>/MONITOR3/ORBIT.X" as the keyword.
 *
 * \note The data in this call are returned as an array of FLTINT objects, containing a data-point 
 * + timestamp pair.
 * 
 * \note If you desire the an entire trace or spectrum of data at a particular timestamp, you
 * should make a call to GetArchivedTraceDataAsFloat() in which case an array of floats is returned at
 * the first timestamp encountered while scanning from start to stop.
 * 
 * \see GetArchivedDataAsText(), GetArchivedData(), GetArchivedTraceDataAsFloat()
 *
 */
EXPORT int GetArchivedDataAsFloat(char *devsrv,time_t start,time_t stop,FLTINT *fiDataArray,int *num);

/* GetArchivedTraceDataAsFloat() */
/**
 * \brief Retrieves archive data from the 'Central Archiver' requested in the call.
 * 
 * This call retrieves archive data from the archiver requested in the call, and will
 * deliver the contents of an archived keyword at a particular timestamp, which is
 * particularly useful when the keyword represents a trace, waveform, or spectrum.
 *
 * \param devsrv [in] must be the keyword-appended full device server name for which the 
 *  archive data is desired.
 * \param start [in] is the start time input (expressed as a UNIX timestamp) for which the
 *  archive data are desired.
 * \param stop [in] is the end time input (expressed as a UNIX timestamp) for which the
 *  archive data are desired.
 * \param fDataArray [out] is a pointer to array of floats to receive the archive data.
 * \param num [in] gives the size of the float buffer which is to receive the archive data, 
 *  and is likewise used to specify the size of the trace, waveform or spectrum which has been
 *  archived.  To this end, this must indeed match the size of the record archived or the call
 *  will fail.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \note You must use the keyword-appended full device name in this call. For 
 * instance, if the horizontal orbit has been archived as keyword "ORBIT.X", a call to get the 
 * entire orbit at a given timestamp is to be made by specifying
 * "/<context>/<archive server>/#0/ORBIT.X" as the keyword.
 *
 * \note The archiver will start searching for a record beginning with the start time specified.
 * It will stop scanning and return the record found as soon as it encounters a timestamp equal
 * to or greater than than the start time and not exceeding the stop time.
 * 
 * \see GetArchivedDataAsText(), GetArchivedData(), GetArchivedDataAsFloat()
 *
 */
EXPORT int GetArchivedTraceDataAsFloat(char *devsrv,time_t start,time_t stop,float *fDataArray,int num);
EXPORT int GetArchivedDataAsSnapshot(char *devsrv,time_t *target,float *fDataArray,int size);
EXPORT int PropertyDataExchange(void *data,short fmt,short siz,DTYPE *dout, DTYPE *din,short access);

/* GetMyServerAddress() */
/**
 * \brief Obtains FEC and Device server information from the Equipment Name Server based on the
 * address and (optional) local equipment module name of the caller.
 * 
 * A diskless (or otherwise 'in-a-box') server which does not know its address parameters such as
 * fec name, export name, context, port offset, etc. can make this call prior to initializing.
 * Upon success, a call to RegisterFecInformation() can then be made to fix these parameters within
 * the server process.
 *
 * \param ctxName (output optional) the Context of local server.
 * \param eqmName (input/output optional)the local equipment module name of local server.
 * \param expName (output optional) the exported device server name of the local server.
 * \param fecName (output optional) is the FEC name of the local server.
 * \param port (output optional) is the port offset of the local server.
 * to the call.
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 */
EXPORT int GetMyServerAddress(char *eqmName,char *expName,char *ctxName,char *fecName,int *port);

/* savePropertyValues() */
/**
 * \brief Saves value settings of the property name given onto disk.
 * 
 * Using this routine will back out the given values of the named property to a disk file
 * named \<name\>-settings.csv.
 *
 * \param name is the name of the property whose values are to be stored
 * \param values is a reference to the values to be stored
 * \param format is the format of the values to be stored (simple format type)
 * \param size is the array size (not the size in bytes) of the values to be stored.
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 */
EXPORT int savePropertyValues(char *name,void *values,short format,int size);

/* restorePropertyValues() */
/**
 * \brief Retrieves the value settings of the property name given from disk.
 * 
 * Using this routine will restore the given values of the named property from the disk file
 * named \<name\>-settings.csv.
 *
 * \param name is the name of the property whose values are to be restored
 * \param values is a reference to the values to be restored
 * \param format is the format of the values to be restored (simple format type)
 * \param size is the array size (not the size in bytes) of the values to be restored.
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 */
EXPORT int restorePropertyValues(char *name,void *values,short format,int size);


#ifndef APPV_MAJOR
# define APPV_MAJOR 1
#endif
#ifndef APPV_MINOR
# define APPV_MINOR 0
#endif
#ifndef APPV_REVISION
# define APPV_REVISION 0
#endif

EXPORT time_t parseCtimeStr(char *cTimeStr);
EXPORT time_t parseCDateAndTimeStr(char *cDateStr,char *cTimeStr);
#define AssignFecApplicationInfo(maj,min,rev) \
  SetAppDate(parseCDateAndTimeStr(__DATE__,__TIME__));\
  SetAppVersion(maj,min,rev);

#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif

#endif /* _TOOLKIT_H_ */
